/*
 * Copyright © 2018-2021 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <assert.h>

#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3baseband/libenatsc3baseband.h>
#include <libenatsc3alp/libenatsc3alp.h>

#include <libenatsc3filereader/libenatsc3filereader.h>

#include <enatsc3cap/enudpcap.h>

struct UserData
{
    int ifd;
    int ofd;
    int32_t verbose;

    ENATSC3BasebandHandle_t hATSC3Baseband;
    ENATSC3ALPHandle_t hATSC3ALP;

    uint64_t plpIds;
};

int32_t fRunningG = 1;

void
IPV4PacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPayload, int32_t payloadLength, int32_t SID, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;
    EntangleCapHeader echeader;

    echeader.length = payloadLength;
    echeader.vIPv4Addr[0] = 0;
    echeader.vIPv4Addr[1] = 0;
    echeader.vIPv4Addr[2] = 0;
    echeader.vIPv4Addr[3] = 0;
    echeader.port = 0;
    echeader.length = htons(echeader.length);
    echeader.sec = ENUtils_ntohll(pPacketTimeUTC->tv_sec);
    echeader.nsec = ntohl(pPacketTimeUTC->tv_nsec);

    if (pUserData->verbose)
    {
        printf("%"PRId64".%09ld %d bytes\n", pPacketTimeUTC->tv_sec, 
            pPacketTimeUTC->tv_nsec, payloadLength);
    }

    if (write(pUserData->ofd, &echeader, sizeof(echeader)) != sizeof(echeader))
    {
        fprintf(stderr, "Error writing ipv4 file: %s\n", 
            strerror(errno));
        goto out;
    }

    if (write(pUserData->ofd, pPayload, payloadLength) != payloadLength) 
    {
        fprintf(stderr, "Error writing ipv4 file: %s\n", 
            strerror(errno));
        goto out;
    }

out:
    return;
}

void
ALPPacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC,
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3ALPProcessALPPacket(pUserData->hATSC3ALP, plpId,
        pPacketTimeUTC, packetNumber, fError, fDiscontinuity, pPacket, 
        sPacket))
    {
        fprintf(stderr, "Failed to process ALP packet\n");
    }
}

void
FileReaderALPPacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    int32_t fError, int32_t fDiscontinuity,
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3ALPProcessALPPacket(pUserData->hATSC3ALP, plpId, pPacketTimeUTC, 
        packetNumber, fError, fDiscontinuity, pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process ALP packet\n");
    }

}

void
FileReaderBasebandPacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3BasebandProcessBasebandPacket(pUserData->hATSC3Baseband,
        plpId, pPacketTimeUTC, packetNumber, fError, fDiscontinuity, pPacket, 
        sPacket))
    {
        fprintf(stderr, "Failed to process baseband packet\n");
    }
}

int32_t
parseArgs(int argc, char **argv, char **ppInFilename, 
    ENATSC3FileType *pInFileType, char **ppOutFilename, 
    int32_t *pTOffsetSeconds, uint64_t *pPLPIds, int32_t *pVerbose)
{
    int32_t err = 0;
    int32_t i;

    *ppInFilename = NULL;
    *ppOutFilename = NULL;
    *pPLPIds = 0;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-b"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_ALP;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-bb"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_Baseband;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-j288alp"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_J288ALP;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-ba"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_ALPRaw;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutFilename = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-plp"))
        {
            if (i + 1 < argc)
            {
                *pPLPIds |= 1ULL << atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-v"))
        {
            if (i + 1 < argc)
            {
                *pVerbose = atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-offset"))
        {
            if (i + 1 < argc)
            {
                *pTOffsetSeconds = atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else err = 1;
    }

    if (*ppInFilename == NULL || *ppOutFilename == NULL)
    {
        err = 1;
    }

    if (err)
    {
        fprintf(stderr, "Usage: %s [-b alpcapturefile] [-bb basebandcapturefile]  -o ipv4file -plp plpid [-plp plpid ...] [-offset seconds] [-v level]\n", argv[0]);
        err = 1;
    }

    return err;
}

void
sighandler(int signum)
{
    if (signum == SIGINT)
    {
        fRunningG = 0;
    }
}

int
main(int argc, char **argv)
{
    int32_t fEnLogInitialized = 0;
    char *pInFilename;
    ENATSC3FileType inFileType;
    char *pOutFilename;
    int32_t tOffsetSeconds = 0;
    struct UserData userdata = 
    {
        .ifd = -1,
        .ofd = -1,
        .verbose = 0,
        .hATSC3Baseband = NULL,
        .hATSC3ALP = NULL,
    };
    struct sigaction act;
    int rval = 0;

    // Graceful exit on ctrl-c
    act.sa_handler = sighandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    sigaction(SIGINT, &act, NULL);

    if (EnLogInitialize(1))
    {
        fprintf(stderr, "Failed to initialize EnLog\n");
        rval = -1;
        goto out;
    }
    fEnLogInitialized = 1;

    if (parseArgs(argc, argv, &pInFilename, &inFileType, &pOutFilename, 
        &tOffsetSeconds, &userdata.plpIds, &userdata.verbose))
    {
        rval = 1;
        goto out;
    }

    if (inFileType == ENATSC3FileType_Baseband)
    {
        if (ENATSC3BasebandOpen(&userdata.hATSC3Baseband))
        {
            fprintf(stderr, "Failed to open baseband processor\n");
            rval = 1;
            goto out;
        }

        if (ENATSC3BasebandRegisterCallbacks(userdata.hATSC3Baseband,
            ALPPacketCallback, &userdata))
        {
            fprintf(stderr, "Failed to register baseband callbacks\n");
            rval = 1;
            goto out;
        }
    }

    if (ENATSC3ALPOpen(&userdata.hATSC3ALP))
    {
        fprintf(stderr, "Failed to open ALP processor\n");
        rval = 1;
        goto out;
    }

    if (ENATSC3ALPRegisterCallbacks(userdata.hATSC3ALP, 
        IPV4PacketCallback, NULL, &userdata))
    {
        fprintf(stderr, "Failed to register callbacks\n");
        rval = 1;
        goto out;
    }

    if (userdata.hATSC3Baseband)
    {
        if (ENATSC3BasebandSetPLPs(userdata.hATSC3Baseband,
            userdata.plpIds))
        {
            fprintf(stderr, "Failed to set baseband PLP mask %"PRIx64"\n",
                userdata.plpIds);
            rval = 1;
            goto out;
        }
    }

    if (ENATSC3ALPSetPLPs(userdata.hATSC3ALP, userdata.plpIds))
    {
        fprintf(stderr, "Failed to set ALP PLP mask %"PRId64"\n", 
            userdata.plpIds);
        rval = 1;
        goto out;
    }

    userdata.ifd = open(pInFilename, O_RDONLY);
    if (userdata.ifd < 0)
    {
        fprintf(stderr, "Failed to open %s for reading: %s\n", 
            pInFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    userdata.ofd = open(pOutFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (userdata.ofd < 0)
    {
        fprintf(stderr, "Failed to open %s for writing: %s\n",
            pOutFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    switch (inFileType)
    {
        case ENATSC3FileType_ALP:
            rval = ENATSC3FileReaderReadALPFile(userdata.ifd, tOffsetSeconds, 
                FileReaderALPPacketCallback, NULL, &fRunningG, &userdata);
            break;
        case ENATSC3FileType_Baseband:
            rval = ENATSC3FileReaderReadBasebandFile(userdata.ifd, 
                tOffsetSeconds, FileReaderBasebandPacketCallback, NULL, 
                &fRunningG, &userdata);
            break;
        case ENATSC3FileType_ALPRaw:
            rval = ENATSC3FileReaderReadALPRawFile(userdata.ifd, 
                FileReaderALPPacketCallback, NULL, &fRunningG, &userdata);
            break;
        case ENATSC3FileType_J288ALP:
            rval = ENATSC3FileReaderReadJ288ALPFile(userdata.ifd, 
                tOffsetSeconds, FileReaderALPPacketCallback, NULL, &fRunningG, 
                &userdata);
            break;
        default:
            fprintf(stderr, "Unhandled input file type %d\n", inFileType);
            rval = 1;
            goto out;
            break;
    }

    if (ENATSC3ALPSetPLPs(userdata.hATSC3ALP, 0))
    {
        fprintf(stderr, "Failed to clear ALP PLP mask\n"); 
        rval = 1;
        goto out;
    }

    if (userdata.hATSC3Baseband)
    {
        if (ENATSC3BasebandSetPLPs(userdata.hATSC3Baseband, 0))
        {
            fprintf(stderr, "Failed to remove clear baseband PLP mask\n");
            rval = 1;
            goto out;
        }
    }
    
out:
    if (userdata.hATSC3ALP)
    {
        ENATSC3ALPClose(userdata.hATSC3ALP);
    }

    if (userdata.hATSC3Baseband)
    {
        ENATSC3BasebandClose(userdata.hATSC3Baseband);
    }

    if (userdata.ofd >= 0)
    {
        close(userdata.ofd);
    }

    if (userdata.ifd >= 0)
    {
        close(userdata.ifd);
    }

    if (fEnLogInitialized)
    {
        if (EnLogShutdown())
        {
            fprintf(stderr, "Failed to shut down EnLog\n");
        }
    }

    return rval;
}
