/*
 * Copyright © 2018-2021 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <libxml2/libxml/xmlreader.h>

#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <libenatsc3utils/enxml.h>
#include <libenatsc3utils/enipfilterset.h>
#include <libenatsc3utils/enatsc3bitratestatistics.h>

#include <enatsc3-common/enatsc3.h>
#include <enatsc3-common/enatsc3_specdefs.h>
#include <libenatsc3baseband/libenatsc3baseband.h>
#include <libenatsc3alp/libenatsc3alp.h>
#include <libenatsc3ipv4/libenatsc3ipv4.h>
#include <libenatsc3ipv4/internetchecksum.h>
#include <libenatsc3route/atsc3routesls_mime.h>
#include <libenatsc3route/atsc3routesls_stsid.h>
#include <libenatsc3route/libenatsc3route.h>

#include <libenatsc3filereader/libenatsc3filereader.h>

struct UserData
{
    ENVerbosity verbosity;
    int32_t fIgnoreChecksumErrors;
    int32_t fBitrateStatistics;
    int ifd;

    EnIPFilterSet *pIPFilterSet;

    ENATSC3BasebandHandle_t hATSC3Baseband;
    ENATSC3ALPHandle_t hATSC3ALP;
    ENATSC3IPV4Handle_t hATSC3IPV4;
    ENATSC3ROUTESessionHandle_t hROUTESession;
    uint32_t vDestinationAddresses[4];
    int32_t nDestinationAddresses;
    uint16_t vDestinationPorts[4];
    int32_t nDestinationPorts;

    struct _LCTChannelInfo
    {
        int32_t fInUse;

        uint32_t tsi;

        int64_t currentTOI;
        uint8_t codepoint;
        char strCurrentURL[256];
        int64_t objectSize;
        int64_t pos;

        int fd;
    } vLCTChannelInfos[32];

    char *pOutPath;

    uint64_t plpIds;

    int32_t fileNum;
    int32_t fNoFileNums;

    // Statistics
    ENATSC3BitrateStatistics stats;
};

int32_t fRunningG = 1;

ENATSC3Status
FindLCTChannelForTSI(struct UserData *pUserData, uint32_t tsi, 
    struct _LCTChannelInfo **ppLCTChannelInfo)
{
    int32_t i;
    ENATSC3Status rval = ENATSC3Status_Failure;

    for (i = 0; i < numof(pUserData->vLCTChannelInfos); i++)
    {
        if (pUserData->vLCTChannelInfos[i].fInUse && 
            pUserData->vLCTChannelInfos[i].tsi == tsi)
        {
            *ppLCTChannelInfo = pUserData->vLCTChannelInfos + i;
            rval = ENATSC3Status_Success;
            break;
        }
    }

    return rval;
}

ENATSC3Status
AddLCTChannelForTSI(struct UserData *pUserData, uint32_t tsi, 
    struct _LCTChannelInfo **ppLCTChannelInfo)
{
    int32_t i;
    ENATSC3Status rval = ENATSC3Status_Failure;

    for (i = 0; i < numof(pUserData->vLCTChannelInfos); i++)
    {
        if (!pUserData->vLCTChannelInfos[i].fInUse)
        {
            pUserData->vLCTChannelInfos[i].fInUse = 1;
            pUserData->vLCTChannelInfos[i].tsi = tsi;
            pUserData->vLCTChannelInfos[i].currentTOI = -1;
            pUserData->vLCTChannelInfos[i].fd = -1;
            *ppLCTChannelInfo = pUserData->vLCTChannelInfos + i;
            rval = ENATSC3Status_Success;
            break;
        }
    }

    return rval;
}

ENATSC3Status
RemoveLCTChannelForTSI(struct UserData *pUserData, uint32_t tsi) 
{
    int32_t i;
    ENATSC3Status rval = ENATSC3Status_Failure;

    for (i = 0; i < numof(pUserData->vLCTChannelInfos); i++)
    {
        if (pUserData->vLCTChannelInfos[i].fInUse && 
            pUserData->vLCTChannelInfos[i].tsi == tsi)
        {
            if (pUserData->vLCTChannelInfos[i].fd >= 0)
            {
                close(pUserData->vLCTChannelInfos[i].fd);
                pUserData->vLCTChannelInfos[i].fd = -1;
            }
            pUserData->vLCTChannelInfos[i].fInUse = 0;
            rval = ENATSC3Status_Success;
            break;
        }
    }

    return rval;
}

ENATSC3Status
MediaFileBeginCallback(struct timespec *pPacketTimeUTC, 
    struct _LCTChannelInfo *pLCTChannelInfo, struct UserData *pUserData)
{
    char strFilename[256];
    ENATSC3Status rval = ENATSC3Status_Success;

    if (pUserData->fNoFileNums)
    {
        if (snprintf(strFilename, sizeof(strFilename), "%s/%s",
            pUserData->pOutPath, pLCTChannelInfo->strCurrentURL) >= 
            sizeof(strFilename))
        {
            fprintf(stderr, "Initialization URL %s too long\n", 
                pLCTChannelInfo->strCurrentURL);
            goto out;
        }
    }
    else
    {
        if (snprintf(strFilename, sizeof(strFilename), "%s/%s-%d",
            pUserData->pOutPath, pLCTChannelInfo->strCurrentURL, 
            pUserData->fileNum++) >= sizeof(strFilename))
        {
            fprintf(stderr, "Initialization URL %s too long\n", 
                pLCTChannelInfo->strCurrentURL);
            goto out;
        }
    }

    pLCTChannelInfo->fd =
        open(strFilename, O_WRONLY | O_CREAT, 0666);
    if (pLCTChannelInfo->fd < 0)
    {
        fprintf(stderr, "Failed to create %s: %s\n", strFilename,
            strerror(errno));
        rval = ENATSC3Status_Failure;
    }

out:
    return rval;
}

void
MediaFileEndCallback(struct timespec *pPacketTimeUTC, 
    struct _LCTChannelInfo *pLCTChannelInfo, struct UserData *pUserData) 
{
    if (pLCTChannelInfo->fd >= 0)
    {
        close(pLCTChannelInfo->fd);
        pLCTChannelInfo->fd = -1;
    }

    return;
}

void
MediaFileAbortCallback(struct timespec *pPacketTimeUTC, 
    struct _LCTChannelInfo *pLCTChannelInfo, struct UserData *pUserData)
{
    if (pLCTChannelInfo->fd >= 0)
    {
        close(pLCTChannelInfo->fd);
        pLCTChannelInfo->fd = -1;
    }

    return;
}

void
MediaFileDataCallback(struct timespec *pPacketTimeUTC, 
    struct _LCTChannelInfo *pLCTChannelInfo, 
    const uint8_t *pPayload, int32_t sPayload, int64_t objectPosition, 
    int64_t objectSize, int32_t fError, struct UserData *pUserData) 
{
    int nb;

    if (lseek(pLCTChannelInfo->fd, objectPosition, SEEK_SET) != objectPosition)
    {
        fprintf(stderr, "Failed to seek to %s:%"PRId64"\n", 
            pLCTChannelInfo->strCurrentURL, objectPosition);
        goto out;
    }

    nb = write(pLCTChannelInfo->fd, pPayload, sPayload);
    if (nb == -1)
    {
        fprintf(stderr, "Failed to write payload for %s: %s\n",
            pLCTChannelInfo->strCurrentURL, strerror(errno));
        goto out;
    }
    else if (nb != sPayload)
    {
        fprintf(stderr, "Short write of payload for %s (%d < %d)\n",
            pLCTChannelInfo->strCurrentURL, nb, sPayload);
        goto out;
    }

out:
    return;
}

void
ObjectDataCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    uint32_t tsi, uint32_t toi, uint8_t codepoint, const char *pId, 
    ENATSC3ROUTESessionIdType idType, const uint8_t *pPayload, 
    int32_t payloadLength, int64_t objectPosition, int64_t tolTransferLength, 
    int64_t ftiTransferLength, int32_t fError, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;
    struct _LCTChannelInfo *pLCTChannelInfo;

    if (FindLCTChannelForTSI(pUserData, tsi, &pLCTChannelInfo))
    {
        fprintf(stderr, "Failed to find LCT channel info for tsi %u\n", tsi);
        goto out;
    }

    if (toi != pLCTChannelInfo->currentTOI && 
        pLCTChannelInfo->currentTOI != -1)
    {
        // If we get here then the object Id has changed prematurely. This
        // would suggest that we didn't receive the end of the previous
        // object.
        fprintf(stderr, 
            "%"PRIu64" %"PRId64".%06lu object changed (%"PRId64":%s) %"PRId64" != %"PRId64"\n",
            packetNumber, 
            pPacketTimeUTC->tv_sec, pPacketTimeUTC->tv_nsec / 1000,
            pLCTChannelInfo->currentTOI,
            pLCTChannelInfo->strCurrentURL,
            pLCTChannelInfo->objectSize,
            pLCTChannelInfo->pos);

        MediaFileAbortCallback(pPacketTimeUTC, pLCTChannelInfo, pUserData);
        pLCTChannelInfo->strCurrentURL[0] = 0;
        pLCTChannelInfo->currentTOI = -1;

#if 0
        if (objectPosition != 0)
        {
            // We need to wait til we're at the start of a new object.
            goto out;
        }
#endif
    }

    // If we've gotten this far then we are processing an object.
    if (pLCTChannelInfo->currentTOI == -1)
    {
        char strContentType[64];

#if 0
        {
            if (payloadLength > LiteralLength("Content-Location: ") &&
                !strncmp((char *)pPayload, "Content-Location: ",
                LiteralLength("Content-Location: ")))
            {
                int32_t payloadOffset = LiteralLength("Content-Location: ");

                while (payloadOffset + 3 < payloadLength &&
                    (pPayload[payloadOffset] != '\r' ||
                     pPayload[payloadOffset + 1] != '\n' ||
                     pPayload[payloadOffset + 2] != '\r' ||
                     pPayload[payloadOffset + 3] != '\n'))
                {
                    payloadOffset++;
                }

                strncpy(pLCTChannelInfo->strCurrentURL, (char *)pPayload +
                    LiteralLength("Content-Location: "), payloadOffset -
                    LiteralLength("Content-Location: "));
                pLCTChannelInfo->strCurrentURL[
                    payloadOffset - LiteralLength("Content-Location: ")] = 0;
                printf("URL: %s\n", pLCTChannelInfo->strCurrentURL);

                pPayload += payloadOffset + 4;
                payloadLength -= payloadOffset + 4;
            }
        }
#else
        if (ENATSC3ROUTESessionGetObjectURLAndContentType(
            pUserData->hROUTESession, 
            tsi, toi, pLCTChannelInfo->strCurrentURL, 
            sizeof(pLCTChannelInfo->strCurrentURL), strContentType, 
            sizeof(strContentType)))
        {
            fprintf(stderr, "Failed to get url for tsi %u toi %u\n",
                tsi, toi);
            goto out;
        }
#endif

        if (pUserData->verbosity >= ENVerbosity_ObjectTrace)
        {
            printf("%"PRIu64" %"PRId64".%06lu MediaFileBegin %s tsi %u codepoint %u\n",
                packetNumber, pPacketTimeUTC->tv_sec, pPacketTimeUTC->tv_nsec / 1000,
                pLCTChannelInfo->strCurrentURL, tsi, codepoint);
        }

        MediaFileBeginCallback(pPacketTimeUTC, pLCTChannelInfo, 
            pUserData);

        pLCTChannelInfo->currentTOI = toi;
        pLCTChannelInfo->codepoint = codepoint;
        pLCTChannelInfo->pos = 0;
        pLCTChannelInfo->objectSize = -1;
    }

    if (pLCTChannelInfo->objectSize == -1)
    {
        if (tolTransferLength != -1)
        {
            pLCTChannelInfo->objectSize = tolTransferLength;
        }
        else if (ftiTransferLength != -1)
        {
            pLCTChannelInfo->objectSize = ftiTransferLength;
        }
    }
    else
    {
        if (tolTransferLength != -1)
        {
            if (pLCTChannelInfo->objectSize != tolTransferLength)
            {
                fprintf(stderr, 
                    "%"PRIu64" %"PRId64".%06lu TOL object size changed %"PRId64" != %"PRId64"\n",
                    packetNumber, 
                    pPacketTimeUTC->tv_sec, pPacketTimeUTC->tv_nsec / 1000,
                    pLCTChannelInfo->objectSize, tolTransferLength);
                printf("MediaFileAbort %s tsi %d\n",
                    pLCTChannelInfo->strCurrentURL, pLCTChannelInfo->tsi);
                MediaFileAbortCallback(pPacketTimeUTC, pLCTChannelInfo, 
                    pUserData); 
                pLCTChannelInfo->strCurrentURL[0] = 0;
                pLCTChannelInfo->currentTOI = -1;
                goto out;
            }
        }

        if (ftiTransferLength != -1)
        {
            if (pLCTChannelInfo->objectSize != ftiTransferLength)
            {
                fprintf(stderr, "%"PRIu64" FTI object size changed %"PRId64" != %"PRId64"\n",
                    packetNumber, pLCTChannelInfo->objectSize,
                    ftiTransferLength);
                printf("MediaFileAbort %s tsi %d\n",
                    pLCTChannelInfo->strCurrentURL, pLCTChannelInfo->tsi);
                MediaFileAbortCallback(pPacketTimeUTC, pLCTChannelInfo, 
                    pUserData);
                pLCTChannelInfo->strCurrentURL[0] = 0;
                pLCTChannelInfo->currentTOI = -1;
                goto out;
            }
        }
    }

    // We currently expect data to be delivered in order.
    //
    // XXXX Strictly speaking data may arrive out of order if @order=0 in the
    // SrcFlow's Payload element.
    if (pLCTChannelInfo->pos != objectPosition)
    {
        fprintf(stderr, "%"PRIu64" %"PRIu64".%.06lu av offset error %"PRId64" != %"PRId64" (%"PRId64" missing) (%s size %"PRId64")\n",
            packetNumber,
            pPacketTimeUTC->tv_sec, pPacketTimeUTC->tv_nsec / 1000000,
            pLCTChannelInfo->pos, objectPosition,
            objectPosition - pLCTChannelInfo->pos,
            pLCTChannelInfo->strCurrentURL, pLCTChannelInfo->objectSize);
    }

    MediaFileDataCallback(pPacketTimeUTC, pLCTChannelInfo,
        pPayload, payloadLength, objectPosition,
        pLCTChannelInfo->objectSize, 0, pUserData);

    pLCTChannelInfo->pos = objectPosition + payloadLength;

    // See if the object is now complete
    if (pLCTChannelInfo->pos == pLCTChannelInfo->objectSize)
    {
        if (pUserData->verbosity >= ENVerbosity_ObjectTrace)
        {
            printf("%"PRIu64" %"PRIu64".%.06lu MediaFileEnd %s tsi %d\n", 
                packetNumber,
                pPacketTimeUTC->tv_sec, pPacketTimeUTC->tv_nsec / 1000000,
                pLCTChannelInfo->strCurrentURL, 
                pLCTChannelInfo->tsi);
        }
        MediaFileEndCallback(pPacketTimeUTC, pLCTChannelInfo, pUserData);
        pLCTChannelInfo->strCurrentURL[0] = 0;
        pLCTChannelInfo->currentTOI = -1;
    }

out:
    return;
}

void
SessionResetCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;
    int32_t i;

    for (i = 0; i < numof(pUserData->vLCTChannelInfos); i++)
    {
        if (pUserData->vLCTChannelInfos[i].fInUse)
        {
            struct _LCTChannelInfo *pLCTChannelInfo;

            pLCTChannelInfo = pUserData->vLCTChannelInfos + i;

            if (pLCTChannelInfo->currentTOI != -1)
            {
                MediaFileAbortCallback(pPacketTimeUTC, pLCTChannelInfo, 
                    pUserData);
                pLCTChannelInfo->currentTOI = -1;
            }

            pUserData->vLCTChannelInfos[i].fInUse = 0;
        }
    }
}

ENATSC3Status
MPDAvailableCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    const char *pMPD, int32_t sMPD, int32_t version, uint32_t crc32, 
    void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;
    char strFilename[256];
    int fd = -1;
    int nb;
    ENATSC3Status rval = ENATSC3Status_Success;

    if (snprintf(strFilename, sizeof(strFilename), "%s/mpd-%d-%"PRIu64,
        pUserData->pOutPath, version, pPacketTimeUTC->tv_sec) >= 
        sizeof(strFilename))
    {
        fprintf(stderr, "MPD filename too long\n");
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        fprintf(stderr, "Failed to open %s: %s\n", strFilename,
            strerror(errno));
        goto out;
    }

    nb = write(fd, pMPD, sMPD);
    if (nb == -1)
    {
        fprintf(stderr, "Failed to write %s: %s\n", strFilename,
            strerror(errno));
        goto out;
    }
    else if (nb != sMPD)
    {
        fprintf(stderr, "Short write of %s (%d < %d)\n", strFilename,
            nb, sMPD);
        goto out;
    }

out:
    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

ENATSC3Status
HELDAvailableCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    const char *pHELD, int32_t sHELD, int32_t version, uint32_t crc32, 
    void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;
    char strFilename[256];
    int fd = -1;
    int nb;
    ENATSC3Status rval = ENATSC3Status_Success;

    if (snprintf(strFilename, sizeof(strFilename), "%s/held-%d-%"PRIu64,
        pUserData->pOutPath, version, pPacketTimeUTC->tv_sec) >= 
        sizeof(strFilename))
    {
        fprintf(stderr, "HELD filename too long\n");
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        fprintf(stderr, "Failed to open %s: %s\n", strFilename,
            strerror(errno));
        goto out;
    }

    nb = write(fd, pHELD, sHELD);
    if (nb == -1)
    {
        fprintf(stderr, "Failed to write %s: %s\n", strFilename,
            strerror(errno));
        goto out;
    }
    else if (nb != sHELD)
    {
        fprintf(stderr, "Short write of %s (%d < %d)\n", strFilename,
            nb, sHELD);
        goto out;
    }

out:
    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

void
SLSAvailableCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    const char *pSLSFilename, const uint8_t *pSLS, int32_t sSLS, 
    int32_t fGzipped, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;
    char strFilename[256];
    int fd = -1;
    int nb;

    if (snprintf(strFilename, sizeof(strFilename), "%s/sls-%s-%d",
        pUserData->pOutPath, pSLSFilename, pUserData->fileNum++) >= 
        sizeof(strFilename))
    {
        fprintf(stderr, "SLS filename too long\n");
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        fprintf(stderr, "Failed to open %s: %s\n", strFilename,
            strerror(errno));
        goto out;
    }

    nb = write(fd, pSLS, sSLS);
    if (nb == -1)
    {
        fprintf(stderr, "Failed to write %s: %s\n", strFilename,
            strerror(errno));
        goto out;
    }
    else if (nb != sSLS)
    {
        fprintf(stderr, "Short write of %s (%d < %d)\n", strFilename,
            nb, sSLS);
        goto out;
    }

out:
    if (fd >= 0)
    {
        close(fd);
    }

    return;
}

int32_t
HandleUDPPacket(int32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint32_t sourceAddress, uint32_t destinationAddress, 
    const uint8_t *pPacket, int32_t packetLength, struct UserData *pUserData)
{
    int32_t sourcePort;
    int32_t destPort;
    int32_t length;
    int32_t checksum;
    ENInternetChecksumContext iccontext;
    uint32_t sum;
    int32_t rval = 0;

    if (pUserData->fBitrateStatistics)
    {
        ENATSC3BitrateStatisticsUpdate(&pUserData->stats, pPacketTimeUTC, 
            packetLength);
    }
    
    if (packetLength < 8)
    {
        fprintf(stderr, "UDP too small, dropping\n");
        rval = 1;
        goto out;
    }

    length = pPacket[4] << 8 | pPacket[5];

    if (packetLength < length)
    {
        fprintf(stderr, "Not enough data for UDP packet: %d < %d\n",
            packetLength, length);
        rval = 1;
        goto out;
    }

    sourcePort = pPacket[0] << 8 | pPacket[1];
    destPort = pPacket[2] << 8 | pPacket[3];
    checksum = pPacket[6] << 8 | pPacket[7];

    if (pUserData->verbosity >= ENVerbosity_Pedantic)
    {
        printf("UDP\n");
        printf("  Source port: %d\n", sourcePort);
        printf("  Dest port: %d\n", destPort);
        printf("  Length: %d\n", length);
        printf("  Checksum: %x\n", checksum);
    }

    if (EnIPFilterSetFilter(destinationAddress, destPort, 
        pUserData->pIPFilterSet))
    {
        goto out;
    }

    if (checksum)
    {
        struct
        {
            uint32_t sourceAddress;
            uint32_t destinationAddress;
            uint8_t zero;
            uint8_t protocol;
            uint8_t lengthHi;
            uint8_t lengthLo;
        } __attribute__((packed)) pseudoHeader =
        {
            .sourceAddress = sourceAddress,
            .destinationAddress = destinationAddress,
            .zero = 0,
            .protocol = 17,
            .lengthHi = packetLength >> 8,
            .lengthLo = packetLength & 0xff,
        };

        ENInternetChecksumInitialize(&iccontext);
        ENInternetChecksumCalc((uint8_t *)&pseudoHeader, sizeof(pseudoHeader), 
            &iccontext);
        ENInternetChecksumCalc(pPacket, packetLength, &iccontext);
        sum = ENInternetChecksumFinalize(&iccontext);
    }
    else
    {
        sum = 0;
    }

    if (sum == 0)
    {
        if (pUserData->verbosity >= ENVerbosity_Pedantic)
        {
            printf("  Checksum verified\n");
        }
    }
    else
    {
        static int32_t fChecksumErrorLogged = 0;

        if (pUserData->verbosity >= ENVerbosity_Pedantic ||
            !fChecksumErrorLogged)
        {
            printf("  UDP CHECKSUM ERROR: %x%s\n", sum,
                pUserData->fIgnoreChecksumErrors ? " (ignored)" : "");
            if (pUserData->verbosity < ENVerbosity_Pedantic)
            {
                printf("(Further checksum errors suppressed; enable pedantic verbosity to view\n");
                fChecksumErrorLogged = 1;
            }
        }
    }

    if (sum == 0 || pUserData->fIgnoreChecksumErrors)
    {
        if (ENATSC3ROUTESessionProcessUDPPacket(pUserData->hROUTESession, 
            pPacketTimeUTC, packetNumber, plpId, pPacket + 8, packetLength - 8))
        {
            fprintf(stderr, "Failed to process UDP packet\n");
            rval = 1;
        }
    }

out:
    return rval;
}

void
TransportPacketCallback(int32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint32_t sourceAddress, uint32_t destinationAddress, 
    uint16_t protocol, const uint8_t *pPacket, int32_t sPacket, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;
    switch (protocol)
    {
        case 17:
            HandleUDPPacket(plpId, pPacketTimeUTC, packetNumber, sourceAddress, 
                destinationAddress, pPacket, sPacket, pUserData);
            break;
        default:
            fprintf(stderr, "Unhandled protocol %d\n", protocol);
            break;
    }
}

void
IPV4PacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPayload, int32_t payloadLength, int32_t SID, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;

    ENATSC3IPV4ProcessIPV4Packet(pUserData->hATSC3IPV4, plpId, pPacketTimeUTC, 
        packetNumber, pPayload, payloadLength);

    return;
}

int32_t
HandleLinkMappingTable(
    ENATSC3ALPLinkLayerSignalingHeader *pLinkLayerSignalingHeader, 
    uint32_t plpId, struct timespec *pPacketTimeUTC, const uint8_t *pPayload, 
    int32_t payloadLength, struct UserData *pUserData)
{
    uint32_t numPLPs;
    int i;
    int j;

    // Skip over signalling header
    pPayload += 5;

    numPLPs = (pPayload[0] >> 2) + 1;
    pPayload++;

    if (pUserData->verbosity >= ENVerbosity_Pedantic)
    {
        printf("num PLPs: %d\n", numPLPs);

        for (i = 0; i < numPLPs; i++)
        {
            uint32_t plpID;
            uint32_t numMulticasts;

            plpID = pPayload[0] >> 2;
            numMulticasts = pPayload[1];
            pPayload += 2;

            printf("  plp: %d\n", plpID);
            printf("    multicasts: %d\n", numMulticasts);

            for (j = 0; j < numMulticasts; j++)
            {
                uint32_t srcIPAddr;
                uint32_t dstIPAddr;
                uint32_t srcUDPPort;
                uint32_t dstUDPPort;
                uint32_t sidFlag;
                uint32_t compressedFlag;
                uint32_t SID;
                uint32_t contextId;

                srcIPAddr = (pPayload[0] << 24) | (pPayload[1] << 16) | 
                    (pPayload[2] << 8) | pPayload[3]; 
                dstIPAddr = (pPayload[4] << 24) | (pPayload[5] << 16) | 
                    (pPayload[6] << 8) | pPayload[7]; 
                srcUDPPort = (pPayload[8] << 8) | pPayload[9];
                dstUDPPort = (pPayload[10] << 8) | pPayload[11];
                sidFlag = pPayload[12] & 0x80;
                compressedFlag = pPayload[12] & 0x40;

                pPayload += 13;

                if (sidFlag)
                {
                    SID = pPayload[0];
                    pPayload++;
                }

                if (compressedFlag)
                {
                    contextId = pPayload[0];
                    pPayload++;
                }

                printf("      Multicast %d\n", j);
                printf("        src: %d.%d.%d.%d:%d\n", 
                    (srcIPAddr >> 24)  & 0xff, (srcIPAddr >> 16)  & 0xff,
                    (srcIPAddr >> 8)  & 0xff, srcIPAddr  & 0xff,
                    srcUDPPort);
                printf("        dst: %d.%d.%d.%d:%d\n", 
                    (dstIPAddr >> 24)  & 0xff, (dstIPAddr >> 16)  & 0xff,
                    (dstIPAddr >> 8)  & 0xff, dstIPAddr  & 0xff,
                    dstUDPPort);
                printf("        SID flag: %d\n", sidFlag);
                if (sidFlag)
                {
                    printf("        SID: %d\n", SID);
                }
                printf("        compressed flag: %d\n", compressedFlag);
                if (compressedFlag)
                {
                    printf("        context id: %d\n", contextId);
                }
            }
        }
    }

    return 0;
}

void
LinkLayerSignalingPacketCallback(uint32_t plpId, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t fError, int32_t fDiscontinuity, const uint8_t *pPayload, 
    int32_t payloadLength, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;
    ENATSC3ALPLinkLayerSignalingHeader llsheader;

    llsheader.signaling_type = pPayload[0];
    llsheader.signaling_type_extension = (pPayload[1] << 8) | pPayload[2];
    llsheader.signaling_version = pPayload[3];
    llsheader.signaling_format = (pPayload[4] >> 6) & 0x3;
    llsheader.signaling_encoding = (pPayload[4] >> 4) & 0x3;

    switch (llsheader.signaling_type)
    {
        case ENATSC3ALPSignalingType_LinkMappingTable:
            HandleLinkMappingTable(&llsheader, plpId, pPacketTimeUTC, 
                pPayload, payloadLength, pUserData);
            break;
        default:
            printf("Unhandled link layer signaling type %d\n", 
                llsheader.signaling_type);
            break;
    }

    return;
}

void
ALPPacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC,
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3ALPProcessALPPacket(pUserData->hATSC3ALP, plpId,
        pPacketTimeUTC, packetNumber, fError, fDiscontinuity, pPacket, 
        sPacket))
    {
        fprintf(stderr, "Failed to process ALP packet\n");
    }
}

void
AddIPAddressCallback(uint32_t addr, uint16_t port, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (EnIPFilterSetAddFilter(addr, port, pUserData->pIPFilterSet))
    {
        fprintf(stderr, "Failed to add filter for %.d.%.d.%.d.%.d:%d\n",
            (addr >> 24) & 0xff, (addr >> 16) & 0xff, (addr >> 8) & 0xff,
            addr & 0xff, port);
    }
}

void
RemoveIPAddressCallback(uint32_t addr, uint16_t port, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (EnIPFilterSetRemoveFilter(addr, port, pUserData->pIPFilterSet))
    {
        fprintf(stderr, "Failed to remove filter for %.d.%.d.%.d.%.d:%d\n",
            (addr >> 24) & 0xff, (addr >> 16) & 0xff, (addr >> 8) & 0xff,
            addr & 0xff, port);
    }
}

void
CommitIPAddressesCallback(void *pCbUserData)
{
    // Nothing to do
}

void
LCTChannelAddedCallback(struct timespec *pPacketTime, uint64_t packetNumber, 
    uint32_t addr, uint16_t port, uint32_t tsi, const char *pId, 
    ENATSC3ROUTESessionIdType idType, void *pCbUserData)
{
    struct _LCTChannelInfo *pLCTChannelInfo;
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (AddLCTChannelForTSI(pUserData, tsi, &pLCTChannelInfo))
    {
        fprintf(stderr, "Failed to allocate LCT Channel info for tsi %u\n",
            tsi);
    }
}

void
LCTChannelRemovedCallback(struct timespec *pPacketTime, uint64_t packetNumber, 
    uint32_t addr, uint16_t port, uint32_t tsi, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (RemoveLCTChannelForTSI(pUserData, tsi))
    {
        fprintf(stderr, "Failed to free LCT Channel info for tsi %u\n",
            tsi);
    }
}

void
FileReaderIPV4PacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3IPV4ProcessIPV4Packet(pUserData->hATSC3IPV4, plpId, 
        pPacketTimeUTC, packetNumber, pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process IPV4 packet\n");
    }
}

void
FileReaderALPPacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    int32_t fError, int32_t fDiscontinuity,
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3ALPProcessALPPacket(pUserData->hATSC3ALP, plpId, pPacketTimeUTC,
        packetNumber, fError, fDiscontinuity, pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process ALP packet\n");
    }
}

void
FileReaderBasebandPacketCallback(uint32_t plpId, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t fError, int32_t fDiscontinuity,
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3BasebandProcessBasebandPacket(pUserData->hATSC3Baseband,
        plpId, pPacketTimeUTC, packetNumber, fError, fDiscontinuity, 
        pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process baseband packet\n");
    }
}

int32_t
parseArgs(int argc, char **argv, char **ppInFilename, 
    ENATSC3FileType *pInFileType, char **ppOutPath, 
    int32_t *pTOffsetSeconds, uint64_t *pPLPIds,
    uint32_t *pDestinationAddrs, int32_t *pNDestinationAddrs, 
    int32_t nMaxDestinationAddrs, uint16_t *pDestinationPorts, 
    int32_t *pNDestinationPorts, int32_t nMaxDestinationPorts, 
    ENVerbosity *pVerbosity, int32_t *pFBitrateStatistics, 
    int32_t *pFIgnoreChecksumErrors, int32_t *pFNoFileNums)
{
    int32_t err = 0;
    int32_t i;

    *ppInFilename = NULL;
    *ppOutPath = NULL;
    *pPLPIds = 0;
    *pNDestinationPorts = 0;
    *pNDestinationAddrs = 0;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-b"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_ALP;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-bb"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_Baseband;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-j288alp"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_J288ALP;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-pcapipv4"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_PCAPIPV4;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutPath = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-plp"))
        {
            if (i + 1 < argc)
            {
                *pPLPIds |= 1ULL << atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-port"))
        {
            if (i + 1 < argc)
            {
                if (*pNDestinationPorts == nMaxDestinationPorts)
                {
                    fprintf(stderr, "Too many destination ports\n");
                    err = 1;
                }
                else
                {
                    pDestinationPorts[*pNDestinationPorts] = atol(argv[i + 1]);
                    (*pNDestinationPorts)++;
                }
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-addr"))
        {
            if (i + 1 < argc)
            {
                struct in_addr addr;
                if (inet_aton(argv[i + 1], &addr))
                {
                    if (*pNDestinationAddrs == nMaxDestinationAddrs)
                    {
                        fprintf(stderr, "Too many destination addrs\n");
                        err = 1;
                    }
                    else
                    {
                        pDestinationAddrs[*pNDestinationAddrs] = 
                            htonl(addr.s_addr);
                        (*pNDestinationAddrs)++;
                    }
                }
                else err = 1;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-v"))
        {
            if (i + 1 < argc)
            {
                *pVerbosity = atol(argv[i + 1 ]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-offset"))
        {
            if (i + 1 < argc)
            {
                *pTOffsetSeconds = atol(argv[i + 1 ]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-ignorechecksum"))
        {
            *pFIgnoreChecksumErrors = 1;
        }
        else if (!strcmp(argv[i], "-bitratestats"))
        {
            *pFBitrateStatistics = 1;
        }
        else if (!strcmp(argv[i], "-nofilenums"))
        {
            *pFNoFileNums = 1;
        }
        else err = 1;
    }

    if (*ppInFilename == NULL || *ppOutPath == NULL || 
        *pNDestinationPorts == 0 || *pNDestinationAddrs == 0)
    {
        err = 1;
    }

    if (err)
    {
        fprintf(stderr, "Usage: %s -b alpcapturefile | -bb basebandcapturefile | -j288alp j288alpcapturefile -o path -plp plpid [-plp plpid ...] -addr sls_multicast_address -port sls_multicast_port [-addr additional_multicast_addr] [-port additional_multicast_port] [-offset seconds] [-v level] [-ignorechecksum] [-bitratestats] [-nofilenums]\n", argv[0]);
        err = 1;
    }

    return err;
}

void
sighandler(int signum)
{
    if (signum == SIGINT)
    {
        fRunningG = 0;
    }
}

int
main(int argc, char **argv)
{
    int32_t fEnLogInitialized = 0;
    int32_t fXML2Initialized = 0;
    int32_t fROUTEObjectCallbacksRegistered = 0;
    int32_t fROUTEMPDAvailableCallbacksRegistered = 0;
    int32_t fROUTELCTChannelCallbacksRegistered = 0;
    int32_t fROUTEHELDAvailableCallbacksRegistered = 0;
    int32_t fROUTESLSAvailableCallbacksRegistered = 0;
    char *pInFilename;
    ENATSC3FileType inFileType;
    struct UserData userdata = 
    {
        .verbosity = ENVerbosity_Terse,
        .fBitrateStatistics = 0,
        .ifd = -1,
        .fIgnoreChecksumErrors = 0,
        .hATSC3IPV4 = NULL,
        .hROUTESession = NULL,
        .pOutPath = NULL,
        .pIPFilterSet = NULL,
        .fileNum = 0,
        .fNoFileNums = 0,
    };
    struct sigaction act;
    int32_t tOffsetSeconds = 0;
    int rval = 0;

    // Graceful exit on ctrl-c
    act.sa_handler = sighandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    sigaction(SIGINT, &act, NULL);

    if (EnLogInitialize(1))
    {
        fprintf(stderr, "Failed to initialize EnLog\n");
        rval = -1;
        goto out;
    }
    fEnLogInitialized = 1;

    xmlInitParser();
    fXML2Initialized = 1;

    if (parseArgs(argc, argv, &pInFilename, &inFileType, &userdata.pOutPath, 
        &tOffsetSeconds, &userdata.plpIds, 
        userdata.vDestinationAddresses, &userdata.nDestinationAddresses, 
        numof(userdata.vDestinationAddresses),
        userdata.vDestinationPorts, &userdata.nDestinationPorts, 
        numof(userdata.vDestinationPorts), &userdata.verbosity, 
        &userdata.fBitrateStatistics, &userdata.fIgnoreChecksumErrors,
        &userdata.fNoFileNums))
    {
        rval = 1;
        goto out;
    }

    ENATSC3BitrateStatisticsInitialize(&userdata.stats, 2000);
    snprintf(userdata.stats.strName, sizeof(userdata.stats.strName), 
        "UDP");
    userdata.stats.strName[sizeof(userdata.stats.strName) - 1] = 0;

    if (EnIPFilterSetAllocate(64, &userdata.pIPFilterSet))
    {
        fprintf(stderr, "Failed to allocate IP filter set\n");
        rval = 1;
        goto out;
    }

    userdata.ifd = open(pInFilename, O_RDONLY);
    if (userdata.ifd < 0)
    {
        fprintf(stderr, "Failed to open %s for reading: %s\n", 
            pInFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (inFileType == ENATSC3FileType_Baseband)
    {
        if (ENATSC3BasebandOpen(&userdata.hATSC3Baseband))
        {
            fprintf(stderr, "Failed to open baseband processor\n");
            rval = 1;
            goto out;
        }

        ENATSC3BasebandSetBitrateStatisticsEnabled(userdata.hATSC3Baseband,
            userdata.fBitrateStatistics);

        if (ENATSC3BasebandRegisterCallbacks(userdata.hATSC3Baseband,
            ALPPacketCallback, &userdata))
        {
            fprintf(stderr, "Failed to register baseband callbacks\n");
            rval = 1;
            goto out;
        }
    }

    if (inFileType != ENATSC3FileType_PCAPIPV4)
    {
        if (ENATSC3ALPOpen(&userdata.hATSC3ALP))
        {
            fprintf(stderr, "Failed to open ALP processor\n");
            rval = 1;
            goto out;
        }

        ENATSC3ALPSetBitrateStatisticsEnabled(userdata.hATSC3ALP, 
            userdata.fBitrateStatistics);

        if (ENATSC3ALPRegisterCallbacks(userdata.hATSC3ALP, IPV4PacketCallback, 
            LinkLayerSignalingPacketCallback, &userdata))
        {
            fprintf(stderr, "Failed to register callbacks\n");
            rval = 1;
            goto out;
        }
    }

    if (userdata.hATSC3Baseband)
    {
        if (ENATSC3BasebandSetPLPs(userdata.hATSC3Baseband, 
            userdata.plpIds))
        {
            fprintf(stderr, "Failed to set baseband PLP mask %"PRIx64"\n",
                userdata.plpIds);
            rval = 1;
            goto out;
        }
    }

    if (userdata.hATSC3ALP)
    {
        if (ENATSC3ALPSetPLPs(userdata.hATSC3ALP, userdata.plpIds))
        {
            fprintf(stderr, "Failed to set ALP PLP mask %"PRIx64"\n",
                userdata.plpIds);
            rval = 1;
            goto out;
        }
    }

    if (ENATSC3IPV4Open(&userdata.hATSC3IPV4))
    {
        fprintf(stderr, "Failed to open ATSC3IPV4 session\n");
        rval = 1;
        goto out;
    }

    ENATSC3IPV4SetVerbosity(userdata.hATSC3IPV4, userdata.verbosity);

    if (ENATSC3IPV4RegisterCallbacks(userdata.hATSC3IPV4, 
        TransportPacketCallback, &userdata))
    {
        fprintf(stderr, "Failed to register ATSC3IPV4 callback\n");
        rval = 1;
        goto out;
    }

    if (ENATSC3ROUTESessionAllocate(ENATSC3ROUTESessionType_ATSC3_DASH, 
        userdata.vDestinationAddresses[0],
        userdata.vDestinationPorts[0], AddIPAddressCallback, 
        RemoveIPAddressCallback, CommitIPAddressesCallback, &userdata, 
        "test", &userdata.hROUTESession))
    {
        fprintf(stderr, "Failed to allocate ROUTE session\n");
        rval = 1;
        goto out;
    }

    ENATSC3ROUTESessionSetVerbosity(userdata.hROUTESession, userdata.verbosity);
    ENATSC3ROUTESessionSetBitrateStatisticsEnabled(userdata.hROUTESession,
        userdata.fBitrateStatistics);

    if (ENATSC3ROUTESessionRegisterLCTChannelCallbacks(userdata.hROUTESession,
        LCTChannelAddedCallback, LCTChannelRemovedCallback, &userdata))
    {
        fprintf(stderr, "Failed to register LCT Channel callbacks\n");
        rval = 1;
        goto out;
    }
    fROUTELCTChannelCallbacksRegistered = 1;

    if (ENATSC3ROUTESessionRegisterObjectCallbacks(userdata.hROUTESession, 
        ObjectDataCallback, SessionResetCallback, &userdata))
    {
        fprintf(stderr, "Failed to register ROUTE object callbacks\n");
        rval = 1;
        goto out;
    }
    fROUTEObjectCallbacksRegistered = 1;

    if (ENATSC3ROUTESessionRegisterMPDAvailableCallback(userdata.hROUTESession, 
        MPDAvailableCallback, &userdata))
    {
        fprintf(stderr, "Failed to register ROUTE MPD callback\n");
        rval = 1;
        goto out;
    }
    fROUTEMPDAvailableCallbacksRegistered = 1;

    if (ENATSC3ROUTESessionRegisterHELDAvailableCallback(userdata.hROUTESession, 
        HELDAvailableCallback, &userdata))
    {
        fprintf(stderr, "Failed to register ROUTE HELD callback\n");
        rval = 1;
        goto out;
    }
    fROUTEHELDAvailableCallbacksRegistered = 1;

    if (ENATSC3ROUTESessionRegisterSLSAvailableCallback(userdata.hROUTESession, 
        SLSAvailableCallback, &userdata))
    {
        fprintf(stderr, "Failed to register ROUTE SLS callback\n");
        rval = 1;
        goto out;
    }
    fROUTESLSAvailableCallbacksRegistered = 1;

    switch (inFileType)
    {
        case ENATSC3FileType_ALP:
            rval = ENATSC3FileReaderReadALPFile(userdata.ifd, tOffsetSeconds, 
                FileReaderALPPacketCallback, NULL, &fRunningG, &userdata);
            break;
        case ENATSC3FileType_Baseband:
            rval = ENATSC3FileReaderReadBasebandFile(userdata.ifd, 
                tOffsetSeconds, FileReaderBasebandPacketCallback, NULL, 
                &fRunningG, &userdata);
            break;
        case ENATSC3FileType_J288ALP:
            rval = ENATSC3FileReaderReadJ288ALPFile(userdata.ifd, 
                tOffsetSeconds, FileReaderALPPacketCallback, NULL, &fRunningG, 
                &userdata);
            break;
        case ENATSC3FileType_PCAPIPV4:
            rval = ENATSC3FileReaderReadPCAPIPV4File(userdata.ifd, 
                tOffsetSeconds, FileReaderIPV4PacketCallback, NULL, &fRunningG, 
                &userdata);
            break;
        default:
            fprintf(stderr, "Unhandled input file type %d\n", inFileType);
            rval = 1;
            goto out;
            break;
    }

    if (userdata.hATSC3ALP)
    {
        if (ENATSC3ALPSetPLPs(userdata.hATSC3ALP, 0))
        {
            fprintf(stderr, "Failed to clear ALP PLP mask\n");
            rval = 1;
            goto out;
        }
    }

    if (userdata.hATSC3Baseband)
    {
        if (ENATSC3BasebandSetPLPs(userdata.hATSC3Baseband, 0)) 
        {
            fprintf(stderr, "Failed to clear baseband PLPs\n");
            rval = 1;
            goto out;
        }
    }
    
out:
    if (fROUTESLSAvailableCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterSLSAvailableCallback(
            userdata.hROUTESession, SLSAvailableCallback, &userdata))
        {
            fprintf(stderr, "Failed to unregister ROUTE SLS callback\n");
        }
    }

    if (fROUTEHELDAvailableCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterHELDAvailableCallback(
            userdata.hROUTESession, HELDAvailableCallback, &userdata))
        {
            fprintf(stderr, "Failed to unregister ROUTE HELD callback\n");
        }
    }

    if (fROUTEMPDAvailableCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterMPDAvailableCallback(
            userdata.hROUTESession, MPDAvailableCallback, &userdata))
        {
            fprintf(stderr, "Failed to unregister ROUTE MPD callback\n");
        }
    }

    if (fROUTELCTChannelCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterLCTChannelCallbacks(
            userdata.hROUTESession, LCTChannelAddedCallback, 
            LCTChannelRemovedCallback, &userdata))
        {
            fprintf(stderr, "Failed to unregister LCT Channel callbacks\n");
        }
    }

    if (fROUTEObjectCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterObjectCallbacks(userdata.hROUTESession, 
            ObjectDataCallback, SessionResetCallback, &userdata))
        {
            fprintf(stderr, 
                "Failed to unregister ROUTE media file callbacks\n");
        }
    }

    if (userdata.hROUTESession)
    {
        if (ENATSC3ROUTESessionFree(userdata.hROUTESession))
        {
            fprintf(stderr, "Failed to free ROUTE session\n");
        }
    }

    if (fXML2Initialized)
    {
        xmlCleanupCharEncodingHandlers();
        xmlDictCleanup();
        xmlCleanupParser();
    }

    if (userdata.hATSC3IPV4)
    {
        if (ENATSC3IPV4Close(userdata.hATSC3IPV4))
        {
            fprintf(stderr, "Failed to close ATSC3IPV4 session\n");
        }
    }

    if (userdata.hATSC3ALP)
    {
        ENATSC3ALPClose(userdata.hATSC3ALP);
    }

    if (userdata.hATSC3Baseband)
    {
        ENATSC3BasebandClose(userdata.hATSC3Baseband);
    }

    if (userdata.pIPFilterSet)
    {
        if (EnIPFilterSetFree(userdata.pIPFilterSet))
        {
            fprintf(stderr, "Failed to free IP filter set\n");
        }
    }

    if (userdata.ifd >= 0)
    {
        close(userdata.ifd);
    }

    if (fEnLogInitialized)
    {
        if (EnLogShutdown())
        {
            fprintf(stderr, "Failed to shut down EnLog\n");
        }
    }

    return rval;
}
