/*
 * Copyright © 2018-2021 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <assert.h>

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3baseband/libenatsc3baseband.h>
#include <libenatsc3alp/libenatsc3alp.h>
#include <libenatsc3ipv4/libenatsc3ipv4.h>
#include <libenatsc3ipv4/internetchecksum.h>

#include <libenatsc3filereader/libenatsc3filereader.h>

#include <enatsc3cap/enudpcap.h>

struct UserData
{
    int ifd;
    int ofd;
    ENVerbosity verbosity;

    int32_t fIgnoreChecksumErrors;

    ENATSC3BasebandHandle_t hATSC3Baseband;
    ENATSC3ALPHandle_t hATSC3ALP;
    ENATSC3IPV4Handle_t hATSC3IPV4;

    uint64_t plpIds;
};

int32_t fRunningG = 1;

int32_t
HandleUDPPacket(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    uint32_t sourceAddress, uint32_t destinationAddress, 
    const uint8_t *pPacket, int32_t packetLength, struct UserData *pUserData)
{
    int32_t sourcePort;
    int32_t destPort;
    int32_t length;
    int32_t checksum;
    ENInternetChecksumContext iccontext;
    uint32_t sum;
    int32_t rval = 0;

    if (packetLength < 8)
    {
        fprintf(stderr, "UDP too small, dropping\n");
        rval = 1;
        goto out;
    }

    length = pPacket[4] << 8 | pPacket[5];

    if (packetLength < length)
    {
        fprintf(stderr, "Not enough data for UDP packet: %d < %d\n",
            packetLength, length);
        rval = 1;
        goto out;
    }

    sourcePort = pPacket[0] << 8 | pPacket[1];
    destPort = pPacket[2] << 8 | pPacket[3];
    checksum = pPacket[6] << 8 | pPacket[7];

    if (checksum)
    {
        struct
        {
            uint32_t sourceAddress;
            uint32_t destinationAddress;
            uint8_t zero;
            uint8_t protocol;
            uint8_t lengthHi;
            uint8_t lengthLo;
        } __attribute__((packed)) pseudoHeader =
        {
            .sourceAddress = htonl(sourceAddress),
            .destinationAddress = htonl(destinationAddress),
            .zero = 0,
            .protocol = 17,
            .lengthHi = packetLength >> 8,
            .lengthLo = packetLength & 0xff,
        };

        if (pUserData->verbosity)
        {
            printf("UDP\n");
            printf("  Source port: %d\n", sourcePort);
            printf("  Dest port: %d\n", destPort);
            printf("  Length: %d\n", length);
            printf("  Checksum: %x\n", checksum);
        }

        ENInternetChecksumInitialize(&iccontext);
        ENInternetChecksumCalc((uint8_t *)&pseudoHeader, sizeof(pseudoHeader), 
            &iccontext);
        ENInternetChecksumCalc(pPacket, packetLength, &iccontext);
        sum = ENInternetChecksumFinalize(&iccontext);
    }
    else
    {
        sum = 0;
    }

    if (sum == 0)
    {
        if (pUserData->verbosity)
        {
            printf("  Checksum verified\n");
        }
    }
    else
    {
        printf("  UDP CHECKSUM ERROR: %x\n", sum);
    }

    if (sum == 0 || pUserData->fIgnoreChecksumErrors)
    {
        EntangleCapHeader echeader;

        echeader.ipv4addr = htonl(destinationAddress);
        echeader.port = htons(destPort);
        echeader.length = htons(length - 8);
        echeader.sec = ENUtils_htonll(pPacketTimeUTC->tv_sec);
        echeader.nsec = htonl(pPacketTimeUTC->tv_nsec);
        if (write(pUserData->ofd, &echeader, sizeof(echeader)) !=
            sizeof(echeader))
        {
            fprintf(stderr, "Failed to write header: %s\n", strerror(errno));
            goto out;
        }

        // We don't write out the UDP header
        write(pUserData->ofd, pPacket + 8, packetLength - 8);
    }

out:
    return rval;
}

void
TransportPacketCallback(int32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint32_t sourceAddress, uint32_t destinationAddress, 
    uint16_t protocol, const uint8_t *pPacket, int32_t sPacket, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;
    switch (protocol)
    {
        case 17:
            HandleUDPPacket(pPacketTimeUTC, packetNumber, sourceAddress, 
                destinationAddress, pPacket, sPacket, pUserData);
            break;
        default:
            fprintf(stderr, "Unhandled protocol %d\n", protocol);
            break;
    }
}

void
IPV4PacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPayload, int32_t payloadLength, int32_t SID, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;

    ENATSC3IPV4ProcessIPV4Packet(pUserData->hATSC3IPV4, plpId, pPacketTimeUTC, 
        packetNumber, pPayload, payloadLength);

    return;
}

void
ALPPacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC,
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3ALPProcessALPPacket(pUserData->hATSC3ALP, plpId,
        pPacketTimeUTC, packetNumber, fError, fDiscontinuity, pPacket, 
        sPacket))
    {
        fprintf(stderr, "Failed to process ALP packet\n");
    }
}

void
FileReaderALPPacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    int32_t fError, int32_t fDiscontinuity,
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3ALPProcessALPPacket(pUserData->hATSC3ALP, plpId, pPacketTimeUTC,
        packetNumber, fError, fDiscontinuity, pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process ALP packet\n");
    }

}

void
FileReaderBasebandPacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    int32_t fError, int32_t fDiscontinuity,
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3BasebandProcessBasebandPacket(pUserData->hATSC3Baseband,
        plpId, pPacketTimeUTC, packetNumber, fError, fDiscontinuity, 
        pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process baseband packet\n");
    }
}

int32_t
parseArgs(int argc, char **argv, char **ppInFilename,  
    ENATSC3FileType *pInFileType, char **ppOutFilename, 
    int32_t *pTOffsetSeconds, uint64_t *pPLPMask, 
    ENVerbosity *pVerbosity, int32_t *pFIgnoreChecksumErrors)
{
    int32_t err = 0;
    int32_t i;

    *ppInFilename = NULL;
    *ppOutFilename = NULL;
    *pPLPMask = 0;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-b"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_ALP;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-bb"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_Baseband;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-j288alp"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_J288ALP;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutFilename = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-plp"))
        {
            if (i + 1 < argc)
            {
                *pPLPMask |= 1ULL << atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-v"))
        {
            if (i + 1 < argc)
            {
                *pVerbosity = atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-offset"))
        {
            if (i + 1 < argc)
            {
                *pTOffsetSeconds = atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-ignorechecksum"))
        {
            *pFIgnoreChecksumErrors = 1;
        }
        else err = 1;
    }

    if (*ppInFilename == NULL || *ppOutFilename == NULL)
    {
        err = 1;
    }

    if (err)
    {
        fprintf(stderr, "Usage: %s -b alpcapturefile -o udpfile -plp plpid [-plp plpid ...] [-offset seconds] [-v] [-ignorechecksum]\n", argv[0]);
        err = 1;
    }

    return err;
}

void
sighandler(int signum)
{
    if (signum == SIGINT)
    {
        fRunningG = 0;
    }
}

int
main(int argc, char **argv)
{
    int32_t fEnLogInitialized = 0;
    char *pInFilename;
    ENATSC3FileType inFileType;
    char *pOutFilename;
    struct UserData userdata = 
    {
        .verbosity = 0,
        .fIgnoreChecksumErrors = 0,
        .hATSC3ALP = NULL,
        .hATSC3IPV4 = NULL,
        .ifd = -1,
        .ofd = -1
    };
    int32_t tOffsetSeconds = 0;
    int rval = 0;

    if (EnLogInitialize(1))
    {
        fprintf(stderr, "Failed to initialize EnLog\n");
        rval = -1;
        goto out;
    }
    fEnLogInitialized = 1;

    if (parseArgs(argc, argv, &pInFilename,&inFileType, &pOutFilename, 
        &tOffsetSeconds, &userdata.plpIds, &userdata.verbosity, 
        &userdata.fIgnoreChecksumErrors))
    {
        rval = 1;
        goto out;
    }

    userdata.ifd = open(pInFilename, O_RDONLY);
    if (userdata.ifd < 0)
    {
        fprintf(stderr, "Failed to open %s for reading: %s\n", 
            pInFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    userdata.ofd = open(pOutFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (userdata.ofd < 0)
    {
        fprintf(stderr, "Failed to open %s for writing: %s\n",
            pOutFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (inFileType == ENATSC3FileType_Baseband)
    {
        if (ENATSC3BasebandOpen(&userdata.hATSC3Baseband))
        {
            fprintf(stderr, "Failed to open baseband processor\n");
            rval = 1;
            goto out;
        }

        if (ENATSC3BasebandRegisterCallbacks(userdata.hATSC3Baseband,
            ALPPacketCallback, &userdata))
        {
            fprintf(stderr, "Failed to register baseband callbacks\n");
            rval = 1;
            goto out;
        }
    }

    if (ENATSC3ALPOpen(&userdata.hATSC3ALP))
    {
        fprintf(stderr, "Failed to open ALP processor\n");
        rval = 1;
        goto out;
    }

    if (ENATSC3ALPRegisterCallbacks(userdata.hATSC3ALP, IPV4PacketCallback, 
        NULL, &userdata))
    {
        fprintf(stderr, "Failed to register callbacks\n");
        rval = 1;
        goto out;
    }

    if (userdata.hATSC3Baseband)
    {
        if (ENATSC3BasebandSetPLPs(userdata.hATSC3Baseband,
            userdata.plpIds))
        {
            fprintf(stderr, "Failed to set baseband PLP mask %"PRIx64"\n",
                userdata.plpIds);
            rval = 1;
            goto out;
        }
    }

    if (ENATSC3ALPSetPLPs(userdata.hATSC3ALP, userdata.plpIds))
    {
        fprintf(stderr, "Failed to set ALP PLP mask %"PRIx64"\n", 
            userdata.plpIds);
        rval = 1;
        goto out;
    }

    if (ENATSC3IPV4Open(&userdata.hATSC3IPV4))
    {
        fprintf(stderr, "Failed to open ATSC3IPV4 session\n");
        rval = 1;
        goto out;
    }

    if (userdata.verbosity)
    {
        ENATSC3IPV4SetVerbosity(userdata.hATSC3IPV4, userdata.verbosity);
    }

    if (ENATSC3IPV4RegisterCallbacks(userdata.hATSC3IPV4,
        TransportPacketCallback, &userdata))
    {
        fprintf(stderr, "Failed to register ATSC3IPV4 callback\n");
        rval = 1;
        goto out;
    }

    switch (inFileType)
    {
        case ENATSC3FileType_ALP:
            rval = ENATSC3FileReaderReadALPFile(userdata.ifd, tOffsetSeconds, 
                FileReaderALPPacketCallback, NULL, &fRunningG, &userdata);
            break;
        case ENATSC3FileType_Baseband:
            rval = ENATSC3FileReaderReadBasebandFile(userdata.ifd, 
                tOffsetSeconds, FileReaderBasebandPacketCallback, NULL, 
                &fRunningG, &userdata);
            break;
        case ENATSC3FileType_J288ALP:
            rval = ENATSC3FileReaderReadJ288ALPFile(userdata.ifd, 
                tOffsetSeconds, FileReaderALPPacketCallback, NULL, &fRunningG, 
                &userdata);
        default:
            fprintf(stderr, "Unhandled input file type %d\n", inFileType);
            rval = 1;
            goto out;
            break;
    }

    if (ENATSC3ALPSetPLPs(userdata.hATSC3ALP, 0))
    {
        fprintf(stderr, "Failed to remove PLPs from ALP\n"); 
        rval = 1;
        goto out;
    }

    if (userdata.hATSC3Baseband)
    {
        if (ENATSC3BasebandSetPLPs(userdata.hATSC3Baseband, 0))
        {
            fprintf(stderr, "Failed to remove PLPs from baseband\n");
            rval = 1;
            goto out;
        }
    }
    
out:
    if (userdata.hATSC3IPV4)
    {
        if (ENATSC3IPV4Close(userdata.hATSC3IPV4))
        {
            fprintf(stderr, "Failed to close ATSC3IPV4 session\n");
        }
    }

    if (userdata.hATSC3ALP)
    {
        ENATSC3ALPClose(userdata.hATSC3ALP);
    }

    if (userdata.hATSC3Baseband)
    {
        ENATSC3BasebandClose(userdata.hATSC3Baseband);
    }

    if (userdata.ofd >= 0)
    {
        close(userdata.ofd);
    }

    if (userdata.ifd >= 0)
    {
        close(userdata.ifd);
    }

    if (fEnLogInitialized)
    {
        if (EnLogShutdown())
        {
            fprintf(stderr, "Failed to shut down EnLog\n");
        }
    }

    return rval;
}
