/*
 * Copyright © 2018-2021 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <assert.h>

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3ipv4/libenatsc3ipv4.h>
#include <libenatsc3ipv4/internetchecksum.h>

#include <enatsc3cap/enudpcap.h>

#define MAX_IPV4_PACKET_SIZE 65536

struct UserData
{
    ENVerbosity verbosity;
    int32_t fIgnoreChecksumErrors;
    int ifd;
    int ofd;
};

int32_t 
HandleUDPPacket(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    uint32_t sourceAddress, uint32_t destinationAddress, 
    const uint8_t *pPacket, int32_t packetLength, struct UserData *pUserData)
{
    int32_t sourcePort;
    int32_t destPort;
    int32_t length;
    int32_t checksum;
    ENInternetChecksumContext iccontext;
    uint32_t sum;
    int32_t rval = 0;

    if (packetLength < 8)
    {
        fprintf(stderr, "UDP too small, dropping\n");
        rval = 1;
        goto out;
    }

    length = pPacket[4] << 8 | pPacket[5];

    if (packetLength < length)
    {
        fprintf(stderr, "Not enough data for UDP packet: %d < %d\n", 
            packetLength, length);
        rval = 1;
        goto out;
    }

    sourcePort = pPacket[0] << 8 | pPacket[1];
    destPort = pPacket[2] << 8 | pPacket[3];
    checksum = pPacket[6] << 8 | pPacket[7];

    if (pUserData->verbosity >= ENVerbosity_PacketTrace)
    {
        printf("UDP\n");
        printf("  Source port: %d\n", sourcePort);
        printf("  Dest port: %d\n", destPort);
        printf("  Length: %d\n", length);
        printf("  Checksum: %x\n", checksum);
    }

    if (checksum)
    {
       struct
        {
            uint32_t sourceAddress;
            uint32_t destinationAddress;
            uint8_t zero;
            uint8_t protocol;
            uint8_t lengthHi;
            uint8_t lengthLo;
        } __attribute__((packed)) pseudoHeader =
        {
            .sourceAddress = htonl(sourceAddress),
            .destinationAddress = htonl(destinationAddress),
            .zero = 0,
            .protocol = 17,
            .lengthHi = packetLength >> 8,
            .lengthLo = packetLength & 0xff,
        };

        ENInternetChecksumInitialize(&iccontext);
        ENInternetChecksumCalc((uint8_t *)&pseudoHeader, sizeof(pseudoHeader), 
            &iccontext);
        ENInternetChecksumCalc(pPacket, packetLength, &iccontext);
        sum = ENInternetChecksumFinalize(&iccontext);

        if (sum == 0)
        {
            if (pUserData->verbosity >= ENVerbosity_PacketTrace)
            {
                printf("  Checksum verified\n");
            }
        }
        else
        {
            printf("  UDP CHECKSUM ERROR: %x\n", sum);
        }
    }
    else
    {
        if (pUserData->verbosity >= ENVerbosity_PacketTrace)
        {
            printf("  No checksum for packet src %d.%d.%d.%d:%d dest %d.%d.%d.%d:%d\n",
                (htonl(sourceAddress) >> 24) & 0xff,
                (htonl(sourceAddress) >> 16) & 0xff,
                (htonl(sourceAddress) >> 8) & 0xff,
                (htonl(sourceAddress) >> 0) & 0xff,
                sourcePort,
                (htonl(destinationAddress) >> 24) & 0xff,
                (htonl(destinationAddress) >> 16) & 0xff,
                (htonl(destinationAddress) >> 8) & 0xff,
                (htonl(destinationAddress) >> 0) & 0xff,
                destPort);
        }
        sum = 0;
    }

    if (sum == 0 || pUserData->fIgnoreChecksumErrors) 
    {
        EntangleCapHeader echeader;

        echeader.ipv4addr = htonl(destinationAddress);
        echeader.port = htons(destPort);
        echeader.length = htons(length - 8);
        echeader.sec = ENUtils_htonll(pPacketTimeUTC->tv_sec);
        echeader.nsec = htonl(pPacketTimeUTC->tv_nsec);
        if (write(pUserData->ofd, &echeader, sizeof(echeader)) != 
            sizeof(echeader))
        {
            fprintf(stderr, "Failed to write header: %s\n", strerror(errno));
            goto out;
        }

        // We don't write out the UDP header 
        write(pUserData->ofd, pPacket + 8, packetLength - 8);
    }

out:
    return rval;
}

void
TransportPacketCallback(int32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint32_t sourceAddress, uint32_t destinationAddress, 
    uint16_t protocol, const uint8_t *pPacket, int32_t sPacket, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;
    switch (protocol)
    {
        case 17:
            HandleUDPPacket(pPacketTimeUTC, packetNumber, sourceAddress, 
                destinationAddress, pPacket, sPacket, pUserData);
            break;
        default:
            fprintf(stderr, "Unhandled protocol %d\n", protocol);
            break;
    }
}

int32_t
parseArgs(int argc, char **argv, char **ppInFilename, char **ppOutFilename,
    ENVerbosity *pVerbosity, int32_t *pFIgnoreChecksumErrors)
{
    int32_t err = 0;
    int32_t i;

    *ppInFilename = NULL;
    *ppOutFilename = NULL;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-b"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutFilename = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-v"))
        {
            if (i + 1 < argc)
            {
                *pVerbosity = atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-ignorechecksum"))
        {
            *pFIgnoreChecksumErrors = 1;
        }
        else err = 1;
    }

    if (*ppInFilename == NULL || *ppOutFilename == NULL)
    {
        err = 1;
    }

    if (err)
    {
        fprintf(stderr, "Usage: %s -b ipv4capturefile -o udpfile [-v verbosity] [-ignorechecksum]\n", argv[0]);
        err = 1;
    }

    return err;
}

int
main(int argc, char **argv)
{
    int32_t fEnLogInitialized = 0;
    char *pInFilename;
    char *pOutFilename;
    uint8_t *pPacket = NULL;
    uint64_t packetNumber = 0;
    struct UserData userdata = 
    {
        .verbosity = ENVerbosity_Terse,
        .fIgnoreChecksumErrors = 0,
        .ifd = -1,
        .ofd = -1,
    };
    ENATSC3IPV4Handle_t hATSC3IPV4 = NULL;
    int rval = 0;

    if (EnLogInitialize(1))
    {
        fprintf(stderr, "Failed to initialize EnLog\n");
        rval = -1;
        goto out;
    }
    fEnLogInitialized = 1;

    if (parseArgs(argc, argv, &pInFilename, &pOutFilename, &userdata.verbosity,
        &userdata.fIgnoreChecksumErrors))
    {
        rval = 1;
        goto out;
    }

    userdata.ifd = open(pInFilename, O_RDONLY);
    if (userdata.ifd < 0)
    {
        fprintf(stderr, "Failed to open %s for reading: %s\n", 
            pInFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    userdata.ofd = open(pOutFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (userdata.ofd < 0)
    {
        fprintf(stderr, "Failed to open %s for writing: %s\n", 
            pInFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    pPacket = (uint8_t *)malloc(MAX_IPV4_PACKET_SIZE);
    if (!pPacket)
    {
        fprintf(stderr, "Out of memory\n");
        rval = 1;
        goto out;
    }

    if (ENATSC3IPV4Open(&hATSC3IPV4))
    {
        fprintf(stderr, "Failed to open ATSC3IPV4 session\n");
        rval = 1;
        goto out;
    }

    ENATSC3IPV4SetVerbosity(hATSC3IPV4, userdata.verbosity);

    if (ENATSC3IPV4RegisterCallbacks(hATSC3IPV4, TransportPacketCallback, 
        &userdata))
    {
        fprintf(stderr, "Failed to regieter ATSC3IPV4 callback\n");
        rval = 1;
        goto out;
    }

    while (1)
    {
        EntangleCapHeader echeader;
        struct timespec packetTimeUTC;
        int nb;

        nb = read(userdata.ifd, &echeader, sizeof(echeader));
        if (!nb)
        {
            // EOF
            break;
        }
        else if (nb != sizeof(echeader))
        {
            fprintf(stderr, "Partial read of ecap header\n");
            rval = 1;
            break;
        }

        echeader.length = ntohs(echeader.length);
        packetTimeUTC.tv_sec = ENUtils_ntohll(echeader.sec);
        packetTimeUTC.tv_nsec = ntohl(echeader.nsec);

        if (echeader.length > MAX_IPV4_PACKET_SIZE)
        {
            fprintf(stderr, "IPv4 packet too large: %d > %d\n",
                echeader.length, MAX_IPV4_PACKET_SIZE);
            rval = 1;
            break;
        }

        nb = read(userdata.ifd, pPacket, echeader.length);
        if (!nb)
        {
            fprintf(stderr, "EOF reading ipv4 packet\n");
            rval = 1;
            break;
        }
        else if (nb != echeader.length)
        {
            fprintf(stderr, "Partial read of ipv4 packet body\n");
            rval = 1;
            break;
        }

        ENATSC3IPV4ProcessIPV4Packet(hATSC3IPV4, 0 /* plpId */, &packetTimeUTC, 
            packetNumber, pPacket, echeader.length);

        packetNumber++;
    }

out:
    if (hATSC3IPV4)
    {
        if (ENATSC3IPV4Close(hATSC3IPV4))
        {
            fprintf(stderr, "Failed to close ATSC3IPV4 session\n");
        }
    }

    if (pPacket)
    {
        free(pPacket);
    }

    if (userdata.ifd >= 0)
    {
        close(userdata.ifd);
    }

    if (userdata.ofd >= 0)
    {
        close(userdata.ofd);
    }

    if (fEnLogInitialized)
    {
        if (EnLogShutdown())
        {
            fprintf(stderr, "Failed to shut down EnLog\n");
        }
    }

    return rval;
}
