/*
 * Copyright © 2018-2021 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <inttypes.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>

#include <libendebug/libendebug.h>
#include <libmultipartparser/multipart_parser.h>
#include <libxml2/libxml/xmlreader.h>
#include <enatsc3-common/enatsc3.h>
#include <libenatsc3utils/enxml.h>

#include <libenatsc3sls/atsc3sls_usbd.h>
#include <libenatsc3sls/atsc3sls_held.h>
#include <libenatsc3route/atsc3routesls_stsid.h>
#include <libenatsc3route/atsc3routesls_envelope.h>
#include <libenatsc3route/atsc3routesls_mime.h>

// This is a fairly loose check for the content-type. Strictly speaking there
// should be no spaces in the content-type value, but in practice white space
// tends to appear after and sometimes before the semicolon separating the
// content-type from its arguments.
#define IsContentType(String, Literal)                          \
  (!strncasecmp(String, Literal, sizeof(Literal) - 1) &&        \
   (String[sizeof(Literal) - 1] == 0 ||                         \
    String[sizeof(Literal) - 1] == ' ' ||                       \
    String[sizeof(Literal) - 1] == ';'))

int32_t 
DocumentAvailableCallback(const char *pDocument, int32_t sDocument, 
    const char *pPart, int32_t sPart, char *pContentType, 
    char *pContentLocation, void *pUserData)
{
    int fd = -1;
    int rval = 0;

    fd = open(pContentLocation, O_CREAT | O_TRUNC | O_WRONLY, 0666);
    if (fd < 0)
    {
        fprintf(stderr, "Failed to open %s for writing\n", pContentLocation);
    }
    else
    {
        if (write(fd, pDocument, sDocument) != sDocument) 
        {
            fprintf(stderr, "Failed to write %s: %s\n", pContentLocation, 
                strerror(errno));
        }
    }

    if (fd >= 0)
    {
        close(fd);
    }

    if (IsContentType(pContentType, "application/dash+xml"))
    {
        printf("dash mpd\n");
    }
    else if (IsContentType(pContentType, "application/s-tsid"))
    {
        ENATSC3ROUTESLSSTSIDNode stsidNode;

        printf("s-tsid\n");
        if (ENATSC3ROUTESLSSTSIDParse(pDocument, sDocument, &stsidNode))
        {
            fprintf(stderr, "Failed to parse s-tsid part\n");
        }
        else
        {
            if (ENATSC3ROUTESLSSTSIDFree(&stsidNode))
            {
                fprintf(stderr, "Failed to free parses stsid\n");
            }
        }
    }
    else if (IsContentType(pContentType, 
        "application/mbms-user-service-description+xml"))
    {
        ENATSC3SLSUSBDBundleDescriptionNode usbdNode;

        printf("usbd\n");
        if (ENATSC3SLSUSBDParse(pDocument, sDocument, &usbdNode))
        {
            fprintf(stderr, "Failed to parse usbd part\n");
        }
        else
        {
            if (ENATSC3SLSUSBDFree(&usbdNode))
            {
                fprintf(stderr, "Failed to free parsed usbd\n");
            }
        }
    }
    else if (IsContentType(pContentType, "application/atsc-held+xml"))
    {
        ENATSC3SLSHELDNode heldNode;

        printf("held\n");
        if (ENATSC3SLSHELDParse(pDocument, sDocument, &heldNode))
        {
            fprintf(stderr, "Failed to parse held part\n");
        }
        else
        {
            if (ENATSC3SLSHELDFree(&heldNode))
            {
                fprintf(stderr, "Failed to free parsed held\n");
            }
        }
    }
    else if (IsContentType(pContentType, "application/mbms-envelope+xml"))
    {
        ENATSC3ROUTESLSEnvelopeNode envelopeNode;

        printf("envelope\n");
        if (ENATSC3ROUTESLSEnvelopeParse(pDocument, sDocument, &envelopeNode))
        {
            fprintf(stderr, "Failed to parse envelope part\n");
        }
    }

    return rval;
}

int32_t
parseArgs(int argc, char **argv, char **ppInFilename)
{
    int32_t err = 0;
    int32_t i;

    *ppInFilename = NULL;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-b"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else err = 1;
    }

    if (!*ppInFilename)
    {
        err = 1;
    }

    if (err)
    {
        fprintf(stderr, "Usage: %s -b xmlfilename\n", argv[0]);
    }

    return err;
}

#define MAX_BODY_LEN (2 * 1024 * 1024)

int
main(int argc, char **argv)
{
    int32_t fEnLogInitialized = 0;
    int32_t fXML2Initialized = 0;
    int fd = -1;
    char *pBody = NULL;
    char *pXMLFilename;
    ENATSC3ROUTESLSMultipartMIMEHeaderInfo_t mimeHeaderInfo;
    const char *pParts;
    int32_t partsLen;
    const char *pBoundary = NULL;
    int32_t boundaryLen = 0;
    int32_t length;
    int32_t i;
    int rval = 0;

    if (parseArgs(argc, argv, &pXMLFilename))
    {
        rval = -1;
        goto out;
    }

    pBody = (char *)malloc(MAX_BODY_LEN);
    if (!pBody)
    {
        fprintf(stderr, "Out of memory\n");
        rval = 1;
        goto out;
    }

    if (EnLogInitialize(1))
    {
        fprintf(stderr, "Failed to initialize EnLog\n");
        rval = -1;
        goto out;
    }
    fEnLogInitialized = 1;

    xmlInitParser();
    fXML2Initialized = 1;

    fd = open(pXMLFilename, O_RDONLY);
    if (fd < 0)
    {
        fprintf(stderr, "Failed to open %s\n", pXMLFilename);
        rval = 1;
        goto out;
    }

    length = read(fd, pBody, MAX_BODY_LEN);
    if (length == MAX_BODY_LEN)
    {
        fprintf(stderr, "message too long\n");
        rval = 1;
        goto out;
    }

    if (ENATSC3ROUTESLSMultipartMIMEHeaderParse(pBody, length, &mimeHeaderInfo, 
        &pParts, &partsLen)) 
    {
        fprintf(stderr, "Failed to parse mime document header\n");
        rval = -1;
        goto out;
    }

    for (i = 0; i < mimeHeaderInfo.nParameters; i++)
    {
        if (mimeHeaderInfo.vParameters[i].nameLen == sizeof("boundary") -1 &&
            !strncmp(mimeHeaderInfo.vParameters[i].pName, "boundary",
                sizeof("boundary") - 1))
        {
            pBoundary = mimeHeaderInfo.vParameters[i].pValue;
            boundaryLen = mimeHeaderInfo.vParameters[i].valueLen;
            break;
        }
    }

    if (!pBoundary)
    {
        fprintf(stderr,  "Failed to parse ROUTE MIME document headers\n");
        rval = 1;
        goto out;
    }

    if (ENATSC3ROUTESLSMultipartMIMEPartsParse(pParts, partsLen, pBoundary, 
        boundaryLen, DocumentAvailableCallback, NULL))
    {
        fprintf(stderr, "Failed to parse mime document parts\n");
        rval = -1;
        goto out;
    }

out:
    if (fd >= 0)
    {
        close(fd);
    }

    if (fXML2Initialized)
    {
        xmlCleanupCharEncodingHandlers();
        xmlDictCleanup();
        xmlCleanupParser();
    }

    if (fEnLogInitialized)
    {
        if (EnLogShutdown())
        {
            fprintf(stderr, "Failed to shut down EnLog\n");
        }
    }

    if (pBody)
    {
        free(pBody);
    }

    return rval;
}
