/*
 * Copyright © 2018-2021 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>

#include <errno.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <arpa/inet.h>

#include <libxml2/libxml/xmlreader.h>

#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <libenatsc3utils/enxml.h>
#include <libenatsc3utils/enipfilterset.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3route/atsc3routesls_mime.h>
#include <libenatsc3route/atsc3routesls_stsid.h>
#include <libenatsc3route/libenatsc3route.h>

#include <enatsc3cap/enudpcap.h>

#define MAX_PAYLOAD_SIZE 65536

struct UserData
{
    ENVerbosity verbosity;
    int32_t fBitrateStatistics;
    int ifd;

    EnIPFilterSet *pIPFilterSet;

    ENATSC3ROUTESessionHandle_t hROUTESession;

    struct _LCTChannelInfo
    {
        int32_t fInUse;

        uint32_t tsi;

        int64_t currentTOI;
        uint8_t codepoint;
        char strCurrentURL[256];
        int64_t objectSize;
        int64_t pos;

        int fd;
    } vLCTChannelInfos[32];

    char *pOutPath;

    int32_t fileNum;
    int32_t fNoFileNums;
};

int32_t fRunningG = 1;

ENATSC3Status
FindLCTChannelForTSI(struct UserData *pUserData, uint32_t tsi, 
    struct _LCTChannelInfo **ppLCTChannelInfo)
{
    int32_t i;
    ENATSC3Status rval = ENATSC3Status_Failure;

    for (i = 0; i < numof(pUserData->vLCTChannelInfos); i++)
    {
        if (pUserData->vLCTChannelInfos[i].fInUse && 
            pUserData->vLCTChannelInfos[i].tsi == tsi)
        {
            *ppLCTChannelInfo = pUserData->vLCTChannelInfos + i;
            rval = ENATSC3Status_Success;
            break;
        }
    }

    return rval;
}

ENATSC3Status
AddLCTChannelForTSI(struct UserData *pUserData, uint32_t tsi, 
    struct _LCTChannelInfo **ppLCTChannelInfo)
{
    int32_t i;
    ENATSC3Status rval = ENATSC3Status_Failure;

    for (i = 0; i < numof(pUserData->vLCTChannelInfos); i++)
    {
        if (!pUserData->vLCTChannelInfos[i].fInUse)
        {
            pUserData->vLCTChannelInfos[i].fInUse = 1;
            pUserData->vLCTChannelInfos[i].tsi = tsi;
            pUserData->vLCTChannelInfos[i].currentTOI = -1;
            pUserData->vLCTChannelInfos[i].fd = -1;
            *ppLCTChannelInfo = pUserData->vLCTChannelInfos + i;
            rval = ENATSC3Status_Success;
            break;
        }
    }

    return rval;
}

ENATSC3Status
RemoveLCTChannelForTSI(struct UserData *pUserData, uint32_t tsi) 
{
    int32_t i;
    ENATSC3Status rval = ENATSC3Status_Failure;

    for (i = 0; i < numof(pUserData->vLCTChannelInfos); i++)
    {
        if (pUserData->vLCTChannelInfos[i].fInUse && 
            pUserData->vLCTChannelInfos[i].tsi == tsi)
        {
            if (pUserData->vLCTChannelInfos[i].fd >= 0)
            {
                close(pUserData->vLCTChannelInfos[i].fd);
                pUserData->vLCTChannelInfos[i].fd = -1;
            }
            pUserData->vLCTChannelInfos[i].fInUse = 0;
            rval = ENATSC3Status_Success;
            break;
        }
    }

    return rval;
}

ENATSC3Status
MediaFileBeginCallback(struct timespec *pPacketTimeUTC, 
    struct _LCTChannelInfo *pLCTChannelInfo, struct UserData *pUserData)
{
    char strFilename[256];
    ENATSC3Status rval = ENATSC3Status_Success;

    if (pUserData->fNoFileNums)
    {
        if (snprintf(strFilename, sizeof(strFilename), "%s/%s",
            pUserData->pOutPath, pLCTChannelInfo->strCurrentURL) >= 
            sizeof(strFilename))
        {
            fprintf(stderr, "Initialization URL %s too long\n", 
                pLCTChannelInfo->strCurrentURL);
            goto out;
        }
    }
    else
    {
        if (snprintf(strFilename, sizeof(strFilename), "%s/%s-%d",
            pUserData->pOutPath, pLCTChannelInfo->strCurrentURL, 
            pUserData->fileNum++) >= sizeof(strFilename))
        {
            fprintf(stderr, "Initialization URL %s too long\n", 
                pLCTChannelInfo->strCurrentURL);
            goto out;
        }
    }

    pLCTChannelInfo->fd =
        open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (pLCTChannelInfo->fd < 0)
    {
        fprintf(stderr, "Failed to create %s: %s\n", strFilename,
            strerror(errno));
        rval = ENATSC3Status_Failure;
    }

out:
    return rval;
}

void
MediaFileEndCallback(struct timespec *pPacketTimeUTC, 
    struct _LCTChannelInfo *pLCTChannelInfo, struct UserData *pUserData) 
{
    if (pLCTChannelInfo->fd >= 0)
    {
        close(pLCTChannelInfo->fd);
        pLCTChannelInfo->fd = -1;
    }

    return;
}

void
MediaFileAbortCallback(struct timespec *pPacketTimeUTC, 
    struct _LCTChannelInfo *pLCTChannelInfo, struct UserData *pUserData)
{
    if (pLCTChannelInfo->fd >= 0)
    {
        close(pLCTChannelInfo->fd);
        pLCTChannelInfo->fd = -1;
    }

    return;
}

void
MediaFileDataCallback(struct timespec *pPacketTimeUTC, 
    struct _LCTChannelInfo *pLCTChannelInfo, 
    const uint8_t *pPayload, int32_t sPayload, int64_t objectPosition, 
    int64_t objectSize, int32_t fError, struct UserData *pUserData) 
{
    int nb;

    if (lseek(pLCTChannelInfo->fd, objectPosition, SEEK_SET) != objectPosition)
    {
        fprintf(stderr, "Failed to seek to %s:%"PRId64"\n", 
            pLCTChannelInfo->strCurrentURL, objectPosition);
        goto out;
    }

    nb = write(pLCTChannelInfo->fd, pPayload, sPayload);
    if (nb == -1)
    {
        fprintf(stderr, "Failed to write payload for %s: %s\n",
            pLCTChannelInfo->strCurrentURL, strerror(errno));
        goto out;
    }
    else if (nb != sPayload)
    {
        fprintf(stderr, "Short write of payload for %s (%d < %d)\n",
            pLCTChannelInfo->strCurrentURL, nb, sPayload);
        goto out;
    }

out:
    return;
}

void
ObjectDataCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    uint32_t tsi, uint32_t toi, uint8_t codepoint, const char *pId, 
    ENATSC3ROUTESessionIdType idType, const uint8_t *pPayload, 
    int32_t payloadLength, int64_t objectPosition, int64_t tolTransferLength, 
    int64_t ftiTransferLength, int32_t fError, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;
    struct _LCTChannelInfo *pLCTChannelInfo;

    if (FindLCTChannelForTSI(pUserData, tsi, &pLCTChannelInfo))
    {
        fprintf(stderr, "Failed to find LCT channel info for tsi %u\n", tsi);
        goto out;
    }

    if (toi != pLCTChannelInfo->currentTOI && 
        pLCTChannelInfo->currentTOI != -1)
    {
        // If we get here then the object Id has changed prematurely. This
        // would suggest that we didn't receive the end of the previous
        // object.
        fprintf(stderr, 
            "%"PRIu64" %"PRId64".%06lu object changed (%"PRId64":%s) %"PRId64" != %"PRId64"\n",
            packetNumber, 
            pPacketTimeUTC->tv_sec, pPacketTimeUTC->tv_nsec / 1000,
            pLCTChannelInfo->currentTOI,
            pLCTChannelInfo->strCurrentURL,
            pLCTChannelInfo->objectSize,
            pLCTChannelInfo->pos);

        MediaFileAbortCallback(pPacketTimeUTC, pLCTChannelInfo, pUserData);
        pLCTChannelInfo->strCurrentURL[0] = 0;
        pLCTChannelInfo->currentTOI = -1;

#if 0
        if (objectPosition != 0)
        {
            // We need to wait til we're at the start of a new object.
            goto out;
        }
#endif
    }

    // If we've gotten this far then we are processing an object.
    if (pLCTChannelInfo->currentTOI == -1)
    {
        char strContentType[64];

#if 0
        {
            if (payloadLength > LiteralLength("Content-Location: ") &&
                !strncmp((char *)pPayload, "Content-Location: ",
                LiteralLength("Content-Location: ")))
            {
                int32_t payloadOffset = LiteralLength("Content-Location: ");

                while (payloadOffset + 3 < payloadLength &&
                    (pPayload[payloadOffset] != '\r' ||
                     pPayload[payloadOffset + 1] != '\n' ||
                     pPayload[payloadOffset + 2] != '\r' ||
                     pPayload[payloadOffset + 3] != '\n'))
                {
                    payloadOffset++;
                }

                strncpy(pLCTChannelInfo->strCurrentURL, (char *)pPayload +
                    LiteralLength("Content-Location: "), payloadOffset -
                    LiteralLength("Content-Location: "));
                pLCTChannelInfo->strCurrentURL[
                    payloadOffset - LiteralLength("Content-Location: ")] = 0;
                printf("URL: %s\n", pLCTChannelInfo->strCurrentURL);

                pPayload += payloadOffset + 4;
                payloadLength -= payloadOffset + 4;
            }
        }
#else
        if (ENATSC3ROUTESessionGetObjectURLAndContentType(
            pUserData->hROUTESession, tsi, toi, pLCTChannelInfo->strCurrentURL, 
            sizeof(pLCTChannelInfo->strCurrentURL), strContentType, 
            sizeof(strContentType)))
        {
            fprintf(stderr, "Failed to get url for tsi %u toi %u\n",
                tsi, toi);
            goto out;
        }
#endif

        printf("%"PRIu64" %"PRId64".%06lu MediaFileBegin %s tsi %u codepoint %u\n",
            packetNumber, pPacketTimeUTC->tv_sec, pPacketTimeUTC->tv_nsec / 1000,
            pLCTChannelInfo->strCurrentURL, tsi, codepoint);

        MediaFileBeginCallback(pPacketTimeUTC, pLCTChannelInfo, 
            pUserData);

        pLCTChannelInfo->currentTOI = toi;
        pLCTChannelInfo->codepoint = codepoint;
        pLCTChannelInfo->pos = 0;
        pLCTChannelInfo->objectSize = -1;
    }

    if (pLCTChannelInfo->objectSize == -1)
    {
        if (tolTransferLength != -1)
        {
            pLCTChannelInfo->objectSize = tolTransferLength;
        }
        else if (ftiTransferLength != -1)
        {
            pLCTChannelInfo->objectSize = ftiTransferLength;
        }
    }
    else
    {
        if (tolTransferLength != -1)
        {
            if (pLCTChannelInfo->objectSize != tolTransferLength)
            {
                fprintf(stderr, 
                    "%"PRIu64" %"PRId64".%06lu TOL object size changed %"PRId64" != %"PRId64"\n",
                    packetNumber, 
                    pPacketTimeUTC->tv_sec, pPacketTimeUTC->tv_nsec / 1000,
                    pLCTChannelInfo->objectSize, tolTransferLength);
                printf("MediaFileAbort %s tsi %d\n",
                    pLCTChannelInfo->strCurrentURL, pLCTChannelInfo->tsi);
                MediaFileAbortCallback(pPacketTimeUTC, pLCTChannelInfo, 
                    pUserData); 
                pLCTChannelInfo->strCurrentURL[0] = 0;
                pLCTChannelInfo->currentTOI = -1;
                goto out;
            }
        }

        if (ftiTransferLength != -1)
        {
            if (pLCTChannelInfo->objectSize != ftiTransferLength)
            {
                fprintf(stderr, "%"PRIu64" FTI object size changed %"PRId64" != %"PRId64"\n",
                    packetNumber, pLCTChannelInfo->objectSize,
                    ftiTransferLength);
                printf("MediaFileAbort %s tsi %d\n",
                    pLCTChannelInfo->strCurrentURL, pLCTChannelInfo->tsi);
                MediaFileAbortCallback(pPacketTimeUTC, pLCTChannelInfo, 
                    pUserData);
                pLCTChannelInfo->strCurrentURL[0] = 0;
                pLCTChannelInfo->currentTOI = -1;
                goto out;
            }
        }
    }

    // We currently expect data to be delivered in order.
    //
    // XXXX Strictly speaking data may arrive out of order if @order=0 in the
    // SrcFlow's Payload element.
    if (pLCTChannelInfo->pos != objectPosition)
    {
        fprintf(stderr, "%"PRIu64" %"PRIu64".%.06lu av offset error %"PRId64" != %"PRId64" (%"PRId64" missing) (%s size %"PRId64")\n",
            packetNumber,
            pPacketTimeUTC->tv_sec, pPacketTimeUTC->tv_nsec / 1000000,
            pLCTChannelInfo->pos, objectPosition,
            objectPosition - pLCTChannelInfo->pos,
            pLCTChannelInfo->strCurrentURL, pLCTChannelInfo->objectSize);
    }

    MediaFileDataCallback(pPacketTimeUTC, pLCTChannelInfo,
        pPayload, payloadLength, objectPosition,
        pLCTChannelInfo->objectSize, 0, pUserData);

    pLCTChannelInfo->pos = objectPosition + payloadLength;

    // See if the object is now complete
    if (pLCTChannelInfo->pos == pLCTChannelInfo->objectSize)
    {
        printf("%"PRIu64" %"PRIu64".%.06lu MediaFileEnd %s tsi %d\n", 
            packetNumber,
            pPacketTimeUTC->tv_sec, pPacketTimeUTC->tv_nsec / 1000000,
            pLCTChannelInfo->strCurrentURL, 
            pLCTChannelInfo->tsi);
        MediaFileEndCallback(pPacketTimeUTC, pLCTChannelInfo, pUserData);
        pLCTChannelInfo->strCurrentURL[0] = 0;
        pLCTChannelInfo->currentTOI = -1;
    }

out:
    return;
}

void
SessionResetCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;
    int32_t i;

    for (i = 0; i < numof(pUserData->vLCTChannelInfos); i++)
    {
        if (pUserData->vLCTChannelInfos[i].fInUse)
        {
            struct _LCTChannelInfo *pLCTChannelInfo;

            pLCTChannelInfo = pUserData->vLCTChannelInfos + i;

            if (pLCTChannelInfo->currentTOI != -1)
            {
                MediaFileAbortCallback(pPacketTimeUTC, pLCTChannelInfo, 
                    pUserData);
                pLCTChannelInfo->currentTOI = -1;
            }

            pUserData->vLCTChannelInfos[i].fInUse = 0;
        }
    }
}

ENATSC3Status
MPDAvailableCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    const char *pMPD, int32_t sMPD, int32_t version, uint32_t crc32, 
    void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;
    char strFilename[256];
    int fd = -1;
    int nb;
    ENATSC3Status rval = ENATSC3Status_Success;

    if (snprintf(strFilename, sizeof(strFilename), "%s/mpd-%d-%"PRIu64,
        pUserData->pOutPath, version, pPacketTimeUTC->tv_sec) >= 
        sizeof(strFilename))
    {
        fprintf(stderr, "MPD filename too long\n");
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        fprintf(stderr, "Failed to open %s: %s\n", strFilename,
            strerror(errno));
        goto out;
    }

    nb = write(fd, pMPD, sMPD);
    if (nb == -1)
    {
        fprintf(stderr, "Failed to write %s: %s\n", strFilename,
            strerror(errno));
        goto out;
    }
    else if (nb != sMPD)
    {
        fprintf(stderr, "Short write of %s (%d < %d)\n", strFilename,
            nb, sMPD);
        goto out;
    }

out:
    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

ENATSC3Status
HELDAvailableCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    const char *pHELD, int32_t sHELD, int32_t version, uint32_t crc32, 
    void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;
    char strFilename[256];
    int fd = -1;
    int nb;
    ENATSC3Status rval = ENATSC3Status_Success;

    if (snprintf(strFilename, sizeof(strFilename), "%s/held-%d-%"PRIu64,
        pUserData->pOutPath, version, pPacketTimeUTC->tv_sec) >= 
        sizeof(strFilename))
    {
        fprintf(stderr, "HELD filename too long\n");
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        fprintf(stderr, "Failed to open %s: %s\n", strFilename,
            strerror(errno));
        goto out;
    }

    nb = write(fd, pHELD, sHELD);
    if (nb == -1)
    {
        fprintf(stderr, "Failed to write %s: %s\n", strFilename,
            strerror(errno));
        goto out;
    }
    else if (nb != sHELD)
    {
        fprintf(stderr, "Short write of %s (%d < %d)\n", strFilename,
            nb, sHELD);
        goto out;
    }

out:
    if (fd >= 0)
    {
        close(fd);
    }

    return rval;
}

void
SLSAvailableCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    const char *pSLSFilename, const uint8_t *pSLS, int32_t sSLS, 
    int32_t fGzipped, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;
    char strFilename[256];
    int fd = -1;
    int nb;

    if (snprintf(strFilename, sizeof(strFilename), "%s/sls-%s-%d",
        pUserData->pOutPath, pSLSFilename, pUserData->fileNum++) >= 
        sizeof(strFilename))
    {
        fprintf(stderr, "SLS filename too long\n");
        goto out;
    }

    fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        fprintf(stderr, "Failed to open %s: %s\n", strFilename,
            strerror(errno));
        goto out;
    }

    nb = write(fd, pSLS, sSLS);
    if (nb == -1)
    {
        fprintf(stderr, "Failed to write %s: %s\n", strFilename,
            strerror(errno));
        goto out;
    }
    else if (nb != sSLS)
    {
        fprintf(stderr, "Short write of %s (%d < %d)\n", strFilename,
            nb, sSLS);
        goto out;
    }

out:
    if (fd >= 0)
    {
        close(fd);
    }

    return;
}

void
AddIPAddressCallback(uint32_t addr, uint16_t port, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (EnIPFilterSetAddFilter(addr, port, pUserData->pIPFilterSet))
    {
        fprintf(stderr, "Failed to add filter for %.d.%.d.%.d.%.d:%d\n",
            (addr >> 24) & 0xff, (addr >> 16) & 0xff, (addr >> 8) & 0xff,
            addr & 0xff, port);
    }
}

void
RemoveIPAddressCallback(uint32_t addr, uint16_t port, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (EnIPFilterSetRemoveFilter(addr, port, pUserData->pIPFilterSet))
    {
        fprintf(stderr, "Failed to remove filter for %.d.%.d.%.d.%.d:%d\n",
            (addr >> 24) & 0xff, (addr >> 16) & 0xff, (addr >> 8) & 0xff,
            addr & 0xff, port);
    }
}

void
CommitIPAddressesCallback(void *pCbUserData)
{
    // Nothing to do
}

void
LCTChannelAddedCallback(struct timespec *pPacketTime, uint64_t packetNumber,
    uint32_t addr, uint16_t port, uint32_t tsi, const char *pId, 
    ENATSC3ROUTESessionIdType idType, void *pCbUserData)
{
    struct _LCTChannelInfo *pLCTChannelInfo;
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (AddLCTChannelForTSI(pUserData, tsi, &pLCTChannelInfo))
    {
        fprintf(stderr, "Failed to allocate LCT Channel info for tsi %u\n",
            tsi);
    }
}

void
LCTChannelRemovedCallback(struct timespec *pPacketTime, uint64_t packetNumber,
    uint32_t addr, uint16_t port, uint32_t tsi, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (RemoveLCTChannelForTSI(pUserData, tsi))
    {
        fprintf(stderr, "Failed to free LCT Channel info for tsi %u\n",
            tsi);
    }
}

int32_t
parseArgs(int argc, char **argv, char **ppInFilename, char **ppOutPath,
    char **ppShadowPath, uint32_t *pDestinationAddr, 
    uint16_t *pDestinationPort, int32_t *pFBitrateStatistics, 
    ENVerbosity *pVerbosity, int32_t *pFNoFileNums)
{
    int32_t err = 0;
    int32_t fHaveDestinationAddr = 0;
    int32_t fHaveDestinationPort = 0;
    int32_t i;

    *ppInFilename = NULL;
    *ppOutPath = NULL;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-b"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutPath = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-shadow"))
        {
            if (i + 1 < argc)
            {
                *ppShadowPath = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-port"))
        {
            if (i + 1 < argc)
            {
                *pDestinationPort = atol(argv[i + 1]);
                fHaveDestinationPort = 1;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-addr"))
        {
            if (i + 1 < argc)
            {
                struct in_addr addr;
                if (inet_aton(argv[i + 1], &addr))
                {
                    *pDestinationAddr = htonl(addr.s_addr);
                    fHaveDestinationAddr = 1;
                }
                else err = 1;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-v"))
        {
            if (i + 1 < argc)
            {
                *pVerbosity = atol(argv[i + 1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-bitratestats"))
        {
            *pFBitrateStatistics = 1;
        }
        else if (!strcmp(argv[i], "-nofilenums"))
        {
            *pFNoFileNums = 1;
        }
        else err = 1;
    }

    if (!*ppInFilename || !*ppOutPath || !fHaveDestinationAddr || 
        !fHaveDestinationPort)
    {
        err = 1;
    }

    if (err)
    {
        fprintf(stderr, "Usage: %s -b udpfilename -o outpath -addr multicast_addr -port mulitcast_port [-shadow path] [-bitratestats] [-v verbosity] [-nofilenums]\n", argv[0]);
    }

    return err;
}

void
sighandler(int signum)
{
    if (signum == SIGINT)
    {
        fRunningG = 0;
    }
}

int
main(int argc, char **argv)
{
    int32_t fEnLogInitialized = 0;
    int32_t fXML2Initialized = 0;
    int32_t fROUTEObjectCallbacksRegistered = 0;
    int32_t fROUTEMPDAvailableCallbacksRegistered = 0;
    int32_t fROUTELCTChannelCallbacksRegistered = 0;
    int32_t fROUTEHELDAvailableCallbacksRegistered = 0;
    int32_t fROUTESLSAvailableCallbacksRegistered = 0;
    char *pInFilename;
    char *pOutPath;
    char *pShadowPath = NULL;
    uint32_t destinationAddr;
    uint16_t destinationPort;
    uint8_t *pPayload = NULL;
    uint64_t udpPacketNumber = 0;
    struct UserData userdata = 
    {
        .verbosity = ENVerbosity_Terse,
        .fBitrateStatistics = 0,
        .ifd = -1,
        .hROUTESession = NULL,
        .pOutPath = NULL,
        .pIPFilterSet = NULL,
        .fileNum = 0,
        .fNoFileNums = 0,
    };
    struct sigaction act;
    int rval = 0;

    // Graceful exit on ctrl-c
    act.sa_handler = sighandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    sigaction(SIGINT, &act, NULL);

    if (EnLogInitialize(1))
    {
        fprintf(stderr, "Failed to initialize EnLog\n");
        rval = -1;
        goto out;
    }
    fEnLogInitialized = 1;

    xmlInitParser();
    fXML2Initialized = 1;

    if (parseArgs(argc, argv, &pInFilename, &pOutPath, &pShadowPath, 
        &destinationAddr, &destinationPort, &userdata.fBitrateStatistics,
        &userdata.verbosity, &userdata.fNoFileNums)) 
    {
        rval = 1;
        goto out;
    }

    if (EnIPFilterSetAllocate(64, &userdata.pIPFilterSet))
    {
        fprintf(stderr, "Failed to allocate IP filter set\n");
        rval = 1;
        goto out;
    }

    if (ENATSC3ROUTESessionAllocate(ENATSC3ROUTESessionType_ATSC3_DASH, 
        destinationAddr, destinationPort, AddIPAddressCallback, 
        RemoveIPAddressCallback, CommitIPAddressesCallback, 
        &userdata, "test", &userdata.hROUTESession))
    {
        fprintf(stderr, "Failed to allocate ROUTE session\n");
        rval = 1;
        goto out;
    }

    ENATSC3ROUTESessionSetVerbosity(userdata.hROUTESession, userdata.verbosity);
    ENATSC3ROUTESessionSetBitrateStatisticsEnabled(userdata.hROUTESession,
        userdata.fBitrateStatistics);

    if (ENATSC3ROUTESessionRegisterLCTChannelCallbacks(userdata.hROUTESession,
        LCTChannelAddedCallback, LCTChannelRemovedCallback, &userdata))
    {
        fprintf(stderr, "Failed to register LCT Channel callbacks\n");
        rval = 1;
        goto out;
    }
    fROUTELCTChannelCallbacksRegistered = 1;

    if (ENATSC3ROUTESessionRegisterObjectCallbacks(userdata.hROUTESession, 
        ObjectDataCallback, SessionResetCallback, &userdata))
    {
        fprintf(stderr, "Failed to register ROUTE object callbacks\n");
        rval = 1;
        goto out;
    }
    fROUTEObjectCallbacksRegistered = 1;

    if (ENATSC3ROUTESessionRegisterMPDAvailableCallback(userdata.hROUTESession, 
        MPDAvailableCallback, &userdata))
    {
        fprintf(stderr, "Failed to register ROUTE MPD callback\n");
        rval = 1;
        goto out;
    }
    fROUTEMPDAvailableCallbacksRegistered = 1;

    if (ENATSC3ROUTESessionRegisterHELDAvailableCallback(userdata.hROUTESession, 
        HELDAvailableCallback, &userdata))
    {
        fprintf(stderr, "Failed to register ROUTE HELD callback\n");
        rval = 1;
        goto out;
    }
    fROUTEHELDAvailableCallbacksRegistered = 1;

    if (ENATSC3ROUTESessionRegisterSLSAvailableCallback(userdata.hROUTESession, 
        SLSAvailableCallback, &userdata))
    {
        fprintf(stderr, "Failed to register ROUTE SLS callback\n");
        rval = 1;
        goto out;
    }
    fROUTESLSAvailableCallbacksRegistered = 1;

    if (pShadowPath)
    {
        ENATSC3ROUTESessionSetShadowFilePath(userdata.hROUTESession, 
            pShadowPath);
    }

    userdata.pOutPath = pOutPath;

    userdata.ifd =  open(pInFilename, O_RDONLY);
    if (userdata.ifd < 0)
    {
        fprintf(stderr, "Failed to open %s for reading: %s\n", pInFilename, 
            strerror(errno));
        rval = 1;
        goto out;
    }

    pPayload = (uint8_t *)malloc(MAX_PAYLOAD_SIZE);
    if (!pPayload)
    {
        fprintf(stderr, "Out of memory\n");
        rval = 1;
        goto out;
    }

    while (1)
    {
        EntangleCapHeader header;
        int32_t payloadLength;
        struct timespec packetTime;
        int nb;

        nb = read(userdata.ifd, &header, sizeof(header));
        if (nb == 0)
        {
            break;
        }
        else if (nb < 0)
        {
            fprintf(stderr, "Failed to read udp file: %s\n", strerror(errno));
            rval = 1;
            break;
        }
        else if (nb != sizeof(header))
        {
            fprintf(stderr, "Partial read of UDP header, exiting\n");
            rval = 1;
            break;
        }

        payloadLength = ntohs(header.length);
        packetTime.tv_sec = ENUtils_ntohll(header.sec);
        packetTime.tv_nsec = ntohl(header.nsec);
        if (payloadLength > MAX_PAYLOAD_SIZE)
        {
            fprintf(stderr, "Payload too long: %d > %d\n", 
                payloadLength, MAX_PAYLOAD_SIZE);
            rval = 1;
            break;
        }

        nb = read(userdata.ifd, pPayload, payloadLength);
        if (nb < 0)
        {
            fprintf(stderr, "Error reading payload: %s\n", strerror(errno));
            rval = 1;
            break;
        }
        else if (nb != payloadLength)
        {
            fprintf(stderr, "Partial read of UDP payload, exiting\n");
            rval = 1;
            break;
        }

        if (htonl(header.ipv4addr) == destinationAddr && 
            ntohs(header.port) == destinationPort)
        {
            if (ENATSC3ROUTESessionProcessUDPPacket(userdata.hROUTESession, 
                &packetTime, udpPacketNumber, 0, pPayload, payloadLength))
            {
                fprintf(stderr, "Failed to process UDP packet\n");
            }
        }

        udpPacketNumber++;
    }

out:
    if (fROUTESLSAvailableCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterSLSAvailableCallback(
            userdata.hROUTESession, SLSAvailableCallback, &userdata))
        {
            fprintf(stderr, "Failed to unregister ROUTE SLS callback\n");
        }
    }

    if (fROUTEHELDAvailableCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterHELDAvailableCallback(
            userdata.hROUTESession, HELDAvailableCallback, &userdata))
        {
            fprintf(stderr, "Failed to unregister ROUTE HELD callback\n");
        }
    }

    if (fROUTEMPDAvailableCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterMPDAvailableCallback(
            userdata.hROUTESession, MPDAvailableCallback, &userdata))
        {
            fprintf(stderr, "Failed to unregister ROUTE MPD callback\n");
        }
    }

    if (fROUTELCTChannelCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterLCTChannelCallbacks(
            userdata.hROUTESession, LCTChannelAddedCallback, 
            LCTChannelRemovedCallback, &userdata))
        {   
            fprintf(stderr, "Failed to unregister LCT Channel callbacks\n");
        }
    }

    if (fROUTEObjectCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterObjectCallbacks(userdata.hROUTESession,
            ObjectDataCallback, SessionResetCallback, &userdata))
        {
            fprintf(stderr,
                "Failed to unregister ROUTE media file callbacks\n");
        }
    }

    if (userdata.hROUTESession)
    {
        if (ENATSC3ROUTESessionFree(userdata.hROUTESession))
        {
            fprintf(stderr, "Failed to free ROUTE session\n");
        }
    }

    if (userdata.pIPFilterSet)
    {
        if (EnIPFilterSetFree(userdata.pIPFilterSet))
        {
            fprintf(stderr, "Failed to free IP filter set\n");
        }
    }

    if (fXML2Initialized)
    {
        xmlCleanupCharEncodingHandlers();
        xmlDictCleanup();
        xmlCleanupParser();
    }

    if (fEnLogInitialized)
    {
        if (EnLogShutdown())
        {
            fprintf(stderr, "Failed to shut down EnLog\n");
        }
    }

    if (userdata.ifd >= 0)
    {
        close(userdata.ifd);
    }

    if (pPayload)
    {
        free(pPayload);
    }

    return rval;
}
