/*
 * Copyright © 2019 - 2022 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/statvfs.h>
#include <pthread.h>
#include <semaphore.h>
#include <signal.h>
#include <assert.h>

#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>

#include <enatsc3-common/enatsc3.h>

#include <librzrsdk2/redzone_c_api.h>
#include <librzrsdk2/props_api.h>

#include "enatsc3cap.h"

#define MAX_BASEBAND_PAYLOAD_SIZE 65536

#define MAX_BASEBAND_RZR_SOURCES 4

#define CAPTURE_FILE_MAX_SIZE ((uint64_t)2 * 1024 * 1024 * 1024)
#define CAPTURE_SESSION_FREE_SPACE_TARGET CAPTURE_FILE_MAX_SIZE

// CORRUPT_PACKETS injects errors into packets
// #define CORRUPT_PACKETS 1
#define CORRUPT_PACKET_FREQUENCY 8
#define CORRUPT_PACKET_BYTES_PER_PACKET 64

typedef enum
{
    ENATSC3BasebandRZRSDK2SourceOperatingMode_Invalid = 0,
    ENATSC3BasebandRZRSDK2SourceOperatingMode_ATSC = 1,
    ENATSC3BasebandRZRSDK2SourceOperatingMode_ATSC3 = 2,
} ENATSC3BasebandRZRSDK2SourceOperatingMode;

typedef struct
{
    int fd;
    int32_t activeBuffer;

    struct _CaptureSessionBuffer
    {
        // If nUsed == 0 the the buffer is available for use.
        uint32_t nUsed;
        uint8_t vData[1024 * 1024];
    } vBuffers[8];

    TaskletRunnerHandle_t hTaskletRunner;
} CaptureSession;

typedef struct
{
    enum 
    {
        CaptureSessionTaskletArgType_Invalid = 0,
        CaptureSessionTaskletArgType_WriteData = 1,
        CaptureSessionTaskletArgType_CloseFile = 2,
    } type;

    CaptureSession *pSession;

    union
    {
        struct
        {
            int fd;
            int bufferNumber;
        } writeDataParams;
        struct
        {
            int fd;
        } closeFileParams;
    };
} CaptureSessionTaskletArg;

typedef struct
{
    char strModel[32];
    char strSerial[32];

    RedZoneCaptureHandle_t hRedZoneCapture;

    int32_t fOpened;

    enum RedZoneOperatingMode operatingMode;
    int32_t frequencyKHz;

    // The currently tuned PLPs and their slot assignemnts. Most ATSC 3 demods
    // can handle only 4 PLPs, as that is all that is required by spec.
    uint8_t vPLPIds[4];
    uint64_t plpIdMask;

    int32_t fPLPIdMaskChanged;

    struct _stats
    {
        uint8_t masterLock;
        uint64_t plpValidMask;
        uint64_t llsValidMask;
        int32_t snr;
        int32_t rssi;

        // Total packets received
        uint32_t nBasebandPackets;
        // Total bits received
        uint32_t nBytes;
        // Number of packets containing uncorrected errors.
        uint32_t nErroredPackets;

        struct _plpStats
        {
            ENATSC3Modulation modulation;
            ENATSC3Coding coding;

            // PLP packets received
            uint32_t nBasebandPackets;
            // PLP bits received
            uint32_t nBytes;
            // Number of continuity counter sequence discontinuities detected
            uint32_t nContinuityCounterErrors;
            // Number of packets containing uncorrected errors.
            uint32_t nErroredPackets;
            int32_t lastContinuityCounter;
        } vPLPStats[64];
    } stats;

    pthread_mutex_t statsLock;

    CaptureSession *pCaptureSession;
    int32_t fDemodRecordChanged;
    // Allocate space for a demod record with stats for one subframe. Note
    // that the subframe stats are declared as a zero-sized array in 
    // ENATSC3CapDemodulatorRecordV2 so we need to append that space.
    struct
    {
        ENATSC3CapDemodulatorRecordV2 demodRecord;
        struct _ATSC3_L1_Detail_Subframe firstSubframe;
    } __attribute__((__packed__));

    pthread_mutex_t lock;
} ENATSC3BasebandRZRSDK2Source;

struct
{
    int32_t fInitialized;

    ENATSC3BasebandRZRSDK2Source vSources[MAX_BASEBAND_RZR_SOURCES];
    int32_t nSources;

    char *pCapturePath;

    pthread_mutex_t lock;
} atsc3BasebandRZRSDK2SourceGlobalsG =
{
    .fInitialized = 0,
    .pCapturePath = NULL,
};

ENATSC3Status
InitializeCaptureFile(char *pFilename, int *pFd)
{
    int fd = -1;
    ENATSC3CapFileHeader fileHeader;
    ENATSC3Status rval = ENATSC3Status_Success;

    fd = open(pFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0)
    {
        fprintf(stderr, "Failed to open %s for writing: %s\n", pFilename, 
            strerror(errno));
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    fileHeader.format = ENATSC3CapFormat_V2;
    fileHeader.type = ENATSC3CapType_ATSC3_Baseband;
    fileHeader.magic = htons(ENATSC3CAP_HEADER_MAGIC);
    if (write(fd, &fileHeader, sizeof(fileHeader)) != sizeof(fileHeader))
    {
        fprintf(stderr, "Failed to write file header: %s\n", strerror(errno));
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    *pFd = fd;

out:
    return rval;

errout:
    if (fd >= 0)
    {
        close(fd);
    }

    goto out;
}

void
CaptureSessionCloseFileTasklet(TaskletRunnerHandle_t hTaskletRunner, 
    TaskletArg *pArg)
{
    CaptureSessionTaskletArg *pCaptureSessionTaskletArg;

    pCaptureSessionTaskletArg = (CaptureSessionTaskletArg *)pArg->vRaw;

    close(pCaptureSessionTaskletArg->closeFileParams.fd);
}

void
CaptureSessionWriteDataTasklet(TaskletRunnerHandle_t hTaskletRunner, 
    TaskletArg *pArg)
{
    CaptureSessionTaskletArg *pCaptureSessionTaskletArg;
    int32_t bufferNumber;
    CaptureSession *pSession;

    pCaptureSessionTaskletArg = (CaptureSessionTaskletArg *)pArg->vRaw;
    bufferNumber = pCaptureSessionTaskletArg->writeDataParams.bufferNumber;
    pSession = pCaptureSessionTaskletArg->pSession;

    if (write(pCaptureSessionTaskletArg->writeDataParams.fd, 
        pSession->vBuffers[bufferNumber].vData, 
        pSession->vBuffers[bufferNumber].nUsed) != 
        pSession->vBuffers[bufferNumber].nUsed)
    {
        fprintf(stderr, "Failed to write data: %s\n ", strerror(errno));
    }
    pSession->vBuffers[bufferNumber].nUsed = 0;
}

ENATSC3Status
CaptureSessionCreate(char *pCaptureFilename, CaptureSession **ppSession)
{
    CaptureSession *pSession = NULL;
    ENATSC3Status rval = ENATSC3Status_Success;

    pSession = (CaptureSession *)calloc(1, sizeof(CaptureSession));
    if (!pSession)
    {
        fprintf(stderr, "Out of memory\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    if (InitializeCaptureFile(pCaptureFilename, &pSession->fd))
    {
        fprintf(stderr, "Failed to open %s for writing: %s\n", 
            pCaptureFilename, strerror(errno));
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    pSession->activeBuffer = -1;

    if (ENUtils_TaskletRunnerCreate("encap",
        TASKLET_PRIORITY_REALTIME_LOW, sizeof(CaptureSessionTaskletArg), 4,
        &pSession->hTaskletRunner))
    {
        fprintf(stderr, "Failed to create tasklet runner\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    *ppSession = pSession;

out:
    return rval;

errout:
    if (pSession)
    {
        if (pSession->hTaskletRunner)
        {
            if (ENUtils_TaskletRunnerDestroy(pSession->hTaskletRunner))
            {
                fprintf(stderr, "Failed to destroy tasklet runner\n");
            }
        }

        free(pSession);
    }

    goto out;
}

ENATSC3Status
CaptureSessionDestroy(CaptureSession *pSession)
{
    ENATSC3Status rval = ENATSC3Status_Success;

    if (pSession->fd >= 0)
    {
        Tasklet *pTasklet;
        CaptureSessionTaskletArg *pCaptureSessionTaskletArg;

        if (ENUtils_TaskletRunnerTaskletAllocate(pSession->hTaskletRunner,
            sizeof(CaptureSessionTaskletArg), &pTasklet))
        {
            fprintf(stderr, "Failed to allocate tasklet\n");
            close(pSession->fd);
        }
        else
        {
            pTasklet->Function = CaptureSessionCloseFileTasklet; 
            pCaptureSessionTaskletArg = 
                (CaptureSessionTaskletArg *)pTasklet->pArg->vRaw;
            pCaptureSessionTaskletArg->type = 
                CaptureSessionTaskletArgType_CloseFile;
            pCaptureSessionTaskletArg->pSession = pSession;
            pCaptureSessionTaskletArg->closeFileParams.fd = pSession->fd;
            if (ENUtils_TaskletRunnerAddTasklet(pSession->hTaskletRunner,
                pTasklet, NULL))
            {
                fprintf(stderr, "Failed to enqueue tasklet\n");
                ENUtils_TaskletRunnerTaskletReturn(pSession->hTaskletRunner,
                    pTasklet);
            }
        }

        pSession->fd = -1;
    }

    if (ENUtils_TaskletRunnerRunAllTasklets(pSession->hTaskletRunner))
    {
        fprintf(stderr, "Failed to run all tasklets\n");
    }

    if (ENUtils_TaskletRunnerDestroy(pSession->hTaskletRunner))
    {
        fprintf(stderr, "Failed to destroy tasklet runner\n");
    }

    free(pSession);

    return rval;
}

ENATSC3Status
CaptureSessionWrite(CaptureSession *pSession, 
    ENATSC3BasebandRZRSDK2Source *pSource, struct iovec *pVecs, int32_t nVecs)
{
    uint32_t nToWrite = 0;
    int32_t i;
    ENATSC3Status rval = ENATSC3Status_Success;

    for (i = 0; i < nVecs; i++)
    {
        nToWrite += pVecs[i].iov_len;
    }

    if (pSession->activeBuffer >= 0)
    {
        if (pSession->vBuffers[pSession->activeBuffer].nUsed + nToWrite >
            sizeof(pSession->vBuffers[pSession->activeBuffer].vData))
        {
            Tasklet *pTasklet;
            CaptureSessionTaskletArg *pCaptureSessionTaskletArg;

            if (ENUtils_TaskletRunnerTaskletAllocate(pSession->hTaskletRunner,
                sizeof(CaptureSessionTaskletArg), &pTasklet))
            {
                fprintf(stderr, "Failed to allocate tasklet\n");
                rval = ENATSC3Status_Failure;
                goto out;
            }

            pTasklet->Function = CaptureSessionWriteDataTasklet; 
            pCaptureSessionTaskletArg = 
                (CaptureSessionTaskletArg *)pTasklet->pArg->vRaw;
            pCaptureSessionTaskletArg->type = 
                CaptureSessionTaskletArgType_WriteData;
            pCaptureSessionTaskletArg->pSession = pSession;
            pCaptureSessionTaskletArg->writeDataParams.fd = pSession->fd;
            pCaptureSessionTaskletArg->writeDataParams.bufferNumber = 
                pSession->activeBuffer;
            if (ENUtils_TaskletRunnerAddTasklet(pSession->hTaskletRunner,
                pTasklet, NULL))
            {
                fprintf(stderr, "Failed to enqueue tasklet\n");
                ENUtils_TaskletRunnerTaskletReturn(pSession->hTaskletRunner,
                    pTasklet);
                rval = ENATSC3Status_Failure;
                goto out;
            }

            pSession->activeBuffer = -1;
        }
    }

    if (pSession->activeBuffer < 0)
    {
        for (i = 0; i < numof(pSession->vBuffers); i++)
        {
            if (pSession->vBuffers[i].nUsed == 0)
            {
                break;
            }
        }

        if (i == numof(pSession->vBuffers))
        {
            fprintf(stderr, "Failed to allocate buffer\n");
            rval = ENATSC3Status_Failure;
            goto out;
        }

        pSession->activeBuffer = i;
    }

    struct _CaptureSessionBuffer *pBuffer;
    pBuffer = pSession->vBuffers + pSession->activeBuffer;
    for (i = 0; i < nVecs; i++)
    {
        ENUtils_CrashAssert(pBuffer->nUsed + pVecs[i].iov_len <= 
            sizeof(pBuffer->vData));
        memcpy(pBuffer->vData + pBuffer->nUsed, pVecs[i].iov_base, 
            pVecs[i].iov_len);
        pBuffer->nUsed += pVecs[i].iov_len;
    }

out:
    return rval;
}

ENATSC3Status
WriteBasebandRecord(ENATSC3BasebandRZRSDK2Source *pSource, 
    RedZoneCaptureBasebandPacket *pPacket)
{
    struct timespec mononow;
    struct timespec rtnow;
    int32_t nPayloadBytes = 0;
    ENATSC3CapRecordHeaderV2 header;
    struct iovec vVecs[numof(pPacket->v0.vVecs) + 1];
    int32_t i;
    ENATSC3Status rval = ENATSC3Status_Success;

    clock_gettime(CLOCK_MONOTONIC, &mononow);
    clock_gettime(CLOCK_REALTIME, &rtnow);

    vVecs[0].iov_base = &header;
    vVecs[0].iov_len = sizeof(header);
    for (i = 0; i < pPacket->v0.nVecs; i++)
    {
        nPayloadBytes += pPacket->v0.vVecs[i].length;
        vVecs[i + 1].iov_base = pPacket->v0.vVecs[i].pData;
        vVecs[i + 1].iov_len = pPacket->v0.vVecs[i].length;
    }

    header.type = ENATSC3CapPayloadType_ATSC3_Baseband;
    header.continuityCounter = pPacket->v0.continuityCounter;
    header.fError = pPacket->v0.fError;
    header.plpId = pPacket->v0.plpId;
    header.length = htonl(nPayloadBytes);
    header.monowhen.sec = ENUtils_htonll(mononow.tv_sec);
    header.monowhen.nsec = htonl(mononow.tv_nsec);
    header.rtwhen.sec = ENUtils_htonll(rtnow.tv_sec);
    header.rtwhen.nsec = htonl(rtnow.tv_nsec);
    header.broadcastTime.L1B_time_info_flag = 
        htonl(pPacket->v0.L1TimePrecision);
    header.broadcastTime.L1D_time_sec = htonl(pPacket->v0.L1TimeSeconds);
    header.broadcastTime.L1D_time_msec = htonl(pPacket->v0.L1TimeMilliseconds);
    header.broadcastTime.L1D_time_usec = htonl(pPacket->v0.L1TimeMicroseconds);
    header.broadcastTime.L1D_time_nsec = htonl(pPacket->v0.L1TimeNanoseconds);
    
    if (CaptureSessionWrite(pSource->pCaptureSession, pSource, vVecs, 
        pPacket->v0.nVecs + 1)) 
    {
        fprintf(stderr, "Failed to submit baseband packet to capture session\n");
        rval = ENATSC3Status_Failure;
        goto out;
    }

out:
    return rval;
}

ENATSC3Status
WriteDemodulatorRecord(ENATSC3BasebandRZRSDK2Source *pSource, 
    ENATSC3CapDemodulatorRecordV2 *pRecord, int32_t nBytes)
{
    struct timespec mononow;
    struct timespec rtnow;
    ENATSC3CapRecordHeaderV2 header;
    struct iovec vVecs[2];
    ENATSC3Status rval = ENATSC3Status_Success;

    clock_gettime(CLOCK_MONOTONIC, &mononow);
    clock_gettime(CLOCK_REALTIME, &rtnow);

    vVecs[0].iov_base = &header;
    vVecs[0].iov_len = sizeof(header);
    vVecs[1].iov_base = pRecord;
    vVecs[1].iov_len = nBytes;

    header.type = ENATSC3CapPayloadType_Demod_V2;
    header.continuityCounter = 0;
    header.fError = 0;
    header.plpId = 0;
    header.length = htonl(nBytes);
    header.monowhen.sec = ENUtils_htonll(mononow.tv_sec);
    header.monowhen.nsec = htonl(mononow.tv_nsec);
    header.rtwhen.sec = ENUtils_htonll(rtnow.tv_sec);
    header.rtwhen.nsec = htonl(rtnow.tv_nsec);
    header.broadcastTime.L1B_time_info_flag = 0;

    if (CaptureSessionWrite(pSource->pCaptureSession, pSource, vVecs, 2)) 
    {
        fprintf(stderr, "Failed to submit demod record to capture session\n");
        rval = ENATSC3Status_Failure;
        goto out;
    }

out:
    return rval;
}

void
ATSC3BasebandRZRSDK2SourceBasebandPacketCallback(
    RedZoneCaptureBasebandPacket *pPacket, void *pUserData)
{
    ENATSC3BasebandRZRSDK2Source *pSource =
        (ENATSC3BasebandRZRSDK2Source *)pUserData;

    if (!pSource->fOpened)
    {
        fprintf(stderr, "Baseband packet on closed source?\n ");
        goto out;
    }

    if (pSource->pCaptureSession)
    {
        if (pSource->fDemodRecordChanged)
        {
            if (pthread_mutex_lock(&pSource->statsLock))
            {
                fprintf(stderr, "Failed to lock\n");
            }

            if (WriteDemodulatorRecord(pSource, &pSource->demodRecord,
                sizeof(ENATSC3CapDemodulatorRecordV2) + 
                sizeof(struct _ATSC3_L1_Detail_Subframe)))
            {
                fprintf(stderr, "Failed to write demod record\n");
            }

            pSource->fDemodRecordChanged = 0;

            if (pthread_mutex_unlock(&pSource->statsLock))
            {
                fprintf(stderr, "Failed to unlock\n");
            }
        }

        if (WriteBasebandRecord(pSource, pPacket))
        {
            fprintf(stderr, "Failed to write baseband packet\n");
        }
    }

out:
    return;
}

static ENATSC3Modulation
RZRToENTASC3Modulation(uint8_t mod)
{
    switch (mod)
    {
        case RZR_ATSC3_PLP_MOD_QPSK:
            return ENATSC3Modulation_QPSK;
        case RZR_ATSC3_PLP_MOD_16QAM:
            return ENATSC3Modulation_16QAM;
        case RZR_ATSC3_PLP_MOD_64QAM:
            return ENATSC3Modulation_64QAM;
        case RZR_ATSC3_PLP_MOD_256QAM:
            return ENATSC3Modulation_256QAM;
        case RZR_ATSC3_PLP_MOD_1024QAM:
            return ENATSC3Modulation_1024QAM;
        case RZR_ATSC3_PLP_MOD_4096QAM:
            return ENATSC3Modulation_4096QAM;
    }

    return ENATSC3Modulation_Unknown;
}

static uint8_t
RZRToENTASC3Coding(uint8_t cod)
{
    switch (cod)
    {
        case RZR_ATSC3_PLP_COD_2_15:
            return 2;
        case RZR_ATSC3_PLP_COD_3_15:
            return 3;
        case RZR_ATSC3_PLP_COD_4_15:
            return 4;
        case RZR_ATSC3_PLP_COD_5_15:
            return 5;
        case RZR_ATSC3_PLP_COD_6_15:
            return 6;
        case RZR_ATSC3_PLP_COD_7_15:
            return 7;
        case RZR_ATSC3_PLP_COD_8_15:
            return 8;
        case RZR_ATSC3_PLP_COD_9_15:
            return 9;
        case RZR_ATSC3_PLP_COD_10_15:
            return 10;
        case RZR_ATSC3_PLP_COD_11_15:
            return 11;
        case RZR_ATSC3_PLP_COD_12_15:
            return 12;
        case RZR_ATSC3_PLP_COD_13_15:
            return 13;
    }

    return 0;
}

static inline void
GetStatistics(ENATSC3BasebandRZRSDK2Source *pSource)
{
    struct _stats stats;
    int32_t fHaveMasterLock = 1;
    int32_t fHavePLPValidMask = 1;
    int32_t fHaveLLSValidMask = 1;
    int32_t fHaveSNR = 1;
    int32_t fHaveRSSI = 1;
    uint64_t plpStatsMask = 0;
    struct RedZonePLPSpecificInfo3 vPLPInfo[64];
    uint64_t mask;
    int32_t i;

    if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
        RedZoneMasterLockProp, &stats.masterLock,
        sizeof(&stats.masterLock) != RZR_SUCCESS))
    {
        fHaveMasterLock = 0;
        fprintf(stderr, "Failed to get master lock\n");
    }

    if (stats.masterLock)
    {
        if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
            RedZonePLPValidBitmaskProp, &stats.plpValidMask, 
            sizeof(stats.plpValidMask)) != RZR_SUCCESS)
        {
            fHavePLPValidMask = 0;
            fprintf(stderr, "Failed to get plp valid mask\n");
        }

        if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
            RedZoneLLSValidBitmaskProp, &stats.llsValidMask, 
            sizeof(stats.llsValidMask)) != RZR_SUCCESS)
        {
            fHaveLLSValidMask = 0;
            fprintf(stderr, "Failed to get lls plps mask\n");
        }

        if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
            RedZoneSNRProp, &stats.snr,
            sizeof(stats.snr)) != RZR_SUCCESS)
        {
            fHaveSNR = 0;
            fprintf(stderr, "Failed to get snr\n");
        }
        if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
            RedZoneRSSIProp, &stats.rssi,
            sizeof(stats.rssi)) != RZR_SUCCESS)
        {
            fHaveSNR = 0;
            fprintf(stderr, "Failed to get rssi\n");
        }

        for (i = 0, mask = 1; i < 64; i++, mask <<= 1)
        {
            if (stats.plpValidMask & mask)
            {
                vPLPInfo[i].index = 255;
                vPLPInfo[i].plp_id = i;
                if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
                    RedZonePLPSpecificInfoProp, vPLPInfo + i, 
                    sizeof(vPLPInfo[0])))
                {
                    fprintf(stderr, "Failed to get plp %d info\n", i);
                    continue;
                }
            }

            plpStatsMask |= mask;
        }
    }
    else
    {
        memset(&stats, 0, sizeof(stats));
    }

    RedZoneCaptureStatistics statistics;
    RedZoneCaptureGetStatistics(pSource->hRedZoneCapture, &statistics);

    printf("lock: %d snr: %d rssi: %d plp: 0x%"PRIX64" lls: 0x%"PRIX64" Baseband frames: %"PRId64" RF errors: %"PRId64" Continuity errors: %"PRId64" Framing errors:%"PRId64"\n",
        stats.masterLock, stats.snr, stats.rssi, stats.plpValidMask, 
        stats.llsValidMask, statistics.nPackets, 
        statistics.nUncorrectedRFErrors,
        statistics.nContinuityCounterErrors, statistics.nFramingErrors);

    if (pthread_mutex_lock(&pSource->statsLock))
    {
        fprintf(stderr, "Failed to lock\n");
    }

    if (fHaveMasterLock)
    {
       pSource->stats.masterLock = stats.masterLock;
       pSource->demodRecord.lock = stats.masterLock;
    }

    if (fHavePLPValidMask)
    {
        pSource->stats.plpValidMask = stats.plpValidMask;
        pSource->demodRecord.atsc3.plpPresentMask = 
            ENUtils_htonll(stats.plpValidMask);
        // XXX We're setting the plpEnabled and plpLocked masks to be the 
        // same as the valid mask
        pSource->demodRecord.atsc3.plpEnabledMask = 
            pSource->demodRecord.atsc3.plpPresentMask; 
        pSource->demodRecord.atsc3.plpLockedMask = 
            pSource->demodRecord.atsc3.plpPresentMask; 
    }

    if (fHaveLLSValidMask)
    {
        pSource->stats.llsValidMask = stats.llsValidMask;
        pSource->demodRecord.atsc3.llsPresentMask = 
            ENUtils_htonll(stats.llsValidMask);
    }

    if (fHaveSNR)
    {
        pSource->stats.snr = stats.snr;
        pSource->demodRecord.snr = htonl(stats.snr);
    }

    if (fHaveRSSI)
    {
        pSource->stats.rssi = stats.rssi;
        pSource->demodRecord.rssi = htonl(stats.rssi);
    }

    for (i = 0, mask = 1; i < 64; i++, mask <<= 1)
    {
        if (plpStatsMask & mask)
        {
            int32_t j;

            pSource->stats.vPLPStats[i].modulation = 
                RZRToENTASC3Modulation(vPLPInfo[i].mod);
            pSource->stats.vPLPStats[i].coding = 
                RZRToENTASC3Coding(vPLPInfo[i].cod);

            // NOTE we're only populating subframe 0
            struct _ATSC3_L1_Detail_Subframe *pSubframe =
                pSource->demodRecord.atsc3.l1_detail_signaling.vSubframes;
            struct _ATSC3_L1_Detail_signaling_subframe_plp *pSubframePLP = 
                pSubframe->vPLPs + i;
            pSubframePLP->L1D_plp_id = i;
            pSubframePLP->L1D_plp_lls_flag = vPLPInfo[i].lls_present;
            pSubframePLP->L1D_plp_layer = vPLPInfo[i].layer;
            pSubframePLP->L1D_plp_start = htonl(vPLPInfo[i].start);
            pSubframePLP->L1D_plp_size = htonl(vPLPInfo[i].size);
            pSubframePLP->L1D_plp_scrambler_type = vPLPInfo[i].scrambler_type;
            pSubframePLP->L1D_plp_fec_type = vPLPInfo[i].fec_type;
            pSubframePLP->L1D_plp_mod = vPLPInfo[i].mod;
            pSubframePLP->L1D_plp_cod = vPLPInfo[i].cod;
            pSubframePLP->L1D_plp_TI_mode = vPLPInfo[i].ti_mode;
            pSubframePLP->L1D_plp_fec_block_start = 
                htonl(vPLPInfo[i].fec_block_start);
            pSubframePLP->L1D_plp_CTI_fec_block_start = 
                htonl(vPLPInfo[i].cti_fec_block_start);
            pSubframePLP->L1D_plp_num_channel_bonded = 
                vPLPInfo[i].num_ch_bonded;
            pSubframePLP->vL1D_plp_bonded_rf_id[0] = 
                vPLPInfo[i].bonded_rf_id_0;
            pSubframePLP->vL1D_plp_bonded_rf_id[1] = 
                vPLPInfo[i].bonded_rf_id_1;
            pSubframePLP->L1D_plp_mimo_stream_combining = 
                vPLPInfo[i].mimo_stream_combine;
            pSubframePLP->L1D_plp_mimo_IQ_interleaving = 
                vPLPInfo[i].mimo_iq_interleave;
            pSubframePLP->L1D_plp_mimo_PH = vPLPInfo[i].mimo_ph;
            pSubframePLP->L1D_plp_type = vPLPInfo[i].plp_type;
            pSubframePLP->L1D_plp_num_subslices = 
                htons(vPLPInfo[i].num_subslice);
            pSubframePLP->L1D_plp_subslice_interval = 
                htonl(vPLPInfo[i].subslice_interval);
            pSubframePLP->L1D_plp_TI_extended_interleaving = 
                vPLPInfo[i].ti_ext_interleave;
            pSubframePLP->L1D_plp_CTI_depth = vPLPInfo[i].cti_depth;
            pSubframePLP->L1D_plp_CTI_start_row = 
                htons(vPLPInfo[i].cti_start_row);
            pSubframePLP->L1D_plp_HTI_inter_subframe = 
                vPLPInfo[i].hti_inter_subframe;
            pSubframePLP->L1D_plp_HTI_num_ti_blocks = 
                vPLPInfo[i].hti_num_ti_blocks;
            pSubframePLP->L1D_plp_HTI_num_fec_blocks_max = 
                htons(vPLPInfo[i].hti_num_fec_block_max);
            pSubframePLP->L1D_plp_HTI_num_fec_blocks = htons(16);
            for (j = 0; j < 16; j++)
            {
                pSubframePLP->vL1D_plp_HTI_num_fec_blocks[j] = 
                    htons(vPLPInfo[j].hti_num_fec_block[j]);
            }
            pSubframePLP->L1D_plp_HTI_cell_interleaver = 
                vPLPInfo[i].hti_cell_interleave;
            pSubframePLP->L1D_plp_ldm_injection_level = 
                vPLPInfo[i].ldm_inj_level;
        }
    }

    pSource->fDemodRecordChanged = 1;

    if (pthread_mutex_unlock(&pSource->statsLock))
    {
        fprintf(stderr, "Failed to unlock\n");
    }
}

ENATSC3Status
ENATSC3BasebandRZRSDK2SourceRun(ENATSC3BasebandRZRSDK2Source *pSource, 
    char *pCaptureFilename, int32_t *pFRunning, int32_t duration)
{
    int32_t i;
    int32_t j;
    struct timespec tNextStats = {.tv_sec = 0, .tv_nsec = 0};
    struct timespec tStop = {.tv_sec = 0x7fffffffffffffff, .tv_nsec = 0};
    ENATSC3Status rval = ENATSC3Status_Success;

    memset(&pSource->stats, 0, sizeof(pSource->stats));
    memset(&pSource->demodRecord, 0, sizeof(pSource->demodRecord));
    pSource->demodRecord.frequencyHz = htonl(pSource->frequencyKHz * 1000);
    for (i = 0; i < numof(pSource->stats.vPLPStats); i++)
    {
        pSource->stats.vPLPStats[i].lastContinuityCounter = -1;
    }

    if (CaptureSessionCreate(pCaptureFilename, &pSource->pCaptureSession))
    {
        fprintf(stderr, "Failed to create capture session\n");
    }

    if (RedZoneCaptureStart(pSource->hRedZoneCapture))
    {
        fprintf(stderr, "Failed to start RedZoneCapture\n");
        rval = ENATSC3Status_Failure;
        goto out;
    }

    if (duration > 0)
    {
        clock_gettime(CLOCK_MONOTONIC, &tStop);
        tStop.tv_sec += duration;
    }

    while (*pFRunning)
    {
        struct timespec mononow;
        usleep(100000);

        clock_gettime(CLOCK_MONOTONIC, &mononow);
        if (mononow.tv_sec > tStop.tv_sec ||
            (mononow.tv_sec == tStop.tv_sec &&
             mononow.tv_nsec >= tStop.tv_nsec))
        {
            break;
        }

        if (mononow.tv_sec > tNextStats.tv_sec ||
            (mononow.tv_sec == tNextStats.tv_sec &&
             mononow.tv_nsec >= tNextStats.tv_nsec))
        {
            GetStatistics(pSource);
            tNextStats = mononow;
            tNextStats.tv_sec += 1;
        }

        if (pSource->fPLPIdMaskChanged)
        {
            uint64_t plpIdMask;
            uint64_t plpIdMaskTemp;

            // We're going to snapshot the mask since processing the change 
            // could take a while and we don't want to hold the lock that
            // long. If someone changes the mask again we'll catch it the next
            // time around.
            if (pthread_mutex_lock(&pSource->lock))
            {
                fprintf(stderr, "Failed to lock\n");
            }

            plpIdMask = pSource->plpIdMask;
            pSource->fPLPIdMaskChanged = 0;

            if (pthread_mutex_unlock(&pSource->lock))
            {
                fprintf(stderr, "Failed to unlock\n");
            }

            // First reconcile the current PLPs with the mask.
            for (i = 0; i < numof(pSource->vPLPIds); i++)
            {
                if (pSource->vPLPIds[i] != 255)
                {
                    if (plpIdMask & 1ULL << pSource->vPLPIds[i])
                    {
                        // Already set, remove from mask
                        plpIdMask &= ~1ULL << pSource->vPLPIds[i];
                    }
                    else
                    {
                        // No longer set, free up slot
                        pSource->vPLPIds[i] = 255;
                    }
                }
            }

            // Next assign new PLPs to slots
            for (i = 0, plpIdMaskTemp = plpIdMask; i < 64; i++)
            {
                if (plpIdMaskTemp & 1)
                {
                    for (j = 0; j < numof(pSource->vPLPIds); j++)
                    {
                        if (pSource->vPLPIds[j] == 255)
                        {
                            pSource->vPLPIds[j] = i;
                            break;
                        }
                    }

                    if (j == numof(pSource->vPLPIds))
                    {
                        fprintf(stderr, "Too many plps: couldn't set %d, mask %"PRIx64", slots %d %d %d %d\n", 
                            i, plpIdMask, pSource->vPLPIds[0], 
                            pSource->vPLPIds[1], pSource->vPLPIds[2], 
                            pSource->vPLPIds[3]);
                    }
                }

                plpIdMaskTemp >>=1;
            }

            printf("Setting PLP mask 0x%"PRIx64" slots %d %d %d %d\n",
                plpIdMask, pSource->vPLPIds[0], pSource->vPLPIds[1], 
                pSource->vPLPIds[2], pSource->vPLPIds[3]); 

            RedZonePLPSet val;

            val.plp0_id = pSource->vPLPIds[0];
            val.plp1_id = pSource->vPLPIds[1];
            val.plp2_id = pSource->vPLPIds[2];
            val.plp3_id = pSource->vPLPIds[3];

            if (RedZoneCaptureSetProp(pSource->hRedZoneCapture,
                RedZonePLPSelectionProp, &val, sizeof(val)))
            {
                fprintf(stderr, "Failed to set PLP selection\n");
            }
        }
    }

    if (RedZoneCaptureStop(pSource->hRedZoneCapture))
    {
        fprintf(stderr, "Failed to stop RedZoneCapture\n");
        rval = ENATSC3Status_Failure;
        goto out;
    }

    if (pSource->pCaptureSession)
    {
        if (CaptureSessionDestroy(pSource->pCaptureSession))
        {
            fprintf(stderr, "Failed to destroy capture ssesion\n");
        }
        pSource->pCaptureSession = NULL;
    }

out:
    return rval;
}

ENATSC3Status
ENATSC3BasebandRZRSDK2SourceOpen(int32_t deviceNumber, 
    ENATSC3BasebandRZRSDK2SourceOperatingMode operatingMode, 
    int32_t frequencyHz, ENATSC3BasebandRZRSDK2Source **ppSource)
{
    ENATSC3BasebandRZRSDK2Source *pSource;
    ENATSC3Status rval = ENATSC3Status_Success;

    if (deviceNumber >= atsc3BasebandRZRSDK2SourceGlobalsG.nSources)
    {
        fprintf(stderr, "Source %d out of range %ld\n", deviceNumber,
            numof(atsc3BasebandRZRSDK2SourceGlobalsG.vSources));
        rval = ENATSC3Status_Failure;
        goto out;
    }

    if (pthread_mutex_lock(&atsc3BasebandRZRSDK2SourceGlobalsG.lock))
    {
        fprintf(stderr, "Failed to lock\n");
    }

    pSource = atsc3BasebandRZRSDK2SourceGlobalsG.vSources + deviceNumber;

    if (pSource->fOpened)
    {
        fprintf(stderr, "Device %d already opened\n", deviceNumber);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    pSource->frequencyKHz = frequencyHz/1000 + 3000;

    // Set all PLPs invalid. Note that we will actually tune the first slot
    // to PLP 0, since one PLP must be set when initially tuning and 0 seems
    // to always be present in emissions.
    pSource->vPLPIds[0] = 255;
    pSource->vPLPIds[1] = 255;
    pSource->vPLPIds[2] = 255;
    pSource->vPLPIds[3] = 255;

    if (operatingMode == ENATSC3BasebandRZRSDK2SourceOperatingMode_ATSC)
    {
        pSource->operatingMode = RedZoneOperatingModeATSC;
    }
    else if (operatingMode == ENATSC3BasebandRZRSDK2SourceOperatingMode_ATSC3)
    {
        pSource->operatingMode = RedZoneOperatingModeATSC3;
    }
    else
    {
        fprintf(stderr, "Invalid operating mode %d\n", operatingMode);
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    if (RedZoneCaptureSetProp(pSource->hRedZoneCapture, 
        RedZoneOperatingModeProp, &pSource->operatingMode, 
        sizeof(pSource->operatingMode)))
    {
        fprintf(stderr, "Failed to set RedZoneOperatingModeProp to %d\n",
            pSource->operatingMode);
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    if (RedZoneCaptureSetProp(pSource->hRedZoneCapture, 
        RedZoneFrequencyKHzProp, &pSource->frequencyKHz, 
        sizeof(pSource->frequencyKHz)))
    {
        fprintf(stderr, "Failed to set frequency %d KHz\n", pSource->frequencyKHz);
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    RedZonePLPSet val;

    val.plp0_id = 0;
    val.plp1_id = 255;
    val.plp2_id = 255;
    val.plp3_id = 255;

    if (RedZoneCaptureSetProp(pSource->hRedZoneCapture, 
        RedZonePLPSelectionProp, &val, sizeof(val)))
    {
        fprintf(stderr, "Failed to set PLP selection\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    if (RedZoneCaptureRegisterCallbacks(pSource->hRedZoneCapture,
        ATSC3BasebandRZRSDK2SourceBasebandPacketCallback, pSource))
    {
        fprintf(stderr, "Failed to register baseband packet callback\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    pSource->fOpened = 1;
    pSource->plpIdMask = 0;
    pSource->fPLPIdMaskChanged = 0;


    *ppSource = pSource;

out:
    if (pthread_mutex_unlock(&atsc3BasebandRZRSDK2SourceGlobalsG.lock))
    {
        fprintf(stderr, "Failed to unlock\n");
    }

    return rval;

errout:
    goto out;
}

ENATSC3Status
ENATSC3BasebandRZRSDK2SourceClose(ENATSC3BasebandRZRSDK2Source *pSource)
{
    ENATSC3Status rval = ENATSC3Status_Success;

    if (pthread_mutex_lock(&atsc3BasebandRZRSDK2SourceGlobalsG.lock))
    {
        fprintf(stderr, "Failed to lock\n");
    }

    pSource->fOpened = 0;

    if (pthread_mutex_unlock(&atsc3BasebandRZRSDK2SourceGlobalsG.lock))
    {
        fprintf(stderr, "Failed to unlock\n");
    }

    return rval;
}

ENATSC3Status
ENATSC3BasebandRZRSDK2SourceSetPLPIds(ENATSC3BasebandRZRSDK2Source *pSource, 
    uint64_t plpIdMask)
{
    ENATSC3Status rval = ENATSC3Status_Success;

    if (pthread_mutex_lock(&pSource->lock))
    {
        fprintf(stderr, "Failed to lock\n");
    }

    pSource->plpIdMask = plpIdMask;
    pSource->fPLPIdMaskChanged = 1;

    if (pthread_mutex_unlock(&pSource->lock))
    {
        fprintf(stderr, "Failed to unlock\n");
    }

    return rval;
}

ENATSC3Status
ENATSC3BasebandRZRSDK2SourceInitialize(void)
{
    int32_t fHaveLock = 0;
    RedZoneCaptureHandle_t hRedZoneCapture = NULL;
    ENATSC3BasebandRZRSDK2Source *pSource;
    int32_t i;
    ENATSC3Status rval = ENATSC3Status_Success;

    atsc3BasebandRZRSDK2SourceGlobalsG.nSources = 0;
    atsc3BasebandRZRSDK2SourceGlobalsG.pCapturePath = 
        getenv("ENATSC3_CAPTURE_PATH");

    if (pthread_mutex_init(&atsc3BasebandRZRSDK2SourceGlobalsG.lock, NULL))
    {
        fprintf(stderr, "Failed to init lock\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }
    fHaveLock = 1;

    pSource = atsc3BasebandRZRSDK2SourceGlobalsG.vSources;
    atsc3BasebandRZRSDK2SourceGlobalsG.nSources = 0;

    while (1)
    {
        int32_t rzrret;

        if (atsc3BasebandRZRSDK2SourceGlobalsG.nSources ==
            numof(atsc3BasebandRZRSDK2SourceGlobalsG.vSources))
        {
            fprintf(stderr, "Too many sources\n");
            break;
        }

        rzrret = RedZoneCaptureOpen(&hRedZoneCapture);
        if (rzrret)
        {
            fprintf(stderr, "RedZoneCaptureOpen failed %d\n", rzrret);
            rval = ENATSC3Status_Failure;
            goto errout;
        }

        rzrret = RedZoneCaptureInit(hRedZoneCapture);
        if (rzrret)
        {
            fprintf(stderr, "Failed to init RedZoneCapture %d\n", rzrret);
            if (RedZoneCaptureClose(hRedZoneCapture))
            {
                fprintf(stderr, "Failed to close RedZoneCapture\n");
            }
            rval = ENATSC3Status_Failure;
            goto errout;
        }

        pSource->hRedZoneCapture = hRedZoneCapture;
#if 0
        if (RedZoneCaptureGetProp(hRedZoneCapture, 
            RedZoneModelNumberStringProp, pSource->strModel,
            sizeof(pSource->strModel) != RZR_SUCCESS))
        {
            fprintf(stderr, "Failed to get model\n");
            pSource->strModel[0] = 0;
        }
        if (RedZoneCaptureGetProp(hRedZoneCapture, 
            RedZoneSerialNumberStringProp, pSource->strSerial,
            sizeof(pSource->strSerial) != RZR_SUCCESS))
        {
            fprintf(stderr, "Failed to get serial\n");
            pSource->strSerial[0] = 0;
        }
#else
        uint32_t model;
        if (RedZoneCaptureGetProp(hRedZoneCapture, 
            RedZoneModelNumberProp, &model,
            sizeof(model) != RZR_SUCCESS))
        {
            fprintf(stderr, "Failed to get model\n");
            pSource->strModel[0] = 0;
        }
        else
        {
            snprintf(pSource->strModel, sizeof(pSource->strModel),
                "RZR-%d", model);
            pSource->strModel[sizeof(pSource->strModel) - 1] = 0;

        }
        pSource->strSerial[0] = 0;
#endif

        pSource->fOpened = 0;
        if (pthread_mutex_init(&pSource->lock, NULL))
        {
            fprintf(stderr, "Failed to init lock\n");
            if (RedZoneCaptureClose(hRedZoneCapture))
            {
                fprintf(stderr, "Failed to close RedZoneCapture\n");
            }
            rval = ENATSC3Status_Failure;
            goto errout;
        }

        if (pthread_mutex_init(&pSource->statsLock, NULL))
        {
            fprintf(stderr, "Failed to init lock\n");
            if (pthread_mutex_destroy(&pSource->lock))
            {
                fprintf(stderr, "Failed to destroy lock\n");
            }
            if (RedZoneCaptureClose(hRedZoneCapture))
            {
                fprintf(stderr, "Failed to close RedZoneCapture\n");
            }
            rval = ENATSC3Status_Failure;
            goto errout;
        }

        atsc3BasebandRZRSDK2SourceGlobalsG.nSources++;
        pSource++;

        break;
    }

out:
    return rval;

errout:
    for (i = 0, pSource = atsc3BasebandRZRSDK2SourceGlobalsG.vSources;
         i < atsc3BasebandRZRSDK2SourceGlobalsG.nSources;
         i++, pSource++)
    {
        if (pthread_mutex_destroy(&pSource->statsLock))
        {
            fprintf(stderr, "Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->lock))
        {
            fprintf(stderr, "Failed to destroy lock\n");
        }
        if (RedZoneCaptureClose(pSource->hRedZoneCapture))
        {
            fprintf(stderr, "Failed to close RedZoneCapture\n");
            pSource->hRedZoneCapture = NULL;
        }
    }

    if (fHaveLock)
    {
        if (pthread_mutex_destroy(&atsc3BasebandRZRSDK2SourceGlobalsG.lock))
        {
            fprintf(stderr, "Failed to destroy lock\n");
        }
    }

    goto out;
}

ENATSC3Status
ENATSC3BasebandRZRSDK2SourceShutdown(void)
{
    ENATSC3BasebandRZRSDK2Source *pSource;
    int32_t i;
    ENATSC3Status rval = ENATSC3Status_Success;

    for (i = 0, pSource = atsc3BasebandRZRSDK2SourceGlobalsG.vSources; 
         i < atsc3BasebandRZRSDK2SourceGlobalsG.nSources; 
         i++, pSource++)
    {
        if (RedZoneCaptureClose(pSource->hRedZoneCapture))
        {
            fprintf(stderr, "Failed to close RedZoneCapture\n");
        }
        pSource->hRedZoneCapture = NULL;

        if (pthread_mutex_destroy(&pSource->statsLock))
        {
            fprintf(stderr, "Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->lock))
        {
            fprintf(stderr, "Failed to destroy lock\n");
        }
    }

    if (pthread_mutex_destroy(&atsc3BasebandRZRSDK2SourceGlobalsG.lock))
    {
        fprintf(stderr, "Failed to destroy lock\n");
    }

    return rval;
}

int32_t
parseArgs(int argc, char **argv, int32_t *pFListDevices, char **ppOutFilename,
    int32_t *pFrequencyHz, int32_t *pDeviceNumber, uint64_t *pPLPMask, 
    int32_t *pDuration)
{
    int32_t err = 0;
    int32_t i;

    *pFListDevices = 0;
    *ppOutFilename = NULL;
    *pFrequencyHz = -1;
    *pDeviceNumber = -1;
    *pPLPMask = 0;
    *pDuration = -1;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-l"))
        {
            *pFListDevices = 1;
        }
        else if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutFilename = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-f"))
        {
            if (i + 1 < argc)
            {
                *pFrequencyHz = atoll(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-device"))
        {
            if (i + 1 < argc)
            {
                *pDeviceNumber = atoll(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-duration"))
        {
            if (i + 1 < argc)
            {
                *pDuration = atoll(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-plp"))
        {
            if (i + 1 < argc)
            {
                int plp = atoll(argv[i+1]);
                if (plp < 0 || plp > 63)
                {
                    fprintf(stderr, "PLP %d out of range\n", plp);
                    err = 1;
                }
                else
                {
                    *pPLPMask |= ((uint64_t)1) << atoll(argv[i+1]);
                }
                i++;
            }
            else err = 1;
        }
        else err = 1;
    }

    if (!*pFListDevices && 
        (!*ppOutFilename || *pFrequencyHz == -1 || *pDeviceNumber == -1))
    {
        fprintf(stderr, "Usage: %s -o outfilename -f frequency -plp plp [-plp plp] [-device device] [-l] [-duration seconds\n", argv[0]);
        fprintf(stderr, "    -plp: specifies the PLPs to tune to in the range 0 - 63.\n"); 
        fprintf(stderr, "          Multiple PLPs may be specified.\n");
        fprintf(stderr, "    -l: list devices\n");
        fprintf(stderr, "    -device: optionally specify device to capture from. If no device is specified\n");
        fprintf(stderr, "     the first device is used.\n"); 
        fprintf(stderr, "    -duration: optionally specify the duration of the capture. If no duration is\n");
        fprintf(stderr, "     specified the capture continues until terminated by ctrl-c\n");
        err = 1;
    }

    return err;
}

int32_t fRunningG;

void
sighandler(int signum)
{
    if (signum == SIGINT)
    {
        fRunningG = 0;
    }
}

int
main(int argc, char *argv[])
{
    int32_t fListDevices = 0;
    int32_t frequencyHz;
    int32_t deviceNumber;
    int32_t duration;
    char *pOutFilename;
    int32_t fInitialized = 0;
    ENATSC3BasebandRZRSDK2Source *pSource = NULL;
    struct sigaction act;
    uint64_t plpMask;
    int32_t i;
    int32_t rval = 0;

    if (parseArgs(argc, argv, &fListDevices, &pOutFilename, &frequencyHz, 
        &deviceNumber, &plpMask, &duration))
    {
        rval = 1;
        goto out;
    }

    if (ENATSC3BasebandRZRSDK2SourceInitialize())
    {
        fprintf(stderr, "Failed to initialize\n");
        rval = 1;
        goto out;
    }
    fInitialized = 1;

    if (fListDevices)
    {
        printf("Devices:\n");
        for (i = 0; i < atsc3BasebandRZRSDK2SourceGlobalsG.nSources; i++)
        {
            printf("  %d: %s / %s\n", i,
                atsc3BasebandRZRSDK2SourceGlobalsG.vSources[i].strModel,
                atsc3BasebandRZRSDK2SourceGlobalsG.vSources[i].strSerial);
        }

        goto out;
    }

    // Graceful exit on ctrl-c
    act.sa_handler = sighandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    sigaction(SIGINT, &act, NULL);

    if (ENATSC3BasebandRZRSDK2SourceOpen(deviceNumber,
        ENATSC3BasebandRZRSDK2SourceOperatingMode_ATSC3, frequencyHz,
        &pSource))
    {
        fprintf(stderr, "Failed to open device %d\n", deviceNumber);
        rval = 1;
        goto out;
    }

    if (ENATSC3BasebandRZRSDK2SourceSetPLPIds(pSource, plpMask))
    {
        fprintf(stderr, "Failed to set PLPs\n");
        rval = 1;
        goto out;
    }

    fRunningG = 1;
    if (ENATSC3BasebandRZRSDK2SourceRun(pSource, pOutFilename, &fRunningG, 
        duration))
    {
        fprintf(stderr, "Failed to start capture\n");
        rval = 1;
        goto out;
    }

out:
    if (pSource)
    {
        if (ENATSC3BasebandRZRSDK2SourceClose(pSource))
        {
            fprintf(stderr, "Failed to close device\n");
        }
    }

    if (fInitialized)
    {
        if (ENATSC3BasebandRZRSDK2SourceShutdown())
        {
            fprintf(stderr, "Failed to shut down\n");
        }
    }

    return rval;
}
