/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <arpa/inet.h>

#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>

#include "enatsc3cap.h"

#define STDOUT 1

int
parseArgs(int argc, char **argv, char **ppInFile, char **ppOutFile)
{
    int err = 0;
    int i;

    *ppInFile = NULL;
    *ppOutFile = NULL;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-b"))
        {
            if (i + 1 < argc)
            {
                *ppInFile = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutFile = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else err = 1;
     }

    if (!*ppInFile) err = 1;
    if (!*ppOutFile) err = 1;

    if (err)
    {
        fprintf(stderr, "Usage: %s -b infile -o outfile\n", argv[0]);
    }

    return err;
}

int
main(int argc, char **argv)
{
    char *pInFilename;
    char *pOutFilename;
    int ifd = -1;
    int ofd = -1;
    off_t pos = 0;
    ENATSC3CapFileHeader fileHeader;
    ENATSC3CapRecordHeaderV2 headerV2;
    char buf[4096];
    int rval = 0;

    if (parseArgs(argc, argv, &pInFilename, &pOutFilename))
    {
        rval = 1;
	goto out;
    }

    ifd = open(pInFilename, O_RDONLY);
    if (ifd < 0)
    {
        fprintf(stderr, "Failed to open %s for reading (%s)\n", pInFilename, 
	    strerror(errno));
	rval = 1;
	goto out;
    }

    if (!strcmp(pOutFilename, "-"))
    {
        ofd = STDOUT;
    }
    else
    {
        ofd = open(pOutFilename, O_CREAT | O_WRONLY | O_TRUNC, 0666);
        if (ofd < 0)
        {
            fprintf(stderr, "Failed to open %s for writing (%s)\n", pOutFilename, 
                strerror(errno));
            rval = 1;
            goto out;
        }
    }

    if (read(ifd, &fileHeader, sizeof(fileHeader)) != sizeof(fileHeader))
    {
        fprintf(stderr, "Failed to read file header: %s\n", strerror(errno));
        rval = 1;
        goto out;
    }

    if (fileHeader.format != ENATSC3CapFormat_V2)
    {
        fprintf(stderr, "Unsupported format %d\n", fileHeader.format);
        rval = 1;
        goto out;
    }

    if (ntohs(fileHeader.magic) != ENATSC3CAP_HEADER_MAGIC)
    {
        fprintf(stderr, "File header has bad magic 0x%.4x\n",
            ntohs(fileHeader.magic));
        rval = 1;
        goto out;
    }

    while (rval == 0)
    {
        off_t startpos = pos;
        int nb;

	nb = read(ifd, &headerV2, sizeof(headerV2));
	if (nb == 0)
	{
	    break;
	}
	else if (nb != sizeof(headerV2))
	{
	    fprintf(stderr, "Failed to read record header: %s\n",
	        strerror(errno));
	    rval = 1;
	    break;
	}

        pos += sizeof(headerV2);

	headerV2.length = ntohl(headerV2.length);
	headerV2.monowhen.sec = ENUtils_ntohll(headerV2.monowhen.sec);
	headerV2.monowhen.nsec = ntohl(headerV2.monowhen.nsec);
	headerV2.rtwhen.sec = ENUtils_ntohll(headerV2.rtwhen.sec);
	headerV2.rtwhen.nsec = ntohl(headerV2.rtwhen.nsec);
        headerV2.broadcastTime.L1D_time_sec = 
            ntohl(headerV2.broadcastTime.L1D_time_sec);
        headerV2.broadcastTime.L1D_time_msec = 
            ntohl(headerV2.broadcastTime.L1D_time_msec);
        headerV2.broadcastTime.L1D_time_usec = 
            ntohl(headerV2.broadcastTime.L1D_time_usec);
        headerV2.broadcastTime.L1D_time_nsec = 
            ntohl(headerV2.broadcastTime.L1D_time_nsec);
	switch (headerV2.type)
	{
            case ENATSC3CapPayloadType_ATSC1_MPEGTS:
                fprintf(stderr, "%"PRId64" MPEGTS: (%d bytes @ rt: %"PRId64".%06ds mono: %"PRId64".%06ds\n",
                    startpos, headerV2.length,
                    headerV2.rtwhen.sec, headerV2.rtwhen.nsec / 1000,
                    headerV2.monowhen.sec, headerV2.monowhen.nsec / 1000);
                break;
            case ENATSC3CapPayloadType_ATSC3_J288ALP:
                fprintf(stderr, "%"PRId64" J288-ALP: (%d bytes @ rt: %"PRId64".%06ds mono: %"PRId64".%06ds\n",
                    startpos, headerV2.length,
                    headerV2.rtwhen.sec, headerV2.rtwhen.nsec / 1000,
                    headerV2.monowhen.sec, headerV2.monowhen.nsec / 1000);
                break;
            case ENATSC3CapPayloadType_ATSC3_Baseband:
                fprintf(stderr, "%"PRId64" baseband: plp: %d (%d bytes @ %"PRId64".%06ds / %"PRId64".%06ds ptp: %u %u %u %u\n",
                    startpos, headerV2.plpId, headerV2.length,
                    headerV2.rtwhen.sec, headerV2.rtwhen.nsec / 1000,
                    headerV2.monowhen.sec, headerV2.monowhen.nsec / 1000,
                    ntohl(headerV2.broadcastTime.L1D_time_sec),
                    ntohl(headerV2.broadcastTime.L1D_time_msec),
                    ntohl(headerV2.broadcastTime.L1D_time_usec),
                    ntohl(headerV2.broadcastTime.L1D_time_nsec));
                break;
            case ENATSC3CapPayloadType_ATSC3_ALP:
                fprintf(stderr, "%"PRId64" ALP: plp: %d (%d bytes @ %"PRId64".%06ds / %"PRId64".%06ds ptp: %u %.03u %.03u %.03u\n",
                    startpos, headerV2.plpId, headerV2.length,
                    headerV2.rtwhen.sec, headerV2.rtwhen.nsec / 1000,
                    headerV2.monowhen.sec, headerV2.monowhen.nsec / 1000,
                    htonl(headerV2.broadcastTime.L1D_time_sec),
                    htonl(headerV2.broadcastTime.L1D_time_msec),
                    htonl(headerV2.broadcastTime.L1D_time_usec),
                    htonl(headerV2.broadcastTime.L1D_time_nsec));
                break;
            case ENATSC3CapPayloadType_Demod_V2:
                fprintf(stderr, "%"PRId64" DEMOD V2: (%d bytes @ %"PRId64".%06ds / %"PRId64".%06ds\n",
                    startpos, headerV2.length,
                    headerV2.rtwhen.sec, headerV2.rtwhen.nsec / 1000,
                    headerV2.monowhen.sec, headerV2.monowhen.nsec / 1000);
                break;
            default:
                fprintf(stderr, "UNHANDLED record type %d\n", headerV2.type);
                break;
	}

	while (headerV2.length)
	{
	    int nToRead;

	    nToRead = MIN(headerV2.length, sizeof(buf));
	    nb = read(ifd, buf, nToRead);
	    if (nb != nToRead)
	    {
	        fprintf(stderr, "Failed to read payload: %s\n", 
		    strerror(errno));
		rval = 1;
		break;
	    }

            if (headerV2.type == ENATSC3CapPayloadType_Demod_V2)
            {
                ENATSC3CapDemodulatorRecordV2 *pDemodRecordV2;

                pDemodRecordV2 = (ENATSC3CapDemodulatorRecordV2 *)buf;
                fprintf(stderr, "  freq %u plppresent %"PRIx64" llspresent %"PRIx64" plpLocked %"PRIx64"\n",
                    ntohl(pDemodRecordV2->frequencyHz),
                    ENUtils_ntohll(pDemodRecordV2->atsc3.plpPresentMask),
                    ENUtils_ntohll(pDemodRecordV2->atsc3.llsPresentMask),
                    ENUtils_ntohll(pDemodRecordV2->atsc3.plpLockedMask));
            }
	    else
	    {
		nb = write(ofd, buf, nToRead);
		if (nb != nToRead)
		{
		    fprintf(stderr, "Failed to write payload: %s\n", 
		        strerror(errno));
		    rval = 1;
		    break;
		}
	    }

            pos += nToRead;
	    headerV2.length -= nToRead;
	}
    }

out:
    if (ifd >= 0)
    {
        close(ifd);
    }

    if (ofd >= 0 && ofd != STDOUT)
    {
        close(ofd);
    }

    return rval;
}
