/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <libenatsc3utils/enatsc3bitratestatistics.h>

#include <enatsc3-common/enatsc3.h>
#include "libenatsc3baseband.h"

static char *ensys = "bbp";
static char *ensubsys = "main";

#define INITIAL_ALP_PACKET_LENGTH 4096
#define MAX_ALP_PACKET_LENGTH (65536 + 1024)

// AUTODETECT_KOREAN_MODE attempts to detect if the ATSC 3 broadcast is 
// according to the 2016 specs, which was adopted in Korea. In "Korean Mode"
// ALP packets containing signaling information are formatted differently.
// In particular, the Additional Header for Signaling Information is treated
// as part of the payload in Korean Mode but as a header (and not part of the
// payload) as of ATSC 2018.
// #define AUTODETECT_KOREAN_MODE 1

typedef struct 
{
    int32_t fEnabled;
    int32_t ALPLength;
    int32_t nALPBytesReceived;

    // UTC time of the first baseband packet comprising the ALP packet
    struct timespec packetTimeUTC;
    // Baseband packet number of the first baseband packet comprising the ALP
    // packet
    uint64_t packetNumber;
    // If any baseband packet comprising the ALP packet contained a 
    // demodulator-reported error
    int32_t fError;

    uint8_t *pALPPacket;
    int32_t nALPPacketBytesAllocated;

    // Statistics
    ENATSC3BitrateStatistics stats;
    ENATSC3BitrateStatistics paddingStats;
} BasebandPLPState;

typedef struct _ENATSC3BasebandHandle_t
{
    ENVerbosity verbosity;

#ifdef AUTODETECT_KOREAN_MODE
    // When Korean Mode autodetect is enabled, fKoreanMode caches the 
    // state.  It is only valid if fKoreanModeDetermined is true.
    int32_t fKoreanMode;

    // When Korean Mode autodetect is enabled, fKoreanModeDetermined 
    // indicates whether the Korean Mode check has been performed.
    int32_t fKoreanModeDetermined;
#endif

    // Per-PLP states for processing the ALP stream. Note that state 
    // information is required since ALP packets can straddle baseband
    // packet boundaries.
    BasebandPLPState vPLPs[64];

    // Callbacks
    ENATSC3BasebandALPCallbackFunction pALPCallback;
    void *pCallbackUserData;

    // Statistics
    int32_t fBitrateStatisticsEnabled;

    // Master lock for this structure
    pthread_mutex_t lock;
} ENATSC3Baseband, *ENATSC3BasebandHandle_t;

#ifdef AUTODETECT_KOREAN_MODE
static ENATSC3Status
ParseLMT(uint8_t *pPayload, int32_t payloadLength)
{
    int32_t numPLPs;
    int32_t i;
    int32_t j;
    ENATSC3Status rval = ENATSC3Status_Success;

    if (!payloadLength)
    {
        rval = ENATSC3Status_Failure;
        goto out;
    }
    numPLPs = (pPayload[0] >> 2) + 1;
    pPayload++;
    payloadLength--;

    for (i = 0; i < numPLPs; i++)
    {
        uint32_t numMulticasts;

        if (payloadLength < 2)
        {
            rval = ENATSC3Status_Failure;
            goto out;
        }
        numMulticasts = pPayload[1];
        pPayload += 2;
        payloadLength -= 2;

        for (j = 0; j < numMulticasts; j++)
        {
            uint32_t sidFlag;
            uint32_t compressedFlag;

            if (payloadLength < 13)
            {
                rval = ENATSC3Status_Failure;
                goto out;
            }

            sidFlag = pPayload[12] & 0x80;
            compressedFlag = pPayload[12] & 0x40;

            pPayload += 13;
            payloadLength -= 13;

            if (sidFlag)
            {
                if (payloadLength < 1)
                {
                    rval = ENATSC3Status_Failure;
                    goto out;
                }
                pPayload++;
                payloadLength--;
            }

            if (compressedFlag)
            {
                if (payloadLength < 1)
                {
                    rval = ENATSC3Status_Failure;
                    goto out;
                }
                pPayload++;
                payloadLength--;
            }
        }
    }

out:
    return rval;
}

static ENATSC3Status
ParseALPSignalingPayload(uint8_t *pPayload, int32_t payloadLength)
{
    ENATSC3Status rval = ENATSC3Status_Success;

    if (payloadLength < 5)
    {
        rval = ENATSC3Status_Failure;
        goto out;
    }

    if (pPayload[0] == 1)
    {
        // We only handle the LMT
        pPayload += 5;
        payloadLength -= 5;
        rval = ParseLMT(pPayload, payloadLength);
    }
    else
    {
        rval = ENATSC3Status_Retry;
    }

out:
    return rval;
}
#endif

// GetALPPacketLength parses ALP packet headers to the extent required to 
// determine the length of the packet. It is only required at the baseband
// packet processing level because it is the only way to determine where
// ALP packet boundaries are.
static ENATSC3Status
GetALPPacketLength(ENATSC3Baseband *pBaseband, uint8_t *pPacket, 
    int32_t packetLength, uint64_t packetNumber, int32_t *pLength)
{
    uint32_t payloadConfiguration;
    int32_t headerLength;
    int32_t payloadLength;
    ENATSC3Status rval = ENATSC3Status_Success;

    // The minimum ALP header length is 2 bytes
    if (packetLength < 2)
    {
        *pLength = 2;
        rval = ENATSC3Status_Retry;
        goto out;
    }

    payloadConfiguration = pPacket[0] & 0x10;

    if (payloadConfiguration == 0)
    {
        // Payload configuration = 0: single packet (with optional header)
        payloadLength = ((pPacket[0] & 0x7) << 8) | pPacket[1];

        if (pPacket[0] & 0x8) // header mode
        {
            // Additional Header for Single Packet

            if (packetLength < 3)
            {
                *pLength = 3;
                rval = ENATSC3Status_Retry;
                goto out;
            }

            payloadLength |= ((pPacket[2] >> 3)) << 11;
            headerLength = 3;

            if (pPacket[2] & 0x2) // Substream Identifer Flag
            {
                headerLength++;
            }

            if (pPacket[2] & 0x1) // Header extension flag
            {
                int32_t extensionLengthMinus1;

                if (headerLength + 2 > packetLength)
                {
                    *pLength = headerLength + 2;
                    rval = ENATSC3Status_Retry;
                    goto out;
                }

                // HEF present
                extensionLengthMinus1 = pPacket[headerLength + 1];

                headerLength += 2 + extensionLengthMinus1 + 1;
            }
        }
        else
        {
            headerLength = 2;
        }
    }
    else
    {
        if (pPacket[0] & 0x8) // Concatenation/segmentation
        {
            int32_t nSegments;

            // Concatenation mode. 

            if (packetLength < 3)
            {
                *pLength = 3;
                rval = ENATSC3Status_Retry;
                goto out;
            }

            payloadLength = ((pPacket[0] & 0x7) << 8) | 
                pPacket[1] | ((pPacket[2] >> 4) << 11);
            nSegments = ((pPacket[2] >> 1) & 0x7) + 2;

            // Advance headerLength over the segment lengths. There are
            // nSegments - 1 lengths. The length of the last segment is
            // implictly signalled as the end of the next-to-last segment until
            // the end of the ALP packet.
            headerLength = 3 + ((nSegments - 2 + 1) * 12 + 4) / 8;

            // Advance over the substream identifier, if present.
            // If present there is one substream identiifer for the ALP packet
            // (i.e. in concatenation mode the SID applies to all input 
            // packets.)
            if (pPacket[2] & 1) // Substream Identifier Flag
            {
                headerLength ++;
            }
        }
        else
        {
            // Segmentation mode.

            if (packetLength < 3)
            {
                *pLength = 3;
                rval = ENATSC3Status_Retry;
                goto out;
            }

            payloadLength = ((pPacket[0] & 0x7) << 8) | pPacket[1];
            headerLength = 3;

            if (pPacket[2] & 0x2) // Substream Identifer Flag
            {
                if (headerLength + 1 > packetLength)
                {
                    *pLength = headerLength + 1;
                    rval = ENATSC3Status_Retry;
                    goto out;
                }

                headerLength++;
            }

            if (pPacket[2] & 0x1) // Header extension flag
            {
                int32_t extensionLengthMinus1;

                if (headerLength + 2 > packetLength)
                {
                    *pLength = headerLength + 2;
                    rval = ENATSC3Status_Retry;
                    goto out;
                }

                // extensionLengthMinus1 is the number of bytes following
                // extensionLengthMinus1 less one byte.
                extensionLengthMinus1 = pPacket[headerLength + 1];

                if (headerLength + 2 + extensionLengthMinus1 + 1 > 
                    packetLength)
                {
                    *pLength = headerLength + 2 + extensionLengthMinus1 + 1;
                    rval = ENATSC3Status_Retry;
                    goto out;
                }

                headerLength += 2 + extensionLengthMinus1 + 1;
            }
        }
    }

    // LLS packets have an Additional Header for Signaling Information that is
    // five bytes long
    if ((pPacket[0] >> 5) == 4)
    {
#ifdef AUTODETECT_KOREAN_MODE
        // To detect Korean Mode we need to determine if the Additional Header
        // For Signaling Information is included in the payload length or not.
        // To do this we are going to assume that it is included in the payload
        // length. If after acquiring the entire packet we are able to 
        // successfully parse the packet then we assume we are in Korean Mode.
        // If we are not, then we are going to assume we are not in Korean 
        // Mode. This test will not work if the ALP packet has five or more
        // bytes of padding at the end of the signaling table, since in that
        // case the signaling table will parse correctly in either case.
        if (!pBaseband->fKoreanModeDetermined)
        {
            ENATSC3Status lret;

            if (packetLength < headerLength + payloadLength)
            {
                *pLength = headerLength + payloadLength;
                rval = ENATSC3Status_Retry;
                goto out;
            }

            lret = ParseALPSignalingPayload(pPacket + headerLength, 
                payloadLength);
            if (lret == ENATSC3Status_Retry)
            {
                // This most likely means that a signaling table other than
                // an LMT arrived. 
                rval = ENATSC3Status_Retry;
                goto out;
            }
            else if (lret == ENATSC3Status_Failure)
            {
                pBaseband->fKoreanMode = 0;
            }
            else if (lret == ENATSC3Status_Success)
            {
                pBaseband->fKoreanMode = 1;
            }
            else
            {
                ERRORPRINT("Unhandled probe return %d\n", lret);
                rval = ENATSC3Status_Retry;
                goto out;
            }
        }

        pBaseband->fKoreanModeDetermined = 1;

        if (pBaseband->fKoreanMode)
        {
            pPacket[1] -= 5;
        }
        else
        {
            headerLength += 5;
        }
#else
        // Assume we are non non-Korean mode
        headerLength += 5;
#endif
    }

    *pLength = headerLength + payloadLength;

    ENUtils_CrashAssert(*pLength > 0);

out:
    return rval;
}

static ENATSC3Status
GrowALPPacketBuffer(BasebandPLPState *pState, int32_t size)
{
    uint8_t *pNew;
    ENATSC3Status rval = ENATSC3Status_Success;

    if (size >= MAX_ALP_PACKET_LENGTH)
    {
        ERRORPRINT("Request to extend ALP packet buffer > %d\n ", size);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    if (pState->nALPPacketBytesAllocated >= size)
    {
        goto out;
    }

    size += 4096;
    if (size > MAX_ALP_PACKET_LENGTH)
    {
        size = MAX_ALP_PACKET_LENGTH;
    }

    pNew = (uint8_t *)realloc(pState->pALPPacket, size);
    if (!pNew)
    {
        ERRORPRINT("Out of memory\n");
        rval = ENATSC3Status_Failure;
    }
    else
    {
        pState->pALPPacket = pNew;
        pState->nALPPacketBytesAllocated = size; 
    }

out:
    return rval;
}

ENATSC3Status
ENATSC3BasebandProcessBasebandPacket(ENATSC3BasebandHandle_t hBaseband, 
    uint32_t plpId, struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t fError, int32_t fDiscontinuity, const uint8_t *pPayload, 
    int32_t payloadLength)
{
    ENATSC3Baseband *pBaseband = (ENATSC3Baseband *)hBaseband;
    int32_t originalPayloadLength = payloadLength;
    int32_t pointer;
    const uint8_t *pPointer = NULL;
    uint8_t mode;
    uint8_t ofi = 0;
    int32_t ext_len = 0;
    uint8_t ext_type = 255;
    BasebandPLPState *pPLPState;
    int32_t fPLPStateLocked = 0;
    int32_t fFirstALPStart = 1;
    ENATSC3Status rval = ENATSC3Status_Success;

    if (plpId > numof(pBaseband->vPLPs))
    {
        ERRORPRINT("Invalid PLP %d\n", plpId);
        goto out;
    }

    if (!payloadLength)
    {
        if (pBaseband->verbosity > ENVerbosity_Terse)
        {
            ERRORPRINT("Short baseband packet\n");
        }
        goto out;
    }
    mode = *pPayload & 0x80;
    pointer = *pPayload & 0x7f;
    pPayload++;
    payloadLength--;

    if (mode)
    {
        if (!payloadLength)
        {
            if (pBaseband->verbosity > ENVerbosity_Terse)
            {
                ERRORPRINT("Short baseband packet\n");
            }
            goto out;
        }

        ofi = *pPayload & 0x3;
        pointer |= (*pPayload >> 2) << 7;
        pPayload++;
        payloadLength--;

        if (ofi == 1)
        {
            if (!payloadLength)
            {
                if (pBaseband->verbosity > ENVerbosity_Terse)
                {
                    ERRORPRINT("Short baseband packet\n");
                }
                goto out;
            }

            ext_type = *pPayload >> 5;
            ext_len = *pPayload & 0x1f;
            pPayload++;
            payloadLength--;

        }
        else if (ofi == 2 || ofi == 3)
        {
            if (!payloadLength)
            {
                if (pBaseband->verbosity > ENVerbosity_Terse)
                {
                    ERRORPRINT("Short baseband packet\n");
                }
                goto out;
            }
            if (ofi == 2)
            {
                ext_type = *pPayload >> 5;
            }
            ext_len= *pPayload & 0x1f;
            pPayload++;
            payloadLength--;

            if (!payloadLength)
            {
                if (pBaseband->verbosity > ENVerbosity_Terse)
                {
                    ERRORPRINT("Short baseband packet\n");
                }
                goto out;
            }
            ext_len |= *pPayload <<5;
            pPayload++;
            payloadLength--;
        }

        if (ext_len > payloadLength)
        {
            if (pBaseband->verbosity > ENVerbosity_Terse)
            {
                DEBUGPRINT("ext_len (%d) > payload len (%d)\n", ext_len, 
                    payloadLength);
            }
            goto out;
        }

        pPayload += ext_len;
        payloadLength -= ext_len;;
    }

    pPLPState = pBaseband->vPLPs + plpId;

    if (pthread_mutex_lock(&pBaseband->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    fPLPStateLocked = 1;

    if (pBaseband->fBitrateStatisticsEnabled)
    {
        ENATSC3BitrateStatisticsUpdate(&pPLPState->stats, pPacketTimeUTC, 
            originalPayloadLength);
    }

    if (!pPLPState->fEnabled)
    {
        goto out;
    }

    // If this is the first ALP packet, or the first after a discontinuity,
    // then we start with the pointer.
    if (pPLPState->nALPBytesReceived == 0)
    {
        if (pointer == 8191)
        {
            // If pointer == 8191 then either there is no ALP packet starting
            // in the baseband packet, or the baseband packet consists entirely 
            // of padding.
            if (pBaseband->fBitrateStatisticsEnabled)
            {
                ENATSC3BitrateStatisticsUpdate(&pPLPState->paddingStats,
                    pPacketTimeUTC, originalPayloadLength);
            }
            goto out;
        }

        if (pointer > payloadLength)
        {
            if (pBaseband->verbosity > ENVerbosity_Terse)
            {
                DEBUGPRINT("Pointer outside packet\n");
            }
            goto out;
        }

        pPayload += pointer;
        payloadLength -= pointer;
        fFirstALPStart = 0;
    }
    else
    {
        if (pointer != 8191)
        {
            if (pointer > payloadLength)
            {
                if (pBaseband->verbosity > ENVerbosity_Terse)
                {
                    DEBUGPRINT("Pointer %d outside packet %d\n", 
                        pointer, payloadLength);
                }
                goto out;
            }
            pPointer = pPayload + pointer;
        }
    }

    while (payloadLength)
    {
        int32_t nToCopy;

        pPLPState->fError |= fError;

        // ALPLength will remain 0 until we have accumulated enough of the ALP 
        // packet to determine its length. This can be as little as two bytes 
        // and varies depending on the presence of additional headers and 
        // extensions.
        if (!pPLPState->ALPLength)
        {
            int32_t alpLength;

            if (pPayload == pPointer && pPLPState->nALPBytesReceived)
            {
                // We hit an ALP boundary while looking for the header?
                if (pBaseband->verbosity > ENVerbosity_Terse)
                {
                    DEBUGPRINT("Aborting ALP packet at pointer\n");
                }
                pPLPState->nALPBytesReceived = 0;
                pPLPState->fError = 0;
                fFirstALPStart = 0;
            }

            if (!pPLPState->nALPBytesReceived)
            {
                pPLPState->packetTimeUTC = *pPacketTimeUTC;
                pPLPState->packetNumber = packetNumber;
            }

            if (MAX(pPLPState->ALPLength, pPLPState->nALPBytesReceived + 1) > 
                pPLPState->nALPPacketBytesAllocated)
            {
                if (GrowALPPacketBuffer(pPLPState, 
                    MAX(pPLPState->ALPLength, 
                        pPLPState->nALPBytesReceived + 1)))
                {
                    ERRORPRINT("Failed to grow ALP packet buffer\n");

                    // We will lose track of where ALP packet boundaries are. 
                    // By setting nALPBytesReceived to 0 we will use the 
                    // pointer in the next baseband packet to re-establish ALP 
                    // packet boundary sync.
                    pPLPState->nALPBytesReceived = 0;
                    pPLPState->fError = 0;
                    goto out;
                }
            }

            pPLPState->pALPPacket[pPLPState->nALPBytesReceived++] = *pPayload;
            pPayload++;
            payloadLength--;

            if (GetALPPacketLength(pBaseband, pPLPState->pALPPacket, 
                pPLPState->nALPBytesReceived, packetNumber, &alpLength))
            {
                // XXX On error length contains the number of bytes needed to
                // compute the packet length. Copy those bytes here if they 
                // are available to avoid copying the ALP header one byte
                // at a time.
                continue;
            }

            pPLPState->ALPLength = alpLength;
            if (pPLPState->ALPLength > pPLPState->nALPPacketBytesAllocated)
            {
                if (GrowALPPacketBuffer(pPLPState, pPLPState->ALPLength))
                {   
                    // We have enough information where we could simply 
                    // discard the current ALP packet, but for the sake of 
                    // simplicity we will discard the remainder of the 
                    // baseband packet and restart ALP processing using the 
                    // pointer in the next baseband packet.
                    pPLPState->nALPBytesReceived = 0;
                    pPLPState->fError = 0;
                    goto out;
                }
            }

            continue;
        }

        nToCopy = MIN(payloadLength, 
            pPLPState->ALPLength - pPLPState->nALPBytesReceived);
        if (pPointer && pPayload < pPointer)
        {
            nToCopy = MIN(nToCopy, pPointer - pPayload);
        }
        memcpy(pPLPState->pALPPacket + pPLPState->nALPBytesReceived, pPayload, 
            nToCopy);
        pPayload += nToCopy;
        payloadLength -= nToCopy;
        pPLPState->nALPBytesReceived += nToCopy;

        if (pPLPState->nALPBytesReceived == pPLPState->ALPLength ||
            pPayload == pPointer)
        {
            if (pPLPState->nALPBytesReceived == pPLPState->ALPLength)
            {
                if (pBaseband->pALPCallback)
                {
                    pBaseband->pALPCallback(plpId, &pPLPState->packetTimeUTC, 
                        pPLPState->packetNumber, pPLPState->fError, 
                        fDiscontinuity, pPLPState->pALPPacket, 
                        pPLPState->ALPLength, pBaseband->pCallbackUserData);
                }
            }
            else
            {
                if (pBaseband->verbosity > ENVerbosity_Terse)
                {
                    DEBUGPRINT("ALP packet (len %d, %d pending, ext_len %d ext_type %d) spans boundary, dropping...\n",
                        pPLPState->ALPLength, 
                        pPLPState->ALPLength - pPLPState->nALPBytesReceived,
                        ext_len, ext_type);
                }
            }
        
            // Get ready to receive the next ALP packet
            pPLPState->nALPBytesReceived = 0;
            pPLPState->ALPLength = 0;
            pPLPState->fError = 0;

            if (fFirstALPStart)
            {
                if (payloadLength && pPayload != pPointer && pointer != 8191)
                {
                    if (pBaseband->verbosity > ENVerbosity_Terse)
                    {
                        DEBUGPRINT("WARN: ALP sync lost: %p != %p (pointer %d payloadLength %d bb pkt %"PRId64" mode %d ofi %d), resync with pointer\n",
                            pPayload, pPointer, pointer, payloadLength, 
                            packetNumber, mode, ofi);
                    }
                    payloadLength += pPayload - pPointer;
                    pPayload = pPointer;
                }
                fFirstALPStart = 0;
            }
        }

        // fDiscontinuties only applies to the first packet
        fDiscontinuity = 0;
    }

out:
    if (fPLPStateLocked)
    {
        if (pthread_mutex_unlock(&pBaseband->lock))
        {
            ERRORPRINT("Failed to unlock\n");
        }
    }

    return rval;
}

ENATSC3Status
ENATSC3BasebandOpen(ENATSC3BasebandHandle_t *phBaseband)
{
    ENATSC3Baseband *pBaseband;
    int32_t fHaveLock = 0;
    BasebandPLPState *pPLPState;
    int32_t i;
    ENATSC3Status rval = ENATSC3Status_Success;

    pBaseband = (ENATSC3Baseband *)calloc(sizeof(ENATSC3Baseband), 1);
    if (!pBaseband)
    {
        ERRORPRINT("Out of memory\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    pBaseband->verbosity = ENVerbosity_Terse;

    if (pthread_mutex_init(&pBaseband->lock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }
    fHaveLock = 1;

    for (i = 0, pPLPState = pBaseband->vPLPs; 
         i < numof(pBaseband->vPLPs); 
         i++, pPLPState++)
    {
        ENATSC3BitrateStatisticsInitialize(&pPLPState->stats, 2000);
        snprintf(pPLPState->stats.strName, sizeof(pPLPState->stats.strName),
            "Baseband PLP %d", i);
        pPLPState->stats.strName[sizeof(pPLPState->stats.strName) - 1] = 0;
        ENATSC3BitrateStatisticsInitialize(&pPLPState->paddingStats, 2000);
        snprintf(pPLPState->paddingStats.strName, 
            sizeof(pPLPState->paddingStats.strName),
            "Baseband Padding PLP %d", i);
        pPLPState->paddingStats.strName[
            sizeof(pPLPState->paddingStats.strName) - 1] = 0;
    }

    *phBaseband = pBaseband;

out:
    return rval;

errout:
    if (pBaseband)
    {
        if (fHaveLock)
        {
            if (pthread_mutex_destroy(&pBaseband->lock))
            {
                ERRORPRINT("Failed to destroy lock\n");
            }
        }

        free(pBaseband);
    }

    goto out;
}

ENATSC3Status
ENATSC3BasebandClose(ENATSC3BasebandHandle_t hBaseband)
{
    ENATSC3Baseband *pBaseband = (ENATSC3Baseband *)hBaseband;
    int32_t i;
    ENATSC3Status rval = ENATSC3Status_Success;

    for (i = 0; i < numof(pBaseband->vPLPs); i++)
    {
        if (pBaseband->vPLPs[i].fEnabled)
        {
            free(pBaseband->vPLPs[i].pALPPacket);
        }
    }

    if (pthread_mutex_destroy(&pBaseband->lock))
    {
        ERRORPRINT("Failed to destroy lock\n");
    }

    free(pBaseband);

    return rval;
}

void
ENATSC3BasebandSetBitrateStatisticsEnabled(ENATSC3BasebandHandle_t hBaseband,
    int32_t fEnabled)
{
    ENATSC3Baseband *pBaseband = (ENATSC3Baseband *)hBaseband;

    pBaseband->fBitrateStatisticsEnabled = fEnabled;
}

ENATSC3Status
ENATSC3BasebandRegisterCallbacks(ENATSC3BasebandHandle_t hBaseband,
    ENATSC3BasebandALPCallbackFunction pALPCallback, void *pUserData)
{
    ENATSC3Baseband *pBaseband = (ENATSC3Baseband *)hBaseband;
    ENATSC3Status rval = ENATSC3Status_Success;

    pBaseband->pALPCallback = pALPCallback;
    pBaseband->pCallbackUserData = pUserData;

    return rval;
}

ENATSC3Status
ENATSC3BasebandSetPLPs(ENATSC3BasebandHandle_t hBaseband, uint64_t plpIdMask)
{
    ENATSC3Baseband *pBaseband = (ENATSC3Baseband *)hBaseband;
    int32_t i;
    ENATSC3Status rval = ENATSC3Status_Success;

    if (pthread_mutex_lock(&pBaseband->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    for (i = 0; i < 64; i++)
    {
        if (plpIdMask & 1)
        {
            if (!pBaseband->vPLPs[i].fEnabled)
            {
                pBaseband->vPLPs[i].nALPPacketBytesAllocated = 
                    INITIAL_ALP_PACKET_LENGTH;
                pBaseband->vPLPs[i].pALPPacket = 
                    (uint8_t *)malloc(pBaseband->vPLPs[i].nALPPacketBytesAllocated);
                if (!pBaseband->vPLPs[i].pALPPacket)
                {
                    ERRORPRINT("Out of memory\n");
                    rval = ENATSC3Status_Failure;
                }
                else
                {
                    pBaseband->vPLPs[i].fEnabled = 1;
                }
            }
        }
        else
        {
            if (!pBaseband->vPLPs[i].fEnabled)
            {
                free(pBaseband->vPLPs[i].pALPPacket);
                pBaseband->vPLPs[i].pALPPacket = NULL;
                pBaseband->vPLPs[i].fEnabled = 0;
            }
        }

        plpIdMask >>= 1;
    }

    if (pthread_mutex_unlock(&pBaseband->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENATSC3Status
ENATSC3BasebandSetVerbosity(ENATSC3BasebandHandle_t hBaseband,
    ENVerbosity verbosity)
{
    ENATSC3Baseband *pBaseband = (ENATSC3Baseband *)hBaseband;
    ENATSC3Status rval = ENATSC3Status_Success;

    pBaseband->verbosity = verbosity;

    return rval;
}
