/*
 * Copyright © 2020-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_defs.h>

#include <libenatsc3utils/enxml.h>

#include "atsc3esg_sggd.h"

static char *ensys = "esg";
static char *ensubsys = "sgdd";

static int32_t fVerbose = 0;

#define VERBOSEPRINT(x, ...) \
    if (fVerbose) \
    { \
        LOGPRINT(x, ##__VA_ARGS__); \
    }

static int32_t
SGDDFragmentNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGSGDDNode *pSGDDNode = (ATSC3ESGSGDDNode *)pUserData;
    ATSC3ESGSGDDFragmentNode *pFragmentNode;
    int depth;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3ESGSGDDNodeType_Fragment);
    pFragmentNode = (ATSC3ESGSGDDFragmentNode *)pParentNode; 

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3ESGSGDDFragmentNodeAttribute_id:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSGDDNode->dataBuffer, (char *)pValue, sValue, 
                &pFragmentNode->idDataOffset))
            {
                ERRORPRINT("Failed to store id\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sid: %.*s\n", depth * 2,  
                "", sValue, pValue);
            break;
        case ATSC3ESGSGDDFragmentNodeAttribute_transportID:
            if (EnXMLParser_antol((char *)pValue, sValue, 
                &pFragmentNode->transportID))
            {
                ERRORPRINT("Failed to parse transportID %.*s\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*stransportID: %.*s\n", depth * 2, "", sValue, 
                pValue);
            break;
        case ATSC3ESGSGDDFragmentNodeAttribute_version:
            if (EnXMLParser_antol((char *)pValue, sValue, 
                &pFragmentNode->version))
            {
                ERRORPRINT("Failed to parse version %.*s\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sversion: %.*s\n", depth * 2, "", sValue, 
                pValue);
            break;
        case ATSC3ESGSGDDFragmentNodeAttribute_fragmentEncoding:
            if (EnXMLParser_antol((char *)pValue, sValue, 
                &pFragmentNode->fragmentEncoding))
            {
                ERRORPRINT("Failed to parse fragmentEncoding %.*s\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sfragmentEncoding: %.*s\n", depth * 2, "", sValue, 
                pValue);
            break;
        case ATSC3ESGSGDDFragmentNodeAttribute_fragmentType:
            if (EnXMLParser_antol((char *)pValue, sValue, 
                &pFragmentNode->fragmentType))
            {
                ERRORPRINT("Failed to parse fragmentType %.*s\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sfragmentType: %.*s\n", depth * 2, "", sValue, 
                pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", depth * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vFragmentAttributes[] = 
{
    {
        .id = ATSC3ESGSGDDFragmentNodeAttribute_transportID,
        .sName = sizeof("transportID") - 1,
        .pName = "transportID",
        .handler = SGDDFragmentNodeAttributeHandler,
    },
    {
        .id = ATSC3ESGSGDDFragmentNodeAttribute_id,
        .sName = sizeof("id") - 1,
        .pName = "id",
        .handler = SGDDFragmentNodeAttributeHandler,
    },
    {
        .id = ATSC3ESGSGDDFragmentNodeAttribute_version,
        .sName = sizeof("version") - 1,
        .pName = "version",
        .handler = SGDDFragmentNodeAttributeHandler,
    },
    {
        .id = ATSC3ESGSGDDFragmentNodeAttribute_fragmentEncoding,
        .sName = sizeof("fragmentEncoding") - 1,
        .pName = "fragmentEncoding",
        .handler = SGDDFragmentNodeAttributeHandler,
    },
    {
        .id = ATSC3ESGSGDDFragmentNodeAttribute_fragmentType,
        .sName = sizeof("fragmentType") - 1,
        .pName = "fragmentType",
        .handler = SGDDFragmentNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vFragmentChildNodes[] =
{
};
static EnXMLParserNode fragmentNodeData = 
{
    .pName = "Fragment",
    .pAttributes = vFragmentAttributes,
    .nAttributes = numof(vFragmentAttributes),
    .pChildNodes = vFragmentChildNodes,
    .nChildNodes = numof(vFragmentChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessFragmentNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGSGDDNode *pSGDDNode = (ATSC3ESGSGDDNode *)pUserData;
    ATSC3ESGSGDDServiceGuideDeliveryUnitNode *pServiceGuideDeliveryUnitNode;
    ATSC3ESGSGDDFragmentNode fragmentNode;
    int32_t fragmentNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == 
        ATSC3ESGSGDDNodeType_ServiceGuideDeliveryUnit);
    pServiceGuideDeliveryUnitNode = 
        (ATSC3ESGSGDDServiceGuideDeliveryUnitNode *)pParentNode;
    memset(&fragmentNode, 0, sizeof(fragmentNode));

    // Allocate space for the node from the data buffer and store the offset 
    // in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pSGDDNode->dataBuffer,
        sizeof(ATSC3ESGSGDDFragmentNode), &fragmentNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for fragment node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pServiceGuideDeliveryUnitNode->fragments, fragmentNodeOffset))
    {
        ERRORPRINT("Failed to store fragment node offset\n");
        rval = 1;
        goto out;
    }

    fragmentNode.base.nodeType = ATSC3ESGSGDDNodeType_Fragment;

    VERBOSEPRINT("%*sFragment\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&fragmentNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)&fragmentNode);

    memcpy(pSGDDNode->dataBuffer.pBuffer + fragmentNodeOffset,
        &fragmentNode, sizeof(fragmentNode));

out:
    return rval;
}

static int32_t
SGDDServiceGuideDeliveryUnitNodeAttributeHandler(
    EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGSGDDNode *pSGDDNode = (ATSC3ESGSGDDNode *)pUserData;
    ATSC3ESGSGDDServiceGuideDeliveryUnitNode *pServiceGuideDeliveryUnitNode;
    int depth;
    int32_t rval = 0;

    assert(pParentNode->nodeType == 
        ATSC3ESGSGDDNodeType_ServiceGuideDeliveryUnit);
    pServiceGuideDeliveryUnitNode = 
        (ATSC3ESGSGDDServiceGuideDeliveryUnitNode *)pParentNode; 

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3ESGSGDDServiceGuideDeliveryUnitNodeAttribute_contentLocation:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSGDDNode->dataBuffer, (char *)pValue, sValue, 
                &pServiceGuideDeliveryUnitNode->contentLocationDataOffset))
            {
                ERRORPRINT("Failed to store contentLocationDataOffset\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*scontentLocationDataOffset: %.*s\n", depth * 2,  
                "", sValue, pValue);
            break;
        case ATSC3ESGSGDDServiceGuideDeliveryUnitNodeAttribute_transportObjectID:
            if (EnXMLParser_antol((char *)pValue, sValue, 
                &pServiceGuideDeliveryUnitNode->transportObjectID))
            {
                ERRORPRINT("Failed to parse transportObjectID %.*s\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*stransportObjectID: %.*s\n", depth * 2, "", sValue, 
                pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", depth * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vServiceGuideDeliveryUnitAttributes[] = 
{
    {
        .id = ATSC3ESGSGDDServiceGuideDeliveryUnitNodeAttribute_contentLocation,
        .sName = sizeof("contentLocation") - 1,
        .pName = "contentLocation",
        .handler = SGDDServiceGuideDeliveryUnitNodeAttributeHandler,
    },
    {
        .id = ATSC3ESGSGDDServiceGuideDeliveryUnitNodeAttribute_transportObjectID,
        .sName = sizeof("transportObjectID") - 1,
        .pName = "transportObjectID",
        .handler = SGDDServiceGuideDeliveryUnitNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vServiceGuideDeliveryUnitChildNodes[] =
{
    {
        .id = ATSC3ESGSGDDServiceGuideDeliveryUnitChildNode_Fragment,
        .sName = sizeof("Fragment") - 1,
        .pName = "Fragment",
        .handler = ProcessFragmentNode,
    },
};
static EnXMLParserNode serviceGuideDeliveryUnitNodeData = 
{
    .pName = "ServiceGuideDeliveryUnit",
    .pAttributes = vServiceGuideDeliveryUnitAttributes,
    .nAttributes = numof(vServiceGuideDeliveryUnitAttributes),
    .pChildNodes = vServiceGuideDeliveryUnitChildNodes,
    .nChildNodes = numof(vServiceGuideDeliveryUnitChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessServiceGuideDeliveryUnitNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGSGDDNode *pSGDDNode = (ATSC3ESGSGDDNode *)pUserData;
    ATSC3ESGSGDDDescriptorEntryNode *pDescriptorEntryNode;
    ATSC3ESGSGDDServiceGuideDeliveryUnitNode *pServiceGuideDeliveryUnitNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3ESGSGDDNodeType_DescriptorEntry);
    pDescriptorEntryNode = (ATSC3ESGSGDDDescriptorEntryNode *)pParentNode;
    pServiceGuideDeliveryUnitNode = 
        &pDescriptorEntryNode->serviceGuideDeliveryUnit;
    memset(pServiceGuideDeliveryUnitNode, 0, 
        sizeof(*pServiceGuideDeliveryUnitNode));

    if (EnXMLParserDataOffsetArrayInitialize(
        &pServiceGuideDeliveryUnitNode->fragments, &pSGDDNode->dataBuffer, 16, 
        16))
    {
        ERRORPRINT("Out of memory\n");
        rval = -1;
        goto out;
    }

    pServiceGuideDeliveryUnitNode->base.nodeType =
        ATSC3ESGSGDDNodeType_ServiceGuideDeliveryUnit;

    VERBOSEPRINT("%*sServiceGuideDeliveryUnit\n", 
        xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&serviceGuideDeliveryUnitNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)pServiceGuideDeliveryUnitNode);

out:
    return rval;
}

static int32_t
SGDDTransportNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGSGDDNode *pSGDDNode = (ATSC3ESGSGDDNode *)pUserData;
    ATSC3ESGSGDDTransportNode *pTransportNode;
    int depth;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3ESGSGDDNodeType_Transport);
    pTransportNode = (ATSC3ESGSGDDTransportNode *)pParentNode; 

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3ESGSGDDTransportNodeAttribute_ipAddress:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSGDDNode->dataBuffer, (char *)pValue, sValue, 
                &pTransportNode->ipAddressDataOffset))
            {
                ERRORPRINT("Failed to store ipAddress\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sipAddress: %.*s\n", depth * 2,  "", sValue, pValue);
            break;
        case ATSC3ESGSGDDTransportNodeAttribute_port:
            if (EnXMLParser_antol((char *)pValue, sValue, 
                &pTransportNode->port))
            {
                ERRORPRINT("Failed to parse port %.*s\n", sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sport: %.*s\n", depth * 2, "", sValue, pValue);
            break;
        case ATSC3ESGSGDDTransportNodeAttribute_transmissionSessionId:
            if (EnXMLParser_antol((char *)pValue, sValue, 
                &pTransportNode->transmissionSessionID))
            {
                ERRORPRINT("Failed to parse transmissionSessionID %.*s\n", 
                    sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*stransmissionSessionID: %.*s\n", depth * 2, "", 
                sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", depth * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vTransportAttributes[] = 
{
    {
        .id = ATSC3ESGSGDDTransportNodeAttribute_ipAddress,
        .sName = sizeof("ipAddress") - 1,
        .pName = "ipAddress",
        .handler = SGDDTransportNodeAttributeHandler,
    },
    {
        .id = ATSC3ESGSGDDTransportNodeAttribute_port,
        .sName = sizeof("port") - 1,
        .pName = "port",
        .handler = SGDDTransportNodeAttributeHandler,
    },
    {
        .id = ATSC3ESGSGDDTransportNodeAttribute_transmissionSessionId,
        .sName = sizeof("transmissionSessionId") - 1,
        .pName = "transmissionSessionId",
        .handler = SGDDTransportNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vTransportChildNodes[] =
{
};
static EnXMLParserNode transportNodeData = 
{
    .pName = "Transport",
    .pAttributes = vTransportAttributes,
    .nAttributes = numof(vTransportAttributes),
    .pChildNodes = vTransportChildNodes,
    .nChildNodes = numof(vTransportChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessTransportNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGSGDDDescriptorEntryNode *pDescriptorEntryNode;
    ATSC3ESGSGDDTransportNode *pTransportNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3ESGSGDDNodeType_DescriptorEntry);
    pDescriptorEntryNode = (ATSC3ESGSGDDDescriptorEntryNode *)pParentNode;
    pTransportNode = &pDescriptorEntryNode->transport;
    memset(pTransportNode, 0, sizeof(*pTransportNode));

    pTransportNode->base.nodeType = ATSC3ESGSGDDNodeType_Transport;

    VERBOSEPRINT("%*sTransport\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&transportNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)pTransportNode);

    return rval;
}

static int32_t
SGDDTimeGroupingCriteriaNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGSGDDTimeGroupingCriteriaNode *pTimeGroupingCriteriaNode;
    int depth;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3ESGSGDDNodeType_TimeGroupingCriteria);
    pTimeGroupingCriteriaNode = 
        (ATSC3ESGSGDDTimeGroupingCriteriaNode *)pParentNode; 

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3ESGSGDDTimeGroupingCriteriaNodeAttribute_startTime:
            if (EnXMLParser_antoll((char *)pValue, sValue, 
                &pTimeGroupingCriteriaNode->startTime))
            {
                ERRORPRINT("Failed to parse startTime %.*s\n", sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sstartTime: %.*s\n", depth * 2, "", 
                sValue, pValue);
            break;
        case ATSC3ESGSGDDTimeGroupingCriteriaNodeAttribute_endTime:
            if (EnXMLParser_antoll((char *)pValue, sValue, 
                &pTimeGroupingCriteriaNode->endTime))
            {
                ERRORPRINT("Failed to parse endTime %.*s\n", sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sendTime: %.*s\n", depth * 2, "", 
                sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", depth * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vTimeGroupingCriteriaAttributes[] = 
{
    {
        .id = ATSC3ESGSGDDTimeGroupingCriteriaNodeAttribute_startTime,
        .sName = sizeof("startTime") - 1,
        .pName = "startTime",
        .handler = SGDDTimeGroupingCriteriaNodeAttributeHandler,
    },
    {
        .id = ATSC3ESGSGDDTimeGroupingCriteriaNodeAttribute_endTime,
        .sName = sizeof("endTime") - 1,
        .pName = "endTime",
        .handler = SGDDTimeGroupingCriteriaNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vTimeGroupingCriteriaChildNodes[] =
{
};
static EnXMLParserNode timeGroupingCriteriaNodeData = 
{
    .pName = "TimeGroupingCriteria",
    .pAttributes = vTimeGroupingCriteriaAttributes,
    .nAttributes = numof(vTimeGroupingCriteriaAttributes),
    .pChildNodes = vTimeGroupingCriteriaChildNodes,
    .nChildNodes = numof(vTimeGroupingCriteriaChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessTimeGroupingCriteriaNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGSGDDGroupingCriteriaNode *pGroupingCriteriaNode;
    ATSC3ESGSGDDTimeGroupingCriteriaNode *pTimeGroupingCriteriaNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3ESGSGDDNodeType_GroupingCriteria);
    pGroupingCriteriaNode = (ATSC3ESGSGDDGroupingCriteriaNode *)pParentNode;
    pTimeGroupingCriteriaNode = &pGroupingCriteriaNode->timeGroupingCriteria;
    memset(pTimeGroupingCriteriaNode, 0, sizeof(*pTimeGroupingCriteriaNode));

    pTimeGroupingCriteriaNode->base.nodeType = 
        ATSC3ESGSGDDNodeType_TimeGroupingCriteria;

    VERBOSEPRINT("%*sTimeGroupingCriteria\n", xmlTextReaderDepth(reader) * 2, 
        "");

    rval = EnXMLParserProcessNode(&timeGroupingCriteriaNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)pTimeGroupingCriteriaNode);

    return rval;
}

static EnXMLParserAttribute vGroupingCriteriaAttributes[] = 
{
};
static EnXMLParserChildNode vGroupingCriteriaChildNodes[] =
{
    {
        .id = ATSC3ESGSGDDNodeType_TimeGroupingCriteria,
        .sName = sizeof("TimeGroupingCriteria") - 1,
        .pName = "TimeGroupingCriteria",
        .handler = ProcessTimeGroupingCriteriaNode,
    },
};
static EnXMLParserNode groupingCriteriaNodeData = 
{
    .pName = "GroupingCriteria",
    .pAttributes = vGroupingCriteriaAttributes,
    .nAttributes = numof(vGroupingCriteriaAttributes),
    .pChildNodes = vGroupingCriteriaChildNodes,
    .nChildNodes = numof(vGroupingCriteriaChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessGroupingCriteriaNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGSGDDDescriptorEntryNode *pDescriptorEntryNode;
    ATSC3ESGSGDDGroupingCriteriaNode *pGroupingCriteriaNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3ESGSGDDNodeType_DescriptorEntry);
    pDescriptorEntryNode = (ATSC3ESGSGDDDescriptorEntryNode *)pParentNode;
    pGroupingCriteriaNode = &pDescriptorEntryNode->groupingCriteria;
    memset(pGroupingCriteriaNode, 0, sizeof(*pGroupingCriteriaNode));

    pGroupingCriteriaNode->base.nodeType = 
        ATSC3ESGSGDDNodeType_GroupingCriteria;

    VERBOSEPRINT("%*sGroupingCriteria\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&groupingCriteriaNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)pGroupingCriteriaNode);

    return rval;
}

static EnXMLParserAttribute vDescriptorEntryAttributes[] = 
{
};

static EnXMLParserChildNode vDescriptorEntryChildNodes[] =
{
    {
        .id = ATSC3ESGSGDDNodeType_GroupingCriteria,
        .sName = sizeof("GroupingCriteria") - 1,
        .pName = "GroupingCriteria",
        .handler = ProcessGroupingCriteriaNode,
    },
    {
        .id = ATSC3ESGSGDDNodeType_Transport,
        .sName = sizeof("Transport") - 1,
        .pName = "Transport",
        .handler = ProcessTransportNode,
    },
    {
        .id = ATSC3ESGSGDDNodeType_ServiceGuideDeliveryUnit,
        .sName = sizeof("ServiceGuideDeliveryUnit") - 1,
        .pName = "ServiceGuideDeliveryUnit",
        .handler = ProcessServiceGuideDeliveryUnitNode,
    },
};
static EnXMLParserNode descriptorEntryNodeData = 
{
    .pName = "DescriptorEntry",
    .pAttributes = vDescriptorEntryAttributes,
    .nAttributes = numof(vDescriptorEntryAttributes),
    .pChildNodes = vDescriptorEntryChildNodes,
    .nChildNodes = numof(vDescriptorEntryChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessDescriptorEntryNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGSGDDNode *pSGDDNode = (ATSC3ESGSGDDNode *)pUserData;
    ATSC3ESGSGDDDescriptorEntryNode descriptorEntryNode;
    int32_t descriptorEntryNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == 
        ATSC3ESGSGDDNodeType_ServiceGuideDeliveryDescriptor);
    memset(&descriptorEntryNode, 0, sizeof(descriptorEntryNode));

    // Allocate space for the node from the data buffer and store the offset 
    // in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pSGDDNode->dataBuffer,
        sizeof(ATSC3ESGSGDDDescriptorEntryNode), &descriptorEntryNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for descriptorEntry node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pSGDDNode->descriptorEntries,
        descriptorEntryNodeOffset))
    {
        ERRORPRINT("Failed to store descriptorEntry node offset\n");
        rval = 1;
        goto out;
    }

    descriptorEntryNode.base.nodeType = ATSC3ESGSGDDNodeType_DescriptorEntry;

    VERBOSEPRINT("%*sDescriptorEntry\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&descriptorEntryNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)&descriptorEntryNode);

    memcpy(pSGDDNode->dataBuffer.pBuffer + descriptorEntryNodeOffset,
        &descriptorEntryNode, sizeof(descriptorEntryNode));

out:
    return rval;
}

static int32_t
SGDDNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3ESGSGDDNode *pSGDDNode = (ATSC3ESGSGDDNode *)pUserData;
    int depth;
    int32_t rval = 0;

    assert(pParentNode->nodeType == 
        ATSC3ESGSGDDNodeType_ServiceGuideDeliveryDescriptor);

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3ESGSGDDNodeAttribute_id:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSGDDNode->dataBuffer, (char *)pValue, sValue, 
                &pSGDDNode->idDataOffset))
            {
                ERRORPRINT("Failed to store id\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sid: %.*s\n", depth * 2,  "", sValue, pValue);
            break;
        case ATSC3ESGSGDDNodeAttribute_version:
            if (EnXMLParser_antoul((char *)pValue, sValue, &pSGDDNode->version))
            {
                ERRORPRINT("Failed to parse version %.*s\n", sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sversion: %.*s\n", depth * 2, "", 
                sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", depth * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vSGDDAttributes[] = 
{
    {
        .id = ATSC3ESGSGDDNodeAttribute_version,
        .sName = sizeof("version") - 1,
        .pName = "version",
        .handler = SGDDNodeAttributeHandler,
    },
    {
        .id = ATSC3ESGSGDDNodeAttribute_id,
        .sName = sizeof("id") - 1,
        .pName = "id",
        .handler = SGDDNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vSGDDChildNodes[] =
{
    {
        .id = ATSC3ESGSGDDNodeType_DescriptorEntry,
        .sName = sizeof("DescriptorEntry") - 1,
        .pName = "DescriptorEntry",
        .handler = ProcessDescriptorEntryNode,
    },
};
static EnXMLParserNode sgddNodeData = 
{
    .pName = "ServiceGuideDeliveryDescriptor",
    .pAttributes = vSGDDAttributes,
    .nAttributes = numof(vSGDDAttributes),
    .pChildNodes = vSGDDChildNodes,
    .nChildNodes = numof(vSGDDChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessSGDDNode(xmlTextReaderPtr reader, ATSC3ESGSGDDNode *pSGDDNode)
{
    int32_t rval = 0;

    VERBOSEPRINT("%*sServiceGuideDeliveryDescriptor\n", 
        xmlTextReaderDepth(reader) * 2, "");

    if (EnXMLParserDataOffsetArrayInitialize(&pSGDDNode->descriptorEntries,
        &pSGDDNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate descriptorEntries array\n");
        rval = -1;
        goto out;
    }

    rval = EnXMLParserProcessNode(&sgddNodeData, reader, pSGDDNode,
        (EnXMLParserNodeBase *)pSGDDNode);

out:
    return rval;
}

int32_t
ATSC3ESGSGDDParse(const char *pXML, int32_t sXML, ATSC3ESGSGDDNode *pParsedSGDD)
{
    xmlTextReaderPtr reader = NULL;
    int32_t fHaveDataBuffer = 0;
    int xmlret;
    int rval = 0;

    memset(pParsedSGDD, 0, sizeof(ATSC3ESGSGDDNode));
    reader = xmlReaderForMemory(pXML, sXML, NULL, NULL, 0);

    if (!reader)
    {
        ERRORPRINT("Failed to create xml reader\n");
        rval = -1;
        goto errout;
    }

    if (EnXMLParserDataBufferInitialize(&pParsedSGDD->dataBuffer, 8192, 8192))
    {
        ERRORPRINT("Failed to initialize SGDD data buffer\n");
        rval = -1;
        goto errout;
    }
    fHaveDataBuffer = 1;

    xmlret = xmlTextReaderRead(reader);
    if (xmlret == 1)
    {
        int nodeType;
        const xmlChar *pLocalName;

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);

        if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "ServiceGuideDeliveryDescriptor"))
        {
            pParsedSGDD->base.nodeType = 
                ATSC3ESGSGDDNodeType_ServiceGuideDeliveryDescriptor;
            ProcessSGDDNode(reader, pParsedSGDD);
        }
        else
        {
            ERRORPRINT("Top-level object isn't ServiceGuideDeliveryDescriptor (%d/%s)\n",
                nodeType, pLocalName);
            rval = -1;
            goto errout;
        }
    }
    else
    {
        ERRORPRINT("Failed to parse SGDD XML\n");
        rval = -1;
        goto errout;
    }

out:
    if (reader)
    {
        xmlFreeTextReader(reader);
    }

    return rval;

errout:
    if (fHaveDataBuffer)
    {
        if (EnXMLParserDataBufferFree(&pParsedSGDD->dataBuffer))
        {
            ERRORPRINT("Failed to free SGDD data buffer\n");
        }
    }

    goto out;
}

int32_t
ATSC3ESGSGDDFree(ATSC3ESGSGDDNode *pParsedSGDD)
{
    int32_t rval = 0;

    if (EnXMLParserDataBufferFree(&pParsedSGDD->dataBuffer))
    {
        ERRORPRINT("Failed to free SGDD data buffer\n");
    }

    return rval;
}

