/*
 * Copyright © 2018-2021 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <assert.h>

#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <enatsc3cap/enatsc3cap.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3j288alp/libenatsc3j288alp.h>
#include <libenatsc3baseband/libenatsc3baseband.h>
#include <libenatsc3alp/libenatsc3alp.h>

#include "libenatsc3filereader.h"

#define MAX_PAYLOAD_SIZE (65536 + 512)

static char *ensys = "frdr";
static char *ensubsys = "main";

int32_t
ENATSC3FileReaderReadALPFile(int fd, int32_t tOffsetSeconds, 
    FileReaderALPPacketCallbackFunction ALPPacketCallback,
    ENATSC3ALPLLSPLPsCallback LLSPLPsCallback, 
    int32_t *pFRunning, void *pUserData)
{
    ENATSC3CapFileHeader fileHeader;
    uint64_t alpPacketNumber = 0;
    uint8_t *pPayload = NULL;
    uint64_t llsValidMask = 0;
    int64_t tStartReading = -1;
    int32_t rval = 0;

    pPayload = (uint8_t *)malloc(MAX_PAYLOAD_SIZE);
    if (!pPayload)
    {
        ERRORPRINT("Out of memory\n");
        rval = 1;
        goto out;
    }

    if (read(fd, &fileHeader, sizeof(fileHeader)) != sizeof(fileHeader))
    {
        ERRORPRINT("Failed to read file header\n");
        rval = 1;
        goto out;
    }

    if (ntohs(fileHeader.magic) != ENATSC3CAP_HEADER_MAGIC)
    {
        ERRORPRINT("File header has bad magic\n");
        rval = 1;
        goto out;
    }
    else
    {
        if (fileHeader.format != ENATSC3CapFormat_V2)
        {
            ERRORPRINT("Unsupported file format %d\n", fileHeader.format);
            rval = 1;
            goto out;
        }

        if (fileHeader.type != ENATSC3CapType_ATSC3_ALP)
        {
            ERRORPRINT("Unsupported file type %d\n", fileHeader.type);
            rval = 1;
            goto out;
        }
    }

    while (*pFRunning)
    {
        ENATSC3CapRecordHeaderV2 headerV2;
        int nb;
        int32_t payloadLength;
        struct timespec packetTime;
        uint8_t plpId;
        uint8_t type;

        if (fileHeader.format == ENATSC3CapFormat_V2)
        {
            nb = read(fd, &headerV2, sizeof(headerV2));
            if (nb != sizeof(headerV2))
            {
                if (nb == 0)
                {
                    break;
                }
                else if (nb < 0)
                {
                    ERRORPRINT("Error reading ALP packet headerV2: %s\n", 
                        strerror(errno));
                    rval = 1;
                    break;
                }
                else
                {
                    ERRORPRINT(
                        "Short read of ALP packet header (%d != %zd)\n", 
                        nb, sizeof(headerV2));
                    rval = 1;
                    break;
                }
            }

            payloadLength = ntohl(headerV2.length);
            packetTime.tv_sec = ENUtils_ntohll(headerV2.rtwhen.sec);
            packetTime.tv_nsec = ntohl(headerV2.rtwhen.nsec);
            plpId = headerV2.plpId;
            type = headerV2.type;
        }
        else
        {
            ERRORPRINT("Unhandled capture format %d\n", 
                fileHeader.format);
            break;
        }

        if (payloadLength > MAX_PAYLOAD_SIZE)
        {
            ERRORPRINT("ALP Packet too large (%d > %d)\n", payloadLength, 
                MAX_PAYLOAD_SIZE);
            rval = 1;
            break;
        }

        nb = read(fd, pPayload, payloadLength);
        if (nb != payloadLength)
        {
            if (nb == 0)
            {
                ERRORPRINT("EOF reading packet payload\n");
            }
            else if (nb < 0)
            {
                ERRORPRINT("Error reading ALP packet: %s\n", 
                    strerror(errno));
            }
            else
            {
                ERRORPRINT("Short read of ALP packet (%d != %d)\n", 
                    nb, payloadLength);
            }

            rval = 1;
            break;
        }

        if (tStartReading == -1)
        {
            tStartReading = packetTime.tv_sec + tOffsetSeconds;
        }

        if (packetTime.tv_sec < tStartReading)
        {
            continue;
        }

        if (type == ENATSC3CapPayloadType_Demod_V2)
        {
            ENATSC3CapDemodulatorRecordV2 *pDemodRecord;
            uint64_t currentLLSValidMask;

            pDemodRecord = (ENATSC3CapDemodulatorRecordV2 *)pPayload;

            currentLLSValidMask = pDemodRecord->atsc3.llsPresentMask;
            if (currentLLSValidMask != llsValidMask)
            {
                if (LLSPLPsCallback)
                {
                    LLSPLPsCallback(currentLLSValidMask, &packetTime, 
                        alpPacketNumber, pUserData);
                }
                llsValidMask = currentLLSValidMask;
            }
            continue;
        }

        // The ALP file format doesn't have RF or continuity counter error
        // information
        ALPPacketCallback(plpId, &packetTime, alpPacketNumber, 0, 0, pPayload, 
            payloadLength, pUserData);

        alpPacketNumber++;
    }

out:
    if (pPayload)
    {
        free(pPayload);
    }

    return rval;
}

struct J288UserData
{
    uint64_t llsValidMask;
    uint64_t alpPacketNumber;

    // The plpId from the record header. It will be 255 if more than one PLP
    // was captured. In that case the PLP Id passed via the J288 ALP callback
    // should be valid. If the capture was for a single PLP then plpId will
    // be valid.
    uint8_t plpId;

    FileReaderALPPacketCallbackFunction ALPPacketCallback;
    void *pUserData;
};

void 
J288ALPPacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, int32_t fError, const uint8_t *pPacket, 
    int32_t sPacket, void *pCbUserData)
{
    struct J288UserData *pUserData = (struct J288UserData *)pCbUserData;

    if (plpId == 255)
    {
        if (pUserData->plpId == 255)
        {
            ERRORPRINT("No PLP from j288 or capture header!\n");
            goto out;
        }
        
        plpId = pUserData->plpId;
    }

    pUserData->ALPPacketCallback(plpId, pPacketTimeUTC, packetNumber, 0, 0,  
        pPacket, sPacket, pUserData->pUserData);

out:
    return;
}

int32_t
ENATSC3FileReaderReadJ288ALPFile(int fd, int32_t tOffsetSeconds, 
    FileReaderALPPacketCallbackFunction ALPPacketCallback,
    ENATSC3ALPLLSPLPsCallback LLSPLPsCallback, 
    int32_t *pFRunning, void *pUserData)
{
    ENATSC3CapFileHeader fileHeader;
    uint8_t *pPayload = NULL;
    int64_t tStartReading = -1;
    struct J288UserData userdata =
    {
        .llsValidMask = 0,
        .alpPacketNumber = 0,
        .plpId = 255,
        .ALPPacketCallback = ALPPacketCallback,
        .pUserData = pUserData,
    };
    ENATSC3J288ALPHandle_t hJ288ALP = NULL;
    int32_t rval = 0;

    if (ENATSC3J288ALPOpen(&hJ288ALP))
    {
        ERRORPRINT("Failed to open hJ288ALP\n");
        rval = 1;
        goto out;
    }

    if (ENATSC3J288ALPRegisterCallbacks(hJ288ALP, J288ALPPacketCallback,
        NULL, NULL, &userdata))
    {
        ERRORPRINT("Failed to register J288 callbacks\n");
        rval = 1;
        goto out;
    }

    pPayload = (uint8_t *)malloc(MAX_PAYLOAD_SIZE);
    if (!pPayload)
    {
        ERRORPRINT("Out of memory\n");
        rval = 1;
        goto out;
    }

    if (read(fd, &fileHeader, sizeof(fileHeader)) != sizeof(fileHeader))
    {
        ERRORPRINT("Failed to read file header\n");
        rval = 1;
        goto out;
    }

    if (fileHeader.format != ENATSC3CapFormat_V2)
    {
        ERRORPRINT("Unsupported format %d\n", fileHeader.format);
        rval = 1;
        goto out;
    }

    if (fileHeader.type != ENATSC3CapType_ATSC3_J288ALP)
    {
        ERRORPRINT("Unsupported file type %d\n", fileHeader.type);
        rval = 1;
        goto out;
    }

    while (*pFRunning)
    {
        ENATSC3CapRecordHeaderV2 headerV2;
        int nb;
        int32_t payloadLength;
        struct timespec packetTime;
        uint8_t type;

        nb = read(fd, &headerV2, sizeof(headerV2));
        if (nb != sizeof(headerV2))
        {
            if (nb == 0)
            {
                break;
            }
            else if (nb < 0)
            {
                ERRORPRINT("Error reading record headerV2: %s\n", 
                    strerror(errno));
                rval = 1;
                break;
            }
            else
            {
                ERRORPRINT("Short read of record header (%d != %zd)\n", 
                    nb, sizeof(headerV2));
                rval = 1;
                break;
            }
        }

        payloadLength = ntohl(headerV2.length);
        packetTime.tv_sec = ENUtils_ntohll(headerV2.rtwhen.sec);
        packetTime.tv_nsec = ntohl(headerV2.rtwhen.nsec);
        type = headerV2.type;
        if (headerV2.plpId != 255)
        {
            userdata.plpId = headerV2.plpId;
        }

        if (payloadLength > MAX_PAYLOAD_SIZE)
        {
            ERRORPRINT("J288 Packet too large (%d > %d)\n", payloadLength, 
                MAX_PAYLOAD_SIZE);
            rval = 1;
            break;
        }

        nb = read(fd, pPayload, payloadLength);
        if (nb != payloadLength)
        {
            if (nb == 0)
            {
                ERRORPRINT("EOF reading packet payload\n");
            }
            else if (nb < 0)
            {
                ERRORPRINT("Error reading ALP packet: %s\n", 
                    strerror(errno));
            }
            else
            {
                ERRORPRINT("Short read of ALP packet (%d != %d)\n", 
                    nb, payloadLength);
            }

            rval = 1;
            break;
        }

        if (tStartReading == -1)
        {
            tStartReading = packetTime.tv_sec + tOffsetSeconds;
        }

        if (packetTime.tv_sec < tStartReading)
        {
            continue;
        }

        if (type == ENATSC3CapPayloadType_Demod_V2)
        {
            ENATSC3CapDemodulatorRecordV2 *pDemodRecord;
            uint64_t currentLLSValidMask;

            pDemodRecord = (ENATSC3CapDemodulatorRecordV2 *)pPayload;

            currentLLSValidMask = pDemodRecord->atsc3.llsPresentMask;
            if (currentLLSValidMask != userdata.llsValidMask)
            {
                if (LLSPLPsCallback)
                {
                    LLSPLPsCallback(currentLLSValidMask, &packetTime, 
                        userdata.alpPacketNumber, pUserData);
                }
                userdata.llsValidMask = currentLLSValidMask;
            }
            continue;
        }
        else if (type == ENATSC3CapPayloadType_Demod_V2)
        {
            ENATSC3CapDemodulatorRecordV2 *pDemodRecord;
            uint64_t currentLLSValidMask;

            pDemodRecord = (ENATSC3CapDemodulatorRecordV2 *)pPayload;

            currentLLSValidMask = 
                ENUtils_ntohll(pDemodRecord->atsc3.llsPresentMask);
            if (currentLLSValidMask != userdata.llsValidMask)
            {
                if (LLSPLPsCallback)
                {
                    LLSPLPsCallback(currentLLSValidMask, &packetTime, 
                        userdata.alpPacketNumber, pUserData);
                }
                userdata.llsValidMask = currentLLSValidMask;
            }
            continue;
        }
        else if (type == ENATSC3CapPayloadType_ATSC3_J288ALP)
        {
            ENATSC3J288ALPProcessJ288Packets(hJ288ALP, &packetTime, 
                pPayload, payloadLength);
        }
        else
        {
            ERRORPRINT("Unhandled record type %d\n", type);
        }

        userdata.alpPacketNumber++;
    }

out:
    if (pPayload)
    {
        free(pPayload);
    }

    if (hJ288ALP)
    {
        if (ENATSC3J288ALPClose(hJ288ALP))
        {
            ERRORPRINT("Failed to close hJ288ALP\n");
        }
    }

    return rval;
}

int32_t
ENATSC3FileReaderReadBasebandFile(int fd, int32_t tOffsetSeconds, 
    FileReaderBasebandPacketCallbackFunction BasebandPacketCallback,
    ENATSC3ALPLLSPLPsCallback LLSPLPsCallback, int32_t *pFRunning,
    void *pUserData)
{
    ENATSC3CapFileHeader fileHeader;
    uint8_t *pPayload = NULL;
    uint64_t basebandPacketNumber = 0;
    uint64_t llsValidMask = 0;
    int64_t tStartReading = -1;
    int32_t lastCC = -1;
    int32_t rval = 0;

    pPayload = (uint8_t *)malloc(MAX_PAYLOAD_SIZE);
    if (!pPayload)
    {
        ERRORPRINT("Out of memory\n");
        rval = 1;
        goto out;
    }

    if (read(fd, &fileHeader, sizeof(fileHeader)) != sizeof(fileHeader))
    {
        ERRORPRINT("Failed to read baseband file header\n");
        rval = 1;
        goto out;
    }

    if (fileHeader.type != ENATSC3CapType_ATSC3_Baseband)
    {
        ERRORPRINT("Unhandled file type %d\n", fileHeader.type);
        rval = 1;
        goto out;
    }

    while (*pFRunning)
    {
        ENATSC3CapRecordHeaderV2 header;
        int nb;
        int32_t payloadLength;
        int32_t fCCError;
        struct timespec packetTime;

        nb = read(fd, &header, sizeof(header));
        if (nb != sizeof(header))
        {
            if (nb == 0)
            {
                break;
            }
            else if (nb < 0)
            {
                ERRORPRINT("Error reading packet header: %s\n",
                    strerror(errno));
                rval = 1;
                break;
            }
            else
            {
                ERRORPRINT(
                    "Short read of packet header (%d != %zd)\n",
                    nb, sizeof(header));
                rval = 1;
                break;
            }
        }

        header.length = ntohl(header.length);
        header.rtwhen.sec = ENUtils_ntohll(header.rtwhen.sec);
        header.rtwhen.nsec = ntohl(header.rtwhen.nsec);

        if (header.length > MAX_PAYLOAD_SIZE)
        {
            ERRORPRINT("Baseband packet too large (%d > %d)\n",
                header.length, MAX_PAYLOAD_SIZE);
            rval = 1;
            break;
        }

        payloadLength = header.length;
        packetTime.tv_sec = header.rtwhen.sec;
        packetTime.tv_nsec = header.rtwhen.nsec;

        nb = read(fd, pPayload, header.length);
        if (nb != header.length)
        {
            if (nb == 0)
            {
                ERRORPRINT("EOF reading packet payload\n");
            }
            else if (nb < 0)
            {
                ERRORPRINT("Error reading baseband packet: %s\n",
                    strerror(errno));
            }
            else
            {
                ERRORPRINT("Short read of baseband packet (%d != %zd)\n",
                    nb, sizeof(header));
            }

            rval = 1;
            break;
        }

        if (tStartReading == -1)
        {
            tStartReading = header.rtwhen.sec + tOffsetSeconds;
        }

        if (header.rtwhen.sec < tStartReading)
        {
            continue;
        }

        if (lastCC != -1)
        {
            if (header.continuityCounter != ((lastCC + 1) & 0xff))
            {
                fCCError = 1;
            }
        }
        lastCC = header.continuityCounter;

        if (header.type == ENATSC3CapPayloadType_Demod_V2)
        {
            uint64_t currentLLSValidMask;

            if (fileHeader.format == ENATSC3CapFormat_V2)
            {
                ENATSC3CapDemodulatorRecordV2 *pDemodRecord;

                // This is a tuner/demod statistics packet
                pDemodRecord = (ENATSC3CapDemodulatorRecordV2 *)pPayload;
                currentLLSValidMask = 
                    ENUtils_ntohll(pDemodRecord->atsc3.llsPresentMask);
            }
            else
            {
                ERRORPRINT("Unhandled format %d\n", fileHeader.format);
                break;
            }

            if (currentLLSValidMask != llsValidMask)
            {
                if (LLSPLPsCallback)
                {
                    LLSPLPsCallback(currentLLSValidMask, 
                        &packetTime, basebandPacketNumber, pUserData);
                }
                llsValidMask = currentLLSValidMask;
            }
            continue;
        }

        if (header.type != ENATSC3CapPayloadType_ATSC3_Baseband)
        {
            ERRORPRINT("Unexpected capture record type %d\n",
                header.type);
            continue;
        }

        BasebandPacketCallback(header.plpId, &packetTime, basebandPacketNumber, 
            header.fError, fCCError, pPayload, payloadLength, pUserData);

        basebandPacketNumber++;
    }
 
out:
    if (pPayload)
    {
        free(pPayload);
    }

    return rval;
}

// GetALPPacketLength parses ALP packet headers to determine the packet length.
// This is required to determine ALP packet boundaries within a raw ALP
// stream.
static int32_t
GetALPPacketLength(uint8_t *pPacket, int32_t packetLength, int32_t *pLength)
{
    uint32_t payloadConfiguration;
    int32_t headerLength;
    int32_t payloadLength;
    int32_t rval = 0;

    // The minimum ALP header length is 2 bytes
    if (packetLength < 2)
    {
        *pLength = 2;
        rval = 2;
        goto out;
    }

    payloadConfiguration = pPacket[0] & 0x10;

    if (payloadConfiguration == 0)
    {
        // Payload configuration = 0: single packet (with optional header)
        payloadLength = ((pPacket[0] & 0x7) << 8) | pPacket[1];

        if (pPacket[0] & 0x8) // header mode
        {
            // Additional Header for Single Packet

            if (packetLength < 3)
            {
                *pLength = 3;
                rval = 2;
                goto out;
            }

            payloadLength |= ((pPacket[2] >> 3)) << 11;
            headerLength = 3;

            if (pPacket[2] & 0x2) // Substream Identifer Flag
            {
                headerLength++;
            }

            if (pPacket[2] & 0x1) // Header extension flag
            {
                int32_t extensionLengthMinus1;

                if (headerLength + 2 > packetLength)
                {
                    *pLength = headerLength + 2;
                    rval = 2;
                    goto out;
                }

                // HEF present
                extensionLengthMinus1 = pPacket[headerLength + 1];

                headerLength += 2 + extensionLengthMinus1 + 1;
            }
        }
        else
        {
            headerLength = 2;
        }
    }
    else
    {
        if (pPacket[0] & 0x8) // Concatenation/segmentation
        {
            int32_t nSegments;

            // Concatenation mode.

            if (packetLength < 3)
            {
                *pLength = 3;
                rval = 2;
                goto out;
            }

            payloadLength = ((pPacket[0] & 0x7) << 8) |
                pPacket[1] | ((pPacket[2] >> 4) << 11);
            nSegments = ((pPacket[2] >> 1) & 0x7) + 2;

            // Advance headerLength over the segment lengths. There are
            // nSegments - 1 lengths. The length of the last segment is
            // implictly signalled as the end of the next-to-last segment until
            // the end of the ALP packet.
            headerLength = 3 + ((nSegments - 2 + 1) * 12 + 4) / 8;

            // Advance over the substream identifiers, if present.
            if (pPacket[2] & 1) // Substream Identifier Flag
            {
                // XXX A/330 is a bit vague on this but suggests that there is
                // one SID per segment.
                headerLength += nSegments;
            }
        }
        else
        {
            // Segmentation mode.

            if (packetLength < 3)
            {
                *pLength = 3;
                rval = 2;
                goto out;
            }

            payloadLength = ((pPacket[0] & 0x7) << 8) | pPacket[1];
            headerLength = 3;

            if (pPacket[2] & 0x2) // Substream Identifer Flag
            {
                if (headerLength + 1 > packetLength)
                {
                    *pLength = headerLength + 1;
                    rval = 2;
                    goto out;
                }

                headerLength++;
            }

            if (pPacket[2] & 0x1) // Header extension flag
            {
                int32_t extensionLengthMinus1;

                if (headerLength + 2 > packetLength)
                {
                    *pLength = headerLength + 2;
                    rval = 2;
                    goto out;
                }

                // extensionLengthMinus1 is the number of bytes following
                // extensionLengthMinus1 less one byte.
                extensionLengthMinus1 = pPacket[headerLength + 1];

                if (headerLength + 2 + extensionLengthMinus1 + 1 >
                    packetLength)
                {
                    *pLength = headerLength + 2 + extensionLengthMinus1 + 1;
                    rval = 2;
                    goto out;
                }

                headerLength += 2 + extensionLengthMinus1 + 1;
            }
        }
    }
    // LLS packets have an Additional Header for Signaling Information that is
    // five bytes long
    if ((pPacket[0] >> 5) == 4)
    {
        headerLength += 5;
    }

    *pLength = headerLength + payloadLength;

    assert(*pLength > 0);

out:
    return rval;
}

int32_t
ENATSC3FileReaderReadALPRawFile(int fd, 
    FileReaderALPPacketCallbackFunction ALPPacketCallback,
    ENATSC3ALPLLSPLPsCallback LLSPLPsCallback, int32_t *pFRunning,
    void *pUserData)
{
    uint64_t alpPacketNumber = 0;
    uint8_t *pPayload = NULL;
    int32_t rval = 0;
    int32_t plpId = 0;
    struct timespec packetTime = 
    {
        .tv_sec = 0,
        .tv_nsec = 0
    };

    pPayload = (uint8_t *)malloc(MAX_PAYLOAD_SIZE);
    if (!pPayload)
    {
        ERRORPRINT("Out of memory\n");
        rval = 1;
        goto out;
    }

    while (*pFRunning)
    {
        int nb;
        int32_t packetLength;
        int32_t offset = 0;

        while (1)
        {
            nb = read(fd, pPayload + offset, 1);
            if (nb != 1)
            {
                if (nb == 0)
                {
                    ERRORPRINT("EOF reading packet payload\n");
                }
                else if (nb < 0)
                {
                    ERRORPRINT("Error reading ALP packet: %s\n", 
                        strerror(errno));
                }
                else
                {
                    ERRORPRINT("Short read of ALP packet (%d != %d)\n", 
                        nb, 2);
                }
                rval = 1;
                break;
            }
            offset++;

            if (GetALPPacketLength(pPayload, offset, &packetLength) == 0)
            {
                break;
            }
        }

        if (rval)
        {
            break;
        }

        nb = read(fd, pPayload + offset, packetLength - offset);
        if (nb != packetLength - offset)
        {
            if (nb == 0)
            {
                ERRORPRINT("EOF reading packet payload\n");
            }
            else if (nb < 0)
            {
                ERRORPRINT("Error reading ALP packet: %s\n", 
                    strerror(errno));
            }
            else
            {
                ERRORPRINT("Short read of ALP packet (%d != %d)\n", 
                    nb, packetLength - offset);
            }

            rval = 1;
            break;
        }

        offset += packetLength - offset;

        ALPPacketCallback(plpId, &packetTime, alpPacketNumber, 0, 0, pPayload, 
            offset, pUserData);

        alpPacketNumber++;
    }

out:
    if (pPayload)
    {
        free(pPayload);
    }

    return rval;
}

#define MAGIC_MICROSECONDS 0xa1b2c3d4
#define MAGIC_NANOSECONDS  0xa1b23c4d

typedef struct
{
    uint32_t magic_number;
    uint16_t version_major;
    uint16_t version_minor;
    int32_t thiszone;
    uint32_t sigfigs;
    uint32_t snaplen;
    uint32_t network;
} __attribute__((__packed__)) pcap_hdr_t;

typedef struct
{
    uint32_t ts_sec;
    uint32_t ts_usec;
    uint32_t incl_len;  // Number of bytes of packet data saved to file.
    uint32_t orig_len;  // The original length of the packet as it appeared on
                        // the network when it was captured. If incl_len and
                        // org)len differ, the actually saved pacekt size was
                        // limited by snaplen.
} __attribute__((__packed__)) pcaprec_hdr_t;

typedef struct
{
    pcaprec_hdr_t header;
    struct
    {
        uint8_t vDestinationMac[6];
        uint8_t vSourceMAC[6];
        uint16_t type;
    } __attribute__((__packed__)) ethernetHeader;
    uint8_t vPacketData[0];
} __attribute__((__packed__)) pcaprec_packet;

int32_t
ENATSC3FileReaderReadPCAPIPV4File(int fd, int32_t tOffsetSeconds, 
    FileReaderIPV4PacketCallbackFunction IPV4PacketCallback,
    ENATSC3ALPLLSPLPsCallback LLSPLPsCallback, 
    int32_t *pFRunning, void *pUserData)
{
    pcap_hdr_t header;
    pcaprec_packet *pPacket;
    int32_t fSwapped = 0;
    int32_t fNanoseconds = 0;
    uint8_t *pPayload = NULL;
    struct timespec tPacket;
    uint8_t vBuffer[4096];
    int64_t packetNumber = 0;
    int32_t fLLSSet = 0;;
    int nb;
    int32_t rval = 0;

    pPacket = (pcaprec_packet *)vBuffer;
    nb = read(fd, &header, sizeof(header));
    if (nb != sizeof(header))
    {
        ERRORPRINT("Failed to read header\n");
        rval = 1;
        goto out;
    }

    if (header.magic_number == MAGIC_MICROSECONDS)
    {
        fSwapped = 0;
        fNanoseconds = 0;
    }
    else if (header.magic_number == MAGIC_NANOSECONDS)
    {
        fSwapped = 0;
        fNanoseconds = 1;
    }
    else if (ntohl(header.magic_number) == MAGIC_MICROSECONDS)
    {
        fSwapped = 1;
        fNanoseconds = 0;
    }
    else if (ntohl(header.magic_number) == MAGIC_NANOSECONDS)
    {
        fSwapped = 1;
        fNanoseconds = 1;
    }
    else
    {
        ERRORPRINT("Unknown magic: %x\n", header.magic_number);
        rval = 1;
        goto out;
    }

    while (*pFRunning)
    {
        nb = read(fd, pPacket, sizeof(pPacket->header));
        if (nb == 0)
        {
            break;
        }
        else if (nb == -1)
        {
            ERRORPRINT("Failed to read packet header: %s\n",
                strerror(errno));
            rval = 1;
            goto out;
        }
        else if (nb < sizeof(pPacket->header))
        {
            ERRORPRINT("Short read of packet header\n");
            rval = 1;
            goto out;
        }

        if (fSwapped)
        {
            pPacket->header.ts_sec = ntohl(pPacket->header.ts_sec);
            pPacket->header.ts_usec = ntohl(pPacket->header.ts_usec);
            pPacket->header.incl_len = ntohl(pPacket->header.incl_len);
            pPacket->header.orig_len = ntohl(pPacket->header.orig_len);
        }

        if (!fNanoseconds)
        {
            pPacket->header.ts_usec *= 1000;
        }

        if (pPacket->header.orig_len > sizeof(vBuffer) -
            sizeof(pPacket->header))
        {
            ERRORPRINT("Packet too large (%d > %zd)\n",
                pPacket->header.orig_len, sizeof(vBuffer) -
                sizeof(pPacket->header));
            rval = 1;
            goto out;
        }

        nb = read(fd, &pPacket->ethernetHeader,
            sizeof(pPacket->ethernetHeader));
        if (nb == 0)
        {
            ERRORPRINT("EOF on ethernet header\n");
            rval = 1;
            break;
        }
        else if (nb == -1)
        {
            ERRORPRINT("Failed to read ethernet header: %s\n",
                strerror(errno));
            rval = 1;
            goto out;
        }
        else if (nb < sizeof(pPacket->ethernetHeader))
        {
            ERRORPRINT("Short read of ethernet header\n");
            rval = 1;
            goto out;
        }

        nb = read(fd, pPacket->vPacketData, pPacket->header.incl_len -
            sizeof(pPacket->ethernetHeader));
        if (nb == 0)
        {
            ERRORPRINT("EOF on packet data\n");
            rval = 1;
            break;
        }
        else if (nb == -1)
        {
            ERRORPRINT("Failed to read packet data: %s\n",
                strerror(errno));
            rval = 1;
            goto out;
        }
        else if (nb < pPacket->header.incl_len -
                sizeof(pPacket->ethernetHeader))
        {
            ERRORPRINT("Short read of packet body\n");
            rval = 1;
            goto out;
        }

        if (ntohs(pPacket->ethernetHeader.type == 0x08))
        {
            tPacket.tv_sec = pPacket->header.ts_sec;
            tPacket.tv_nsec = pPacket->header.ts_usec;

            if (!fLLSSet)
            {
                // We don't know which PLPs are present, so assume PLP 0
                if (LLSPLPsCallback)
                {
                    LLSPLPsCallback(1, &tPacket, packetNumber, pUserData);
                }
                fLLSSet = 1;
            }

            IPV4PacketCallback(0, &tPacket, packetNumber, 
                pPacket->vPacketData, 
                pPacket->header.orig_len - sizeof(pPacket->ethernetHeader),  
                pUserData);
        }

        packetNumber++;
    }

out:
    if (pPayload)
    {
        free(pPayload);
    }

    return rval;
}
