/*
 * Copyright © 2020-2022 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <enatsc3-common/enatsc3.h>
#include "libenatsc3j288alp.h"

static char *ensys = "j288alp";
static char *ensubsys = "main";

#define INITIAL_ALP_PACKET_LENGTH 2048
#define MAX_ALP_PACKET_LENGTH (65536 + 1024)

#define INVALID_PLP_ID 0xffffffff

// All packets from the Quatro 4K seem to come on this pid.
#define DEMOD_PID_VALUE 0x2D

// T-REC J.288 encapsulates TLV data in MPEG2-TS packets.  J.288 TS packets
// have the following structure:
//
// Packet Header: 3 bytes
// top_pointer_field: 0 or 1 bytes (present if TLV_start_indicator is true in 
//     the packet header)
// Payload (184 bytes if the top_pointer_field is present, 185 bytes otherwise)
//
// The Packet Header has the following structure:
// sync_byte                    8 bits  0x47
// transport_error_indicator    1 bit   Indicates that at least one error has
//                                      occurred in the packet
// TLV_start_indicator          1 bit   A value of '1' indicates the presence
//                                      of at least one start of a TLV packet.
//                                      A value of '0' indicates no start of a
//                                      TLV packet.
// '0'                          1 bit
// PID                          13 bits A uniqe value for a fragmented TLV
//                              packet.
//
// In addition to encapsulating the ALP packets, it appears that the ALP
// stream is modified to include additional ALP packets with user-private
// extension headers. These ALP packets do not occur in the emission but are
// generated from L1 information. There appear to be two generated extension 
// headers, one carrying the PTP and one that signals the PLP ID of the next 
// ALP packet.
//
// The PTP extension payload for the most part follows A/322 except that the
// L1B_time_info_flag is prepended to the packed L1D_time_info. The 
// L1B_time_info_flag indicates the precision of PTP time and the validity of 
// the fields that follow. See A/322.
//
// The PTP payload is 8 bytes (64 bits) packed as follows:
//
// L1B_time_info_flag          2 bits
// L1D_time_sec                32 bits
// L1D_time_msec               10 bits
// L1D_time_usec               10 bits
// L1D_time_nsec               10 bits
//
// The PLP extension payload is a single byte (8 bits). The upper 8 bits carry
// a PLP ID. The lower two bits appear to be stuffing.
//
// PLP_ID                       6 bits
// ?????                        2 bits
//
// The PLP extension appears to only be present if more than one PLP is 
// captured. (If only a single PLP is captured then strictly speaking the 
// PLP extension header is not required since we know which PLP it came from.)

#define TRANSPORT_ERROR_INDICATOR       0x80
#define TLV_START_INDICATOR             0x40

// The demod seems to manufacture ALP packets to carry information like the PTP
// and PLP Id of the ALP packet being transmitted. The PTP would normally be
// transmitted as part of L1D, and the PLP would be implicit in the PLP that 
// the ALP came across. 
//
// These ALP packets appear to have 0xF0 and 0xF1 extension headers only and
// no other headers or payload. 
#define ALP_EXTENSION_HEADER_USER_PRIVATE_PTP 0xF0
#define ALP_EXTENSION_HEADER_USER_PRIVATE_PLP 0xF1

typedef struct _ENATSC3J288ALPHandle_t
{
    ENVerbosity verbosity;

    // Callbacks
    ENATSC3J288ALPCallbackFunction pALPCallback;
    ENATSC3J288BasebandCallbackFunction pBasebandCallback;
    ENATSC3J288PTPCallbackFunction pPTPCallback;
    void *pCallbackUserData;

    // Indicates that we are receiving one of the demod-generated ALP
    // packets noted above.
    int32_t fDemodPacket;

    // The PLP Id of the ALP packet being received. It is assumed that once
    // we receive a PLP extension header that the PLP Id is "sticky" and
    // applies to all ALP packets until we get another PLP Id extension 
    // header.
    uint32_t plpId;

    // Whether we have locked to the TS sync byte
    int32_t fTSLock;

    // Whether we are locked to ALP frames
    int32_t fALPLock;

    // The TS packet buffer is used to accumulate TS data, either while
    // acquiring lock or when we receive a partial TS Packet
    uint8_t vTSPacketBuffer[188 * 3];
    int32_t nTSPacketBufferBytesUsed;

    uint8_t vALPPacket[MAX_ALP_PACKET_LENGTH];
    int32_t fError;
    int32_t alpPacketBytesReceived;
    int32_t alpPacketLength;

    // The time at which we get the first bit of data for the ALP packet.
    struct timespec tALP;

    // A counter of ALP packets processed
    uint64_t packetNumber;
} ENATSC3J288ALP, *ENATSC3J288ALPHandle_t;

#ifdef AUTODETECT_KOREAN_MODE
static ENATSC3Status
ParseLMT(uint8_t *pPayload, int32_t payloadLength)
{
    int32_t numPLPs;
    int32_t i;
    int32_t j;
    ENATSC3Status rval = ENATSC3Status_Success;

    if (!payloadLength)
    {
        rval = ENATSC3Status_Failure;
        goto out;
    }
    numPLPs = (pPayload[0] >> 2) + 1;
    pPayload++;
    payloadLength--;

    for (i = 0; i < numPLPs; i++)
    {
        uint32_t numMulticasts;

        if (payloadLength < 2)
        {
            rval = ENATSC3Status_Failure;
            goto out;
        }
        numMulticasts = pPayload[1];
        pPayload += 2;
        payloadLength -= 2;

        for (j = 0; j < numMulticasts; j++)
        {
            uint32_t sidFlag;
            uint32_t compressedFlag;

            if (payloadLength < 13)
            {
                rval = ENATSC3Status_Failure;
                goto out;
            }

            sidFlag = pPayload[12] & 0x80;
            compressedFlag = pPayload[12] & 0x40;

            pPayload += 13;
            payloadLength -= 13;

            if (sidFlag)
            {
                if (payloadLength < 1)
                {
                    rval = ENATSC3Status_Failure;
                    goto out;
                }
                pPayload++;
                payloadLength--;
            }

            if (compressedFlag)
            {
                if (payloadLength < 1)
                {
                    rval = ENATSC3Status_Failure;
                    goto out;
                }
                pPayload++;
                payloadLength--;
            }
        }
    }

out:
    return rval;
}

static ENATSC3Status
ParseALPSignalingPayload(uint8_t *pPayload, int32_t payloadLength)
{
    ENATSC3Status rval = ENATSC3Status_Success;

    if (payloadLength < 5)
    {
        rval = ENATSC3Status_Failure;
        goto out;
    }

    if (pPayload[0] == 1)
    {
        // We only handle the LMT
        pPayload += 5;
        payloadLength -= 5;
        rval = ParseLMT(pPayload, payloadLength);
    }
    else
    {
        rval = ENATSC3Status_Retry;
    }

out:
    return rval;
}
#endif

// GetALPPacketLength parses ALP packet headers to the extent required to 
// deterine the length of the packet. It is only required at the baseband
// packet processing level and below because it is the only way to determine 
// where ALP packet boundaries are.
static ENATSC3Status
GetALPPacketLength(ENATSC3J288ALP *pJ288ALP, struct timespec *pPacketTimeUTC,
    uint8_t *pPacket, int32_t packetLength, int32_t *pLength)
{
    uint32_t payloadConfiguration;
    int32_t headerLength;
    int32_t payloadLength;
    ENATSC3Status rval = ENATSC3Status_Success;

    // The minimum ALP header length is 2 bytes
    if (packetLength < 2)
    {
        rval = ENATSC3Status_Retry;
        goto out;
    }

    payloadConfiguration = pPacket[0] & 0x10;

    if (payloadConfiguration == 0)
    {
        // Payload configuration = 0: single packet (with optional header)
        payloadLength = ((pPacket[0] & 0x7) << 8) | pPacket[1];

        if (pPacket[0] & 0x8) // header mode
        {
            // Additional Header for Single Packet

            if (packetLength < 3)
            {
                rval = ENATSC3Status_Retry;
                goto out;
            }

            payloadLength |= ((pPacket[2] >> 3)) << 11;
            headerLength = 3;

            if (pPacket[2] & 0x2) // Substream Identifer Flag
            {
                headerLength++;
            }

            if (pPacket[2] & 0x1) // Header extension flag
            {
                int32_t extensionType;
                int32_t extensionLengthMinus1;

                if (headerLength + 2 > packetLength)
                {
                    rval = ENATSC3Status_Retry;
                    goto out;
                }

                // HEF present
                extensionType = pPacket[headerLength];
                extensionLengthMinus1 = pPacket[headerLength + 1];

                if (headerLength + 2 + extensionLengthMinus1 + 1 >
                    packetLength)
                {
                    rval = ENATSC3Status_Retry;
                    goto out;
                }

                if (extensionType == ALP_EXTENSION_HEADER_USER_PRIVATE_PTP)
                {
                    ENATSC3J288PTPInfo ptpInfo;
                    uint64_t raw;

                    if (!payloadLength)
                    {
                        pJ288ALP->fDemodPacket = 1;
                    }
                    else
                    {
                        // We're not going to mark this as a demod packet and
                        // just pass it up.
                        ERRORPRINT("WARN: Got PTP packet with payload!\n");
                    }

                    raw =
                        ((uint64_t)(pPacket[headerLength + 2])) << 56 |
                        ((uint64_t)(pPacket[headerLength + 3])) << 48 |
                        ((uint64_t)(pPacket[headerLength + 4])) << 40 |
                        ((uint64_t)(pPacket[headerLength + 5])) << 32 |
                        ((uint64_t)(pPacket[headerLength + 6])) << 24 |
                        ((uint64_t)(pPacket[headerLength + 7])) << 16 |
                        ((uint64_t)(pPacket[headerLength + 8])) << 8 |
                        ((uint64_t)(pPacket[headerLength + 9]));
                    ptpInfo.L1B_time_info_flag = raw >> 62;
                    ptpInfo.L1D_time_sec = (raw >> 30) & 0xffffffff;
                    ptpInfo.L1D_time_msec = (raw >> 20) & 0x3ff;
                    ptpInfo.L1D_time_usec = (raw >> 10) & 0x3ff;
                    ptpInfo.L1D_time_nsec = raw & 0x3ff;
                    if (pJ288ALP->pPTPCallback)
                    {
                        pJ288ALP->pPTPCallback(pPacketTimeUTC, 
                            pJ288ALP->packetNumber, &ptpInfo, 
                            pJ288ALP->pCallbackUserData);
                    }
                }
                else if (extensionType == ALP_EXTENSION_HEADER_USER_PRIVATE_PLP)
                {
                    if (!payloadLength)
                    {
                        pJ288ALP->fDemodPacket = 1;
                    }
                    else
                    {
                        // We're not going to mark this as a demod packet and
                        // just pass it up.
                        ERRORPRINT("WARN: Got PLP packet with payload!\n");
                    }

                    pJ288ALP->plpId = pPacket[headerLength + 2] >> 2;
                }

                headerLength += 2 + extensionLengthMinus1 + 1;
            }
        }
        else
        {
            headerLength = 2;
        }
    }
    else
    {
        if (pPacket[0] & 0x8) // Concatenation/segmentation
        {
            int32_t nSegments;

            // Concatenation mode. 

            if (packetLength < 3)
            {
                rval = ENATSC3Status_Retry;
                goto out;
            }

            payloadLength = ((pPacket[0] & 0x7) << 8) | 
                pPacket[1] | ((pPacket[2] >> 4) << 11);
            nSegments = ((pPacket[2] >> 1) & 0x7) + 2;

            // Advance headerLength over the segment lengths. There are
            // nSegments - 1 lengths. The length of the last segment is
            // implictly signalled as the end of the next-to-last segment until
            // the end of the ALP packet.
            headerLength = 3 + ((nSegments - 2 + 1) * 12 + 4) / 8;

            // Advance over the substream identifiers, if present.
            if (pPacket[2] & 1) // Substream Identifier Flag
            {
                // XXX A/330 is a bit vague on this but suggests that there is
                // one SID per segment.
                headerLength += nSegments;
            }
        }
        else
        {
            // Segmentation mode.

            if (packetLength < 3)
            {
                rval = ENATSC3Status_Retry;
                goto out;
            }

            payloadLength = ((pPacket[0] & 0x7) << 8) | pPacket[1];
            headerLength = 3;

            if (pPacket[2] & 0x2) // Substream Identifer Flag
            {
                if (headerLength + 1 > packetLength)
                {
                    rval = ENATSC3Status_Retry;
                    goto out;
                }

                headerLength++;
            }

            if (pPacket[2] & 0x1) // Header extension flag
            {
                int32_t extensionLengthMinus1;

                if (headerLength + 2 > packetLength)
                {
                    rval = ENATSC3Status_Retry;
                    goto out;
                }

                // extensionLengthMinus1 is the number of bytes following
                // extensionLengthMinus1 less one byte.
                extensionLengthMinus1 = pPacket[headerLength + 1];

                if (headerLength + 2 + extensionLengthMinus1 + 1 > 
                    packetLength)
                {
                    rval = ENATSC3Status_Retry;
                    goto out;
                }

                headerLength += 2 + extensionLengthMinus1 + 1;
            }
        }
    }

    // LLS packets have an Additional Header for Signaling Information that is
    // five bytes long
    if ((pPacket[0] >> 5) == 4)
    {
#ifdef AUTODETECT_KOREAN_MODE
        // To detect Korean Mode we need to determine if the Additional Header
        // For Signaling Information is included in the payload length or not.
        // To do this we assume that it is included in the payload length. If 
        // after acquiring the entire packet we are able to successfully parse 
        // it then we assume we are in Korean Mode. If we are not able to parse
        // the packet, then we assume we are not in Korean Mode. This test will 
        // not work if the ALP packet has five or more bytes of padding at the 
        // end of the signaling table, since in that case the signaling table 
        // will parse correctly in either case.
        if (!pJ288ALP->fKoreanModeDetermined)
        {
            ENATSC3Status lret;

            if (packetLength < headerLength + payloadLength)
            {
                rval = ENATSC3Status_Retry;
                goto out;
            }

            lret = ParseALPSignalingPayload(pPacket + headerLength, 
                payloadLength);
            if (lret == ENATSC3Status_Retry)
            {
                // This most likely means that a signaling table other than
                // an LMT arrived. 
                rval = ENATSC3Status_Retry;
                goto out;
            }
            else if (lret == ENATSC3Status_Failure)
            {
                pJ288ALP->fKoreanMode = 0;
            }
            else if (lret == ENATSC3Status_Success)
            {
                pJ288ALP->fKoreanMode = 1;
            }
            else
            {
                ERRORPRINT("Unhandled probe return %d\n", lret);
                rval = ENATSC3Status_Retry;
                goto out;
            }
        }

        pJ288ALP->fKoreanModeDetermined = 1;

        if (pJ288ALP->fKoreanMode)
        {
            pPacket[1] -= 5;
        }
        else
        {
            headerLength += 5;
        }
#else
        // Assume we are in non-Korean mode
        headerLength += 5;
#endif
    }

    *pLength = headerLength + payloadLength;
    //LOGPRINT_NOFILE("HEADERLEN %d PAYLOAD %d\n", headerLength, payloadLength);

    ENUtils_CrashAssert(*pLength > 0);

out:
    return rval;
}

void
StartNewALP(ENATSC3J288ALP *pJ288ALP)
{
    pJ288ALP->plpId = 255;
    pJ288ALP->alpPacketBytesReceived = 0;
    pJ288ALP->alpPacketLength = 0;
    pJ288ALP->fDemodPacket = 0;
    pJ288ALP->fError = 0;
}

void
SendALPPacketAsBaseband(ENATSC3J288ALP *pJ288ALP, uint32_t plpId, 
    uint8_t *pALPPacket, int32_t alpPacketLength, 
    struct timespec *pPacketTimeUTC)
{
    int32_t fFirst = 1;
    uint8_t vBasebandPacket[8192];

    while (alpPacketLength)
    {
        int32_t nToCopy = MIN(alpPacketLength, 8190);

        // The baseband packet header is two bytes consisting of the mode and
        // ofi, and a pointer to the first (ALP) packet within the baseband 
        // packet. For simplicity we will use mode=1 and ofi=0. We're not going 
        // to try to merge multiple ALP packets into a single baseband, so the 
        // first byte of the ALP packet starts a baseband packet, and subsequent
        // packets have pointer = 8191 (n ALP packets starts in the baseband
        // packet)
        if (fFirst)
        {
            vBasebandPacket[0] = 0x80;
            vBasebandPacket[1] = 0;
            fFirst = 0;
        }
        else
        {
            vBasebandPacket[0] = 0x80 | (8191 >> 6);
            vBasebandPacket[1] = (8191 & 0x3f) << 2;
        }

        memcpy(vBasebandPacket + 2, pALPPacket, nToCopy);
        pALPPacket += nToCopy;
        alpPacketLength -= nToCopy;

        pJ288ALP->pBasebandCallback(plpId, pPacketTimeUTC, 
            pJ288ALP->packetNumber, pJ288ALP->fError, vBasebandPacket, 8192, 
            pJ288ALP->pCallbackUserData);
    }
}

void
SendALPPacket(ENATSC3J288ALP *pJ288ALP, uint32_t plpId, uint8_t *pALPPacket, 
    int32_t alpPacketLength, struct timespec *pPacketTimeUTC)
{
    if (plpId != INVALID_PLP_ID && !pJ288ALP->fDemodPacket)
    {
        if (pJ288ALP->pALPCallback)
        {
            pJ288ALP->pALPCallback(plpId, pPacketTimeUTC, 
                pJ288ALP->packetNumber, pJ288ALP->fError, pALPPacket, 
                alpPacketLength, pJ288ALP->pCallbackUserData);
        }

        if (pJ288ALP->pBasebandCallback)
        {
            SendALPPacketAsBaseband(pJ288ALP, plpId, pALPPacket, 
                alpPacketLength, pPacketTimeUTC);
        }
    }
}

void
ALPPacketLockLost(ENATSC3J288ALP *pJ288ALP)
{
    // If we've received a partial ALP packet then send it off.
    if (pJ288ALP->alpPacketBytesReceived)
    {
        SendALPPacket(pJ288ALP, pJ288ALP->plpId,  pJ288ALP->vALPPacket,
            pJ288ALP->alpPacketBytesReceived, &pJ288ALP->tALP);
    }

    StartNewALP(pJ288ALP);
}

ENATSC3Status
ProcessJ288PacketPayload(ENATSC3J288ALP *pJ288ALP, 
    struct timespec *pPacketTimeUTC, const uint8_t *pPayload, 
    int32_t payloadLength, int32_t fError, int32_t fEnd,
    int32_t *pNConsumed)
{
    int32_t nConsumed = 0;
    ENATSC3Status rval = ENATSC3Status_Success;

    ENUtils_CrashAssert(payloadLength >= 0);

    // Check if the ALP payload has overflowed. This means that there is an
    // error in the stream. Rather than discarding the packet we will pass it
    // up and note that the data may contain an error. We will also discard
    // future bytes until a J288 packet is received with the pointer set, 
    // indicating the start of an ALP packet.
    if (pJ288ALP->alpPacketBytesReceived + payloadLength >
        MAX_ALP_PACKET_LENGTH)
    {
        ERRORPRINT("ALP packet overflows buffer (len %d of %d)\n", 
            pJ288ALP->alpPacketBytesReceived + payloadLength,
            pJ288ALP->alpPacketLength);
        payloadLength = MAX_ALP_PACKET_LENGTH - 
            pJ288ALP->alpPacketBytesReceived;
        pJ288ALP->fError = 1;
    }

    if (!pJ288ALP->alpPacketBytesReceived)
    {
        pJ288ALP->tALP = *pPacketTimeUTC;
    }

    memcpy(pJ288ALP->vALPPacket + pJ288ALP->alpPacketBytesReceived, 
        pPayload, payloadLength);
    pJ288ALP->alpPacketBytesReceived += payloadLength;
    nConsumed = payloadLength;
    pJ288ALP->fError |= fError;

    // Check if we need to compute the ALP packet length. The only way to 
    // determine it is to parse the ALP packet headers. It is possible that
    // not enough of the ALP packet has been received to determine the length,
    // in which case we will need to retyr later.
    if (!pJ288ALP->alpPacketLength)
    {
        ENATSC3Status lret;

        lret = GetALPPacketLength(pJ288ALP, pPacketTimeUTC, 
            pJ288ALP->vALPPacket, pJ288ALP->alpPacketBytesReceived, 
            &pJ288ALP->alpPacketLength);
        if (lret)
        {
            if (lret != ENATSC3Status_Retry)
            {
                ERRORPRINT("Error determining ALP packet length\n");
                rval = ENATSC3Status_Failure;
                StartNewALP(pJ288ALP);
            }
            else
            {
                if (fEnd)
                {
                    // We reached the end of the packet without even being able 
                    // to parse the header? We're going to discard.
                    ERRORPRINT("Failed to determine ALP length before end\n");
                    StartNewALP(pJ288ALP);
                }
            }

            goto out;
        }
    }

    if (pJ288ALP->alpPacketBytesReceived > pJ288ALP->alpPacketLength)
    {
        // If fEnd is set then we know where the packet ended based on the
        // pointer in the TS packet. (This is a special case - In general we 
        // need to rely on the ALP packet length determined by parsing the ALP 
        // headers.) 
        if (fEnd)
        {
            ERRORPRINT("ALP packet too large (%d > %d)\n", 
                pJ288ALP->alpPacketBytesReceived, pJ288ALP->alpPacketLength);
            pJ288ALP->fError = 1;
        }
        else
        {
            // While blindly copying the TS payload it appears we copied too
            // much. This can happen since a TS payload can straddle more
            // than two ALP packets - for example the last few bytes of one 
            // ALP packet, the entirety of a second small ALP packet, and 
            // the beginning of a third ALP packet. The beginning of the 
            // second ALP packet will be signaled by the pointer field, 
            // however the beginning of the third ALP packet must be 
            // inferred from the size of the second ALP packet as indicated
            // by its headers.
            //
            // "Un-receive" the extra bytes
            nConsumed -= pJ288ALP->alpPacketBytesReceived -
                pJ288ALP->alpPacketLength;
            ENUtils_CrashAssert(nConsumed > 0);
            pJ288ALP->alpPacketBytesReceived = pJ288ALP->alpPacketLength;
        }
    }

    // At this point the ALP packet length is known
    if (pJ288ALP->alpPacketBytesReceived >= pJ288ALP->alpPacketLength || 
        fEnd)
    {
        if (pJ288ALP->alpPacketBytesReceived > pJ288ALP->alpPacketLength)
        {
            ERRORPRINT("LONG ALP packet received length %d > length %d\n",
                pJ288ALP->alpPacketBytesReceived, 
                pJ288ALP->alpPacketLength);
        }
        else if (fEnd && 
            pJ288ALP->alpPacketBytesReceived != pJ288ALP->alpPacketLength)
        {
            ERRORPRINT("ALP packet ended prematurely %d < %d\n", 
                pJ288ALP->alpPacketBytesReceived, 
                pJ288ALP->alpPacketLength);
        }

        SendALPPacket(pJ288ALP, pJ288ALP->plpId,  pJ288ALP->vALPPacket, 
            pJ288ALP->alpPacketBytesReceived, &pJ288ALP->tALP);
        StartNewALP(pJ288ALP);
    }

out:
    *pNConsumed = nConsumed;

    return rval;
}

void
TSPacketLockLost(ENATSC3J288ALP *pJ288ALP, struct timespec *pPacketTimeUTC)
{
    int32_t nConsumed;

    LOGPRINT_NOFILE("TS SYNC LOST\n");
    ProcessJ288PacketPayload(pJ288ALP, pPacketTimeUTC, NULL, 0, 0, 1, 
        &nConsumed);

    // If we lost TS lock then we implicitly lost ALP lock as well.
    pJ288ALP->fALPLock = 0;
}

void
HandleJ288Packet(ENATSC3J288ALP *pJ288ALP, struct timespec *pPacketTimeUTC, 
    const uint8_t *pPacket)
{
    int32_t fError = 0;
    int32_t pid;

    if (pPacket[1] & TRANSPORT_ERROR_INDICATOR)
    {
        fError = 1;
    }

    pid = (pPacket[1] << 8 | pPacket[2]) & 0x1fff;

    // We are assuming there is only one stream
    if (pid != DEMOD_PID_VALUE)
    {
        ERRORPRINT("Unexpected PID value 0x%x\n", pid);
    }

    if (pPacket[1] & TLV_START_INDICATOR)
    {
        uint8_t pointer = pPacket[3];

        if (pointer > 184)
        {
            LOGPRINT_NOFILE("Illegal pointer value %d\n", pointer);
            fError = 1;
            pointer = 184;
        }

        if (pJ288ALP->fALPLock)
        {
            int32_t nConsumed;

            if (pointer)
            {
                if (ProcessJ288PacketPayload(pJ288ALP, pPacketTimeUTC, 
                    pPacket + 4, pointer, fError, 1, &nConsumed))
                {
                    LOGPRINT_NOFILE("Lost ALP lock\n");
                    pJ288ALP->fALPLock = 0;
                    goto out;
                }

                if (nConsumed != pointer)
                {
                    ERRORPRINT("SNH: Partial data consumed\n");
                }
            }
        }

        while (pointer != 184)
        {
            int32_t nConsumed;

            if (ProcessJ288PacketPayload(pJ288ALP, pPacketTimeUTC, 
                pPacket + 4 + pointer, 188 - pointer - 4, fError, 0, 
                &nConsumed))
            {
                LOGPRINT_NOFILE("Lost ALP lock\n");
                pJ288ALP->fALPLock = 0;
                goto out;
            }
            pointer += nConsumed;
        }

        pJ288ALP->fALPLock = 1;
    }
    else
    {
        if (pJ288ALP->fALPLock)
        {
            int32_t nConsumed;

            if (ProcessJ288PacketPayload(pJ288ALP, pPacketTimeUTC, pPacket + 3, 
                188 - 3, fError, 0, &nConsumed))
            {
                pJ288ALP->fALPLock = 0;
                goto out;
            }

            if (nConsumed != 188 - 3)
            {
                ERRORPRINT("SNH: Partial data consumed\n");
            }
        }
    }

out:
    return;
}

ENATSC3Status
ENATSC3J288ALPProcessJ288Packets(ENATSC3J288ALPHandle_t hATSC3J288ALP,
    struct timespec *pPacketTimeUTC, const uint8_t *pData, int32_t nBytes)
{
    ENATSC3J288ALP *pJ288ALP = (ENATSC3J288ALP *)hATSC3J288ALP;
    ENATSC3Status rval = ENATSC3Status_Success;

    while (nBytes)
    {
        // If we do not have TS lock then try to acquire it. We assume we have
        // TS lock if we have see three TS packets with sync bytes at the 
        // expected location.
        if (!pJ288ALP->fTSLock)
        {
            int32_t nToCopy;

            nToCopy = MIN(nBytes, 
                sizeof(pJ288ALP->vTSPacketBuffer) - 
                pJ288ALP->nTSPacketBufferBytesUsed);
            memcpy(
                pJ288ALP->vTSPacketBuffer + pJ288ALP->nTSPacketBufferBytesUsed, 
                pData, nToCopy);
            pJ288ALP->nTSPacketBufferBytesUsed += nToCopy;
            pData += nToCopy;
            nBytes -= nToCopy;
            if (pJ288ALP->nTSPacketBufferBytesUsed == 
                sizeof(pJ288ALP->vTSPacketBuffer))
            {
                int32_t i;

                ENUtils_CrashAssert(sizeof(pJ288ALP->vTSPacketBuffer) >= 
                    188 * 3);

                for (i = 0; i < 188; i++)
                {
                    if (pJ288ALP->vTSPacketBuffer[i] == 0x47 && 
                        pJ288ALP->vTSPacketBuffer[i + 188] == 0x47 &&
                        pJ288ALP->vTSPacketBuffer[i + 188 * 2] == 0x47)
                    {
                        memmove(pJ288ALP->vTSPacketBuffer, 
                            pJ288ALP->vTSPacketBuffer + i,
                            pJ288ALP->nTSPacketBufferBytesUsed - i);
                        pJ288ALP->nTSPacketBufferBytesUsed -= i;
                        pJ288ALP->fTSLock = 1;
                        break;
                    }
                }

                if (i == 188)
                {
                    memmove(pJ288ALP->vTSPacketBuffer, 
                        pJ288ALP->vTSPacketBuffer + 188,
                        pJ288ALP->nTSPacketBufferBytesUsed - 188);
                    pJ288ALP->nTSPacketBufferBytesUsed -= 188;
                    continue;
                }
            }

            StartNewALP(pJ288ALP);
        }

        // If we got here then we have TS lock. See if we have buffered TS
        // packets.
        if (pJ288ALP->nTSPacketBufferBytesUsed)
        {
            if (pJ288ALP->nTSPacketBufferBytesUsed >= 188)
            {
                int32_t nSent = 0;
                int32_t i;

                for (i = 0; 
                     i + 188 <= pJ288ALP->nTSPacketBufferBytesUsed; 
                     i += 188)
                {
                    if (pJ288ALP->vTSPacketBuffer[i] != 0x47)
                    {
                        ERRORPRINT("TS sync lost\n");
                        pJ288ALP->fTSLock = 0;
                        break;
                    }
                    HandleJ288Packet(pJ288ALP, pPacketTimeUTC, 
                        pJ288ALP->vTSPacketBuffer + i);
                    nSent += 188;
                }

                if (nSent)
                {
                    memmove(pJ288ALP->vTSPacketBuffer, 
                        pJ288ALP->vTSPacketBuffer + nSent,
                        pJ288ALP->nTSPacketBufferBytesUsed - nSent);
                    pJ288ALP->nTSPacketBufferBytesUsed -= nSent;
                }
            }

            // Try to top up any partial packet
            if (pJ288ALP->nTSPacketBufferBytesUsed && pJ288ALP->fTSLock)
            {
                int32_t nToCopy;
                
                nToCopy = MIN(nBytes, 188 - pJ288ALP->nTSPacketBufferBytesUsed);
                memcpy(pJ288ALP->vTSPacketBuffer + 
                    pJ288ALP->nTSPacketBufferBytesUsed, pData, nToCopy);
                pJ288ALP->nTSPacketBufferBytesUsed += nToCopy;
                nBytes -= nToCopy;
                pData += nToCopy;

                if (pJ288ALP->nTSPacketBufferBytesUsed < 188)
                {
                    break;
                }

                if (pJ288ALP->vTSPacketBuffer[0] != 0x47)
                {
                    ERRORPRINT("TS sync lost\n");
                    pJ288ALP->fTSLock = 0;
                }
                else
                {
                    HandleJ288Packet(pJ288ALP, pPacketTimeUTC, 
                        pJ288ALP->vTSPacketBuffer);
                    pJ288ALP->nTSPacketBufferBytesUsed = 0;
                }
            }

            if (!pJ288ALP->fTSLock)
            {
                TSPacketLockLost(pJ288ALP, pPacketTimeUTC);
                continue;
            }
        }

        // If we got here then there is no buffered TS.
        while (nBytes >= 188)
        {
            HandleJ288Packet(pJ288ALP, pPacketTimeUTC, pData);
            pData += 188;
            nBytes -= 188; 
        }

        if (nBytes)
        {
            memcpy(pJ288ALP->vTSPacketBuffer, pData, nBytes);
            pJ288ALP->nTSPacketBufferBytesUsed = nBytes;
            break;
        }
    }

    return rval;
}

ENATSC3Status
ENATSC3J288ALPOpen(ENATSC3J288ALPHandle_t *phJ288ALP)
{
    ENATSC3J288ALP *pJ288ALP;
    ENATSC3Status rval = ENATSC3Status_Success;

    pJ288ALP = (ENATSC3J288ALP *)calloc(sizeof(ENATSC3J288ALP), 1);
    if (!pJ288ALP)
    {
        ERRORPRINT("Out of memory\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    pJ288ALP->verbosity = ENVerbosity_Terse;
    pJ288ALP->plpId = INVALID_PLP_ID;
    pJ288ALP->fTSLock = 0;

    *phJ288ALP = pJ288ALP;

out:
    return rval;

errout:
    if (pJ288ALP)
    {
        free(pJ288ALP);
    }

    goto out;
}

ENATSC3Status
ENATSC3J288ALPClose(ENATSC3J288ALPHandle_t hJ288ALP)
{
    ENATSC3J288ALP *pJ288ALP = (ENATSC3J288ALP *)hJ288ALP;
    ENATSC3Status rval = ENATSC3Status_Success;

    free(pJ288ALP);

    return rval;
}

ENATSC3Status
ENATSC3J288ALPRegisterCallbacks(ENATSC3J288ALPHandle_t hJ288ALP,
    ENATSC3J288ALPCallbackFunction pALPCallback,
    ENATSC3J288BasebandCallbackFunction pBasebandCallback, 
    ENATSC3J288PTPCallbackFunction pPTPCallback,
    void *pUserData)
{
    ENATSC3J288ALP *pJ288ALP = (ENATSC3J288ALP *)hJ288ALP;
    ENATSC3Status rval = ENATSC3Status_Success;

    pJ288ALP->pALPCallback = pALPCallback;
    pJ288ALP->pBasebandCallback = pBasebandCallback;
    pJ288ALP->pPTPCallback = pPTPCallback;
    pJ288ALP->pCallbackUserData = pUserData;

    return rval;
}

ENATSC3Status
ENATSC3J288ALPSetVerbosity(ENATSC3J288ALPHandle_t hJ288ALP,
    ENVerbosity verbosity)
{
    ENATSC3J288ALP *pJ288ALP = (ENATSC3J288ALP *)hJ288ALP;
    ENATSC3Status rval = ENATSC3Status_Success;

    pJ288ALP->verbosity = verbosity;

    return rval;
}
