/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libendebug/libendebug.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3utils/enxml.h>

#include "atsc3lls_rrt.h"

static char *ensys = "atsc30lls";
static char *ensubsys = "slt";

static int32_t fVerbose = 0;

#define MIN(x, y) ( ( ( x) < (y) ) ? (x) : (y) )
#define MAX(x, y) ( ( ( x) > (y) ) ? (x) : (y) )
#define numof(x) (sizeof(x) / sizeof(x[0]))

static int32_t
TextTypeNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue,
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData,
    EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSRRT_RatingRegionTablesNode *pRatingRegionTablesNode = 
        (ATSC3LLSRRT_RatingRegionTablesNode *)pUserData;
    ATSC3LLSRRT_TextTypeNode *pTextTypeNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3LLSRRTNodeType_TextType);
    pTextTypeNode = (ATSC3LLSRRT_TextTypeNode *)pParentNode;

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(
        &pRatingRegionTablesNode->dataBuffer, (char *)pValue, sValue, 
        &pTextTypeNode->stringDataOffset))
    {
        ERRORPRINT("Failed to store text\n");
        rval = -1;
        goto out; 
    }

    if (fVerbose)
    {
        LOGPRINT("%*seventCode: %.*s\n", xmlTextReaderDepth(reader) * 2, "", 
            sValue, pValue);
    }

out:
    return rval;
}

static int32_t
TextTypeNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSRRT_TextTypeNode *pTextTypeNode;
    ATSC3LLSRRT_RatingRegionTablesNode *pRatingRegionTablesNode = 
        (ATSC3LLSRRT_RatingRegionTablesNode *)pUserData;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3LLSRRTNodeType_TextType);
    pTextTypeNode = (ATSC3LLSRRT_TextTypeNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ATSC3LLSRRTTextTypeNodeAttribute_lang:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pRatingRegionTablesNode->dataBuffer, (char *)pValue, sValue, 
                &pTextTypeNode->langDataOffset))
            {
                ERRORPRINT("Failed to store lang\n");
                rval = -1;
                break;
            }

            if (fVerbose)
            {
                LOGPRINT("%*slang: %.*s\n", xmlTextReaderDepth(reader) * 2, 
                    "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

    return rval;
}

EnXMLParserAttribute vTextTypeAttributes[] = 
{
    {
        .id = ATSC3LLSRRTTextTypeNodeAttribute_lang,
        .sName = sizeof("lang") - 1,
        .pName = "lang",
        .handler = TextTypeNodeAttributeHandler,
    },
};
EnXMLParserChildNode vTextTypeChildNodes[] = 
{
};
EnXMLParserNode textTypeNodeData = 
{
    .pName = "TextType",
    .pAttributes = vTextTypeAttributes,
    .nAttributes = numof(vTextTypeAttributes),
    .pChildNodes = vTextTypeChildNodes,
    .nChildNodes = numof(vTextTypeChildNodes),
    .textHandler = TextTypeNodeTextHandler,
};


static int32_t
RatingNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSRRT_RatingNode *pRatingNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3LLSRRTNodeType_Rating);

    pRatingNode = (ATSC3LLSRRT_RatingNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ATSC3LLSRRTRatingNodeAttribute_ratingValue:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pRatingNode->ratingValue))
            {
                ERRORPRINT("Failed to parse ratingValue %.*s\n", 
                    sValue, pValue);
                rval = -1;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sratingValue: %.*s\n", 
                    xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
                break;
            }
    }

out:
    return rval;
}

static int32_t
ProcessRatingValueAbbrevNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSRRT_RatingRegionTablesNode *pRatingRegionTablesNode = 
       (ATSC3LLSRRT_RatingRegionTablesNode *)pUserData;
    ATSC3LLSRRT_RatingNode *pRatingNode;
    ATSC3LLSRRT_TextTypeNode ratingValueAbbrevNode;
    int32_t ratingValueAbbrevNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3LLSRRTNodeType_Rating);
    pRatingNode = (ATSC3LLSRRT_RatingNode *)pParentNode;

    memset(&ratingValueAbbrevNode, 0, sizeof(ratingValueAbbrevNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(
        &pRatingRegionTablesNode->dataBuffer, sizeof(ATSC3LLSRRT_TextTypeNode), 
        &ratingValueAbbrevNodeOffset))
    {
        ERRORPRINT(
            "Failed to allocate space for ratingValueAbbrev TextType node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pRatingNode->ratingValueAbbrevs, ratingValueAbbrevNodeOffset))
    {
        ERRORPRINT("Failed to store ratingValueAbbrev offset\n");
        rval = 1;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sRatingValueAbbrev\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary 
    // ratingValueAbbrevNode that lives on the stack. Once we are done 
    // processing the node it is committed to the data buffer.
    ratingValueAbbrevNode.base.nodeType = ATSC3LLSRRTNodeType_TextType;
    rval = EnXMLParserProcessNode(&textTypeNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)&ratingValueAbbrevNode);

    memcpy(pRatingRegionTablesNode->dataBuffer.pBuffer + 
        ratingValueAbbrevNodeOffset, &ratingValueAbbrevNode, 
        sizeof(ratingValueAbbrevNode));

out:
    return rval;
}

static int32_t
ProcessRatingValueStringNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSRRT_RatingRegionTablesNode *pRatingRegionTablesNode = 
        (ATSC3LLSRRT_RatingRegionTablesNode *)pUserData;
    ATSC3LLSRRT_RatingNode *pRatingNode;
    ATSC3LLSRRT_TextTypeNode ratingValueStringNode;
    int32_t ratingValueStringNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3LLSRRTNodeType_Rating);
    pRatingNode = (ATSC3LLSRRT_RatingNode *)pParentNode;

    memset(&ratingValueStringNode, 0, sizeof(ratingValueStringNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(
        &pRatingRegionTablesNode->dataBuffer, sizeof(ATSC3LLSRRT_TextTypeNode), 
        &ratingValueStringNodeOffset))
    {
        ERRORPRINT(
            "Failed to allocate space for ratingValueStrings TextType node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pRatingNode->ratingValueStrings, ratingValueStringNodeOffset))
    {
        ERRORPRINT("Failed to store ratingValueString offset\n");
        rval = 1;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sRatingValueString\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary 
    // ratingValueStringNode that lives on the stack. Once we are done 
    // processing the node it is committed to the data buffer.
    ratingValueStringNode.base.nodeType = ATSC3LLSRRTNodeType_TextType;
    rval = EnXMLParserProcessNode(&textTypeNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)&ratingValueStringNode);

    memcpy(pRatingRegionTablesNode->dataBuffer.pBuffer + 
        ratingValueStringNodeOffset, &ratingValueStringNode, 
        sizeof(ratingValueStringNode));

out:
    return rval;
}

EnXMLParserAttribute vRatingAttributes[] = 
{
    {
        .id = ATSC3LLSRRTRatingNodeAttribute_ratingValue,
        .sName = sizeof("ratingValue") - 1,
        .pName = "ratingValue",
        .handler = RatingNodeAttributeHandler,
    },
};
EnXMLParserChildNode vRatingChildNodes[] = 
{
    {
        .id = ATSC3LLSRRTRatingChildNode_RatingValueAbbrev,
        .sName = sizeof("RatingValueAbbrev") - 1,
        .pName = "RatingValueAbbrev",
        .handler = ProcessRatingValueAbbrevNode,
    },
    {
        .id = ATSC3LLSRRTRatingChildNode_RatingValueString,
        .sName = sizeof("RatingValueString") - 1,
        .pName = "RatingValueString",
        .handler = ProcessRatingValueStringNode,
    },
};
EnXMLParserNode ratingNodeData = 
{
    .pName = "Rating",
    .pAttributes = vRatingAttributes,
    .nAttributes = numof(vRatingAttributes),
    .pChildNodes = vRatingChildNodes,
    .nChildNodes = numof(vRatingChildNodes),
    .textHandler = NULL,
};

static int32_t
DimensionNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSRRT_DimensionNode *pDimensionNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3LLSRRTNodeType_Dimension);

    pDimensionNode = (ATSC3LLSRRT_DimensionNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ATSC3LLSRRTDimensionNodeAttribute_dimensionLevels:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pDimensionNode->dimensionLevels))
            {
                ERRORPRINT("Failed to parse dimensionLevels %.*s\n", 
                    sValue, pValue);
                rval = -1;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sdimensionLevels: %.*s\n", 
                    xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSRRTDimensionNodeAttribute_dimensionGraduated:
            if (EnXMLParserStringToBool((char *)pValue, sValue, 
                &pDimensionNode->dimensionGraduated))
            {
                ERRORPRINT("Failed to parse dimensionsGraduated: %.*s\n",
                    sValue, pValue);
                rval = -1;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sdimensionGraduated: %.*s\n", 
                    xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
    }

out:
    return rval;
}

static int32_t
ProcessDimensionTitleNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSRRT_RatingRegionTablesNode *pRatingRegionTablesNode = 
        (ATSC3LLSRRT_RatingRegionTablesNode *)pUserData;
    ATSC3LLSRRT_DimensionNode *pDimensionNode;
    ATSC3LLSRRT_TextTypeNode dimensionTitleNode;
    int32_t dimensionTitleNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3LLSRRTNodeType_Dimension);
    pDimensionNode = (ATSC3LLSRRT_DimensionNode *)pParentNode;

    memset(&dimensionTitleNode, 0, sizeof(dimensionTitleNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(
        &pRatingRegionTablesNode->dataBuffer, sizeof(ATSC3LLSRRT_TextTypeNode), 
        &dimensionTitleNodeOffset))
    {
        ERRORPRINT(
            "Failed to allocate space for dimensionTitle TextType node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pDimensionNode->dimensionTitles, dimensionTitleNodeOffset))
    {
        ERRORPRINT("Failed to store dimensionTitle node offset\n");
        rval = 1;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sDimensionTitle\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary 
    // dimensionTitleNode that lives on the stack. Once we are done processing 
    // the node it is committed to the data buffer.
    dimensionTitleNode.base.nodeType = ATSC3LLSRRTNodeType_TextType;
    rval = EnXMLParserProcessNode(&textTypeNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)&dimensionTitleNode);

    memcpy(pRatingRegionTablesNode->dataBuffer.pBuffer + 
        dimensionTitleNodeOffset, &dimensionTitleNode, 
        sizeof(dimensionTitleNode));

out:
    return rval;
}

static int32_t
ProcessRatingNode(EnXMLParserChildNode *pChildNode, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSRRT_RatingRegionTablesNode *pRatingRegionTablesNode = 
        (ATSC3LLSRRT_RatingRegionTablesNode *)pUserData;
    ATSC3LLSRRT_DimensionNode *pDimensionNode;
    ATSC3LLSRRT_RatingNode ratingNode;
    int32_t ratingNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3LLSRRTNodeType_Dimension);
    pDimensionNode = (ATSC3LLSRRT_DimensionNode *)pParentNode;

    memset(&ratingNode, 0, sizeof(ratingNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(
        &pRatingRegionTablesNode->dataBuffer, sizeof(ATSC3LLSRRT_RatingNode), 
        &ratingNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for rating node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pDimensionNode->ratings, 
        ratingNodeOffset))
    {
        ERRORPRINT("Failed to store rating node offset\n");
        rval = 1;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sRating\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary 
    // ratingNode that lives on the stack. Once we are done processing the 
    // node it is committed to the data buffer.
    if (EnXMLParserDataOffsetArrayInitialize(&ratingNode.ratingValueAbbrevs,
        &pRatingRegionTablesNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate ratingValueAbbrevs array\n");
        rval = -1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayInitialize(&ratingNode.ratingValueStrings,
        &pRatingRegionTablesNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate ratingValueStrings array\n");
        rval = -1;
        goto out;
    }

    ratingNode.base.nodeType = ATSC3LLSRRTNodeType_Rating;
    rval = EnXMLParserProcessNode(&ratingNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)&ratingNode);

    memcpy(pRatingRegionTablesNode->dataBuffer.pBuffer + ratingNodeOffset, 
        &ratingNode, sizeof(ratingNode));

out:
    return rval;
}

EnXMLParserAttribute vDimensionAttributes[] = 
{
    {
        .id = ATSC3LLSRRTDimensionNodeAttribute_dimensionLevels,
        .sName = sizeof("dimensionLevels") - 1,
        .pName = "dimensionLevels",
        .handler = DimensionNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSRRTDimensionNodeAttribute_dimensionGraduated,
        .sName = sizeof("dimensionGraduated") - 1,
        .pName = "dimensionGraduated",
        .handler = DimensionNodeAttributeHandler,
    },
};
EnXMLParserChildNode vDimensionChildNodes[] = 
{
    {
        .id = ATSC3LLSRRTDimensionChildNode_DimensionTitle,
        .sName = sizeof("DimensionTitle") - 1,
        .pName = "DimensionTitle",
        .handler = ProcessDimensionTitleNode,
    },
    {
        .id = ATSC3LLSRRTDimensionChildNode_Rating,
        .sName = sizeof("Rating") - 1,
        .pName = "Rating",
        .handler = ProcessRatingNode,
    },
};
EnXMLParserNode dimensionNodeData = 
{
    .pName = "Dimension",
    .pAttributes = vDimensionAttributes,
    .nAttributes = numof(vDimensionAttributes),
    .pChildNodes = vDimensionChildNodes,
    .nChildNodes = numof(vDimensionChildNodes),
    .textHandler = NULL,
};

static int32_t
RatingRegionTableNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSRRT_RatingRegionTableNode *pRatingRegionTableNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3LLSRRTNodeType_RatingRegionTable);
    pRatingRegionTableNode = (ATSC3LLSRRT_RatingRegionTableNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ATSC3LLSRRTRatingRegionTableNodeAttribute_regionIdentifier:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pRatingRegionTableNode->regionIdentifier))
            {
                ERRORPRINT("Failed to parse regionIdentifier %.*s\n", 
                    sValue, pValue);
                rval = -1;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sregionIdentifier: %.*s\n", xmlTextReaderDepth
                    (reader) * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
                break;
            }
    }

out:
    return rval;
}

static int32_t
ProcessRegionIdTextNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSRRT_RatingRegionTablesNode *pRatingRegionTablesNode = 
        (ATSC3LLSRRT_RatingRegionTablesNode *)pUserData;
    ATSC3LLSRRT_RatingRegionTableNode *pRatingRegionTableNode;
    ATSC3LLSRRT_TextTypeNode regionIdTextNode;
    int32_t regionIdTextNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3LLSRRTNodeType_RatingRegionTable);
    pRatingRegionTableNode = (ATSC3LLSRRT_RatingRegionTableNode *)pParentNode;

    memset(&regionIdTextNode, 0, sizeof(regionIdTextNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(
        &pRatingRegionTablesNode->dataBuffer, 
        sizeof(ATSC3LLSRRT_TextTypeNode), &regionIdTextNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for regionIdText node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pRatingRegionTableNode->regionIdTexts, regionIdTextNodeOffset))
    {
        ERRORPRINT("Failed to store regionIdText node offset\n");
        rval = 1;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*RegionIdText\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary 
    // regionIdTextNode that lives on the stack. Once we are done 
    // processing the node it is committed to the data buffer.
    regionIdTextNode.base.nodeType = ATSC3LLSRRTNodeType_TextType;
    rval = EnXMLParserProcessNode(&textTypeNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)&regionIdTextNode);

    memcpy(pRatingRegionTablesNode->dataBuffer.pBuffer + 
        regionIdTextNodeOffset, &regionIdTextNode, sizeof(regionIdTextNode));

out:
    return rval;
}

static int32_t
ProcessDimensionNode(EnXMLParserChildNode *pChildNode, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSRRT_RatingRegionTablesNode *pRatingRegionTablesNode = 
       (ATSC3LLSRRT_RatingRegionTablesNode *)pUserData;
    ATSC3LLSRRT_RatingRegionTableNode *pRatingRegionTableNode;
    ATSC3LLSRRT_DimensionNode dimensionNode;
    int32_t dimensionNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3LLSRRTNodeType_RatingRegionTable);
    pRatingRegionTableNode = (ATSC3LLSRRT_RatingRegionTableNode *)pParentNode;

    memset(&dimensionNode, 0, sizeof(dimensionNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(
        &pRatingRegionTablesNode->dataBuffer, 
        sizeof(ATSC3LLSRRT_DimensionNode), &dimensionNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for dimension node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pRatingRegionTableNode->dimensions, dimensionNodeOffset))
    {
        ERRORPRINT("Failed to store dimension node offset\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayInitialize(&dimensionNode.dimensionTitles, 
        &pRatingRegionTablesNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate dimensionTitles array\n");
        rval = -1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayInitialize(&dimensionNode.ratings, 
        &pRatingRegionTablesNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate dimensionTitles array\n");
        rval = -1;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sDimension\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary 
    // dimensionNode that lives on the stack. Once we are done 
    // processing the node it is committed to the data buffer.
    dimensionNode.base.nodeType = ATSC3LLSRRTNodeType_Dimension;
    rval = EnXMLParserProcessNode(&dimensionNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)&dimensionNode);

    memcpy(pRatingRegionTablesNode->dataBuffer.pBuffer + dimensionNodeOffset, 
        &dimensionNode, sizeof(dimensionNode));

out:
    return rval;
}

EnXMLParserAttribute vRatingRegionTableAttributes[] = 
{
    {
        .id = ATSC3LLSRRTRatingRegionTableNodeAttribute_regionIdentifier,
        .sName = sizeof("regionIdentifier") - 1,
        .pName = "regionIdentifier",
        .handler = RatingRegionTableNodeAttributeHandler,
    },
};
EnXMLParserChildNode vRatingRegionTableChildNodes[] = 
{
    {
        .id = ATSC3LLSRRTRatingRegionTableChildNode_RegionIdText,
        .sName = sizeof("RegionIdText") - 1,
        .pName = "RegionIdText",
        .handler = ProcessRegionIdTextNode,
    },
    {
        .id = ATSC3LLSRRTRatingRegionTableChildNode_Dimension,
        .sName = sizeof("Dimension") - 1,
        .pName = "Dimension",
        .handler = ProcessDimensionNode,
    },
};
EnXMLParserNode ratingRegionTableNodeData = 
{
    .pName = "RatingRegionTable",
    .pAttributes = vRatingRegionTableAttributes,
    .nAttributes = numof(vRatingRegionTableAttributes),
    .pChildNodes = vRatingRegionTableChildNodes,
    .nChildNodes = numof(vRatingRegionTableChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessRatingRegionTableNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSRRT_RatingRegionTablesNode *pRatingRegionTablesNode;
    ATSC3LLSRRT_RatingRegionTableNode ratingRegionTableNode;
    int32_t ratingRegionTableNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3LLSRRTNodeType_RatingRegionTables);
    pRatingRegionTablesNode = (ATSC3LLSRRT_RatingRegionTablesNode *)pParentNode;

    memset(&ratingRegionTableNode, 0, sizeof(ratingRegionTableNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(
        &pRatingRegionTablesNode->dataBuffer, 
        sizeof(ratingRegionTableNode), &ratingRegionTableNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for ratingRegionTable node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pRatingRegionTablesNode->ratingRegionTables, 
        ratingRegionTableNodeOffset))
    {
        ERRORPRINT("Failed to store ratingRegionTable node offset\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayInitialize(
        &ratingRegionTableNode.regionIdTexts,
        &pRatingRegionTablesNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate regionIdTexts array\n");
        rval = -1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayInitialize(
        &ratingRegionTableNode.dimensions,
        &pRatingRegionTablesNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate dimensions array\n");
        rval = -1;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*RatingRegionTable\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary 
    // ratingRegionTableNode that lives on the stack. Once we are done 
    // processing the node it is committed to the data buffer.
    ratingRegionTableNode.base.nodeType = 
        ATSC3LLSRRTNodeType_RatingRegionTable;
    rval = EnXMLParserProcessNode(&ratingRegionTableNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)&ratingRegionTableNode);

    memcpy(pRatingRegionTablesNode->dataBuffer.pBuffer + 
        ratingRegionTableNodeOffset, &ratingRegionTableNode, 
        sizeof(ratingRegionTableNode));

out:
    return rval;
}

EnXMLParserAttribute vRRTAttributes[] = 
{
};
EnXMLParserChildNode vRRTChildNodes[] = 
{
    {
        .id = ATSC3LLSRRTChildNode_RatingRegionTable,
        .sName = sizeof("RatingRegionTable") - 1,
        .pName = "RatingRegionTable",
        .handler = ProcessRatingRegionTableNode,
    },
};
EnXMLParserNode rrtNodeData = 
{
    .pName = "RRT",
    .pAttributes = vRRTAttributes,
    .nAttributes = numof(vRRTAttributes),
    .pChildNodes = vRRTChildNodes,
    .nChildNodes = numof(vRRTChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessRatingRegionTablesNode(xmlTextReaderPtr reader, 
    ATSC3LLSRRT_RatingRegionTablesNode *pRatingRegionTablesNode)
{
    int32_t rval = 0;

    if (fVerbose)
    {
        LOGPRINT("RRT\n");
    }

    if (EnXMLParserDataOffsetArrayInitialize(
        &pRatingRegionTablesNode->ratingRegionTables, 
        &pRatingRegionTablesNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate ratingRegionTables array\n");
        rval = -1;
        goto out;
    }

    rval = EnXMLParserProcessNode(&rrtNodeData, reader, 
        pRatingRegionTablesNode, 
        (EnXMLParserNodeBase *)pRatingRegionTablesNode);

out:
    return rval;
}

ENATSC3Status
ATSC3LLSParseRRT(const char *pXML, int32_t sXML, 
    ATSC3LLSRRT_RatingRegionTablesNode *pRatingRegionTablesNode)
{
    xmlTextReaderPtr reader = NULL;
    int32_t fHaveDataBuffer = 0;
    int xmlret;
    ENATSC3Status rval = ENATSC3Status_Success;

    memset(pRatingRegionTablesNode, 0, 
        sizeof(ATSC3LLSRRT_RatingRegionTablesNode));
    pRatingRegionTablesNode->base.nodeType = 
        ATSC3LLSRRTNodeType_RatingRegionTables;

    if (EnXMLParserDataBufferInitialize(&pRatingRegionTablesNode->dataBuffer, 
        16384, 8192))
    {
        ERRORPRINT("Failed to initialize RRT data buffer\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }
    fHaveDataBuffer = 1;

    reader = xmlReaderForMemory(pXML, sXML, NULL, NULL, 0);
    if (!reader)
    {
        ERRORPRINT("Failed to create xml reader\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    xmlret = xmlTextReaderRead(reader);
    if (xmlret == 1)
    {
        int nodeType;
        const xmlChar *pLocalName;

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);

        if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "RatingRegionTables"))
        {
            int32_t lret;

            lret = ProcessRatingRegionTablesNode(reader, 
                pRatingRegionTablesNode);
            if (lret)
            {
                rval = ENATSC3Status_Failure;
                goto errout;
            }
        }
        else
        {
            ERRORPRINT("Top-level object isn't RatingRegionTables (%d/%s)\n",
                nodeType, pLocalName);
            rval = ENATSC3Status_Failure;
            goto errout;
        }
    }
    else
    {
        ERRORPRINT("Failed to parse RRT XML\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }

out:
    if (reader)
    {
        xmlFreeTextReader(reader);
    }

    return rval;

errout:
    if (fHaveDataBuffer)
    {
        if (EnXMLParserDataBufferFree(&pRatingRegionTablesNode->dataBuffer))
        {
            ERRORPRINT("Failed to free RRT data buffer\n");
        }
    }

    goto out;
}

ENATSC3Status
ATSC3LLSFreeParsedRRT(
    ATSC3LLSRRT_RatingRegionTablesNode *pRatingRegionTablesNode)
{
    ENATSC3Status rval = ENATSC3Status_Success;

    if (EnXMLParserDataBufferFree(&pRatingRegionTablesNode->dataBuffer))
    {
        ERRORPRINT("Failed to free RRT string buffer\n");
    }

    return rval;
}

