/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 * All Rights Reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_defs.h>

#include <libenatsc3utils/enxml.h>

#include <enatsc3-common/enatsc3.h>
#include "atsc3mmtpsls_usbd.h"

static char *ensys = "mmtpsls";
static char *ensubsys = "usbd";

static int32_t fVerbose = 0;

static int32_t
MPUComponentNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3MMTPSLSUSBDBundleDescriptionNode *pUSBDNode = 
        (ATSC3MMTPSLSUSBDBundleDescriptionNode *)pUserData;
    ATSC3MMTPSLSUSBDMPUComponentNode *pMPUComponentNode;
    int depth;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3MMTPSLSUSBDNodeType_MPUComponent);

    pMPUComponentNode = (ATSC3MMTPSLSUSBDMPUComponentNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3MMTPSLSUSBDMPUComponentNodeAttribute_mmtPackageId:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pUSBDNode->dataBuffer, (char *)pValue, sValue,
                &pMPUComponentNode->mmtPackageIdDataOffset))
            {
                ERRORPRINT(
                    "Failed to store componentId in string buffer\n");
                rval = -1;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*smmtPackageId: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

    return rval;
}

EnXMLParserAttribute vMPUComponentAttributes[] = 
{
    {
        .id = ATSC3MMTPSLSUSBDMPUComponentNodeAttribute_mmtPackageId,
        .sName = sizeof("mmtPackageId") - 1,
        .pName = "mmtPackageId",
        .handler = MPUComponentNodeAttributeHandler,
    },
};
EnXMLParserChildNode vMPUComponentChildNodes[] =
{
};
EnXMLParserNode mpuComponentNodeData = 
{
    .pName = "MPUComponent",
    .pAttributes = vMPUComponentAttributes,
    .nAttributes = numof(vMPUComponentAttributes),
    .pChildNodes = vMPUComponentChildNodes,
    .nChildNodes = numof(vMPUComponentChildNodes),
    .textHandler = NULL,
};

static int32_t
ComponentInfoNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3MMTPSLSUSBDBundleDescriptionNode *pUSBDNode = 
        (ATSC3MMTPSLSUSBDBundleDescriptionNode *)pUserData;
    ATSC3MMTPSLSUSBDComponentInfoNode *pComponentInfoNode;
    int depth;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3MMTPSLSUSBDNodeType_ComponentInfo);

    pComponentInfoNode = (ATSC3MMTPSLSUSBDComponentInfoNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3MMTPSLSUSBDComponentInfoNodeAttribute_componentType:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pComponentInfoNode->componentType))
            {
                ERRORPRINT("Failed to parse componentType %.*s\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*scomponentType: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3MMTPSLSUSBDComponentInfoNodeAttribute_componentRole:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pComponentInfoNode->componentRole))
            {
                ERRORPRINT("Failed to parse componentRole %.*s\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*scomponentRole: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3MMTPSLSUSBDComponentInfoNodeAttribute_componentProtectedFlag:
            if (EnXMLParserStringToBool((char *)pValue, sValue, 
                &pComponentInfoNode->componentProtectedFlag))
            {
                ERRORPRINT("Failed to parse componentProtectedFlag %.*s\n", 
                    sValue, pValue);
                rval = -1;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*scomponentProtectedFlag: %.*s\n", depth * 2, "", 
                    sValue, pValue);
            }
            break;
        case ATSC3MMTPSLSUSBDComponentInfoNodeAttribute_componentId:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pUSBDNode->dataBuffer, (char *)pValue, sValue,
                &pComponentInfoNode->componentIdDataOffset))
            {
                ERRORPRINT(
                    "Failed to store componentId in string buffer\n");
                rval = -1;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*scomponentId: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3MMTPSLSUSBDComponentInfoNodeAttribute_componentName:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pUSBDNode->dataBuffer, (char *)pValue, sValue,
                &pComponentInfoNode->componentNameDataOffset))
            {
                ERRORPRINT(
                    "Failed to store componentId in string buffer\n");
                rval = -1;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*scomponentName: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

out:
    return rval;
}

EnXMLParserAttribute vComponentInfoAttributes[] = 
{
    {
        .id = ATSC3MMTPSLSUSBDComponentInfoNodeAttribute_componentType,
        .sName = sizeof("componentType") - 1,
        .pName = "componentType",
        .handler = ComponentInfoNodeAttributeHandler,
    },
    {
        .id = ATSC3MMTPSLSUSBDComponentInfoNodeAttribute_componentRole,
        .sName = sizeof("componentRole") - 1,
        .pName = "componentRole",
        .handler = ComponentInfoNodeAttributeHandler,
    },
    {
        .id = ATSC3MMTPSLSUSBDComponentInfoNodeAttribute_componentProtectedFlag,
        .sName = sizeof("componentProtectedFlag") - 1,
        .pName = "componentProtectedFlag",
        .handler = ComponentInfoNodeAttributeHandler,
    },
    {
        .id = ATSC3MMTPSLSUSBDComponentInfoNodeAttribute_componentId,
        .sName = sizeof("componentId") - 1,
        .pName = "componentId",
        .handler = ComponentInfoNodeAttributeHandler,
    },
    {
        .id = ATSC3MMTPSLSUSBDComponentInfoNodeAttribute_componentName,
        .sName = sizeof("componentName") - 1,
        .pName = "componentName",
        .handler = ComponentInfoNodeAttributeHandler,
    },
};
EnXMLParserChildNode vComponentInfoChildNodes[] =
{
};
EnXMLParserNode componentInfoNodeData = 
{
    .pName = "ComponentInfo",
    .pAttributes = vComponentInfoAttributes,
    .nAttributes = numof(vComponentInfoAttributes),
    .pChildNodes = vComponentInfoChildNodes,
    .nChildNodes = numof(vComponentInfoChildNodes),
    .textHandler = NULL,
};

static int32_t
ChannelNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3MMTPSLSUSBDBundleDescriptionNode *pUSBDNode = 
        (ATSC3MMTPSLSUSBDBundleDescriptionNode *)pUserData;
    ATSC3MMTPSLSUSBDChannelNode *pChannelNode;
    int depth;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ATSC3MMTPSLSUSBDNodeType_Channel);

    pChannelNode = (ATSC3MMTPSLSUSBDChannelNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3MMTPSLSUSBDChannelNodeAttribute_serviceIcon:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pUSBDNode->dataBuffer, (char *)pValue, sValue,
                &pChannelNode->serviceIconDataOffset))
            {
                ERRORPRINT(
                    "Failed to store componentId in string buffer\n");
                rval = -1;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sserviceIcon: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3MMTPSLSUSBDChannelNodeAttribute_serviceGenre:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pChannelNode->serviceGenre))
            {
                ERRORPRINT("Failed to parse serviceGenre %.*s\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sserviceGener: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

out:
    return rval;
}

EnXMLParserAttribute vChannelAttributes[] = 
{
    {
        .id = ATSC3MMTPSLSUSBDChannelNodeAttribute_serviceGenre,
        .sName = sizeof("serviceGenre") - 1,
        .pName = "serviceGenre",
        .handler = ChannelNodeAttributeHandler,
    },
    {
        .id = ATSC3MMTPSLSUSBDChannelNodeAttribute_serviceIcon,
        .sName = sizeof("serviceIcon") - 1,
        .pName = "serviceIcon",
        .handler = ChannelNodeAttributeHandler,
    },
};
EnXMLParserChildNode vChannelChildNodes[] =
{
};
EnXMLParserNode channelNodeData = 
{
    .pName = "Channel",
    .pAttributes = vChannelAttributes,
    .nAttributes = numof(vChannelAttributes),
    .pChildNodes = vChannelChildNodes,
    .nChildNodes = numof(vChannelChildNodes),
    .textHandler = NULL,
};

static int32_t
UserServiceDescriptionNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3MMTPSLSUSBDUserServiceDescriptionNode *pUserServiceDescriptionNode;
    int depth;
    int32_t rval = 0;

    assert(pParentNode->nodeType == 
        ATSC3MMTPSLSUSBDNodeType_UserServiceDescription);

    pUserServiceDescriptionNode = 
        (ATSC3MMTPSLSUSBDUserServiceDescriptionNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3MMTPSLSUSBDUserServiceDescriptionNodeAttribute_serviceId:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pUserServiceDescriptionNode->serviceId))
            {
                ERRORPRINT("Failed to parse serviceId %.*s\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sserviceId: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3MMTPSLSUSBDUserServiceDescriptionNodeAttribute_serviceStatus:
            if (EnXMLParserStringToBool((char *)pValue, sValue, 
                &pUserServiceDescriptionNode->serviceStatus))
            {
                ERRORPRINT("Failed to parse serviceStatus %.*s\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sserviceStatus: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

out:
    return rval;
}

static int32_t
ProcessChannelNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3MMTPSLSUSBDUserServiceDescriptionNode *pUserServiceDescriptionNode;
    ATSC3MMTPSLSUSBDChannelNode *pChannelNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == 
        ATSC3MMTPSLSUSBDNodeType_UserServiceDescription);

    pUserServiceDescriptionNode =
        (ATSC3MMTPSLSUSBDUserServiceDescriptionNode *)pParentNode;
    
    pChannelNode = &pUserServiceDescriptionNode->channel;
    pChannelNode->base.nodeType = ATSC3MMTPSLSUSBDNodeType_Channel;

    if (fVerbose)
    {
        LOGPRINT("%*sChannel\n", xmlTextReaderDepth(reader) * 2, "");
    }

    rval = EnXMLParserProcessNode(&channelNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)pChannelNode);

    return rval;
}

static int32_t
ProcessMPUComponentNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3MMTPSLSUSBDUserServiceDescriptionNode *pUserServiceDescriptionNode;
    ATSC3MMTPSLSUSBDMPUComponentNode *pMPUComponentNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == 
        ATSC3MMTPSLSUSBDNodeType_UserServiceDescription);

    pUserServiceDescriptionNode =
        (ATSC3MMTPSLSUSBDUserServiceDescriptionNode *)pParentNode;
    
    pMPUComponentNode = &pUserServiceDescriptionNode->mpuComponent;
    pMPUComponentNode->base.nodeType = ATSC3MMTPSLSUSBDNodeType_MPUComponent;

    if (fVerbose)
    {
        LOGPRINT("%*sMPUComponent\n", xmlTextReaderDepth(reader) * 2, "");
    }

    rval = EnXMLParserProcessNode(&mpuComponentNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)pMPUComponentNode);

    return rval;
}

static int32_t
ProcessComponentInfoNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3MMTPSLSUSBDUserServiceDescriptionNode *pUserServiceDescriptionNode;
    ATSC3MMTPSLSUSBDComponentInfoNode *pComponentInfoNode;
    int32_t rval = 0;


    assert(pParentNode->nodeType == 
        ATSC3MMTPSLSUSBDNodeType_UserServiceDescription);

    pUserServiceDescriptionNode =
        (ATSC3MMTPSLSUSBDUserServiceDescriptionNode *)pParentNode;
 
    if (pUserServiceDescriptionNode->nComponentInfos ==
        numof(pUserServiceDescriptionNode->vComponentInfos))
    {
        ERRORPRINT("Too name ComponentInfos\n");
        rval = -1;
        goto out;
    }
    
    pComponentInfoNode = pUserServiceDescriptionNode->vComponentInfos +
        pUserServiceDescriptionNode->nComponentInfos;
    pUserServiceDescriptionNode->nComponentInfos++;
    pComponentInfoNode->base.nodeType = ATSC3MMTPSLSUSBDNodeType_ComponentInfo;

    if (fVerbose)
    {
        LOGPRINT("%*sComponentInfo\n", xmlTextReaderDepth(reader) * 2, "");
    }

    rval = EnXMLParserProcessNode(&componentInfoNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)pComponentInfoNode);

out:
    return rval;
}

EnXMLParserAttribute vUserServiceDescriptionAttributes[] = 
{
    {
        .id = ATSC3MMTPSLSUSBDUserServiceDescriptionNodeAttribute_serviceId,
        .sName = sizeof("serviceId") - 1,
        .pName = "serviceId",
        .handler = UserServiceDescriptionNodeAttributeHandler,
    },
};
EnXMLParserChildNode vUserServiceDescriptionChildNodes[] =
{
    {
        .id = ATSC3MMTPSLSUSBDUserServiceDescriptionChildNode_Channel,
        .sName = sizeof("Channel") - 1,
        .pName = "Channel",
        .handler = ProcessChannelNode,
    },
    {
        .id = ATSC3MMTPSLSUSBDUserServiceDescriptionChildNode_MPUComponent,
        .sName = sizeof("MPUComponent") - 1,
        .pName = "MPUComponent",
        .handler = ProcessMPUComponentNode,
    },
    {
        .id = ATSC3MMTPSLSUSBDUserServiceDescriptionChildNode_ComponentInfo,
        .sName = sizeof("ComponentInfo") - 1,
        .pName = "ComponentInfo",
        .handler = ProcessComponentInfoNode,
    },
};
EnXMLParserNode userServiceDescriptionNodeData = 
{
    .pName = "UserServiceDescription",
    .pAttributes = vUserServiceDescriptionAttributes,
    .nAttributes = numof(vUserServiceDescriptionAttributes),
    .pChildNodes = vUserServiceDescriptionChildNodes,
    .nChildNodes = numof(vUserServiceDescriptionChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessUserServiceDescriptionNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3MMTPSLSUSBDBundleDescriptionNode *pBundleDescriptionNode;
    ATSC3MMTPSLSUSBDUserServiceDescriptionNode *pUserServiceDescriptionNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == 
        ATSC3MMTPSLSUSBDNodeType_BundleDescriptionMMT);

    pBundleDescriptionNode =
        (ATSC3MMTPSLSUSBDBundleDescriptionNode *)pParentNode;
    pUserServiceDescriptionNode = 
        &pBundleDescriptionNode->userServiceDescription;
    pUserServiceDescriptionNode->base.nodeType = 
        ATSC3MMTPSLSUSBDNodeType_UserServiceDescription;

    if (fVerbose)
    {
        LOGPRINT("%*sUserServiceDescription\n", xmlTextReaderDepth(reader) * 2, 
            "");
    }

    rval = EnXMLParserProcessNode(&userServiceDescriptionNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)pUserServiceDescriptionNode);

    return rval;
}

EnXMLParserAttribute vBundleDescriptionMMTAttributes[] = 
{
};
EnXMLParserChildNode vBundleDescriptionMMTChildNodes[] =
{
    {
        .id = ATSC3MMTPSLSUSBDNodeType_UserServiceDescription,
        .sName = sizeof("UserServiceDescription") - 1,
        .pName = "UserServiceDescription",
        .handler = ProcessUserServiceDescriptionNode,
    },
};
EnXMLParserNode bundleDescriptionMMTNodeData = 
{
    .pName = "BundleDescriptionMMT",
    .pAttributes = vBundleDescriptionMMTAttributes,
    .nAttributes = numof(vBundleDescriptionMMTAttributes),
    .pChildNodes = vBundleDescriptionMMTChildNodes,
    .nChildNodes = numof(vBundleDescriptionMMTChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessUSBDBundleDescriptionMMTNode(xmlTextReaderPtr reader, 
    ATSC3MMTPSLSUSBDBundleDescriptionNode *pParsedUSBD)
{
    int32_t rval = 0;

    if (fVerbose)
    {
        LOGPRINT("BundleDescriptionMMT\n");
    }

    rval = EnXMLParserProcessNode(&bundleDescriptionMMTNodeData, reader, 
        pParsedUSBD, (EnXMLParserNodeBase *)pParsedUSBD);

    return rval;
}

ENATSC3Status
ENATSC3MMTPSLSUSBDParse(const char *pXML, int32_t sXML, 
    ATSC3MMTPSLSUSBDBundleDescriptionNode *pParsedUSBD)
{
    xmlTextReaderPtr reader = NULL;
    int32_t fHaveDataBuffer = 0;
    int xmlret;
    ENATSC3Status rval = ENATSC3Status_Success;

    memset(pParsedUSBD, 0, sizeof(ATSC3MMTPSLSUSBDBundleDescriptionNode));
    if (EnXMLParserDataBufferInitialize(&pParsedUSBD->dataBuffer, 4096, 4096))
    {
        ERRORPRINT("Failed to allocate USBD data buffer\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }
    fHaveDataBuffer = 1;

    reader = xmlReaderForMemory(pXML, sXML, NULL, NULL, 0);
    if (!reader)
    {
        ERRORPRINT("Failed to create xml reader\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    xmlret = xmlTextReaderRead(reader);
    if (xmlret == 1)
    {
        int nodeType;
        const xmlChar *pLocalName;

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);

        if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "BundleDescriptionMMT"))
        {
            pParsedUSBD->base.nodeType = 
                ATSC3MMTPSLSUSBDNodeType_BundleDescriptionMMT;
            ProcessUSBDBundleDescriptionMMTNode(reader, pParsedUSBD);
        }
        else
        {
            ERRORPRINT("Top-level object isn't USBD (%d/%s)\n",
                nodeType, pLocalName);
            rval = ENATSC3Status_Failure;
            goto errout;
        }
    }
    else
    {
        ERRORPRINT("Couldn't parse XML\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }

out:
    if (reader)
    {
        xmlFreeTextReader(reader);
    }

    return rval;

errout:
    if (fHaveDataBuffer)
    {
        if (EnXMLParserDataBufferFree(&pParsedUSBD->dataBuffer))
        {
            ERRORPRINT("Failed to free USBD data buffer\n");
        }
    }

    goto out;
}

ENATSC3Status
ENATSC3MMTPSLSUSBDFree(ATSC3MMTPSLSUSBDBundleDescriptionNode *pParsedUSBD)
{
    ENATSC3Status rval = ENATSC3Status_Success;

    if (EnXMLParserDataBufferFree(&pParsedUSBD->dataBuffer))
    {
        ERRORPRINT("Failed to free USBD data buffer\n");
    }

    return rval;
}

