/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 * All Rights Reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>

#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <libenatsc3utils/engunzip.h>
#include <enatsc3-common/enatsc3.h>
#include "libenatsc3mmtp.h"

static char *ensys = "mmtp";
static char *ensubsys = "main";

#define MEDIA_SAMPLE_BUFFER_SIZE (4 * 1024 * 1024)
#define MAX_MEDIA_SAMPLES 300

#define MAX_PAYLOAD_SIZE 65536
#define NANOSECONDS_PER_SECOND 1000000000

#define MAX_BOX_SIZE 0x7fffffff

#define INVALID_PAYLOAD_SEQUENCE_NUMBER -1

#define  NTP_UNIX_OFFSET_YEARS  (uint64_t)70
#define  DAYS_IN_YEAR           (uint64_t)365
#define  NUMBER_OF_LEAP_YEARS   (uint64_t)17
#define  SECONDS_IN_DAY         (uint64_t)86400
#define  NTP_UNIX_OFFSET_SECONDS  \
        ((NTP_UNIX_OFFSET_YEARS * DAYS_IN_YEAR + NUMBER_OF_LEAP_YEARS) * SECONDS_IN_DAY)
#define NTP_TIME_TO_UTC(x) ((x) - NTP_UNIX_OFFSET_SECONDS)

#define MMTP_TIME_INVALID 0xffffffffffffffff

#define BOX_TYPE_MOOV 0x6d6f6f76
#define BOX_TYPE_MVHD 0x6d766864
#define BOX_TYPE_MOOF 0x6d6f6f66
#define BOX_TYPE_TRAF 0x74726166
#define BOX_TYPE_TFDT 0x74666474
#define BOX_TYPE_FTYP 0x66747970
#define BOX_TYPE_TRAK 0x7472616b
#define BOX_TYPE_MDIA 0x6d646961
#define BOX_TYPE_MDHD 0x6d646864
#define BOX_TYPE_MDAT 0x6d646174
#define BOX_TYPE_MMPU 0x6d6d7075
#define BOX_TYPE_MFHD 0x6d666864
#define BOX_TYPE_TFHD 0x74666864
#define BOX_TYPE_TRUN 0x7472756e
#define BOX_TYPE_HDLR 0x68646c72

#define HANDLER_SUBT 0x73756274
#define HANDLER_VIDE 0x76696465
#define HANDLER_SOUN 0x736f756e

// The maximum mdat box size we accept. If we see a box larger than 
// MAX_MDAT_BOX_SIZE_ACCEPTED we assume a bitstream error has occurred.
#define MAX_MDAT_BOX_SIZE_ACCEPTED (24 * 1024 *1024)

// The maximum sample duration we accept. Samples with larger durations will be
// rejected. This is intended to avoid aberrant behavior if there is corrption
// of the boxes leading to wildly invalid sample_duration values. It may lead
// to rejection of valid sample durations (music choice channels for exmaple),
// and the way we determine valid sample durations may need to be examined
// in the future - possibly by looking at the duration of a segment as a whole.
#define MAX_SAMPLE_DURATION(timescale) (timescale / 10)

// DEBUG_ISOBMFF enables output ISOBMFF-related output (box validation, etc.)
// #define DEBUG_ISOBMFF 1

// DEBUG_SHADOWFILE enables additional logging that is useful in identifying
// data written to the shadow file. DEBUG_ISOBMFF may also be useful.
// #define DEBUG_SHADOWFILE 1

static uint8_t vPaddingG[4096] = 
{
    0,
};

typedef struct 
{
    uint32_t asset_ID_scheme;
    char strAssetID[128];
    uint32_t handler;

    int32_t packet_id;

    // The packet time of the first packet for the current fragment.
    struct timespec fragmentPacketTime;

    // The packet number of the first paket for the current fragment
    uint64_t fragmentPacketNumber;

    // MMTP Packet Sequence Number for this asset - should increment by one 
    // for each MMTP packet received.
    int64_t mmtpPacketSeqno;

    // MMTP packet counter - counts *down* the number of packets in the MFU
    int32_t mmtpPayloadFragmentationCounter;

    // The number of samples in the current MPU, extracted from the trun box
    struct
    {
        int32_t sample_size;
        int64_t sample_composition_time_offset;
        int32_t sample_duration;

        // sample_composition_time_offset + decoding_time (relative to this
        // fragment's base_decoding_time)
        int32_t frag_composition_time;
    } vSampleInfos[300];
    int32_t nSampleInfos;

    // the minimum composition time (relative to this fragment's 
    // base_decoding_time)
    int32_t min_frag_composition_time;

    // The expected offset of the next MMTP packet, relative to the start
    // of the sample it contains.
    uint32_t expectedSampleOffset; 

    // The size of the MFU header, which is added to the sample size for
    // fragmented packets only.
    uint32_t mfuHeaderLength;

    // The expected sample number of the next MMTP packet.
    uint32_t expectedSampleNumber; 

    // The mmtp_payload_sequence_number we are currently processing, 
    // initialized to -1
    int64_t expectedPayloadSequenceNumber;

    // The timescale from the mvhd box
    uint32_t mvhdTimescale;

    // The timescale from the mdhd box
    uint32_t mdhdTimescale;

    // The size of the mdat box
    uint64_t mdatBoxSize;
    // The size of the mdat box header
    uint32_t mdatBoxHeaderSize;
    
    // The number of extra bytes found after the mdat box header in the 
    // movie fragment metadata MPU
    int32_t nExtraMovieFragmentMetadataBytes;

    // The version of the last user service description fragment processed
    int32_t usbdVersion;
    uint32_t usbdCrc32;

    // The version of the last held fragment processed
    int32_t heldVersion;

    // The version of the last video stream properties fragment processed
    int32_t vspdVersion;

    // The version of the last audio stream properties fragment processed
    int32_t aspdVersion;

    // The version of the last caption asset descriptor fragment processed
    int32_t cadVersion;

    // Extracted from the MPU timestamp descriptor
    struct
    {
        int32_t fPresent;
        uint32_t mpu_sequence_number;
        uint64_t mpu_presentation_time;
    } mpu_timestamp_descriptor;

    // Packet aggregation buffer for type 0 or 1 (moov / moof) box
    // data
    uint8_t vMPUMetadataDUBuffer[8192];
    int32_t sMPUMetadataDUBuffer;
    int32_t nMPUMetadataDUBufferBytesUsed;
    uint32_t moovSequenceNumber;
    int32_t fHaveMPUMetadataDU;
    uint8_t vMovieFragmentMetadataDUBuffer[8192];
    int32_t sMovieFragmentMetadataDUBuffer;
    int32_t nMovieFragmentMetadataDUBufferBytesUsed;
    int32_t fHaveMovieFragmentMetadataDU;
    uint8_t *pMediaSampleBuffer;
    int32_t sMediaSampleBuffer;
    int32_t nMediaSampleBufferBytesUsed;
    struct _MediaSample
    {
        // A pointer into pMediaSampleBuffer
        uint8_t *pData;
        // Size of the media sample in bytes
        int32_t nData;
    } vMediaSamples[MAX_MEDIA_SAMPLES];
    int32_t nMediaSamples;

    // the presentation time of the first AU in an MPU, in mdhdTimescale units.
    uint64_t mpuPresentationTime;
    // The last good mpuPresentationTime received
    uint64_t lastGoodMPUPresentationTime;
    // The packet time when mpuPresentationTime was acquired
    struct timespec lastGoodMPUPresentationTimePacketTime;
    // The number of MPUs since the last MPU with a good base media decode
    // time.
    int32_t nRejectedBaseMediaDecodeTimes;

    // The duration of the mpu being processed
    uint32_t mpuDuration;

    // The number of bytes pushed for this asset
    uint64_t nBytesPushed;

    int shadowFd;
    char strShadowFilename[256];
} MMTPAssetInfo;

typedef struct _MMTPSession 
{
    ENVerbosity verbosity;
    char strDebugName[16];
    int32_t sessionId;

    uint32_t slsAddr;
    uint16_t slsPort;

    MMTPIPAddressAddCallback IPAddressAddCallback;
    MMTPIPAddressRemoveCallback IPAddressRemoveCallback;
    void *pIPAddrCallbackUserData;

    ENATSC3MMTPUSBDAvailableCallback USBDAvailableCallback;
    ENATSC3MMTPHELDAvailableCallback HELDAvailableCallback;
    ENATSC3MMTPVSPDAvailableCallback VSPDAvailableCallback;
    ENATSC3MMTPASPDAvailableCallback ASPDAvailableCallback;
    ENATSC3MMTPCADAvailableCallback CADAvailableCallback;
    ENATSC3MMTPMPTSignalingMessageCallback MPTSignalingMessageCallback;
    ENATSC3MMTPMediaDataCallback MediaDataCallback;
    void *pCallbackUserData;

    MMTPAssetInfo vAssetInfos[16];
}  MMTPSession, *MMTPSessionHandle_t;

// The following are parsed out of an MMTP packet. The PacketInfo also contains
// intormation on iterating through fragments which are aggregated in the 
// packet. 
typedef struct
{
    uint32_t length;
    uint32_t type;
    uint32_t timed_flag;
    uint32_t fragmentation_indicator;
    uint32_t aggregation_flag;
    uint32_t fragmentation_counter;
    uint32_t sequence_number;

    int32_t offset;

    int32_t dataUnitLength;

    const uint8_t *pData;
    uint32_t sData;
} MMTPPayloadInfo;

typedef struct
{
    uint32_t version;
    uint32_t packet_counter_flag;
    uint32_t fec_type;
    uint32_t header_extension_flag;
    uint32_t rap_flag;
    uint32_t qos_classifier_flag;
    uint32_t flow_identifier_flag;
    uint32_t flow_extension_flag;
    uint32_t compression_flag;
    uint32_t indicator_flag;
    uint32_t packet_type;
    uint32_t packet_id;
    uint32_t timestamp;
    uint32_t sequence_number;
    uint32_t packet_counter;
    uint32_t reliability_flag;
    uint32_t type_of_bitrate;
    uint32_t delay_sensitivity;
    uint32_t transport_priority;
    uint32_t flow_label;
    uint32_t header_extension_type;
    uint32_t header_extension_length;

    int32_t headerLength;
} MMTPProtocolPacketInfo;

struct 
{
    int32_t sessionCounter;
} enatsc3mmtpGlobalsG =
{
    .sessionCounter = 0,
};

int32_t
GetAssetInfo(uint16_t packet_id, MMTPSession *pSession, 
    MMTPAssetInfo **ppAssetInfo)
{
    int32_t i;
    int32_t rval = 0;

    for (i = 0; i < numof(pSession->vAssetInfos); i++)
    {
        if (pSession->vAssetInfos[i].packet_id == packet_id)
        {
            *ppAssetInfo = pSession->vAssetInfos + i;
            break;
        }
        else if (pSession->vAssetInfos[i].packet_id == -1)
        {
            MMTPAssetInfo *pAssetInfo = pSession->vAssetInfos + i;

            pAssetInfo->packet_id = packet_id;
            *ppAssetInfo = pAssetInfo;

            if (getenv("MMTP_SHADOW_PREFIX"))
            {
                if (snprintf(pAssetInfo->strShadowFilename, 
                    sizeof(pAssetInfo->strShadowFilename), "%s-%d-%s-%d.mp4", 
                    getenv("MMTP_SHADOW_PREFIX"), pSession->sessionId, 
                    pSession->strDebugName, packet_id) >= 
                    sizeof(pAssetInfo->strShadowFilename))
                {
                    ERRORPRINT("Filename too long\n");
                }
                else
                {
                    pAssetInfo->shadowFd = open(pAssetInfo->strShadowFilename,
                        O_WRONLY | O_CREAT | O_TRUNC, 0666);
                }
            }
            break;
        }
    }
   
    if (i == numof(pSession->vAssetInfos))
    {
        *ppAssetInfo = NULL;
        rval = 1;
    }

    return rval;
}

int32_t
PushData(struct timespec *pPacketTime, uint64_t packetNumber, 
    MMTPAssetInfo *pAssetInfo, MMTPSession *pSession, uint8_t *pData, 
    int32_t sData, uint32_t mmtp_payload_type, 
    uint32_t mmtp_payload_sequence_number)
{
    int32_t rval = 0;

    if (pAssetInfo->shadowFd >= 0)
    {
#ifdef DEBUG_SHADOWFILE
        LOGPRINT_NOFILE("%s:%s:%s WRITING %d @ %"PRIu64"\n",
            pSession->strDebugName, pAssetInfo->strAssetID,
            pAssetInfo->strShadowFilename, sData, lseek(pAssetInfo->shadowFd, 
            0, SEEK_CUR));
#endif
        if (write(pAssetInfo->shadowFd, pData, sData) != sData)
        {
            LOGPRINT_NOFILE("%s:%s Failed to write to shadow file\n",
                pSession->strDebugName, pAssetInfo->strAssetID);
        }
    }

    if (pSession->MediaDataCallback)
    {
        if (pSession->verbosity >= ENVerbosity_ObjectTrace)
        {
            LOGPRINT("%s:%s: Push %d @ %"PRIu64"\n", pSession->strDebugName,
                pAssetInfo->strAssetID, sData, pAssetInfo->nBytesPushed);
        }

        pSession->MediaDataCallback(pPacketTime, packetNumber, 
            pAssetInfo->packet_id, pAssetInfo->strAssetID, mmtp_payload_type, 
            mmtp_payload_sequence_number, pAssetInfo->mpuPresentationTime, 
            pAssetInfo->mpuDuration, pAssetInfo->mdhdTimescale, pData, sData, 
            pSession->pCallbackUserData);

        pAssetInfo->nBytesPushed += sData;
    }

    return rval;
}

int32_t
PadMDAT(struct timespec *pPacketTime, uint64_t packetNumber, 
    MMTPAssetInfo *pAssetInfo, MMTPSession *pSession, int32_t sPadding, 
    uint32_t mmtp_payload_type, uint32_t mmtp_payload_sequence_number)
{
    int32_t rval = 0;

    if (sPadding < 0)
    {
        ERRORPRINT("%s:%s NEGATIVE PADDING %d\n", pSession->strDebugName, 
            pAssetInfo->strAssetID, sPadding);
        return 0;
    }

    ENUtils_CrashAssert(sPadding >= 0);

    while (sPadding)
    {
        int32_t nb;

        nb = MIN(sPadding, sizeof(vPaddingG));
        if (PushData(pPacketTime, packetNumber, pAssetInfo, pSession, 
            vPaddingG, nb, mmtp_payload_type, mmtp_payload_sequence_number))
        {
            ERRORPRINT("%s:%s Failed to write padding\n", 
                pSession->strDebugName, pAssetInfo->strAssetID);
            rval = 1;
            break;
        }

        sPadding -= nb;
    }

    return rval;
}

int32_t
MMTPPayloadParse(MMTPSession *pSession, const uint8_t *pData, uint32_t sData, 
    MMTPPayloadInfo *pPayloadInfo)
{
    int32_t rval = 0;

    if (sData < 8)
    {
        rval = 1;
        goto out;
    }

    pPayloadInfo->pData = pData;
    pPayloadInfo->sData = sData;

    pPayloadInfo->length = (pData[0] << 8) | pData[1];
    pPayloadInfo->type = (pData[2] >> 4) & 0xf;
    pPayloadInfo->timed_flag = (pData[2] >> 3) & 0x1;
    pPayloadInfo->fragmentation_indicator = (pData[2] >> 1) & 0x3;
    pPayloadInfo->aggregation_flag = pData[2] & 0x1;
    pPayloadInfo->fragmentation_counter = pData[3];
    pPayloadInfo->sequence_number = (pData[4] << 24) | 
        (pData[5] << 16) | (pData[6] << 8) | pData[7]; 

    pPayloadInfo->offset = 8;

    if (pPayloadInfo->aggregation_flag)
    {
        // ENUtils_CrashAssert("Unverified code path" && 0);
        if (pPayloadInfo->offset + 2 > sData)
        {
            ENUtils_CrashAssert(0);
            rval = 1;
            goto out;
        }

        pPayloadInfo->dataUnitLength = 
            (pData[pPayloadInfo->offset] << 8) |
            pData[pPayloadInfo->offset + 1];
        pPayloadInfo->offset += 2;
    }
    else
    {
        pPayloadInfo->dataUnitLength = pPayloadInfo->length - 
            pPayloadInfo->offset;
    }

    if (pPayloadInfo->offset + pPayloadInfo->dataUnitLength > 
        pPayloadInfo->sData)
    {
        LOGPRINT("%s: MMTP Payload overflow @ %d (n = %d)\n",  
            pSession->strDebugName, pPayloadInfo->offset + 2, 
            pPayloadInfo->sData);
        rval = 1;
        goto out;
    }

out:
    return rval;
}

int32_t
MMTPPayloadAdvance(MMTPSession *pSession, MMTPPayloadInfo *pPayloadInfo)
{
    int32_t rval = 0;

    pPayloadInfo->offset += pPayloadInfo->dataUnitLength;

    if (!pPayloadInfo->aggregation_flag)
    {
        goto out;
    }

    if (pPayloadInfo->offset == pPayloadInfo->length)
    {
        goto out;
    }

    if (pPayloadInfo->offset + 2 > pPayloadInfo->sData)
    {
        LOGPRINT("%s: MMTP Payload overflow @ %d (n = %d)\n",
            pSession->strDebugName, pPayloadInfo->offset + 2, 
            pPayloadInfo->sData);
        rval = 1;
        goto out;
    }

    pPayloadInfo->dataUnitLength = 
        (pPayloadInfo->pData[pPayloadInfo->offset] << 8) |
        pPayloadInfo->pData[pPayloadInfo->offset + 1];
    pPayloadInfo->offset += 2;

    if (pPayloadInfo->offset + pPayloadInfo->dataUnitLength > 
        pPayloadInfo->sData)
    {
        LOGPRINT("%s: MMTP Payload overflow @ %d (n = %d)\n", 
            pSession->strDebugName, pPayloadInfo->offset + 2, 
            pPayloadInfo->sData);
        rval = 1;
        goto out;
    }

out:
    return rval;
}

int32_t
MMTPProtocolPacketParse(MMTPSession *pSession, const uint8_t *pPayload, 
    MMTPProtocolPacketInfo *pProtocolPacketInfo)
{
    int32_t offset = 0;
    int32_t rval = 0;

    pProtocolPacketInfo->version = pPayload[0] >> 6;
    
    if (pProtocolPacketInfo->version == 1)
    {
        pProtocolPacketInfo->packet_counter_flag = (pPayload[0] >> 5) & 1;
        pProtocolPacketInfo->fec_type = (pPayload[0] >> 3) & 0x3;
        pProtocolPacketInfo->header_extension_flag = (pPayload[0] >> 2) & 0x1;
        pProtocolPacketInfo->rap_flag = (pPayload[0] >> 1) & 0x1;
        pProtocolPacketInfo->qos_classifier_flag = pPayload[0] & 0x1;
        pProtocolPacketInfo->flow_identifier_flag = (pPayload[1] >> 7) & 0x1;
        pProtocolPacketInfo->flow_extension_flag = (pPayload[1] >> 6) & 0x1;
        pProtocolPacketInfo->compression_flag = (pPayload[1] >> 5) & 0x1;
        pProtocolPacketInfo->indicator_flag = (pPayload[1] >> 4) & 0x01;
        pProtocolPacketInfo->packet_type = pPayload[1] & 0x0f;
        pProtocolPacketInfo->packet_id = (pPayload[2] << 8) | pPayload[3];
        pProtocolPacketInfo->timestamp = 
            (pPayload[4] << 24) | (pPayload[5] << 16) | 
            (pPayload[6] << 8) | pPayload[7];
        pProtocolPacketInfo->sequence_number = 
            (pPayload[8] << 24) | (pPayload[9] << 16) | 
            (pPayload[10] << 8) | pPayload[11];
        offset = 12;
        if (pProtocolPacketInfo->packet_counter_flag)
        {
            pProtocolPacketInfo->packet_counter = 
                (pPayload[12] << 24) | (pPayload[13] << 16) | 
                (pPayload[14] << 8) | pPayload[15];
            offset += 4;
        }
        pProtocolPacketInfo->reliability_flag =  (pPayload[offset] >> 7) & 0x1;
        pProtocolPacketInfo->type_of_bitrate = (pPayload[offset] >> 5) & 0x3;
        pProtocolPacketInfo->delay_sensitivity = (pPayload[offset] >> 2) & 0x7;
        pProtocolPacketInfo->transport_priority = 
            ((pPayload[offset] & 0x3) << 1) | (pPayload[offset + 1] >> 7);
        pProtocolPacketInfo->flow_label = pPayload[offset + 1] & 0x7f;

        offset += 2;
        pProtocolPacketInfo->headerLength = offset;
        if (pProtocolPacketInfo->header_extension_flag)
        {
            pProtocolPacketInfo->header_extension_type = 
                (pPayload[offset] << 8) | pPayload[offset + 1];
            pProtocolPacketInfo->header_extension_length = 
                (pPayload[offset + 2] << 8) | pPayload[offset + 3];

            pProtocolPacketInfo->headerLength += 
                4 + pProtocolPacketInfo->header_extension_length;
        }

    }
    else
    {
        ERRORPRINT("%s Unhandled MMTP Protocol Version %d\n", 
            pSession->strDebugName, pProtocolPacketInfo->version);
        rval = 1;
        goto out;
    }

out:
    return rval;
}

ENATSC3Status
FindBox(uint8_t *pData, int64_t sData, uint32_t boxType, int32_t *pBoxOffset,
    uint32_t *pBoxHeaderSize, uint64_t *pBoxSize)
{
    uint32_t cursor = 0;
    ENATSC3Status rval = ENATSC3Status_Failure;

    while (cursor + 8 <= sData)
    {
        uint64_t size;
        uint32_t type;

        size = ntohl(*((uint32_t *)(pData + cursor)));
        type = ntohl(*((uint32_t *)(pData + cursor + 4)));
        if (size == 1)
        {
            if (cursor + 16 > sData)
            {
                break;
            }

            size = 
                (((uint64_t)ntohl(*((uint32_t *)(pData + cursor + 8)))) << 32) |
                ntohl(*((uint32_t *)(pData + cursor + 12)));
            *pBoxHeaderSize = 16;
        }
        else
        {
            *pBoxHeaderSize = 8;
        }

        // A box size of 0 is invalid (and would result in an infinite loop)
        if (size == 0)
        {
            break;
        }

        // Also discard boxes that would cause the cursor to overflow. We
        // aren't expecting any 2GB boxes...
        if (size + cursor > MAX_BOX_SIZE)
        {
            break;
        }

        if (type == boxType)
        {
            *pBoxOffset = cursor;
            *pBoxSize = size;
            rval = ENATSC3Status_Success;
            break;
        }

        cursor += size;
    }

    return rval;
}

ENATSC3Status
ParseAssetInfoFromMmpu(MMTPSession *pSession, uint8_t *pData, int32_t sData, 
    MMTPAssetInfo *pAssetInfo)
{
    int32_t cursor;
    uint32_t boxHeaderSize;
    uint64_t mmpuBoxSize;
    ENATSC3Status rval = ENATSC3Status_Success;

    if (FindBox(pData, sData, BOX_TYPE_MMPU, &cursor, &boxHeaderSize, 
        &mmpuBoxSize))
    {
        LOGPRINT("%s:%s Failed to find mmpu box\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    uint32_t asset_ID_scheme;
    uint32_t asset_ID_length;
    uint32_t boxVersion;
    uint32_t boxFlags;
    uint32_t mpuSequenceNumber;
    uint32_t assetType_IsADCPresent_Reserved;

    (void)boxVersion;
    (void)boxFlags;
    (void)mpuSequenceNumber;
    (void)assetType_IsADCPresent_Reserved;

    if (cursor + 25 >= sData)
    {
        LOGPRINT("%s:%s short AssetIdentifierBox\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    boxVersion = pData[cursor + 8];
    boxFlags = (pData[cursor + 9] << 16) | (pData[cursor + 10] << 8) | 
        pData[cursor + 11];
    assetType_IsADCPresent_Reserved = pData[cursor + 12];
    mpuSequenceNumber = ntohl(*((uint32_t *)(pData + cursor + 13)));

    asset_ID_scheme = ntohl(*((uint32_t *)(pData + cursor + 17)));
    asset_ID_length = ntohl(*((uint32_t *)(pData + cursor + 21)));

    if (cursor + 25 + asset_ID_length >= sData)
    {
        LOGPRINT("%s:%s short AssetIdentifierBox\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    if (asset_ID_length >= sizeof(pAssetInfo->strAssetID))
    {
        LOGPRINT("%s:%s Asset name too long: %.*s\n", pSession->strDebugName, 
            pAssetInfo->strAssetID, asset_ID_length, pData + cursor + 25);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    strncpy(pAssetInfo->strAssetID, (char *)(pData + cursor + 25), 
        asset_ID_length);
    pAssetInfo->strAssetID[asset_ID_length] = 0;
    pAssetInfo->asset_ID_scheme = asset_ID_scheme;

out:
    return rval;
}

typedef struct
{
    // FullBox 
    uint8_t version;
    uint8_t flags[3];

    union
    {
        struct
        {
            uint32_t creation_time;
            uint32_t modification_time;
            uint32_t timescale;
            uint32_t duration;
            uint32_t rate;
            uint16_t volume;
            uint16_t reserved1;
            uint32_t reserved2[2];
            uint32_t matrix[9];
            uint32_t pre_defined[6]; 
            uint32_t next_track_ID;
        } __attribute__((__packed__)) v0;

        struct
        {
            uint64_t creation_time;
            uint64_t modification_time;
            uint32_t timescale;
            uint64_t duration;
            uint32_t rate;
            uint16_t volume;
            uint16_t reserved1;
            uint32_t reserved2[2];
            uint32_t matrix[9];
            uint32_t pre_defined[6];
            uint32_t next_track_ID;
        } __attribute__((__packed__)) v1;
    } __attribute__((__packed__));

} __attribute__((__packed__)) MvhdBoxBody;

typedef struct
{
    // FullBox 
    uint8_t version;
    uint8_t flags[3];

    union
    {
        struct
        {
            uint32_t creation_time;
            uint32_t modification_time;
            uint32_t timescale;
            uint32_t duration;
            uint16_t language;
            uint16_t pre_defined;
        } __attribute__((__packed__)) v0;

        struct
        {
            uint64_t creation_time;
            uint64_t modification_time;
            uint32_t timescale;
            uint64_t duration;
            uint16_t language;
            uint16_t pre_defined;
        } __attribute__((__packed__)) v1;
    } __attribute__((__packed__));

} __attribute__((__packed__)) MdhdBoxBody;

ENATSC3Status
VerifyBoxStructure(MMTPSession *pSession, MMTPAssetInfo *pAssetInfo, 
    uint8_t *pData, int32_t sData)
{
    struct Box
    {
        uint32_t size;
        uint32_t type;
    } __attribute((__packed__)) *pBox;
    ENATSC3Status rval = ENATSC3Status_Failure;

    // We are simply going to walk the top-level boxes until we see the
    // mdat box. see
    while (sData >= sizeof(struct Box))
    {
        uint64_t boxSize;
        uint32_t boxType;

        pBox = (struct Box *)pData;
        boxSize = ntohl(pBox->size);
        boxType = ntohl(pBox->type);

        // If the box size == 1 then the box size is actually a 64-bit
        // value following the type.
        if (boxSize == 1)
        {
            if (sData >= sizeof(struct Box) + 8)
            {
                boxSize =
                    ENUtils_ntohll((uint64_t)(pData + sizeof(struct Box)));
            }
        }

#ifdef DEBUG_ISOBMFF
        LOGPRINT_NOFILE("%s:%s: VERIFYBOX %c%c%c%c size %"PRIu64"\n",
            pSession->strDebugName, pAssetInfo->strAssetID, 
            (boxType >> 24) & 0xff, (boxType >> 16) & 0xff, 
            (boxType >> 8) & 0xff, boxType & 0xff, boxSize);
#endif

        // A box size of 0 is invalid (and would result in an infinite loop)
        if (boxSize == 0)
        {
            break;
        }

        // If we hit an mdat box then we are done - it is the last box in a 
        // fragment. The payload of the mdat is not provided here so its 
        // size must be verified elsewhere.
        if (boxType == BOX_TYPE_MDAT)
        {
            rval = ENATSC3Status_Success;
            break;
        }

        // This means we have corrupt box structures...
        if (boxSize > sData)
        {
            break;
        }

        sData -= boxSize;
        pData += boxSize;
    }

    if (sData == 0)
    {
        rval = ENATSC3Status_Success;
    }

    return rval;
}

ENATSC3Status
ParseAssetInfoFromMoov(MMTPSession *pSession, uint8_t *pData, int32_t sData, 
    MMTPAssetInfo *pAssetInfo)
{
    int32_t cursor;
    uint64_t moovBoxSize;
    uint32_t moovBoxHeaderSize;
    uint8_t *pMoovBoxData;
    uint64_t mvhdBoxSize;
    uint32_t mvhdBoxHeaderSize;
    uint8_t *pMvhdBoxData;
    MvhdBoxBody *pMvhdBoxBody;
    uint64_t trakBoxSize;
    uint32_t trakBoxHeaderSize;
    uint8_t *pTrakBoxData;
    uint64_t mdiaBoxSize;
    uint32_t mdiaBoxHeaderSize;
    uint8_t *pMdiaBoxData;
    uint64_t hdlrBoxSize;
    uint32_t hdlrBoxHeaderSize;
    uint8_t *pHdlrBoxData;
    uint64_t mdhdBoxSize;
    uint32_t mdhdBoxHeaderSize;
    uint8_t *pMdhdBoxData;
    MdhdBoxBody *pMdhdBoxBody;
    ENATSC3Status rval = ENATSC3Status_Success;

    if (FindBox(pData, sData, BOX_TYPE_MOOV, &cursor, &moovBoxHeaderSize, 
        &moovBoxSize))
    {
        LOGPRINT("%s:%s Failed to find moov box\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    if (cursor + moovBoxSize > sData)
    {
        LOGPRINT("%s:%s moov box too large\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    pMoovBoxData = pData + cursor + moovBoxHeaderSize;

    if (FindBox(pMoovBoxData, moovBoxSize - moovBoxHeaderSize, BOX_TYPE_MVHD,
        &cursor, &mvhdBoxHeaderSize, &mvhdBoxSize))
    {
        LOGPRINT("%s:%s Failed to find mvhd box\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    if (cursor + mvhdBoxSize > moovBoxSize - moovBoxHeaderSize)
    {
        LOGPRINT("%s:%s mvhd box exceeds moov\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    pMvhdBoxData = pMoovBoxData + cursor + mvhdBoxHeaderSize;
    pMvhdBoxBody = (MvhdBoxBody *)pMvhdBoxData;
    if (pMvhdBoxBody->version == 0)
    {
        pAssetInfo->mvhdTimescale = ntohl(pMvhdBoxBody->v0.timescale);
    }
    else if (pMvhdBoxBody->version == 1)
    {
        pAssetInfo->mvhdTimescale = ntohl(pMvhdBoxBody->v1.timescale);
    }
    else
    {
        LOGPRINT("%s:%s Unhandled mvhd version %d\n", pSession->strDebugName, 
            pAssetInfo->strAssetID, pMvhdBoxBody->version);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    if (FindBox(pMoovBoxData, moovBoxSize - moovBoxHeaderSize, BOX_TYPE_TRAK,
        &cursor, &trakBoxHeaderSize, &trakBoxSize))
    {
        LOGPRINT("%s:%s Failed to find trak box\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    if (cursor + trakBoxSize > moovBoxSize - moovBoxHeaderSize)
    {
        LOGPRINT("%s:%s trak box exceeds moov\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    pTrakBoxData = pMoovBoxData + cursor + trakBoxHeaderSize;

    if (FindBox(pTrakBoxData, trakBoxSize - trakBoxHeaderSize, BOX_TYPE_MDIA,
        &cursor, &mdiaBoxHeaderSize, &mdiaBoxSize))
    {
        LOGPRINT("%s:%s Failed to find mdia box\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    if (cursor + mdiaBoxSize > trakBoxSize - trakBoxHeaderSize)
    {
        LOGPRINT("%s:%s trak mdia exceeds moov\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    pMdiaBoxData = pTrakBoxData + cursor + mdiaBoxHeaderSize;

    if (FindBox(pMdiaBoxData, mdiaBoxSize - mdiaBoxHeaderSize, BOX_TYPE_HDLR,
        &cursor, &hdlrBoxHeaderSize, &hdlrBoxSize))
    {
        LOGPRINT("%s:%s Failed to find hdlr box\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
    }
    else
    {
        if (cursor + hdlrBoxSize > mdiaBoxSize - mdiaBoxHeaderSize)
        {
            LOGPRINT("%s:%s hdlr exceeds mdia\n", pSession->strDebugName,
                pAssetInfo->strAssetID);
        }
        else
        {
            pHdlrBoxData = pMdiaBoxData + cursor + hdlrBoxHeaderSize;
            uint32_t handler;
            handler =htonl(*((uint32_t *)(pHdlrBoxData + 8)));
            if (handler == HANDLER_SUBT ||
                handler == HANDLER_VIDE ||
                handler == HANDLER_SOUN)
            {
                pAssetInfo->handler = handler;
            }
        }
    }

    if (FindBox(pMdiaBoxData, mdiaBoxSize - mdiaBoxHeaderSize, BOX_TYPE_MDHD,
        &cursor, &mdhdBoxHeaderSize, &mdhdBoxSize))
    {
        LOGPRINT("%s:%s Failed to find mdhd box\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    if (cursor + mdhdBoxSize > mdiaBoxSize - mdiaBoxHeaderSize)
    {
        LOGPRINT("%s:%s mdia box exceeds moov\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    pMdhdBoxData = pMdiaBoxData + cursor + mdhdBoxHeaderSize;
    pMdhdBoxBody = (MdhdBoxBody *)pMdhdBoxData;

    if (pMdhdBoxBody->version == 0)
    {
        pAssetInfo->mdhdTimescale = ntohl(pMdhdBoxBody->v0.timescale);
    }
    else if (pMdhdBoxBody->version == 1)
    {
        pAssetInfo->mdhdTimescale = ntohl(pMdhdBoxBody->v1.timescale);
    }
    else
    {
        LOGPRINT("%s:%s Unhandled mdhd version %d\n", pSession->strDebugName, 
            pAssetInfo->strAssetID, pMdhdBoxBody->version);
        rval = ENATSC3Status_Failure;
        goto out;
    }
    
out:
    return rval;
}

int32_t
ParseSampleInfoFromMoof(MMTPSession *pSession, struct timespec *pPacketTime, 
    uint8_t *pData, int32_t sData, MMTPAssetInfo *pAssetInfo, 
    MMTPPayloadInfo *pMMTPPayloadInfo)
{
    int32_t cursor;
    uint64_t moofBoxSize;
    uint32_t moofBoxHeaderSize;
    uint8_t *pMoofBoxData;
    uint64_t mfhdBoxSize;
    uint32_t mfhdBoxHeaderSize;
    uint8_t *pMfhdBoxData;
    uint64_t trafBoxSize;
    uint32_t trafBoxHeaderSize;
    uint8_t *pTrafBoxData;
    uint64_t trunBoxSize;
    uint32_t trunBoxHeaderSize;
    uint8_t *pTrunBoxData;
    int32_t trunOffset;
    uint8_t trunVersion;
    uint32_t trunFlags;
    uint32_t trunSampleCount;
    uint32_t trunDataOffset = 0;
    uint32_t trunFirstSampleFlags = 0;
    uint64_t tfhdBoxSize;
    uint32_t tfhdBoxHeaderSize;
    uint8_t *pTfhdBoxData;
    int32_t tfhdOffset;
    uint32_t tfhdFlags;
    uint32_t tfhdTrackID;
    uint64_t tfhdBaseDataOffset;
    uint32_t tfhdSampleDescriptionIndex;
    uint32_t tfhdDefaultSampleDuration;
    uint32_t tfhdDefaultSampleSize;
    uint32_t tfhdDefaultSampleFlags;
    uint64_t tfdtBoxSize;
    uint32_t tfdtBoxHeaderSize;
    uint8_t *pTfdtBoxData;
    uint64_t mdatBoxSize;
    uint32_t mdatBoxHeaderSize;
    int32_t decoding_time;
    uint64_t presentationTime;
    int32_t fHavePresentationTime = 0;
    uint32_t mpuDuration = 0;
    int32_t i;
    int32_t rval = 0;

    (void)trunDataOffset;
    (void)tfhdTrackID;
    (void)tfhdBaseDataOffset;
    (void)tfhdSampleDescriptionIndex;
    (void)tfhdDefaultSampleFlags;

    if (!pAssetInfo->mpu_timestamp_descriptor.fPresent)
    {
        LOGPRINT("%s:%s ERROR NO MPU TIMETAMP DESCRIPTOR!!!!!!!!\n",
            pSession->strDebugName, pAssetInfo->strAssetID);
        presentationTime = 0;
    }
    else if (pAssetInfo->mpu_timestamp_descriptor.mpu_sequence_number != 
        pMMTPPayloadInfo->sequence_number)
    {
        LOGPRINT("%s:%s ERROR MPU TIMESTAMP DESCRIPTOR FROM WRONG SEQUENCE %d != %d\n",
            pSession->strDebugName, pAssetInfo->strAssetID,  
            pAssetInfo->mpu_timestamp_descriptor.mpu_sequence_number,
            pMMTPPayloadInfo->sequence_number);
        presentationTime = 0;
    }
    else
    {
        presentationTime = 
            pAssetInfo->mpu_timestamp_descriptor.mpu_presentation_time;
        fHavePresentationTime = 1;
    }

    // Find the moof box.
    if (FindBox(pData, sData, BOX_TYPE_MOOF, &cursor, &moofBoxHeaderSize, 
        &moofBoxSize))
    {
        LOGPRINT("%s:%s Failed to find moof box\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = 1;
        goto out;
    }

    if (cursor + moofBoxSize > sData)
    {
        LOGPRINT("%s:%s moof box exceeds data\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = 1;
        goto out;
    }

    pMoofBoxData = pData + cursor + moofBoxHeaderSize;

    // Find the mfhd box contained in the moof box
    if (FindBox(pMoofBoxData, moofBoxSize - moofBoxHeaderSize, BOX_TYPE_MFHD,
        &cursor, &mfhdBoxHeaderSize, &mfhdBoxSize))
    {
        LOGPRINT("%s:%s Failed to find mfhd box\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = 1;
        goto out;
    }

    if (cursor + mfhdBoxSize > moofBoxSize - moofBoxHeaderSize)
    {
        LOGPRINT("%s:%s mfhd box exceeds data\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = 1;
        goto out;
    }

    pMfhdBoxData = pMoofBoxData + cursor + mfhdBoxHeaderSize;

    if (mfhdBoxSize - mfhdBoxHeaderSize < 4)
    {
        LOGPRINT("%s:%s mfhd box too small (%"PRIu64")\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, mfhdBoxSize);
        rval = 1;
        goto out;
    }

    // Rewrite the mfhd sequence number using the MMTP payload sequence number
    pMfhdBoxData[0] = (pMMTPPayloadInfo->sequence_number >> 24) & 0xff; 
    pMfhdBoxData[1] = (pMMTPPayloadInfo->sequence_number >> 16) & 0xff; 
    pMfhdBoxData[2] = (pMMTPPayloadInfo->sequence_number >> 8) & 0xff; 
    pMfhdBoxData[3] = (pMMTPPayloadInfo->sequence_number >> 0) & 0xff; 

    // Find the first traf box contained in the moof box.
    //
    // XXX We really want the traf box corresponding to the audio or video
    // samples in the movie fragment. There may be additional traf boxes, 
    // in particular a hint traf box. The track ID should be used to determine
    // which traf box we need. 
    if (FindBox(pMoofBoxData, moofBoxSize - moofBoxHeaderSize, BOX_TYPE_TRAF,
        &cursor, &trafBoxHeaderSize, &trafBoxSize))
    {
        LOGPRINT("%s:%s Failed to find traf box\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = 1;
        goto out;
    }

    if (cursor + trafBoxSize > moofBoxSize - moofBoxHeaderSize)
    {
        LOGPRINT("%s:%s mfhd box exceeds data\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = 1;
        goto out;
    }

    pTrafBoxData = pMoofBoxData + cursor + trafBoxHeaderSize;

    // Find the tfhd box contained in the traf box.
    if (FindBox(pTrafBoxData, trafBoxSize - trafBoxHeaderSize, BOX_TYPE_TFHD,
        &cursor, &tfhdBoxHeaderSize, &tfhdBoxSize))
    {
        LOGPRINT("%s:%s Failed to find tfhd box\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = 1;
        goto out;
    }

    if (cursor + tfhdBoxSize > trafBoxSize - trafBoxHeaderSize)
    {
        LOGPRINT("%s:%s mfhd box exceeds data\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = 1;
        goto out;
    }

    pTfhdBoxData = pTrafBoxData + cursor + tfhdBoxHeaderSize;
    tfhdOffset = 0;

    tfhdFlags = 
        (pTfhdBoxData[tfhdOffset + 1] << 16) | 
        (pTfhdBoxData[tfhdOffset + 2] << 8) | 
        pTfhdBoxData[tfhdOffset + 3];
    tfhdTrackID = 
        (pTfhdBoxData[tfhdOffset + 4] << 24) | 
        (pTfhdBoxData[tfhdOffset + 5] << 16) | 
        (pTfhdBoxData[tfhdOffset + 6] << 8) | 
        pTfhdBoxData[tfhdOffset + 7];
    tfhdOffset += 8;

    if (tfhdFlags & 0x01)
    {
        tfhdBaseDataOffset = 
            ((uint64_t)pTfhdBoxData[tfhdOffset + 0] << 56) | 
            ((uint64_t)pTfhdBoxData[tfhdOffset + 1] << 48) | 
            ((uint64_t)pTfhdBoxData[tfhdOffset + 2] << 40) | 
            ((uint64_t)pTfhdBoxData[tfhdOffset + 3] << 32) |
            (pTfhdBoxData[tfhdOffset + 4] << 24) | 
            (pTfhdBoxData[tfhdOffset + 5] << 16) | 
            (pTfhdBoxData[tfhdOffset + 6] << 8) | 
            pTfhdBoxData[tfhdOffset + 7];
        tfhdOffset += 8;
    }
    if (tfhdFlags & 0x02)
    {
        tfhdSampleDescriptionIndex = 
            (pTfhdBoxData[tfhdOffset + 0] << 24) | 
            (pTfhdBoxData[tfhdOffset + 1] << 16) | 
            (pTfhdBoxData[tfhdOffset + 2] << 8) | 
            pTfhdBoxData[tfhdOffset + 3];
        tfhdOffset += 4;
    }
    if (tfhdFlags & 0x08)
    {
        tfhdDefaultSampleDuration =
            (pTfhdBoxData[tfhdOffset + 0] << 24) | 
            (pTfhdBoxData[tfhdOffset + 1] << 16) | 
            (pTfhdBoxData[tfhdOffset + 2] << 8) | 
            pTfhdBoxData[tfhdOffset + 3];
        if (pAssetInfo->handler == HANDLER_VIDE ||
            pAssetInfo->handler == HANDLER_SOUN)
        {
            if (tfhdDefaultSampleDuration > 
                MAX_SAMPLE_DURATION(pAssetInfo->mdhdTimescale))
            {
                ERRORPRINT(
                    "%s: Invalid tfhd default sample duration %u, using %u\n",
                    pAssetInfo->strAssetID, tfhdDefaultSampleDuration, 
                    pAssetInfo->mdhdTimescale / 120);
                tfhdDefaultSampleDuration = pAssetInfo->mdhdTimescale / 120;
                pTfhdBoxData[tfhdOffset + 0] = 
                    ((pAssetInfo->mdhdTimescale / 120) >> 24) & 0xff;
                pTfhdBoxData[tfhdOffset + 1] = 
                    ((pAssetInfo->mdhdTimescale / 120) >> 16) & 0xff;
                pTfhdBoxData[tfhdOffset + 2] = 
                    ((pAssetInfo->mdhdTimescale / 120) >> 8) & 0xff;
                pTfhdBoxData[tfhdOffset + 3] = 
                    ((pAssetInfo->mdhdTimescale / 120) >> 0) & 0xff;
            }
        }
        tfhdOffset += 4;
    }
    else
    {
        tfhdDefaultSampleDuration = 0;
    }
    if (tfhdFlags & 0x10)
    {
        tfhdDefaultSampleSize = 
            (pTfhdBoxData[tfhdOffset + 0] << 24) | 
            (pTfhdBoxData[tfhdOffset + 1] << 16) | 
            (pTfhdBoxData[tfhdOffset + 2] << 8) | 
            pTfhdBoxData[tfhdOffset + 3];
        tfhdOffset += 4;
    }
    else
    {
        tfhdDefaultSampleSize = 0;
    }
    if (tfhdFlags & 0x20)
    {
        tfhdDefaultSampleFlags = 
            (pTfhdBoxData[tfhdOffset + 0] << 24) | 
            (pTfhdBoxData[tfhdOffset + 1] << 16) | 
            (pTfhdBoxData[tfhdOffset + 2] << 8) | 
            pTfhdBoxData[tfhdOffset + 3];
        tfhdOffset += 4;
    }
    else
    {
        tfhdDefaultSampleFlags = 0;
    }

    // Find the tfdt box contained in the traf box.
    if (FindBox(pTrafBoxData, trafBoxSize - trafBoxHeaderSize, BOX_TYPE_TFDT,
        &cursor, &tfdtBoxHeaderSize, &tfdtBoxSize))
    {
        LOGPRINT("%s:%s Failed to find tfdt box\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = 1;
        goto out;
    }

    if (cursor + tfdtBoxSize > trafBoxSize - trafBoxHeaderSize)
    {
        LOGPRINT("%s:%s mfhd box exceeds data\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = 1;
        goto out;
    }

    pTfdtBoxData = pTrafBoxData + cursor + tfdtBoxHeaderSize;

    if (!fHavePresentationTime)
    {
        uint8_t *pTfdtBox;

        // We don't know the presentation time. Nuke the tfdt box, which 
        // should cause this fragemnt to play contiguously with the preceding
        // one. 
        //
        // XXX We could try to use the last available sequence_number/
        // presentation_time pair and extrapolate using estimated fragment
        // durations. This might allow slightly faster recovery however
        // since we are in this condition, the bitstream has been damaged
        // already.
        pTfdtBox = pTfdtBoxData - tfdtBoxHeaderSize;
        pTfdtBox[4] = 's';
        pTfdtBox[5] = 'k';
        pTfdtBox[6] = 'i';
        pTfdtBox[7] = 'p';
    }

    // Find the trun box contained in the traf box.
    if (FindBox(pTrafBoxData, trafBoxSize - trafBoxHeaderSize, BOX_TYPE_TRUN,
        &cursor, &trunBoxHeaderSize, &trunBoxSize))
    {
        LOGPRINT("%s:%s Failed to find trun box\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = 1;
        goto out;
    }

    if (cursor + trunBoxSize > trafBoxSize - trafBoxHeaderSize)
    {
        LOGPRINT("%s:%s mfhd box exceeds data\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = 1;
        goto out;
    }

    pTrunBoxData = pTrafBoxData + cursor + trunBoxHeaderSize;
    trunOffset = 0;

    trunVersion = pTrunBoxData[trunOffset];
    trunFlags = 
        (pTrunBoxData[trunOffset + 1] << 16) | 
        (pTrunBoxData[trunOffset + 2] << 8) | 
        pTrunBoxData[trunOffset + 3];
    trunSampleCount = 
        (pTrunBoxData[trunOffset + 4] << 24) | 
        (pTrunBoxData[trunOffset + 5] << 16) | 
        (pTrunBoxData[trunOffset + 6] << 8) | 
        pTrunBoxData[trunOffset + 7];
    trunOffset += 8;

    if (trunFlags & 0x01)
    {
        trunDataOffset = 
            (pTrunBoxData[trunOffset + 0] << 24) | 
            (pTrunBoxData[trunOffset + 1] << 16) | 
            (pTrunBoxData[trunOffset + 2] << 8) | 
            pTrunBoxData[trunOffset + 3];
        trunOffset += 4;
    }

    if (trunFlags & 0x04)
    {
        trunFirstSampleFlags = 
            (pTrunBoxData[trunOffset + 0] << 24) | 
            (pTrunBoxData[trunOffset + 1] << 16) | 
            (pTrunBoxData[trunOffset + 2] << 8) | 
            pTrunBoxData[trunOffset + 3];
        trunOffset += 4;
    }
    else
    {
        trunFirstSampleFlags = trunFlags; 
    }

    pAssetInfo->nSampleInfos = trunSampleCount;
    if (trunSampleCount > numof(pAssetInfo->vSampleInfos))
    {
        ERRORPRINT("%s:%s Too many samples (%d > %lu)\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, trunSampleCount, 
            numof(pAssetInfo->vSampleInfos));
        rval = 1;
        goto out;
    }

    decoding_time = 0;
    for (i = 0; i < trunSampleCount; i++)
    {
        if (trunFlags & 0x100)
        {
            // sample duration
            pAssetInfo->vSampleInfos[i].sample_duration = 
                (pTrunBoxData[trunOffset + 0] << 24) | 
                (pTrunBoxData[trunOffset + 1] << 16) | 
                (pTrunBoxData[trunOffset + 2] << 8) | 
                pTrunBoxData[trunOffset + 3];
            if (pAssetInfo->vSampleInfos[i].sample_duration <= 0 ||
                pAssetInfo->vSampleInfos[i].sample_duration > 
                MAX_SAMPLE_DURATION(pAssetInfo->mdhdTimescale))
            {
                uint32_t sampleDuration = pAssetInfo->mdhdTimescale / 120;
                pAssetInfo->vSampleInfos[i].sample_duration =
                    sampleDuration;
                pTrunBoxData[trunOffset + 0] = (sampleDuration >> 24) & 0xff;
                pTrunBoxData[trunOffset + 1] = (sampleDuration >> 16) & 0xff;
                pTrunBoxData[trunOffset + 2] = (sampleDuration >> 8) & 0xff;
                pTrunBoxData[trunOffset + 3] = sampleDuration & 0xff;
            }
            trunOffset += 4;
        }
        else if (tfhdFlags & 0x08)
        {
            pAssetInfo->vSampleInfos[i].sample_duration = 
                tfhdDefaultSampleDuration;
        }
        else
        {
            pAssetInfo->vSampleInfos[i].sample_duration = 1;
            ERRORPRINT("%s:%s No sample duration!\n",
                pSession->strDebugName, pAssetInfo->strAssetID);
        }

        mpuDuration += pAssetInfo->vSampleInfos[i].sample_duration;

        if (trunFlags & 0x200)
        {
            // sample size
            pAssetInfo->vSampleInfos[i].sample_size = 
                (pTrunBoxData[trunOffset + 0] << 24) | 
                (pTrunBoxData[trunOffset + 1] << 16) | 
                (pTrunBoxData[trunOffset + 2] << 8) | 
                pTrunBoxData[trunOffset + 3];
             trunOffset += 4;
        }
        else if (tfhdFlags & 0x10)
        {
            pAssetInfo->vSampleInfos[i].sample_size = tfhdDefaultSampleSize;
        }
        else
        {
            LOGPRINT("%s:%s Couldn't determine sample size\n",
                pSession->strDebugName, pAssetInfo->strAssetID);
            rval = 1;
            goto out;
        }

        if (trunFlags & 0x400)
        {
            // sample flags
            trunOffset += 4;
        }

        if (trunFlags & 0x800)
        {
            // sample composition time offset
            if (trunVersion == 0)
            {
                pAssetInfo->vSampleInfos[i].sample_composition_time_offset = 
                    (pTrunBoxData[trunOffset + 0] << 24) | 
                    (pTrunBoxData[trunOffset + 1] << 16) | 
                    (pTrunBoxData[trunOffset + 2] << 8) | 
                    pTrunBoxData[trunOffset + 3];
            }
            else if (trunVersion == 1)
            {
                pAssetInfo->vSampleInfos[i].sample_composition_time_offset = 
                    (int32_t)((pTrunBoxData[trunOffset + 0] << 24) | 
                              (pTrunBoxData[trunOffset + 1] << 16) | 
                              (pTrunBoxData[trunOffset + 2] << 8) | 
                              pTrunBoxData[trunOffset + 3]);
            }
            trunOffset += 4;
        }
        else
        {
            pAssetInfo->vSampleInfos[i].sample_composition_time_offset = 0;
        }
        pAssetInfo->vSampleInfos[i].frag_composition_time = decoding_time + 
            pAssetInfo->vSampleInfos[i].sample_composition_time_offset;
        decoding_time += pAssetInfo->vSampleInfos[i].sample_duration;

        if (i == 0 || 
            pAssetInfo->vSampleInfos[i].frag_composition_time <
            pAssetInfo->min_frag_composition_time)
        {
            pAssetInfo->min_frag_composition_time = 
                pAssetInfo->vSampleInfos[i].frag_composition_time;
        }
    }

    pAssetInfo->mpuDuration = mpuDuration;

    if (fHavePresentationTime)
    {
        uint64_t ticks;

        // Convert the presentation time from NTP to UTC, then scale to
        // mdhdTimescale. NTP time is a fixed point value with the upper 32
        // bits representing seconds and lower 32 bits representing 
        // fractional seconds. 
        ticks = NTP_TIME_TO_UTC(presentationTime >> 32) * 
            pAssetInfo->mdhdTimescale;
        ticks += 
            ((presentationTime & 0xffffffff) * pAssetInfo->mdhdTimescale) >> 32;

        int32_t fInvalidPresentationTime = 0;
        if (pAssetInfo->lastGoodMPUPresentationTimePacketTime.tv_sec)
        {
            uint64_t expectedPresentationTime;

#ifdef DEBUG_ISOBMFF
            LOGPRINT_NOFILE("%s:%s mdhdTimescale %u\n", 
                pSession->strDebugName, pAssetInfo->strAssetID,
                pAssetInfo->mdhdTimescale);
#endif

            expectedPresentationTime = (pPacketTime->tv_sec - 
                pAssetInfo->lastGoodMPUPresentationTimePacketTime.tv_sec) * 
                pAssetInfo->mdhdTimescale;
            expectedPresentationTime += (pPacketTime->tv_nsec - 
                pAssetInfo->lastGoodMPUPresentationTimePacketTime.tv_nsec) *
                pAssetInfo->mdhdTimescale /  NANOSECONDS_PER_SECOND;
            expectedPresentationTime += pAssetInfo->lastGoodMPUPresentationTime;

            if (ticks < pAssetInfo->lastGoodMPUPresentationTime)
            {
                ERRORPRINT("%s:%s presentation time went backwards? %"PRIu64" < %"PRIu64"\n",
                    pSession->strDebugName, pAssetInfo->strAssetID, ticks, 
                    pAssetInfo->lastGoodMPUPresentationTime);
                // mpu presentation time went backwards?
                fInvalidPresentationTime = 1;
            }
            else
            {
                uint64_t discrepancyMilliseconds;

                if (expectedPresentationTime > ticks)
                {
                    discrepancyMilliseconds = expectedPresentationTime - ticks;
                }
                else
                {
                    discrepancyMilliseconds = ticks - expectedPresentationTime;
                }
                discrepancyMilliseconds = discrepancyMilliseconds * 
                    MILLISECONDS_PER_SECOND / pAssetInfo->mdhdTimescale;
                if (discrepancyMilliseconds > MILLISECONDS_PER_SECOND)
                {
                    ERRORPRINT("%s:%s LARGE DISCREPANCY %"PRIu64" ms\n", 
                        pSession->strDebugName, pAssetInfo->strAssetID,
                        discrepancyMilliseconds);
                    fInvalidPresentationTime = 1;
                }
            }
        }

        if (fInvalidPresentationTime)
        {
            uint8_t *pTfdtBox;

            // Nuke the tfdt box, which should cause this fragemnt to play 
            // contiguously with the preceding one. 
            pTfdtBox = pTfdtBoxData - tfdtBoxHeaderSize;
            pTfdtBox[4] = 's';
            pTfdtBox[5] = 'k';
            pTfdtBox[6] = 'i';
            pTfdtBox[7] = 'p';
        }
        else
        {
            // Try to validate the version based on the size
            if (pTfdtBoxData[0] != 0 && pTfdtBoxData[0] != 1)
            {
                if (tfdtBoxSize == 20)
                {
                    LOGPRINT_NOFILE("Rewriting tfdt version %d -> 1\n", 
                        pTfdtBoxData[0]);
                    pTfdtBoxData[0] = 1;
                } 
                else if (tfdtBoxSize == 16)
                {
                    LOGPRINT_NOFILE("Rewriting tfdt version %d -> 0\n", 
                        pTfdtBoxData[0]);
                    pTfdtBoxData[0] = 0;
                }
                else
                {
                    LOGPRINT_NOFILE("Unable to determine tfdt box version (%d) size %d\n", pTfdtBoxData[0], tfdtBoxSize);
                    rval = 1;
                    goto out;
                }
            }

#if defined(DEBUG_ISOBMFF) || defined(DEBUG_SHADOWFILE)
            LOGPRINT_NOFILE("%s:%s baseMediaDecodeTime %"PRIu64"\n", 
                pSession->strDebugName, pAssetInfo->strAssetID, ticks);
#endif
            if (pTfdtBoxData[0] == 1)
            {
                pTfdtBoxData[4] = (ticks >> 56) & 0xff; 
                pTfdtBoxData[5] = (ticks >> 48) & 0xff; 
                pTfdtBoxData[6] = (ticks >> 40) & 0xff; 
                pTfdtBoxData[7] = (ticks >> 32) & 0xff; 
                pTfdtBoxData[8] = (ticks >> 24) & 0xff; 
                pTfdtBoxData[9] = (ticks >> 16) & 0xff; 
                pTfdtBoxData[10] = (ticks >> 8) & 0xff; 
                pTfdtBoxData[11] = ticks & 0xff; 
            }
            else
            {
                pTfdtBoxData[4] = (ticks >> 24) & 0xff; 
                pTfdtBoxData[5] = (ticks >> 16) & 0xff; 
                pTfdtBoxData[6] = (ticks >> 8) & 0xff; 
                pTfdtBoxData[7] = ticks & 0xff; 
            }

            pAssetInfo->mpuPresentationTime = ticks;
            pAssetInfo->lastGoodMPUPresentationTime = ticks;
            pAssetInfo->lastGoodMPUPresentationTimePacketTime = *pPacketTime;
        }
    }
    else
    {
        pAssetInfo->mpuPresentationTime = 0;
    }

    // Find the mdat box.
    if (FindBox(pData, sData, BOX_TYPE_MDAT, &cursor, &mdatBoxHeaderSize, 
        &mdatBoxSize))
    {
        LOGPRINT("%s:%s Failed to find mdat box\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = 1;
        goto out;
    }

    pAssetInfo->nExtraMovieFragmentMetadataBytes = sData - cursor - 8;
    if (pAssetInfo->nExtraMovieFragmentMetadataBytes)
    {
        LOGPRINT("%s:%s movie fragment metadata MPU (%d != %d)\n",
            pSession->strDebugName, pAssetInfo->strAssetID, cursor + 8, sData);
    }

    if (mdatBoxSize > MAX_MDAT_BOX_SIZE_ACCEPTED)
    {
        LOGPRINT("%s:%s mdat box too large (%"PRIu64" > %d)\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, mdatBoxSize, 
            MAX_MDAT_BOX_SIZE_ACCEPTED);
        rval = 1;
        goto out;
    }

    pAssetInfo->mdatBoxSize = mdatBoxSize;
    pAssetInfo->mdatBoxHeaderSize = mdatBoxHeaderSize;
#ifdef DEBUG_ISOBMFF
    LOGPRINT_NOFILE("%s:%s MDAT BOX SIZE %"PRIu64" at %d of %d\n", 
        pSession->strDebugName, pAssetInfo->strAssetID, mdatBoxSize,
        cursor, sData);
#endif

out:
    return rval;
}

static inline void
ResetSampleData(MMTPAssetInfo *pAssetInfo)
{
    pAssetInfo->expectedSampleOffset = 0;
    pAssetInfo->mfuHeaderLength = 0;
}

static inline void
ResetFragmentData(MMTPAssetInfo *pAssetInfo)
{
    pAssetInfo->nMediaSamples = 1;
    pAssetInfo->nMediaSampleBufferBytesUsed = 0;
    pAssetInfo->fHaveMPUMetadataDU = 0;
    pAssetInfo->fHaveMovieFragmentMetadataDU = 0;
    pAssetInfo->expectedSampleNumber = 1;
    pAssetInfo->vMediaSamples[0].pData = pAssetInfo->pMediaSampleBuffer;
    pAssetInfo->vMediaSamples[0].nData = 0;
    ResetSampleData(pAssetInfo);
}

int32_t
HandleMFUDU(uint64_t packetNumber, const uint8_t *pPayload,
    int32_t payloadLength, MMTPSession *pSession, MMTPAssetInfo *pAssetInfo,
    MMTPProtocolPacketInfo *pMMTPProtocolPacketInfo, 
    MMTPPayloadInfo *pMMTPPayloadInfo)
{
    int32_t offset = 0;
    int32_t rval = 0;

    // FT = 2 boxes contain mdat media samples.
    // The MMTP payload data
    uint32_t movie_fragment_sequence_number = 
        (pPayload[offset + 0] << 24) | 
        (pPayload[offset + 1] << 16) |
        (pPayload[offset + 2] << 8) | 
        pPayload[offset + 3];
    uint32_t movie_fragment_sample_number = 
        (pPayload[offset + 4] << 24) | 
        (pPayload[offset + 5] << 16) |
        (pPayload[offset + 6] << 8) | 
        pPayload[offset + 7];
    uint32_t movie_fragment_sample_offset =
        (pPayload[offset + 8] << 24) | 
        (pPayload[offset + 9] << 16) |
        (pPayload[offset + 10] << 8) | 
        pPayload[offset + 11];
    uint32_t priority = pPayload[offset + 12];
    uint32_t dependency_counter = pPayload[offset + 13];

    if (pSession->verbosity >= ENVerbosity_ObjectTrace)
    {
        if (pMMTPProtocolPacketInfo->packet_id)
        {
            LOGPRINT("%s:%s %"PRIu64"  movie_fragment_seqno %d movie_fragment_sampleno %d mdat_offset %d priority %d depcount %d\n",
                pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
                movie_fragment_sequence_number, 
                movie_fragment_sample_number, 
                movie_fragment_sample_offset, priority, 
                dependency_counter);
        }
    }

    offset += 14;

    // Check if this mmtp packet belongs to a previous sample. This
    // could occur due to bitstream errors or because we deliberately
    // dropped a sample.
    if (movie_fragment_sample_number < pAssetInfo->expectedSampleNumber)
    {
        // Just drop the packet.
        if (pSession->verbosity >= ENVerbosity_Terse)
        {
            LOGPRINT("%s:%s %"PRIu64" %d DROP packet (sample number %d < %d)\n",
                pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
                pMMTPProtocolPacketInfo->packet_id, 
                movie_fragment_sample_number, 
                pAssetInfo->expectedSampleNumber);
        }
        rval = 1;
        goto out;
    }

    // Check for missing samples.
    if (movie_fragment_sample_number != pAssetInfo->expectedSampleNumber)
    {
        // Because we may or may not have the moof metadata, we cannot rely
        // on it to determine when we have completely received a sample. 
        // Instead we will check to see if the sample offset has gone to zero, 
        // which would signal we are starting a new mfu.
        if (movie_fragment_sample_offset == 0 &&
            movie_fragment_sample_number != 
            pAssetInfo->expectedSampleNumber + 1)
        {
            if (pSession->verbosity >= ENVerbosity_Terse)
            {
                LOGPRINT_NOFILE("%s:%s %"PRIu64" sample number discrepancy (%d != %u, padding)\n",
                    pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
                    movie_fragment_sample_number, 
                    pAssetInfo->expectedSampleNumber);
            }
        }

        // Terminate the current sample
        if (pAssetInfo->nMediaSamples == numof(pAssetInfo->vMediaSamples))
        {
            ERRORPRINT("%s:%s Too any media samples\n", pSession->strDebugName,
                pAssetInfo->strAssetID);
            rval = 1;
            goto out;
        }
        pAssetInfo->nMediaSamples++;
        pAssetInfo->vMediaSamples[pAssetInfo->nMediaSamples - 1].pData = 
            pAssetInfo->pMediaSampleBuffer + 
            pAssetInfo->nMediaSampleBufferBytesUsed;
        pAssetInfo->vMediaSamples[pAssetInfo->nMediaSamples - 1].nData = 0;
        ResetSampleData(pAssetInfo);
        pAssetInfo->expectedSampleNumber++;

        // Now handle any missing samples
        while (movie_fragment_sample_number != 
            pAssetInfo->expectedSampleNumber)
        {
            if (pAssetInfo->nMediaSamples == numof(pAssetInfo->vMediaSamples))
            {
                ERRORPRINT("%s:%s Too any media samples\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID);
                rval = 1;
                goto out;
            }
            pAssetInfo->nMediaSamples++;
            pAssetInfo->vMediaSamples[pAssetInfo->nMediaSamples - 1].pData = 
                pAssetInfo->pMediaSampleBuffer + 
                pAssetInfo->nMediaSampleBufferBytesUsed;
            pAssetInfo->vMediaSamples[pAssetInfo->nMediaSamples - 1].nData = 0;
            pAssetInfo->expectedSampleNumber++;
        }
    }

    // Check for missing MFUs in a sample. Pad the mdat box as
    // appropriate.
    if (movie_fragment_sample_offset > 
        pAssetInfo->expectedSampleOffset + pAssetInfo->mfuHeaderLength)
    {
        uint32_t nPaddingBytes;

        if (pAssetInfo->expectedSampleOffset == 0)
        {
            LOGPRINT("%s:%s %"PRIu64" DROP packet - first packet in sample missing\n",
                pSession->strDebugName, pAssetInfo->strAssetID, packetNumber);

            // We lost one or more mmtp packets at the beginning of the 
            // sample. If we can figure out mfuHeaderLength we could 
            // pad the mdat with only the amount lost. It's 
            // questionable whether a decoder could do anything with
            // the sample though. Since it's not clear how to 
            // determine mfuHeaderLength we're just going to set the
            // entire sample as padding and drop the packet.
            if (pAssetInfo->nMediaSamples == numof(pAssetInfo->vMediaSamples))
            {
                ERRORPRINT("%s:%s Too any media samples\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID);
                rval = 1;
                goto out;
            }
            pAssetInfo->nMediaSamples++;
            pAssetInfo->vMediaSamples[pAssetInfo->nMediaSamples - 1].pData = 
                pAssetInfo->pMediaSampleBuffer + 
                pAssetInfo->nMediaSampleBufferBytesUsed;
            pAssetInfo->vMediaSamples[pAssetInfo->nMediaSamples - 1].nData = 0;
            ResetSampleData(pAssetInfo);
            pAssetInfo->expectedSampleNumber++;
#ifdef DEBUG_ISOBMFF
            LOGPRINT_NOFILE("%s:%s expectedSampleNumber %u\n", 
                pSession->strDebugName, pAssetInfo->strAssetID,
                pAssetInfo->expectedSampleNumber);
#endif
            rval = 1;
            goto out;
        }

        nPaddingBytes = movie_fragment_sample_offset - 
            pAssetInfo->expectedSampleOffset - 
            pAssetInfo->mfuHeaderLength;
        if (pSession->verbosity >= ENVerbosity_Terse)
        {
            LOGPRINT_NOFILE("%s:%s %"PRIu64" Missing sample data (%d < %d): padding packet_id %d\n",
                pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
                pAssetInfo->expectedSampleOffset + 
                pAssetInfo->mfuHeaderLength, movie_fragment_sample_offset,
                pMMTPProtocolPacketInfo->packet_id);
        }
        if (pAssetInfo->nMediaSampleBufferBytesUsed + nPaddingBytes > 
            pAssetInfo->sMediaSampleBuffer)
        {
            ERRORPRINT("%s:%s Media sample buffer full (%d + %d > %d)\n", 
                pSession->strDebugName, pAssetInfo->strAssetID,
                pAssetInfo->nMediaSampleBufferBytesUsed, nPaddingBytes,
                pAssetInfo->sMediaSampleBuffer);
            rval = 1;
            goto out;
        }
        memset(pAssetInfo->vMediaSamples[pAssetInfo->nMediaSamples - 1].pData + 
            pAssetInfo->vMediaSamples[pAssetInfo->nMediaSamples - 1].nData, 0, 
            nPaddingBytes);
        pAssetInfo->nMediaSampleBufferBytesUsed += nPaddingBytes;
        pAssetInfo->vMediaSamples[pAssetInfo->nMediaSamples - 1].nData += 
            nPaddingBytes;
        pAssetInfo->expectedSampleOffset +=
            movie_fragment_sample_offset - 
            pAssetInfo->expectedSampleOffset;
    }
    else if (movie_fragment_sample_offset < 
        pAssetInfo->expectedSampleOffset)
    {
        LOGPRINT("%s:%s %"PRIu64" Offset error: cursor ahead of write position (%d > %d) packet_id %d\n",
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
            pAssetInfo->expectedSampleOffset, movie_fragment_sample_offset, 
            pMMTPProtocolPacketInfo->packet_id);
    }
    
    // MFU header, first packet or complete MPU
    if (pMMTPPayloadInfo->fragmentation_indicator == 1 || 
        pMMTPPayloadInfo->fragmentation_indicator == 0)
    {
        // MMTH Sample. Contents differs depending on whether the 
        // mmtp payload contains timed data.

        // NOTE: The MMTH Sample is included in the mdat box size,
        // even though the mdat offsets assume it is *not* present
        // in the mdat box. To recreate the original mdat box
        // (without the MMTP hint track samples), eliminate
        // the mmtp samples and pad the box with zeros. That way
        // the media sample references into the mdat box will
        // be correct, and the mdat box will still have the
        // correct size.
        uint32_t sample_sequence_number;

        if (offset + 4 >= payloadLength)
        {
            ERRORPRINT("%s:%s Offset beyond packet\n", pSession->strDebugName,
                pAssetInfo->strAssetID);
            rval = 1;
            goto out;
        }

        sample_sequence_number = 
            (pPayload[offset + 0] << 24) | 
            (pPayload[offset + 1] << 16) |
            (pPayload[offset + 2] << 8) | 
            pPayload[offset + 3];

        offset += 4;
        if (pMMTPPayloadInfo->timed_flag)
        {
            uint32_t trackrefindex;
            uint32_t movie_fragment_sequence_number;
            uint32_t sample_number;
            uint32_t priority;
            uint32_t dependency_counter;
            uint32_t mdat_offset;
            uint32_t length;

            if (offset + 18 >= payloadLength)
            {
                ERRORPRINT("%s:%s Offset beyond packet\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID);
                rval = 1;
                goto out;
            }

            trackrefindex = pPayload[offset+ 0];
            movie_fragment_sequence_number = 
                (pPayload[offset + 1] << 24) | 
                (pPayload[offset + 2] << 16) |
                (pPayload[offset + 3] << 8) | 
                pPayload[offset + 4];
            sample_number = (pPayload[offset + 5] << 24) | 
                (pPayload[offset + 6] << 16) |
                (pPayload[offset + 7] << 8) | 
                pPayload[offset + 8];
            priority = pPayload[offset + 9];
            dependency_counter = pPayload[offset + 10];
            mdat_offset = (pPayload[offset + 11] << 24) | 
                (pPayload[offset + 12] << 16) |
                (pPayload[offset + 13] << 8) | 
                pPayload[offset + 14];
            length = (pPayload[offset + 15] << 24) | 
                (pPayload[offset + 16] << 16) |
                (pPayload[offset + 17] << 8) | 
                pPayload[offset + 18];

            if (pSession->verbosity >= ENVerbosity_ObjectTrace)
            {
                LOGPRINT("%s:%s %"PRIu64"  sampleseqno %d trackrefidx %d mov_frag_seqno %d sampleno %d priority %d depcount %d mdat offset %d len %d\n",
                    pSession->strDebugName, pAssetInfo->strAssetID, 
                    packetNumber, sample_sequence_number, trackrefindex, 
                    movie_fragment_sequence_number, 
                    sample_number, priority, dependency_counter, 
                    mdat_offset, length);
            }

            offset += 19;
            pAssetInfo->mfuHeaderLength = 23;

            if (offset + 3 >= payloadLength)
            {
                ERRORPRINT("%s:%s Offset beyond packet\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID);
                rval = 1;
                goto out;
            }

            // multilayer info
            uint32_t multilayerBoxSize = 
                (pPayload[offset + 0] << 24) | 
                (pPayload[offset + 1] << 16) |
                (pPayload[offset + 2] << 8) | 
                pPayload[offset + 3]; 

            if (multilayerBoxSize > payloadLength - offset)
            {
                multilayerBoxSize = payloadLength - offset;
            }

            offset += multilayerBoxSize;
            pAssetInfo->mfuHeaderLength += multilayerBoxSize;

            // XXX If an MFU has been fragmented into multiple packets then the 
            // MFU header is included in the media sample size, otherwise it 
            // isn't.
            if (pMMTPPayloadInfo->fragmentation_counter == 0)
            {
                pAssetInfo->mfuHeaderLength = 0;
            }
        }
        else
        {
            LOGPRINT("%s:%s Ignoring packet with timed flag set\n",
                pSession->strDebugName, pAssetInfo->strAssetID);
            rval = 1;
            goto out;
        }
    }

    // Update the expected mdat offset based on the information
    // from this packet. 
    pAssetInfo->expectedSampleOffset = movie_fragment_sample_offset + 
        payloadLength - offset;

    if (pAssetInfo->nMediaSampleBufferBytesUsed + payloadLength - offset > 
        pAssetInfo->sMediaSampleBuffer)
    {
        ERRORPRINT("%s:%s Media sample buffer full\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = 1;
        goto out;
    }

    if (offset > payloadLength)
    {
        ERRORPRINT("%s:%s Offset beyond packet\n", pSession->strDebugName,
            pAssetInfo->strAssetID);
        rval = 1;
        goto out;
    }

    memcpy(pAssetInfo->vMediaSamples[pAssetInfo->nMediaSamples - 1].pData + 
        pAssetInfo->vMediaSamples[pAssetInfo->nMediaSamples - 1].nData, 
        pPayload + offset, payloadLength - offset);
    pAssetInfo->nMediaSampleBufferBytesUsed += payloadLength - offset;
    pAssetInfo->vMediaSamples[pAssetInfo->nMediaSamples - 1].nData += 
        payloadLength - offset;

out:
    return rval;
}

int32_t
ProcessUserServiceDescriptionSignalingMessage(struct timespec *pPacketTime, 
    uint64_t packetNumber, const uint8_t *pPayload, int32_t payloadLength, 
    MMTPSession *pSession, MMTPAssetInfo *pAssetInfo, 
    MMTPProtocolPacketInfo *pMMTPProtocolPacketInfo, 
    ENATSC3MMTPATSC3SignalingInfo *pMMTPATSC3SignalingInfo)
{
    uint32_t usbdCrc32;
    int32_t rval = 0;

    usbdCrc32 = ENUtils_CalcCRC32(pPayload, payloadLength);

    if (pAssetInfo->usbdVersion == pMMTPATSC3SignalingInfo->content_version)
    {
        if (usbdCrc32 == pAssetInfo->usbdCrc32)
        {
            // Nothing has changed
            goto out;
        }
        else
        {
            LOGPRINT("%s:%s %"PRIu64" usbd crc mismatch\n",
                pSession->strDebugName, pAssetInfo->strAssetID, packetNumber);
        }
    }

    if (payloadLength < pMMTPATSC3SignalingInfo->content_length)
    {
        LOGPRINT("%s:%s %"PRIu64" USBD truncated (len %d payloadLen %d)\n",
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
            pMMTPATSC3SignalingInfo->content_length, payloadLength);
        rval = 1;
        goto out;
    }

    if ((((pAssetInfo->usbdVersion + 1) & 0xff) != 
        pMMTPATSC3SignalingInfo->content_version) && 
        pAssetInfo->usbdVersion != -1)
    {
        LOGPRINT("%s:%s %"PRIu64" Unexpected usbd signaling message version (%d last %d)\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
            pMMTPATSC3SignalingInfo->version, pAssetInfo->usbdVersion);
    }
    pAssetInfo->usbdVersion = pMMTPATSC3SignalingInfo->content_version;
    pAssetInfo->usbdCrc32 = usbdCrc32;

    if (pSession->USBDAvailableCallback)
    {
        if (pMMTPATSC3SignalingInfo->content_compression == 1)
        {
            // No compression
            pSession->USBDAvailableCallback(pPacketTime, packetNumber, 
                pMMTPATSC3SignalingInfo, (char *)pPayload, 
                pMMTPATSC3SignalingInfo->content_length, 
                pAssetInfo->usbdVersion, pSession->pCallbackUserData);
        }
        else if (pMMTPATSC3SignalingInfo->content_compression == 2)
        {
            char strXML[16384];
            int32_t xmlLength;

            // gzip
            EnGunzipUnpipe((uint8_t *)pPayload, 
                pMMTPATSC3SignalingInfo->content_length, 
                (unsigned char *)strXML, sizeof(strXML), &xmlLength);
            pSession->USBDAvailableCallback(pPacketTime,  packetNumber,
                pMMTPATSC3SignalingInfo, strXML, xmlLength,
                pAssetInfo->usbdVersion, pSession->pCallbackUserData);
        }
        else
        {
            ERRORPRINT("%s:%s Unhandled usbd compression %d\n", 
                pSession->strDebugName, pAssetInfo->strAssetID,
                pMMTPATSC3SignalingInfo->content_compression);
        }
    }

out:
    return rval;
}

int32_t
ProcessHELDSignalingMessage(struct timespec *pPacketTime, 
    uint64_t packetNumber, const uint8_t *pPayload, int32_t payloadLength, 
    MMTPSession *pSession, MMTPAssetInfo *pAssetInfo, 
    MMTPProtocolPacketInfo *pMMTPProtocolPacketInfo, 
    ENATSC3MMTPATSC3SignalingInfo *pMMTPATSC3SignalingInfo)
{
    int32_t rval = 0;

    if (pAssetInfo->heldVersion == pMMTPATSC3SignalingInfo->content_version)
    {
        // Nothing has changed
        goto out;
    }

    if (payloadLength < pMMTPATSC3SignalingInfo->content_length)
    {
        LOGPRINT("%s:%s %"PRIu64" HELD truncated (len %d payloadLen %d)\n",
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
            pMMTPATSC3SignalingInfo->content_length, payloadLength);
        rval = 1;
        goto out;
    }

    if ((((pAssetInfo->heldVersion + 1) & 0xff) != 
        pMMTPATSC3SignalingInfo->content_version) && 
        pAssetInfo->heldVersion != -1)
    {
        LOGPRINT("%s:%s %"PRIu64" Unexpected held signaling message version (%d last %d)\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
            pMMTPATSC3SignalingInfo->version, pAssetInfo->heldVersion);
    }
    pAssetInfo->heldVersion = pMMTPATSC3SignalingInfo->content_version;

    if (pSession->HELDAvailableCallback)
    {
        if (pMMTPATSC3SignalingInfo->content_compression == 1)
        {
            // No compression
            pSession->HELDAvailableCallback(pPacketTime,  packetNumber,
                pMMTPATSC3SignalingInfo, (char *)pPayload, 
                pMMTPATSC3SignalingInfo->content_length, 
                pAssetInfo->heldVersion, pSession->pCallbackUserData);
        }
        else if (pMMTPATSC3SignalingInfo->content_compression == 2)
        {
            char strXML[16384];
            int32_t xmlLength;

            // gzip
            EnGunzipUnpipe((uint8_t *)pPayload, 
                pMMTPATSC3SignalingInfo->content_length, 
                (unsigned char *)strXML, sizeof(strXML), &xmlLength);
            pSession->HELDAvailableCallback(pPacketTime, packetNumber, 
                pMMTPATSC3SignalingInfo, strXML, xmlLength,
                pAssetInfo->heldVersion, pSession->pCallbackUserData);
        }
        else
        {
            ERRORPRINT("%s:%s Unhandled held compression %d\n", 
                pSession->strDebugName, pAssetInfo->strAssetID, 
                pMMTPATSC3SignalingInfo->content_compression);
        }
    }

out:
    return rval;
}

int32_t
ProcessVideoStreamPropertiesDescriptorMessage(struct timespec *pPacketTime, 
    uint64_t packetNumber, const uint8_t *pPayload, int32_t payloadLength, 
    MMTPSession *pSession, MMTPAssetInfo *pAssetInfo, 
    MMTPProtocolPacketInfo *pMMTPProtocolPacketInfo, 
    ENATSC3MMTPATSC3SignalingInfo *pMMTPATSC3SignalingInfo)
{
    int32_t rval = 0;

    if (pAssetInfo->vspdVersion == pMMTPATSC3SignalingInfo->content_version)
    {
        // Nothing has changed
        goto out;
    }

    if (payloadLength < pMMTPATSC3SignalingInfo->content_length)
    {
        LOGPRINT("%s:%s %"PRIu64" VSPD truncated (len %d payloadLen %d)\n",
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
            pMMTPATSC3SignalingInfo->content_length, payloadLength);
        rval = 1;
        goto out;
    }

    if ((((pAssetInfo->vspdVersion + 1) & 0xff) != 
        pMMTPATSC3SignalingInfo->content_version) && 
        pAssetInfo->vspdVersion != -1)
    {
        LOGPRINT("%s:%s %"PRIu64" Unexpected vspd signaling message version (%d last %d)\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
            pMMTPATSC3SignalingInfo->version, pAssetInfo->vspdVersion);
    }
    pAssetInfo->vspdVersion = pMMTPATSC3SignalingInfo->content_version;

    if (pSession->VSPDAvailableCallback)
    {
        if (pMMTPATSC3SignalingInfo->content_compression == 1)
        {
            // No compression
            pSession->VSPDAvailableCallback(pPacketTime, packetNumber,
                pMMTPATSC3SignalingInfo, pPayload, 
                pMMTPATSC3SignalingInfo->content_length, 
                pAssetInfo->vspdVersion, pSession->pCallbackUserData);
        }
        else
        {
            ERRORPRINT("%s:%s Unhandled vspd compression %d\n", 
                pSession->strDebugName, pAssetInfo->strAssetID,  
                pMMTPATSC3SignalingInfo->content_compression);
        }
    }

out:
    return rval;
}

int32_t
ProcessCaptionAssetDescriptorMessage(struct timespec *pPacketTime, 
    uint64_t packetNumber, const uint8_t *pPayload, int32_t payloadLength, 
    MMTPSession *pSession, MMTPAssetInfo *pAssetInfo, 
    MMTPProtocolPacketInfo *pMMTPProtocolPacketInfo, 
    ENATSC3MMTPATSC3SignalingInfo *pMMTPATSC3SignalingInfo)
{
    int32_t rval = 0;

    if (pAssetInfo->cadVersion == pMMTPATSC3SignalingInfo->content_version)
    {
        // Nothing has changed
        goto out;
    }

    if (payloadLength < pMMTPATSC3SignalingInfo->content_length)
    {
        LOGPRINT("%s:%s %"PRIu64" Caption asset descriptor truncated (len %d payloadLen %d)\n",
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
            pMMTPATSC3SignalingInfo->content_length, payloadLength);
        rval = 1;
        goto out;
    }

    if ((((pAssetInfo->cadVersion + 1) & 0xff) != 
        pMMTPATSC3SignalingInfo->content_version) && 
        pAssetInfo->cadVersion != -1)
    {
        LOGPRINT("%s:%s %"PRIu64" Unexpected caption asset descriptor signaling message version (%d last %d)\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
            pMMTPATSC3SignalingInfo->version, pAssetInfo->cadVersion);
    }
    pAssetInfo->cadVersion = pMMTPATSC3SignalingInfo->content_version;

    if (pSession->CADAvailableCallback)
    {
        if (pMMTPATSC3SignalingInfo->content_compression == 1)
        {
            // No compression
            pSession->CADAvailableCallback(pPacketTime, packetNumber, 
                pMMTPATSC3SignalingInfo, pPayload, 
                pMMTPATSC3SignalingInfo->content_length, 
                pAssetInfo->cadVersion, pSession->pCallbackUserData);
        }
        else
        {
            ERRORPRINT(
                "%s:%s Unhandled caption asset descriptor  compression %d\n", 
                pSession->strDebugName, pAssetInfo->strAssetID,  
                pMMTPATSC3SignalingInfo->content_compression);
        }
    }

out:
    return rval;
}

int32_t
ProcessAudioStreamPropertiesDescriptorMessage(struct timespec *pPacketTime, 
    uint64_t packetNumber, const uint8_t *pPayload, int32_t payloadLength, 
    MMTPSession *pSession, MMTPAssetInfo *pAssetInfo, 
    MMTPProtocolPacketInfo *pMMTPProtocolPacketInfo, 
    ENATSC3MMTPATSC3SignalingInfo *pMMTPATSC3SignalingInfo)
{
    int32_t rval = 0;

    if (pAssetInfo->aspdVersion == pMMTPATSC3SignalingInfo->content_version)
    {
        // Nothing has changed
        goto out;
    }

    if (payloadLength < pMMTPATSC3SignalingInfo->content_length)
    {
        LOGPRINT("%s:%s %"PRIu64" ASPD truncated (len %d payloadLen %d)\n",
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
            pMMTPATSC3SignalingInfo->content_length, payloadLength);
        rval = 1;
        goto out;
    }

    if ((((pAssetInfo->aspdVersion + 1) & 0xff) != 
        pMMTPATSC3SignalingInfo->content_version) && 
        pAssetInfo->aspdVersion != -1)
    {
        LOGPRINT("%s:%s %"PRIu64" Unexpected aspd signaling message version (%d last %d)\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
            pMMTPATSC3SignalingInfo->version, pAssetInfo->aspdVersion);
    }
    pAssetInfo->aspdVersion = pMMTPATSC3SignalingInfo->content_version;

    if (pSession->ASPDAvailableCallback)
    {
        if (pMMTPATSC3SignalingInfo->content_compression == 1)
        {
            // No compression
            pSession->ASPDAvailableCallback(pPacketTime, packetNumber, 
                pMMTPATSC3SignalingInfo, pPayload, 
                pMMTPATSC3SignalingInfo->content_length, 
                pAssetInfo->aspdVersion, pSession->pCallbackUserData);
        }
        else
        {
            ERRORPRINT("%s:%s Unhandled aspd compression %d\n", 
                pSession->strDebugName, pAssetInfo->strAssetID,  
                pMMTPATSC3SignalingInfo->content_compression);
        }
    }

out:
    return rval;
}

int32_t
ProcessIdentifierMapping(uint64_t packetNumber, const uint8_t *pPayload, 
    int32_t payloadLength, MMTPSession *pSession, MMTPAssetInfo *pAssetInfo, 
    MMTPProtocolPacketInfo *pMMTPProtocolPacketInfo, 
    ENATSC3MMTPIdentifier_mapping *pIdentifierMapping)
{
    int32_t i;
    int32_t offset = 0;

    if (payloadLength < 1)
    {
        LOGPRINT("%s:%s %"PRIu64" short Identifier Mapping\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber);
        offset = -1;
        goto out;
    }

    pIdentifierMapping->type = pPayload[0];
    offset = 1;

    switch (pIdentifierMapping->type)
    {
        case 0: // asset id
            if (payloadLength < offset + 8)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                offset = -1;
                goto out;
            }
            pIdentifierMapping->asset.asset_id_scheme = 
                (pPayload[offset + 0] << 24) | (pPayload[offset + 1] << 16) | 
                (pPayload[offset + 2] << 8) | pPayload[offset + 3];
            pIdentifierMapping->asset.asset_id_length = 
                (pPayload[offset + 4] << 24) | (pPayload[offset + 5] << 16) | 
                (pPayload[offset + 6] << 8) | pPayload[offset + 7];
            pIdentifierMapping->asset.pAssetId = pPayload + offset + 8;
            offset += 8 + pIdentifierMapping->asset.asset_id_length;
            break;
        case 1: // list of URLs
            if (payloadLength < offset + 2)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                offset = -1;
                goto out;
            }
            pIdentifierMapping->urlList.url_count = 
                (pPayload[offset] << 8) | pPayload[offset + 1];
            offset += 2;
            pIdentifierMapping->urlList.pRawUrlList = pPayload + offset;
            for (i = 0; i < pIdentifierMapping->urlList.url_count; i++)
            {
                uint16_t length;

                if (payloadLength < offset + 2)
                {
                    LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                        pSession->strDebugName, pAssetInfo->strAssetID, 
                        packetNumber);
                    offset = -1;
                    goto out;
                }
                length = (pPayload[offset] << 8) | pPayload[offset + 1];
                if (i >= numof(pIdentifierMapping->urlList.vUrls))
                {
                    LOGPRINT("%s:%s %"PRIu64" Too many urls\n", 
                        pSession->strDebugName, pAssetInfo->strAssetID,
                        packetNumber);
                }
                else
                {
                    pIdentifierMapping->urlList.vUrls[i].url_length = length;
                    pIdentifierMapping->urlList.vUrls[i].pUrl = 
                        pPayload + offset + 2;
                }

                offset += 2 + length;
            }
            break;
        case 2: // regex
            if (payloadLength < offset + 2)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                offset = -1;
                goto out;
            }
            pIdentifierMapping->regex.regex_length =
                (pPayload[offset] << 8) | pPayload[offset + 1];
            pIdentifierMapping->regex.pRegex = pPayload + offset + 2;
            offset += 2 + pIdentifierMapping->regex.regex_length;
            break;
        case 3: // DASH represention@id
            if (payloadLength < offset + 4)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                offset = -1;
                goto out;
            }
            pIdentifierMapping->dashRepresentationId.representation_id_length =
                (pPayload[1] << 24) | (pPayload[2] << 16) | 
                (pPayload[3] << 8) | pPayload[4];
            pIdentifierMapping->dashRepresentationId.pRepresentationId = 
                pPayload + offset + 4;
            offset += 4 + pIdentifierMapping->dashRepresentationId.representation_id_length;
            break;
    }

    if (payloadLength < offset)
    {
        LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber);
        offset = -1;
        goto out;
    }

out:
    return offset;
}

int32_t
ProcessLocationInfo(uint64_t packetNumber, const uint8_t *pPayload, 
    int32_t payloadLength, MMTPSession *pSession, MMTPAssetInfo *pAssetInfo, 
    MMTPProtocolPacketInfo *pMMTPProtocolPacketInfo, 
    ENATSC3MMTPLocation_info *pLocationInfo)
{
    uint8_t url_length;
    uint8_t location_type;
    int32_t offset = 0;

    if (payloadLength < 1)
    {
        LOGPRINT("%s:%s %"PRIu64": short Location info)\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber);
        offset = -1;
        goto out;
    }

    location_type = pPayload[0];
    if (pLocationInfo)
    {
        pLocationInfo->location_type = location_type;
    }
    offset = 1;

    switch (location_type)
    {
        case 0: // packet_id
            if (payloadLength < offset + 2)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID, 
                    packetNumber);
                offset = -1;
                goto out;
            }
            if (pLocationInfo)
            {
                pLocationInfo->packet_id = 
                    (pPayload[offset] << 8) | pPayload[offset + 1];
            }
            offset += 2;
            break;
        case 1: // ipv4 addr + packet id
            if (payloadLength < offset + 12)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                offset = -1;
                goto out;
            }
            if (pLocationInfo)
            {
                memcpy(&pLocationInfo->ipv4_packet_id.ipv4_src_addr, 
                    pPayload + offset, 4); 
                memcpy(&pLocationInfo->ipv4_packet_id.ipv4_dst_addr, 
                    pPayload + offset + 4, 4); 
                pLocationInfo->ipv4_packet_id.dst_port = 
                    (pPayload[offset + 8] << 8) | pPayload[offset + 9];
                pLocationInfo->ipv4_packet_id.packet_id = 
                    (pPayload[offset + 10] << 8) | pPayload[offset + 11];
            }
            offset += 12;
            break;
        case 2: // ipv6 addr + packet id
            if (payloadLength < offset + 36)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                offset = -1;
                goto out;
            }
            if (pLocationInfo)
            {
                memcpy(&pLocationInfo->ipv6_packet_id.ipv6_src_addr, 
                    pPayload + offset, 16); 
                memcpy(&pLocationInfo->ipv6_packet_id.ipv6_dst_addr, 
                    pPayload + offset + 16, 16); 
                pLocationInfo->ipv6_packet_id.dst_port = 
                    (pPayload[offset + 32] << 8) | pPayload[offset + 33];
                pLocationInfo->ipv6_packet_id.packet_id = 
                    (pPayload[offset + 34] << 8) | pPayload[offset + 35];
            }
            offset += 36;
            break;
        case 3: // mpegts
            if (payloadLength < offset + 6)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                offset = -1;
                goto out;
            }
            if (pLocationInfo)
            {
                pLocationInfo->mpegts_pid.network_id =
                    (pPayload[offset] << 8) | pPayload[offset + 1];
                pLocationInfo->mpegts_pid.MPEG_2_transport_stream_id =
                    (pPayload[offset + 2] << 8) | pPayload[offset + 3];
                pLocationInfo->mpegts_pid.MPEG_2_PID =
                    (pPayload[offset + 4] << 8) | pPayload[offset + 5];
            }
            offset += 6;
            break;
        case 4: // ipv6 addr + MPEG 2 PID
            if (payloadLength < offset + 34)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                offset = -1;
                goto out;
            }
            if (pLocationInfo)
            {
                memcpy(&pLocationInfo->ipv6_pid.ipv6_src_addr, 
                    pPayload + offset, 16); 
                memcpy(&pLocationInfo->ipv6_pid.ipv6_dst_addr, 
                    pPayload + offset + 16, 16); 
                pLocationInfo->ipv6_pid.MPEG_2_PID = 
                    (pPayload[offset + 32] << 8) | pPayload[offset + 33];
            }
            offset += 34;
            break;
        case 5: // URL
            if (payloadLength < offset + 1)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                offset = -1;
                goto out;
            }
            url_length = pPayload[offset];
            if (pLocationInfo)
            {
                pLocationInfo->url.url_length = url_length;
                pLocationInfo->url.pUrl = pPayload + 1;
            }
            offset += 1 + url_length;
            break;
        case 6: // URL with offset
            if (payloadLength < offset + 1)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                offset = -1;
                goto out;
            }
            url_length = pPayload[offset];
            if (pLocationInfo)
            {
                pLocationInfo->url_offset.url_length = url_length;
                pLocationInfo->url_offset.pUrl = pPayload + 1;
            }
            offset += 1 + url_length;
            if (payloadLength < offset + 4)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                offset = -1;
                goto out;
            }
            if (pLocationInfo)
            {
                pLocationInfo->url_offset.byte_offset = 
                    (pPayload[offset] << 8) | pPayload[offset + 1];
                pLocationInfo->url_offset.length = 
                    (pPayload[offset + 2] << 8) | pPayload[offset + 3];
            }
            offset += 4;
            break;
        case 7:
            break;
        case 8: // message_id
            if (payloadLength < offset + 1)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                offset = -1;
                goto out;
            }
            if (pLocationInfo)
            {
                pLocationInfo->message_id = pPayload[offset];
            }
            offset += 1;
            break;
        case 9: // packet_id + message_id
            if (payloadLength < offset + 3)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                offset = -1;
                goto out;
            }
            if (pLocationInfo)
            {
                pLocationInfo->packet_id_message_id.packet_id = 
                    (pPayload[offset] << 8) | pPayload[offset + 1];
                pLocationInfo->packet_id_message_id.message_id = 
                    pPayload[offset + 2];
            }
            offset += 3;
            break;
        case 0xA: // ipv4 addr + packet id + message_id
            if (payloadLength < offset + 13)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                offset = -1;
                goto out;
            }
            if (pLocationInfo)
            {
                memcpy(&pLocationInfo->ipv4_packet_id_message_id.ipv4_src_addr, 
                    pPayload + offset, 4); 
                memcpy(&pLocationInfo->ipv4_packet_id_message_id.ipv4_dst_addr, 
                    pPayload + offset + 4, 4); 
                pLocationInfo->ipv4_packet_id_message_id.dst_port = 
                    (pPayload[offset + 8] << 8) | pPayload[offset + 9];
                pLocationInfo->ipv4_packet_id_message_id.packet_id = 
                    (pPayload[offset + 10] << 8) | pPayload[offset + 11];
                pLocationInfo->ipv4_packet_id_message_id.message_id = 
                    pPayload[offset + 12];
            }
            offset += 13;
            break;
        case 0xB: // ipv6 addr + packet id + message_id
            if (payloadLength < offset + 37)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                offset = -1;
                goto out;
            }
            if (pLocationInfo)
            {
                memcpy(&pLocationInfo->ipv6_packet_id_message_id.ipv6_src_addr, 
                    pPayload + offset, 16); 
                memcpy(&pLocationInfo->ipv6_packet_id_message_id.ipv6_dst_addr, 
                    pPayload + offset + 16, 16); 
                pLocationInfo->ipv6_packet_id_message_id.dst_port = 
                    (pPayload[offset + 32] << 8) | pPayload[offset + 33];
                pLocationInfo->ipv6_packet_id_message_id.packet_id = 
                    (pPayload[offset + 34] << 8) | pPayload[offset + 35];
                pLocationInfo->ipv6_packet_id_message_id.message_id = 
                    pPayload[offset + 36];
            }
            offset += 37;
            break;
        case 0xC: // ipv4 addr + mpeg pid
            if (payloadLength < offset + 10)
            {
                LOGPRINT("%s:%s %"PRIu64": short Identifier Mapping\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                offset = -1;
                goto out;
            }
            if (pLocationInfo)
            {
                memcpy(&pLocationInfo->ipv4_pid.ipv4_src_addr, 
                    pPayload + offset, 4); 
                memcpy(&pLocationInfo->ipv4_pid.ipv4_dst_addr, 
                    pPayload + offset + 4, 4); 
                pLocationInfo->ipv4_pid.MPEG_2_PID =
                    (pPayload[offset + 8] << 8) | pPayload[offset + 9];
            }
            offset += 10;
            break;
    }

    if (payloadLength < offset)
    {
        LOGPRINT("%s:%s %"PRIu64": short Location info\n", 
            pSession->strDebugName, pAssetInfo->strAssetID,
            packetNumber);
        offset = -1;
        goto out;
    }

out:
    return offset;
}

int32_t
ProcessMPTSignalingMessageDescriptors(uint64_t packetNumber, 
    const uint8_t *pDescriptorBytes, int32_t nDescriptorBytes, 
    MMTPSession *pSession, MMTPAssetInfo *pAssetInfo, 
    MMTPProtocolPacketInfo *pMMTPProtocolPacketInfo,
    ENATSC3MMTPMPTSignalingInfoAsset *pAsset)
{
    int32_t rval = 0;

    while (nDescriptorBytes >= 3)
    {
        uint32_t descriptorId;
        uint8_t descriptorLength;

        descriptorId = (pDescriptorBytes[0] << 8) | pDescriptorBytes[1];
        descriptorLength = pDescriptorBytes[2];
        pDescriptorBytes += 3;
        nDescriptorBytes -= 3;

        if (descriptorLength > nDescriptorBytes)
        {
            LOGPRINT("%s:%s %"PRIu64" descriptor too long (%d > %d)\n", 
                pSession->strDebugName, pAssetInfo->strAssetID,
                packetNumber, descriptorLength, 
                nDescriptorBytes);
            rval = 1;
            goto out;
        }

        switch (descriptorId)
        {
            case 0x0001: // mpu timestamp descriptor
                if (descriptorLength != 12)
                {
                    LOGPRINT("%s:%s %"PRIu64" Invalid mpu_timestamp_descriptor length %d\n",
                        pSession->strDebugName, pAssetInfo->strAssetID,
                        packetNumber, descriptorLength);
                    rval = 1;
                    goto out;
                }
                pAssetInfo->mpu_timestamp_descriptor.fPresent = 1;
                pAssetInfo->mpu_timestamp_descriptor.mpu_sequence_number = 
                    (pDescriptorBytes[0] << 24) | (pDescriptorBytes[1] << 16) |
                    (pDescriptorBytes[2] << 8) | pDescriptorBytes[3];
                pAssetInfo->mpu_timestamp_descriptor.mpu_presentation_time =
                    ((uint64_t)pDescriptorBytes[4] << 56) | 
                    ((uint64_t)pDescriptorBytes[5] << 48) |
                    ((uint64_t)pDescriptorBytes[6] << 40) | 
                    ((uint64_t)pDescriptorBytes[7] << 32) |
                    ((uint64_t)pDescriptorBytes[8] << 24) | 
                    ((uint64_t)pDescriptorBytes[9] << 16) |
                    ((uint64_t)pDescriptorBytes[10] << 8) | 
                    (uint64_t)pDescriptorBytes[11];

                pDescriptorBytes += descriptorLength;
                nDescriptorBytes -= descriptorLength;

                if (pSession->verbosity >= ENVerbosity_ObjectTrace)
                {
                    LOGPRINT("%s:%s %"PRIu64" presentation time packet_id %d seqno %d %"PRIu64".%.6"PRIu64"\n",
                        pSession->strDebugName, pAssetInfo->strAssetID,
                        packetNumber, 
                        pAsset->vLocations[0].packet_id,
                        pAssetInfo->mpu_timestamp_descriptor.mpu_sequence_number,
                        pAssetInfo->mpu_timestamp_descriptor.mpu_presentation_time >> 32,
                        ((pAssetInfo->mpu_timestamp_descriptor.mpu_presentation_time & 0xffffffff) * 1000000) >> 32); 
                }
                break;
            default:
                LOGPRINT("%s:%s %"PRIu64" Unhandled MPT descriptor 0x%.4x\n",
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber, descriptorId);
                pDescriptorBytes += descriptorLength;
                nDescriptorBytes -= descriptorLength;
                break;
        }
    }

    if (nDescriptorBytes)
    {
        LOGPRINT("%s:%s %"PRIu64" %d extra descriptor bytes\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
            nDescriptorBytes);
    }

out:
    return rval;
}

int32_t
ProcessMPTSignalingMessage(struct timespec *pPacketTime, 
    uint64_t packetNumber, const uint8_t *pPayload, int32_t payloadLength, 
    MMTPSession *pSession, MMTPAssetInfo *pAssetInfo, 
    MMTPProtocolPacketInfo *pMMTPProtocolPacketInfo)
{
    ENATSC3MMTPMPTSignalingInfo mmtpMPTSignalingInfo;
    ENATSC3MMTPMPTSignalingInfoAsset *pAsset;
    uint32_t messageLength;
    uint32_t tableLength;
    int32_t offset;
    int32_t asset;
    int32_t location;
    int32_t rval = 0;

    (void)messageLength;
    (void)tableLength;

    memset(&mmtpMPTSignalingInfo, 0, sizeof(mmtpMPTSignalingInfo));

    mmtpMPTSignalingInfo.payloadCRC32 = 
        ENUtils_CalcCRC32(pPayload, payloadLength);

    if (10 > payloadLength)
    {
        LOGPRINT("%s:%s %"PRIu64": short MPT Signaling Message\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber);
        rval = 1;
        goto out;
    }

    mmtpMPTSignalingInfo.message_id = (pPayload[0] << 8) | pPayload[1];
    mmtpMPTSignalingInfo.message_version = pPayload[2];
    messageLength = (pPayload[3] << 8) | pPayload[4];
    mmtpMPTSignalingInfo.table_id = pPayload[5];
    mmtpMPTSignalingInfo.table_version = pPayload[6];
    tableLength = (pPayload[7] << 8) | pPayload[8];
    mmtpMPTSignalingInfo.MP_table_mode = pPayload[9] & 0x03;
    offset = 10;

    if (mmtpMPTSignalingInfo.table_id == 0x20 || 
        mmtpMPTSignalingInfo.table_id == 0x11)
    {
        if (offset + 1 > payloadLength)
        {
            LOGPRINT("%s:%s %"PRIu64": short MPT Signaling Message\n", 
                pSession->strDebugName, pAssetInfo->strAssetID, packetNumber);
            rval = 1;
            goto out;
        }

        mmtpMPTSignalingInfo.MMTP_package_id_length = pPayload[offset];
        mmtpMPTSignalingInfo.pPackageId = pPayload + offset + 1;
        offset += 1 + mmtpMPTSignalingInfo.MMTP_package_id_length;

        if (offset + 2 > payloadLength)
        {
            LOGPRINT("%s:%s %"PRIu64": short MPT Signaling Message\n", 
                pSession->strDebugName, pAssetInfo->strAssetID, packetNumber);
            rval = 1;
            goto out;
        }

        mmtpMPTSignalingInfo.MP_table_descriptors_length = 
            (pPayload[offset] << 8) | pPayload[offset + 1];
        mmtpMPTSignalingInfo.pTableDescriptors = pPayload + offset + 2;
        offset += 2 + mmtpMPTSignalingInfo.MP_table_descriptors_length;
    }

    if (offset + 1 > payloadLength)
    {
        LOGPRINT("%s:%s %"PRIu64": short MPT Signaling Message\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber);
        rval = 1;
        goto out;
    }

    mmtpMPTSignalingInfo.number_of_assets = pPayload[offset];
    offset++;

    if (mmtpMPTSignalingInfo.number_of_assets > 
        numof(mmtpMPTSignalingInfo.vAssets))
    {
        ERRORPRINT("%s: %"PRIu64" WARN: Too many assets (%d > %lu), message 0x%x table 0x%x offset %d)\n", 
            pSession->strDebugName, packetNumber, 
            mmtpMPTSignalingInfo.number_of_assets, 
            numof(mmtpMPTSignalingInfo.vAssets), mmtpMPTSignalingInfo.message_id,
            mmtpMPTSignalingInfo.table_id, offset);
    }

    for (asset = 0, pAsset = mmtpMPTSignalingInfo.vAssets; 
         asset < mmtpMPTSignalingInfo.number_of_assets && 
         asset < numof(mmtpMPTSignalingInfo.vAssets); 
         asset++, pAsset++)
    {
        int32_t nb;

        nb = ProcessIdentifierMapping(packetNumber, pPayload + offset,
            payloadLength - offset, pSession, pAssetInfo, 
            pMMTPProtocolPacketInfo, &pAsset->Identifier_mapping);
        if (nb < 0)
        {
            rval = -1;
            goto out;
        }

        offset += nb;

        if (offset + 5 > payloadLength)
        {
            LOGPRINT("%s:%s %"PRIu64": short MPT Signaling Message\n", 
                pSession->strDebugName, pAssetInfo->strAssetID, packetNumber);
            rval = 1;
            goto out;
        }

        memcpy(pAsset->asset_type, pPayload + offset, 4);
        pAsset->default_asset_flag = (pPayload[offset + 4] >> 1) & 0x01;
        pAsset->asset_clock_relation_flag = pPayload[offset + 4] & 0x01;
        offset += 5;

        if (pAsset->asset_clock_relation_flag)
        {
            if (offset + 2 > payloadLength)
            {
                LOGPRINT("%s:%s %"PRIu64": short MPT Signaling Message\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    packetNumber);
                rval = 1;
                goto out;
            }

            pAsset->asset_clock_relation_id = pPayload[offset];
            pAsset->asset_timescale_flag = pPayload[offset + 1] & 0x01;
            offset += 2;
            if (pAsset->asset_timescale_flag)
            {
                if (offset + 4 > payloadLength)
                {
                    LOGPRINT("%s:%s %"PRIu64": short MPT Signaling Message\n", 
                        pSession->strDebugName, pAssetInfo->strAssetID,
                        packetNumber);
                    rval = 1;
                    goto out;
                }

                pAsset->asset_timescale = 
                    (pPayload[offset + 0] << 24) | 
                    (pPayload[offset + 1] << 16) | 
                    (pPayload[offset + 2] << 8) | 
                    (pPayload[offset + 3]);
                offset += 4;
            }
        }

        if (offset + 1 > payloadLength)
        {
            LOGPRINT("%s:%s %"PRIu64": short MPT Signaling Message\n", 
                pSession->strDebugName, pAssetInfo->strAssetID, packetNumber);
            rval = 1;
            goto out;
        }

        pAsset->location_count = pPayload[offset];
        offset++;

        for (location = 0; location < pAsset->location_count; location++)
        {
            int32_t nb;

            nb = ProcessLocationInfo(packetNumber, pPayload + offset,
                payloadLength - offset, pSession, pAssetInfo, 
                pMMTPProtocolPacketInfo, 
                location < numof(pAsset->vLocations) ? 
                    pAsset->vLocations + location : 
                    NULL);
            if (nb < 0)
            {
                rval = -1;
                goto out;
            }

            offset += nb;
        }

        pAsset->nLocations = 
            MIN(numof(pAsset->vLocations), pAsset->location_count);

        if (offset + 2 > payloadLength)
        {
            LOGPRINT("%s:%s %"PRIu64": short MPT Signaling Message\n", 
                pSession->strDebugName, pAssetInfo->strAssetID, packetNumber);
            rval = 1;
            goto out;
        }
        pAsset->asset_descriptors_length = 
            (pPayload[offset] << 8) | pPayload[offset + 1];

        if (offset + 2 + pAsset->asset_descriptors_length > payloadLength)
        {
            LOGPRINT("%s:%s %"PRIu64": short MPT Signaling Message\n", 
                pSession->strDebugName, pAssetInfo->strAssetID, packetNumber);
            rval = 1;
            goto out;
        }
        pAsset->pAssetDescriptors = pPayload + offset + 2;
        offset += 2 + pAsset->asset_descriptors_length;

        if (ProcessMPTSignalingMessageDescriptors(packetNumber, 
            pAsset->pAssetDescriptors, pAsset->asset_descriptors_length, 
            pSession, pAssetInfo, pMMTPProtocolPacketInfo, pAsset))
        {
            LOGPRINT("%s:%s %"PRIu64" Failed to process descriptors\n", 
                pSession->strDebugName, pAssetInfo->strAssetID, packetNumber);
            rval = 1;
            goto out;
        }
    }

    mmtpMPTSignalingInfo.nAssets = MIN(numof(mmtpMPTSignalingInfo.vAssets), 
        mmtpMPTSignalingInfo.number_of_assets);

    if (pSession->MPTSignalingMessageCallback)
    {
        pSession->MPTSignalingMessageCallback(pPacketTime, packetNumber, 
            &mmtpMPTSignalingInfo, pSession->pCallbackUserData);
    }

out:
    return rval;
}

int32_t
ProcessHRBMSignalingMessage(struct timespec *pPacketTime, 
    uint64_t packetNumber, const uint8_t *pPayload, int32_t payloadLength, 
    MMTPSession *pSession, MMTPAssetInfo *pAssetInfo, 
    MMTPProtocolPacketInfo *pMMTPProtocolPacketInfo)
{
    int32_t rval = 0;

    return rval;
}

int32_t
ProcessHRBRemovalSignalingMessage(struct timespec *pPacketTime, 
    uint64_t packetNumber, const uint8_t *pPayload, int32_t payloadLength, 
    MMTPSession *pSession, MMTPAssetInfo *pAssetInfo, 
    MMTPProtocolPacketInfo *pMMTPProtocolPacketInfo)
{
    int32_t rval = 0;

    return rval;
}

int32_t
ProcessATSC3SignalingMessage(struct timespec *pPacketTime, 
    uint64_t packetNumber, const uint8_t *pPayload, int32_t payloadLength, 
    MMTPSession *pSession, MMTPAssetInfo *pAssetInfo, 
    MMTPProtocolPacketInfo *pMMTPProtocolPacketInfo)
{
    ENATSC3MMTPATSC3SignalingInfo mmtpATSC3SignalingInfo;
    int32_t rval = 0;

    mmtpATSC3SignalingInfo.packet_id = pAssetInfo->packet_id;
    mmtpATSC3SignalingInfo.pAssetId = pAssetInfo->strAssetID;
    mmtpATSC3SignalingInfo.version = pPayload[2];
    mmtpATSC3SignalingInfo.length = 
        (pPayload[3] << 24) | (pPayload[4] << 16) | 
        (pPayload[5] << 8) | pPayload[6];
    mmtpATSC3SignalingInfo.service_id = (pPayload[7] << 8) | pPayload[8];
    mmtpATSC3SignalingInfo.content_type = (pPayload[9] << 8) | pPayload[10];
    mmtpATSC3SignalingInfo.content_version = pPayload[11];
    mmtpATSC3SignalingInfo.content_compression = pPayload[12];
    mmtpATSC3SignalingInfo.uri_length = pPayload[13];
    mmtpATSC3SignalingInfo.pURI = pPayload + 14;
    mmtpATSC3SignalingInfo.content_length = 
        (pPayload[14 + mmtpATSC3SignalingInfo.uri_length] << 24) | 
        (pPayload[15 + mmtpATSC3SignalingInfo.uri_length] << 16) |
        (pPayload[16 + mmtpATSC3SignalingInfo.uri_length] << 8) | 
        pPayload[17 + mmtpATSC3SignalingInfo.uri_length];

    switch (mmtpATSC3SignalingInfo.content_type)
    {
        case 0x0001: // UserServiceDescription
            ProcessUserServiceDescriptionSignalingMessage(pPacketTime, 
                packetNumber, pPayload + 18 + mmtpATSC3SignalingInfo.uri_length, 
                payloadLength - 18 - mmtpATSC3SignalingInfo.uri_length,
                pSession, pAssetInfo, pMMTPProtocolPacketInfo, 
                &mmtpATSC3SignalingInfo);
            break;
        case 0x0003: // HELD
            ProcessHELDSignalingMessage(pPacketTime,
                packetNumber, pPayload + 18 + mmtpATSC3SignalingInfo.uri_length,
                payloadLength - 18 - mmtpATSC3SignalingInfo.uri_length,
                pSession, pAssetInfo, pMMTPProtocolPacketInfo,
                &mmtpATSC3SignalingInfo);
            break;
        case 0x0005: // video stream properties descriptor
            ProcessVideoStreamPropertiesDescriptorMessage(pPacketTime, 
                packetNumber, pPayload + 18 + mmtpATSC3SignalingInfo.uri_length, 
                payloadLength - 18 - mmtpATSC3SignalingInfo.uri_length,
                pSession, pAssetInfo, pMMTPProtocolPacketInfo, 
                &mmtpATSC3SignalingInfo);
            break;
        case 0x0008: // Caption Asset Descriptor
            ProcessCaptionAssetDescriptorMessage(pPacketTime, 
                packetNumber, pPayload + 18 + mmtpATSC3SignalingInfo.uri_length, 
                payloadLength - 18 - mmtpATSC3SignalingInfo.uri_length,
                pSession, pAssetInfo, pMMTPProtocolPacketInfo, 
                &mmtpATSC3SignalingInfo);
            break;
        case 0x0009: // Audio Stream Properties Descriptor
            ProcessAudioStreamPropertiesDescriptorMessage(pPacketTime, 
                packetNumber, pPayload + 18 + mmtpATSC3SignalingInfo.uri_length, 
                payloadLength - 18 - mmtpATSC3SignalingInfo.uri_length,
                pSession, pAssetInfo, pMMTPProtocolPacketInfo, 
                &mmtpATSC3SignalingInfo);
            break;
        default:
            LOGPRINT("%s:%s %"PRIu64" Unhandled ATSC 3 signaling message content_type %d\n", 
                pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
                mmtpATSC3SignalingInfo.content_type);
            rval = 1; 
            break;
    }

    return rval;
}

int32_t
ProcessV1SignalingMessage(struct timespec *pPacketTime, 
    uint64_t packetNumber, const uint8_t *pPayload, int32_t payloadLength, 
    MMTPSession *pSession, MMTPAssetInfo *pAssetInfo, 
    MMTPProtocolPacketInfo *pMMTPProtocolPacketInfo)
{
    uint16_t message_id;
    int32_t rval = 0;

    if (payloadLength < 2)
    {
        LOGPRINT("%s:%s %"PRIu64" Signaling message payload too short (%d)\n",
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
            payloadLength);
        rval = 1;
        goto out;
    }

    message_id = (pPayload[0] << 8) | pPayload[1];

    if (message_id >= 0x11 && message_id <= 0x20)
    {
        rval = ProcessMPTSignalingMessage(pPacketTime, packetNumber, 
            pPayload, payloadLength, pSession, pAssetInfo, 
            pMMTPProtocolPacketInfo);
    }
    else if (message_id == 0x0204)
    {
        rval = ProcessHRBMSignalingMessage(pPacketTime, packetNumber, 
            pPayload, payloadLength, pSession, pAssetInfo, 
            pMMTPProtocolPacketInfo);
    }
    else if (message_id == 0x020a)
    {
        rval = ProcessHRBRemovalSignalingMessage(pPacketTime, packetNumber, 
            pPayload, payloadLength, pSession, pAssetInfo, 
            pMMTPProtocolPacketInfo);
    }
    else if (message_id == 0x8100)
    {
        rval = ProcessATSC3SignalingMessage(pPacketTime, packetNumber, 
            pPayload, payloadLength, pSession, pAssetInfo, 
            pMMTPProtocolPacketInfo);
    }
    else
    {
        LOGPRINT("%s:%s %"PRIu64" Unhandled signaling message id 0x%.4x\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
            message_id);
        rval = 1;
    }

out:
    return rval;
}

void
WriteISOBMFFFragment(struct timespec *pPacketTime, uint64_t packetNumber, 
    MMTPSession *pSession, MMTPAssetInfo *pAssetInfo, int32_t sequenceNumber)
{
    struct _MediaSample *pMediaSample;
    uint64_t mdatBoxBytesReceived = 0;
    int32_t i;

    if (!pAssetInfo->fHaveMPUMetadataDU)
    {
        LOGPRINT("%s:%s Did not receive moov - discarding fragment %d\n",
            pSession->strDebugName, pAssetInfo->strAssetID, sequenceNumber);
        goto out;
    }

    if (!pAssetInfo->fHaveMovieFragmentMetadataDU)
    {
        LOGPRINT("%s:%s: Did not receive moof - discarding fragment\n",
            pSession->strDebugName, pAssetInfo->strAssetID);
        goto out;
    }

#ifdef DEBUG_SHADOWFILE
    LOGPRINT_NOFILE("%s:%s PUSHING FRAGMENT\n",
        pSession->strDebugName, pAssetInfo->strAssetID);
#endif

    mdatBoxBytesReceived = pAssetInfo->mdatBoxHeaderSize + 
        pAssetInfo->nExtraMovieFragmentMetadataBytes;

#ifdef DEBUG_SHADOWFILE
    LOGPRINT_NOFILE("%s%s PUSHING MPU METADATA\n", 
        pSession->strDebugName, pAssetInfo->strAssetID);
#endif
    PushData(&pAssetInfo->fragmentPacketTime, pAssetInfo->fragmentPacketNumber, 
        pAssetInfo, pSession, pAssetInfo->vMPUMetadataDUBuffer, 
        pAssetInfo->nMPUMetadataDUBufferBytesUsed, 0,
        pAssetInfo->moovSequenceNumber);

#ifdef DEBUG_SHADOWFILE
    LOGPRINT_NOFILE("%s:%s PUSHING FRAGMENT METADATA\n",
        pSession->strDebugName, pAssetInfo->strAssetID);
#endif
    PushData(&pAssetInfo->fragmentPacketTime, pAssetInfo->fragmentPacketNumber,
        pAssetInfo, pSession, pAssetInfo->vMovieFragmentMetadataDUBuffer, 
        pAssetInfo->nMovieFragmentMetadataDUBufferBytesUsed, 1, sequenceNumber);

#ifdef DEBUG_SHADOWFILE
    LOGPRINT_NOFILE("%s:%s PUSHING SAMPLES\n",
        pSession->strDebugName, pAssetInfo->strAssetID);
#endif

    for (i = 0, pMediaSample = pAssetInfo->vMediaSamples; 
         i < pAssetInfo->nMediaSamples; 
         i++, pMediaSample++)
    {
#ifdef DEBUG_SHADOWFILE
        LOGPRINT_NOFILE(
            "%s:%s Pushing sample %d size %d @ %"PRIu64" of %"PRIu64"\n",
            pSession->strDebugName, pAssetInfo->strAssetID, i, 
            pMediaSample->nData, mdatBoxBytesReceived, 
            pAssetInfo->mdatBoxSize);
#endif
        if (mdatBoxBytesReceived + pMediaSample->nData > 
            pAssetInfo->mdatBoxSize)
        {
            LOGPRINT("%s:%s: Sample %d/%d exceeds mdat box %"PRIu64" (received %"PRIu64" sample size %d)\n", 
                pSession->strDebugName, pAssetInfo->strAssetID,
                i, pAssetInfo->nMediaSamples, 
                pAssetInfo->mdatBoxSize, mdatBoxBytesReceived,
                pMediaSample->nData);
            break;
        }

        PushData(&pAssetInfo->fragmentPacketTime, 
            pAssetInfo->fragmentPacketNumber, pAssetInfo, pSession, 
            pMediaSample->pData, pMediaSample->nData, 2, sequenceNumber); 
        mdatBoxBytesReceived += pMediaSample->nData;

        if (pAssetInfo->vSampleInfos[i].sample_size >= pMediaSample->nData)
        {
            int32_t nPaddingBytes;

            nPaddingBytes = pAssetInfo->vSampleInfos[i].sample_size - 
                pMediaSample->nData;
            if (mdatBoxBytesReceived + nPaddingBytes > pAssetInfo->mdatBoxSize)
            {
                LOGPRINT("%s:%s Sample %d/%d padding %d exceeds mdat box size %"PRIu64" (received %d)\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    i, pAssetInfo->nMediaSamples, nPaddingBytes,
                    pAssetInfo->mdatBoxSize, mdatBoxBytesReceived);
                break;
            }
            PadMDAT(&pAssetInfo->fragmentPacketTime, 
                pAssetInfo->fragmentPacketNumber, pAssetInfo, pSession, 
                nPaddingBytes, 2, sequenceNumber);
            mdatBoxBytesReceived += nPaddingBytes;
        }
        else
        {
            LOGPRINT("%s:%s received sample %d/%d size %d > indicated size %d\n",
                pSession->strDebugName, pAssetInfo->strAssetID,
                i, pAssetInfo->nMediaSamples,
                pMediaSample->nData, pAssetInfo->vSampleInfos[i].sample_size);
        }
    }

    if (PadMDAT(&pAssetInfo->fragmentPacketTime, 
        pAssetInfo->fragmentPacketNumber, pAssetInfo, pSession, 
        pAssetInfo->mdatBoxSize - mdatBoxBytesReceived, 2, sequenceNumber))
    {
        ERRORPRINT("%s:%s %"PRIu64" Failed to pad mdat box packet_id %d\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
            pAssetInfo->packet_id);
    }

out:
    return;
}

int32_t
ProcessV1ISOBMFFPayload(struct timespec *pPacketTime, uint64_t packetNumber, 
    const uint8_t *pPayload, int32_t payloadLength, MMTPSession *pSession, 
    MMTPAssetInfo *pAssetInfo, MMTPProtocolPacketInfo *pMMTPProtocolPacketInfo)
{
    MMTPPayloadInfo mmtpPayloadInfo;
    int32_t offset = 0;
    int rval = 0;

    if (pMMTPProtocolPacketInfo->packet_id < 0x10 || 
        pMMTPProtocolPacketInfo->packet_id > 0x1ffe)
    {
        // A/331 8.1.2.1.3: The value of the packet_id field of MMTPP packets 
        // shall be between 0x0010 and 0x1ffe for easy conversion of an MMTPP 
        // stream to an MPEG-2 TS.
        ERRORPRINT("%s:%s %"PRIu64" SNH: Invalid packet id 0x%x\n", 
            pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
            pMMTPProtocolPacketInfo->packet_id);
    }

    if (MMTPPayloadParse(pSession, pPayload + offset, payloadLength - offset, 
        &mmtpPayloadInfo))
    {
        LOGPRINT("%s: %"PRIu64" Failed to parse MMTP Payload packet_id %d\n", 
            pSession->strDebugName, packetNumber,
            pMMTPProtocolPacketInfo->packet_id);
        rval = 1;
        goto out;
    }

    offset += mmtpPayloadInfo.offset;

    if (pSession->verbosity >= ENVerbosity_ObjectTrace)
    {
        LOGPRINT("%s:%s   payload_len %d payload_type %d frag_ind %d agg_flag %d frag_counter %d seq_num %d timed %d\n", 
           pSession->strDebugName, pAssetInfo->strAssetID, 
           mmtpPayloadInfo.length, mmtpPayloadInfo.type, 
           mmtpPayloadInfo.fragmentation_indicator, 
           mmtpPayloadInfo.aggregation_flag, 
           mmtpPayloadInfo.fragmentation_counter, 
           mmtpPayloadInfo.sequence_number, 
           mmtpPayloadInfo.timed_flag);
    }

    if (mmtpPayloadInfo.fragmentation_indicator == 0 || 
        mmtpPayloadInfo.fragmentation_indicator == 1)
    {
        // This packet contains an entire MPU, or is the first packet of a 
        // fragment
        if (pAssetInfo->mmtpPayloadFragmentationCounter != 0 &&
            pAssetInfo->mmtpPayloadFragmentationCounter != -1)
        {
            LOGPRINT("%s:%s %"PRIu64" Start of MFU but fragmentation counter not 0 packet_id %d!\n",
                pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
                pMMTPProtocolPacketInfo->packet_id);
        }
    }
    else 
    {
        // This packet is part of a fragment and not the first packet in the
        // fragment.
        if (pAssetInfo->mmtpPayloadFragmentationCounter - 1 != 
            mmtpPayloadInfo.fragmentation_counter &&
            pAssetInfo->mmtpPayloadFragmentationCounter != -1)
        {
            if (pSession->verbosity >= ENVerbosity_Terse)
            {
                LOGPRINT_NOFILE("%s:%s %"PRIu64" MMTP Payload Fragmentation Counter Discrepancy (%d != %d) packet_id %d\n", 
                    pSession->strDebugName, pAssetInfo->strAssetID, 
                    packetNumber, mmtpPayloadInfo.fragmentation_counter, 
                    pAssetInfo->mmtpPayloadFragmentationCounter - 1,
                    pMMTPProtocolPacketInfo->packet_id);
            }
        }
    }

    if (mmtpPayloadInfo.fragmentation_indicator &&
        mmtpPayloadInfo.aggregation_flag)
    {
        // CHECK SPEC: This would appear to be disallowed. If it is allowed 
        // then we need to determine where the fragment is (before the 
        // aggregated data units? One of the data units?)
        ERRORPRINT("%s:%s Fragmentation and aggregation in same packet - discarding\n",
            pSession->strDebugName, pAssetInfo->strAssetID);
        rval = 1;
        goto out;
    }

    pAssetInfo->mmtpPayloadFragmentationCounter = 
        mmtpPayloadInfo.fragmentation_counter;

    // A discrepancy in the expectedPayloadSequenceNumber could indicate that 
    // complete MPUs are missing, or that we did not receive the tail end of 
    // an MPU. Either way we need to terminate processing of the current MPU 
    // (if any) and begin processing a new MPU.
    if (pAssetInfo->expectedPayloadSequenceNumber != 
        mmtpPayloadInfo.sequence_number)
    {
        // Note that we're assuming that it we see a change in the sequence
        // number then we are starting a new fragment. Strictly speaking
        // fragments could be delivered out-of-order. 
        if (pAssetInfo->expectedPayloadSequenceNumber != 
            INVALID_PAYLOAD_SEQUENCE_NUMBER)
        {
            WriteISOBMFFFragment(&pAssetInfo->fragmentPacketTime, 
                packetNumber, pSession, pAssetInfo, 
                pAssetInfo->expectedPayloadSequenceNumber);

            if (mmtpPayloadInfo.sequence_number != 
                pAssetInfo->expectedPayloadSequenceNumber + 1)
            {
                LOGPRINT_NOFILE("%s:%s WARN: sequence discrepancy %u -> %u\n",
                    pSession->strDebugName, pAssetInfo->strAssetID,
                    pAssetInfo->expectedPayloadSequenceNumber,
                    mmtpPayloadInfo.sequence_number) ;
            }
            ResetFragmentData(pAssetInfo);
        }

        pAssetInfo->fragmentPacketTime = *pPacketTime;
        pAssetInfo->fragmentPacketNumber = packetNumber;
    }

    pAssetInfo->expectedPayloadSequenceNumber = mmtpPayloadInfo.sequence_number;

    // An MMTP payload can carry fragments of data units or aggregate data 
    // units. (The definition of a data unit depends on the data type. For 
    // example for media samples it may be an MPU fragment.) In the case of 
    // aggregation, an MMTP payload carries data units from a single asset. 
    //
    // Iterate over the data unit (or data unit fragments) contained within 
    // the MMTP payload.
    do
    {
        if (mmtpPayloadInfo.type == 0)
        {
            if (pAssetInfo->fHaveMPUMetadataDU)
            {
                // If we already have a complete MPU metadata DU.
                // Ignore this one. 
                LOGPRINT("%s:%s WARN: Already have mpu metadata DU, ignoring\n",
                    pSession->strDebugName, pAssetInfo->strAssetID);
            }
            else
            {
                // FT = 0 packets carry boxes which apply to the MPU as a whole. 
                // XXX: If the MPU consists of one movie fragment, then can the 
                // moof also be contained in an FT = 0 packet?
                if (mmtpPayloadInfo.fragmentation_indicator == 0 || 
                    mmtpPayloadInfo.fragmentation_indicator == 1)
                {
                    pAssetInfo->nMPUMetadataDUBufferBytesUsed = 0;
                }

                if (pAssetInfo->nMPUMetadataDUBufferBytesUsed + payloadLength - 
                    offset > pAssetInfo->sMPUMetadataDUBuffer)
                {
                    LOGPRINT_NOFILE("%s:%s %"PRIu64" MP4 box data too large (%d > %d)\n", 
                        pSession->strDebugName, pAssetInfo->strAssetID, 
                        packetNumber,
                        pAssetInfo->nMPUMetadataDUBufferBytesUsed + 
                            payloadLength - 
                        offset, pAssetInfo->sMPUMetadataDUBuffer);
                    pAssetInfo->expectedPayloadSequenceNumber = 
                        mmtpPayloadInfo.sequence_number;
                    ResetFragmentData(pAssetInfo);
                    rval = 1;
                    break;
                }

                memcpy(pAssetInfo->vMPUMetadataDUBuffer + 
                    pAssetInfo->nMPUMetadataDUBufferBytesUsed,
                    pPayload + offset, payloadLength - offset);
                pAssetInfo->nMPUMetadataDUBufferBytesUsed += 
                    payloadLength - offset;

                if (mmtpPayloadInfo.fragmentation_indicator == 0 ||
                    mmtpPayloadInfo.fragmentation_indicator == 3)
                {
                    if (VerifyBoxStructure(pSession, pAssetInfo, 
                        pAssetInfo->vMPUMetadataDUBuffer, 
                        pAssetInfo->nMPUMetadataDUBufferBytesUsed))
                    {
                        LOGPRINT_NOFILE("%s:%s %"PRIu64" Failed verify box structure packet_id %d\n",
                            pSession->strDebugName, pAssetInfo->strAssetID, 
                            packetNumber, pMMTPProtocolPacketInfo->packet_id);
                        ResetFragmentData(pAssetInfo);
                        rval = 1;
                        break;
                    }

                    if (ParseAssetInfoFromMmpu(pSession, 
                        pAssetInfo->vMPUMetadataDUBuffer,
                        pAssetInfo->nMPUMetadataDUBufferBytesUsed, pAssetInfo))
                    {
                        LOGPRINT_NOFILE(
                            "%s:%s %"PRIu64" Failed to parse mmpu packet_id %d\n", 
                            pSession->strDebugName, pAssetInfo->strAssetID, 
                            packetNumber, pMMTPProtocolPacketInfo->packet_id);
                        ResetFragmentData(pAssetInfo);
                        rval = 1;
                        break;
                    }
                    if (ParseAssetInfoFromMoov(pSession, 
                        pAssetInfo->vMPUMetadataDUBuffer,
                        pAssetInfo->nMPUMetadataDUBufferBytesUsed, pAssetInfo))
                    {
                        LOGPRINT_NOFILE("%s:%s %"PRIu64" Failed to parse moov packet_id %d\n", 
                            pSession->strDebugName, pAssetInfo->strAssetID, 
                            packetNumber, pMMTPProtocolPacketInfo->packet_id);
                        ResetFragmentData(pAssetInfo);
                        rval = 1;
                        break;
                    }

                    pAssetInfo->moovSequenceNumber = 
                        mmtpPayloadInfo.sequence_number;
                    pAssetInfo->fHaveMPUMetadataDU = 1;
                }
            }
        }
        else if (mmtpPayloadInfo.type == 1)
        {
            if (pAssetInfo->fHaveMovieFragmentMetadataDU)
            {
                // If we already have a complete movie fragment metadata DU.
                // Ignore this one. 
                LOGPRINT("%s:%s WARN: Already have movie fragment metadata DU, ignoring\n",
                    pSession->strDebugName, pAssetInfo->strAssetID);
            }
            else
            {
                // FT = 1 packets carry movie fragments (moof) and related boxes. 
                // The last box is the mdat box header.
                if (mmtpPayloadInfo.fragmentation_indicator == 0 || 
                    mmtpPayloadInfo.fragmentation_indicator == 1)
                {
                    pAssetInfo->nMovieFragmentMetadataDUBufferBytesUsed = 0;
                }

                if (pAssetInfo->nMovieFragmentMetadataDUBufferBytesUsed + 
                    payloadLength - offset >
                    pAssetInfo->sMovieFragmentMetadataDUBuffer)
                {
                    LOGPRINT_NOFILE("%s:%s %"PRIu64" MP4 box data too large (%d > %d)\n", 
                        pSession->strDebugName, pAssetInfo->strAssetID, 
                        packetNumber, 
                        pAssetInfo->nMovieFragmentMetadataDUBufferBytesUsed + 
                        payloadLength - offset,
                        pAssetInfo->sMovieFragmentMetadataDUBuffer);
                    pAssetInfo->expectedPayloadSequenceNumber = 
                        mmtpPayloadInfo.sequence_number;
                    ResetFragmentData(pAssetInfo);
                    rval = 1;
                    break;
                }

                memcpy(pAssetInfo->vMovieFragmentMetadataDUBuffer + 
                    pAssetInfo->nMovieFragmentMetadataDUBufferBytesUsed,
                    pPayload + offset, payloadLength - offset);
                pAssetInfo->nMovieFragmentMetadataDUBufferBytesUsed += 
                    payloadLength - offset;

                if (mmtpPayloadInfo.fragmentation_indicator == 0 ||
                    mmtpPayloadInfo.fragmentation_indicator == 3)
                {
                    if (VerifyBoxStructure(pSession, pAssetInfo, 
                        pAssetInfo->vMPUMetadataDUBuffer, 
                        pAssetInfo->nMPUMetadataDUBufferBytesUsed))
                    {
                        LOGPRINT_NOFILE("%s:%s %"PRIu64" Failed verify box structure packet_id %d\n",
                            pSession->strDebugName, pAssetInfo->strAssetID, 
                            packetNumber, pMMTPProtocolPacketInfo->packet_id);
                        ResetFragmentData(pAssetInfo);
                        rval = 1;
                        break;
                    }

                    if (ParseSampleInfoFromMoof(pSession, pPacketTime,  
                        pAssetInfo->vMovieFragmentMetadataDUBuffer,
                        pAssetInfo->nMovieFragmentMetadataDUBufferBytesUsed, 
                        pAssetInfo, &mmtpPayloadInfo))
                    {
                        LOGPRINT_NOFILE("%s:%s: %"PRIu64" Failed to parse moof packet_id %d\n", 
                            pSession->strDebugName, pAssetInfo->strAssetID, 
                            packetNumber, pMMTPProtocolPacketInfo->packet_id);
                        ResetFragmentData(pAssetInfo);
                        rval = 1;
                        break;
                    }
                    pAssetInfo->fHaveMovieFragmentMetadataDU = 1;
                }
            }
        }
        else if (mmtpPayloadInfo.type == 2) 
        {
            rval = HandleMFUDU(packetNumber, pPayload + offset,
                payloadLength - offset, pSession, pAssetInfo, 
                pMMTPProtocolPacketInfo, &mmtpPayloadInfo);
        }

        if (MMTPPayloadAdvance(pSession, &mmtpPayloadInfo))
        {
            LOGPRINT_NOFILE("%s:%s %"PRIu64" Failed to advance MMTP payload to next DU packet_id %d\n", 
                pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
                pMMTPProtocolPacketInfo->packet_id);
            rval = 1;
            ResetFragmentData(pAssetInfo);
            break;
        }
    } while (mmtpPayloadInfo.offset != mmtpPayloadInfo.length);

out:
    return rval;
}

ENATSC3Status
ENATSC3MMTPSessionProcessUDPPacket(MMTPSessionHandle_t hMMTPSession, 
    struct timespec *pPacketTime, uint64_t packetNumber, 
    const uint8_t *pPayload, int32_t payloadLength)
{
    MMTPSession *pSession = (MMTPSession *)hMMTPSession;
    int32_t offset;
    MMTPProtocolPacketInfo mmtpProtocolPacketInfo;
    MMTPAssetInfo *pAssetInfo = NULL;
    ENATSC3Status rval = ENATSC3Status_Success;

    // For ATSC 3.0, a UDP packet contains an MMTP protocol packet. 
    //
    // A single MMTP protocol packet in turn carries a single MMTP payload. 
    // (ISO 23008-1 WD 7.3.1)

    if (MMTPProtocolPacketParse(pSession, pPayload, &mmtpProtocolPacketInfo))
    {
        LOGPRINT("%s: %"PRIu64" Failed to parse MMTPP protocol packet\n",
            pSession->strDebugName, packetNumber);
        rval = ENATSC3Status_Failure;
        goto out;
    }

    // A/331 requires MMTP protocol version 1, however there appear to be some
    // version 0 packets floating around.
    if (mmtpProtocolPacketInfo.version == 1)
    {
        offset = mmtpProtocolPacketInfo.headerLength;

        if (pSession->verbosity >= ENVerbosity_ObjectTrace)
        {
            LOGPRINT("%s: %"PRIu64" %.5d: v1: packet_type %d packet_id %d timestamp %d seqno %d counter %d, header_extension_flag %d compression_flag %d\n",
                pSession->strDebugName, packetNumber, payloadLength, 
                mmtpProtocolPacketInfo.packet_type, 
                mmtpProtocolPacketInfo.packet_id, 
                mmtpProtocolPacketInfo.timestamp, 
                mmtpProtocolPacketInfo.sequence_number, 
                mmtpProtocolPacketInfo.packet_counter_flag ? 
                    mmtpProtocolPacketInfo.packet_counter: -1,
                mmtpProtocolPacketInfo.header_extension_flag, 
                mmtpProtocolPacketInfo.compression_flag);
        }

        if (offset > payloadLength)
        {
            if (pSession->verbosity >= ENVerbosity_ObjectTrace)
            {
                LOGPRINT_NOCONSOLE("Bad offset %d payloadLength %d\n", 
                    offset, payloadLength);
            }
            rval = ENATSC3Status_Failure;
            goto out;
        }

        if (GetAssetInfo(mmtpProtocolPacketInfo.packet_id, pSession, 
            &pAssetInfo))
        {
            LOGPRINT(
                "%s %"PRIu64": Failed to get asset info for packet_id %d\n",
                pSession->strDebugName, packetNumber, 
                mmtpProtocolPacketInfo.packet_id);
            rval = ENATSC3Status_Failure;
            goto out;
        }

        if (((pAssetInfo->mmtpPacketSeqno + 1) & 0xffffffff) != 
            mmtpProtocolPacketInfo.sequence_number &&
            pAssetInfo->mmtpPacketSeqno != -1)
        {
            if (pSession->verbosity >= ENVerbosity_Terse)
            {
                LOGPRINT("%s:%s %"PRIu64" MMTP Packet Sequence number error %"PRId64" != %"PRIu32", packet_id %d, type %d\n",
                    pSession->strDebugName, pAssetInfo->strAssetID, 
                    packetNumber, 
                    (pAssetInfo->mmtpPacketSeqno + 1) & 0xffffffff, 
                    mmtpProtocolPacketInfo.sequence_number, 
                    mmtpProtocolPacketInfo.packet_id,
                    mmtpProtocolPacketInfo.packet_type);
            }
        }

        pAssetInfo->mmtpPacketSeqno = mmtpProtocolPacketInfo.sequence_number;

        if (mmtpProtocolPacketInfo.packet_type == 0) // ISOBMFF file
        {
            rval = ProcessV1ISOBMFFPayload(pPacketTime, packetNumber,   
                pPayload + offset, payloadLength - offset, pSession, 
                pAssetInfo, &mmtpProtocolPacketInfo);
            if (rval)
            {
                rval = ENATSC3Status_Failure;
                goto out;
            }
        }
        else if (mmtpProtocolPacketInfo.packet_type == 1) // Generic File Delivery (GFD)
        {
            // A/331 8.1.2.1.3: The value of '0x01' for the type field of an 
            // MMTPP packet header shall not be used, i.e. the Generic File 
            // Delivery (GFD) mode specified in subclauses 9.3.3 and 9.3.4 of
            // [37] shall not be used. 
            ERRORPRINT("%s:%s SNH: GFD mode not allowed in A/331\n",
                pSession->strDebugName, pAssetInfo->strAssetID);
            rval = ENATSC3Status_Failure;
        }
        else if (mmtpProtocolPacketInfo.packet_type == 2) // signaling message
        {
            // WHY IS THIS NEEDED? WHAT ARE WE SKIPPING OVER?  CHECK 23008-1
            // If this contains aggregation/frag info then it should be 
            // passed to ProcessV1SignalingMessage so it can be handled.
            offset += 2;
            rval = ProcessV1SignalingMessage(pPacketTime, packetNumber, 
                pPayload + offset, payloadLength - offset, pSession, 
                pAssetInfo, &mmtpProtocolPacketInfo);
            if (rval)
            {
                rval = ENATSC3Status_Failure;
                goto out;
            }
        }
        else
        {
            ERRORPRINT("%s:%s %"PRIu64" unhandled packet type %d\n", 
                pSession->strDebugName, pAssetInfo->strAssetID, packetNumber, 
                mmtpProtocolPacketInfo.packet_type);
        }
    }
    else if (mmtpProtocolPacketInfo.version == 0)
    {
        // A/331 8.1.2.1.3: The value of the version field of MMTPP packets 
        // shall be '01'
        ERRORPRINT("%s:%s Invalid MMTPP Protocol Version 0\n",
            pSession->strDebugName, pAssetInfo->strAssetID);
        goto out;
    }
    else
    {
        ERRORPRINT("%s:%s Unhandled version %d\n", pSession->strDebugName, 
            pAssetInfo->strAssetID, mmtpProtocolPacketInfo.version);
        goto out;
    }

out:
    return rval;
}

ENATSC3Status
ENATSC3MMTPSessionAllocate(uint32_t addr, uint16_t port, 
    MMTPIPAddressAddCallback IPAddressAddCallback,
    MMTPIPAddressRemoveCallback IPAddressRemoveCallback,
    void *pIPAddrCallbackUserData, char *pDebugName, 
    MMTPSessionHandle_t *phMMTPSession)
{
    MMTPSession *pMMTPSession = NULL;
    int32_t i;
    ENATSC3Status rval = ENATSC3Status_Success;

    pMMTPSession = (MMTPSession *)calloc(sizeof(MMTPSession), 1);
    if (!pMMTPSession)
    {
        ERRORPRINT("Out of memory\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    strncpy(pMMTPSession->strDebugName, pDebugName, 
        sizeof(pMMTPSession->strDebugName) - 1);
    pMMTPSession->strDebugName[sizeof(pMMTPSession->strDebugName) - 1] = 0;

    for (i = 0; i < numof(pMMTPSession->vAssetInfos); i++)
    {
        pMMTPSession->vAssetInfos[i].packet_id = -1;
        pMMTPSession->vAssetInfos[i].mmtpPacketSeqno = -1;
        pMMTPSession->vAssetInfos[i].mmtpPayloadFragmentationCounter = -1;
        pMMTPSession->vAssetInfos[i].usbdVersion = -1;
        pMMTPSession->vAssetInfos[i].usbdCrc32 = 0;
        pMMTPSession->vAssetInfos[i].vspdVersion = -1;
        pMMTPSession->vAssetInfos[i].aspdVersion = -1;
        pMMTPSession->vAssetInfos[i].cadVersion = -1;
        pMMTPSession->vAssetInfos[i].expectedPayloadSequenceNumber = 
            INVALID_PAYLOAD_SEQUENCE_NUMBER;
        pMMTPSession->vAssetInfos[i].sMPUMetadataDUBuffer = 
            sizeof(pMMTPSession->vAssetInfos[i].vMPUMetadataDUBuffer);
        pMMTPSession->vAssetInfos[i].nMPUMetadataDUBufferBytesUsed = 0;
        pMMTPSession->vAssetInfos[i].sMovieFragmentMetadataDUBuffer = 
            sizeof(pMMTPSession->vAssetInfos[i].vMovieFragmentMetadataDUBuffer);
        pMMTPSession->vAssetInfos[i].nMovieFragmentMetadataDUBufferBytesUsed = 0;
        pMMTPSession->vAssetInfos[i].sMediaSampleBuffer = 
            MEDIA_SAMPLE_BUFFER_SIZE;
        pMMTPSession->vAssetInfos[i].nMediaSampleBufferBytesUsed = 0;
        pMMTPSession->vAssetInfos[i].pMediaSampleBuffer = 
            (uint8_t *)malloc(MEDIA_SAMPLE_BUFFER_SIZE);
        pMMTPSession->vAssetInfos[i].mpuPresentationTime = 0;
        pMMTPSession->vAssetInfos[i].lastGoodMPUPresentationTimePacketTime.tv_sec = 0;
        pMMTPSession->vAssetInfos[i].nRejectedBaseMediaDecodeTimes = 0;
        if (!pMMTPSession->vAssetInfos[i].pMediaSampleBuffer)
        {
            ERRORPRINT("%s: Out of memory\n", pMMTPSession->strDebugName);
            rval = ENATSC3Status_Failure;
            goto errout;
        }

        pMMTPSession->vAssetInfos[i].mvhdTimescale = 0;
        pMMTPSession->vAssetInfos[i].mdhdTimescale = 0;
        pMMTPSession->vAssetInfos[i].nBytesPushed = 0;

        pMMTPSession->vAssetInfos[i].shadowFd = -1;

        strcpy(pMMTPSession->vAssetInfos[i].strAssetID, "unnamed");
        ResetFragmentData(pMMTPSession->vAssetInfos + i);
    }

    pMMTPSession->verbosity = 0;

    pMMTPSession->IPAddressAddCallback = IPAddressAddCallback;
    pMMTPSession->IPAddressRemoveCallback = IPAddressRemoveCallback;
    pMMTPSession->pIPAddrCallbackUserData = pIPAddrCallbackUserData;
    pMMTPSession->slsAddr = addr;
    pMMTPSession->slsPort = port;
    pMMTPSession->sessionId = enatsc3mmtpGlobalsG.sessionCounter++;
    IPAddressAddCallback(addr, port, pIPAddrCallbackUserData);

    *phMMTPSession = pMMTPSession;

out:
    return rval;

errout:
    if (pMMTPSession)
    {
        for (i = 0; i < numof(pMMTPSession->vAssetInfos); i++)
        {
            if (pMMTPSession->vAssetInfos[i].pMediaSampleBuffer)
            {
                free(pMMTPSession->vAssetInfos[i].pMediaSampleBuffer);
            }
        }

        free(pMMTPSession);
    }

    goto out;
}

ENATSC3Status
ENATSC3MMTPSessionFree(MMTPSessionHandle_t hMMTPSession)
{
    MMTPSession *pMMTPSession = (MMTPSession *)hMMTPSession;
    int32_t i;
    ENATSC3Status rval = ENATSC3Status_Success;

    for (i = 0; i < numof(pMMTPSession->vAssetInfos); i++)
    {
        if (pMMTPSession->vAssetInfos[i].pMediaSampleBuffer)
        {
            free(pMMTPSession->vAssetInfos[i].pMediaSampleBuffer);
        }

        if (pMMTPSession->vAssetInfos[i].shadowFd >= 0)
        {
            close(pMMTPSession->vAssetInfos[i].shadowFd);
            pMMTPSession->vAssetInfos[i].shadowFd = -1;
        }
    }

    pMMTPSession->IPAddressRemoveCallback(pMMTPSession->slsAddr, 
        pMMTPSession->slsPort, pMMTPSession->pIPAddrCallbackUserData);

    free(pMMTPSession);

    return rval;
}

ENATSC3Status
ENATSC3MMTPSessionSetVerbosity(MMTPSessionHandle_t hMMTPSession, 
    ENVerbosity verbosity)
{
    MMTPSession *pMMTPSession = (MMTPSession *)hMMTPSession;
    ENATSC3Status rval = ENATSC3Status_Success;

    pMMTPSession->verbosity = verbosity;

    return rval;
}

ENATSC3Status
ENATSC3MMTPSessionRegisterCallbacks(MMTPSessionHandle_t hMMTPSession,
    ENATSC3MMTPUSBDAvailableCallback USBDAvailableCallback,
    ENATSC3MMTPHELDAvailableCallback HELDAvailableCallback,
    ENATSC3MMTPVSPDAvailableCallback VSPDAvailableCallback,
    ENATSC3MMTPASPDAvailableCallback ASPDAvailableCallback,
    ENATSC3MMTPCADAvailableCallback CADAvailableCallback,
    ENATSC3MMTPMPTSignalingMessageCallback MPTSignalingMessageCallback,
    ENATSC3MMTPMediaDataCallback MediaDataCallback,
    void *pUserData)
{
    MMTPSession *pMMTPSession = (MMTPSession *)hMMTPSession;
    ENATSC3Status rval = ENATSC3Status_Success;

    pMMTPSession->USBDAvailableCallback = USBDAvailableCallback;
    pMMTPSession->HELDAvailableCallback = HELDAvailableCallback;
    pMMTPSession->VSPDAvailableCallback = VSPDAvailableCallback;
    pMMTPSession->ASPDAvailableCallback = ASPDAvailableCallback;
    pMMTPSession->CADAvailableCallback = CADAvailableCallback;
    pMMTPSession->MPTSignalingMessageCallback = MPTSignalingMessageCallback;
    pMMTPSession->MediaDataCallback = MediaDataCallback;
    pMMTPSession->pCallbackUserData = pUserData;

    return rval;
}

ENATSC3Status
ENATSC3MMTPSessionUnregisterCallbacks(MMTPSessionHandle_t hMMTPSession,
    ENATSC3MMTPUSBDAvailableCallback USBDAvailableCallback,
    ENATSC3MMTPHELDAvailableCallback HELDAvailableCallback,
    ENATSC3MMTPVSPDAvailableCallback VSPDAvailableCallback,
    ENATSC3MMTPASPDAvailableCallback ASPDAvailableCallback,
    ENATSC3MMTPCADAvailableCallback CADAvailableCallback,
    ENATSC3MMTPMPTSignalingMessageCallback MPTSignalingMessageCallback,
    ENATSC3MMTPMediaDataCallback MediaDataCallback,
    void *pUserData)
{
    MMTPSession *pMMTPSession = (MMTPSession *)hMMTPSession;
    ENATSC3Status rval = ENATSC3Status_Success;

    if (pMMTPSession->USBDAvailableCallback == USBDAvailableCallback &&
        pMMTPSession->HELDAvailableCallback == HELDAvailableCallback &&
        pMMTPSession->VSPDAvailableCallback == VSPDAvailableCallback &&
        pMMTPSession->ASPDAvailableCallback == ASPDAvailableCallback &&
        pMMTPSession->CADAvailableCallback == CADAvailableCallback &&
        pMMTPSession->MPTSignalingMessageCallback == 
            MPTSignalingMessageCallback &&
        pMMTPSession->MediaDataCallback == MediaDataCallback &&
        pMMTPSession->pCallbackUserData == pUserData)
    {
        pMMTPSession->USBDAvailableCallback = NULL;
        pMMTPSession->HELDAvailableCallback = NULL;
        pMMTPSession->VSPDAvailableCallback = NULL;
        pMMTPSession->ASPDAvailableCallback = NULL;
        pMMTPSession->CADAvailableCallback = NULL;
        pMMTPSession->MPTSignalingMessageCallback = NULL;
        pMMTPSession->MediaDataCallback = NULL;
        pMMTPSession->pCallbackUserData = NULL;
    }
    else
    {
        rval = ENATSC3Status_Failure;
    }

    return rval;
}
