/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_defs.h>

#include <libenatsc3utils/enxml.h>

#include <enatsc3-common/enatsc3.h>
#include "atsc3routesls_stsid.h"

static char *ensys = "routesls";
static char *ensubsys = "stsid";

static int32_t fVerbose = 0;

#define VERBOSEPRINT(x, ...) \
    if (fVerbose) \
    { \
        LOGPRINT(x, ##__VA_ARGS__); \
    }

static int32_t
PayloadNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDPayloadNode *pPayloadNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_Payload);

    pPayloadNode = (ENATSC3ROUTESLSSTSIDPayloadNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ENATSC3ROUTESLSSTSIDPayloadNodeAttribute_codePoint:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pPayloadNode->codePoint))
            {
                ERRORPRINT("Failed to parse codePoint %.*s\n", sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*scodePoint: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDPayloadNodeAttribute_formatID:
            if (EnXMLParser_antoul((char *)pValue, sValue,
                &pPayloadNode->formatID))
            {
                ERRORPRINT("Failed to parse formatID %.*s\n", sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sformatID: %.*s\n", xmlTextReaderDepth(reader) * 2,
                "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDPayloadNodeAttribute_frag:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pPayloadNode->frag)) 
            {
                ERRORPRINT("Failed to parse frag %.*s\n", sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sfrag: %.*s\n", xmlTextReaderDepth(reader) * 2, 
                "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDPayloadNodeAttribute_order:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pPayloadNode->order))
            {
                ERRORPRINT("Failed to parse order %.*s\n", sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sorder: %.*s\n", xmlTextReaderDepth(reader) * 2,
                "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDPayloadNodeAttribute_srcFecPayloadID:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pPayloadNode->srcFecPayloadID))
            {
                ERRORPRINT("Failed to parse srcFecPayloadID %.*s\n", 
                    sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*ssrcFecPayloadID: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", 
                xmlTextReaderDepth(reader) * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vPayloadAttributes[] = 
{
    {
        .id = ENATSC3ROUTESLSSTSIDPayloadNodeAttribute_codePoint,
        .sName = sizeof("codePoint") - 1,
        .pName = "codePoint",
        .handler = PayloadNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDPayloadNodeAttribute_formatID,
        .sName = sizeof("formatID") - 1,
        .pName = "formatID",
        .handler = PayloadNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDPayloadNodeAttribute_frag,
        .sName = sizeof("frag") - 1,
        .pName = "frag",
        .handler = PayloadNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDPayloadNodeAttribute_order,
        .sName = sizeof("order") - 1,
        .pName = "order",
        .handler = PayloadNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDPayloadNodeAttribute_srcFecPayloadID,
        .sName = sizeof("srcFecPayloadID") - 1,
        .pName = "srcFecPayloadID",
        .handler = PayloadNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vPayloadChildNodes[] =
{
};
static EnXMLParserNode payloadNodeData = 
{
    .pName = "Payload",
    .pAttributes = vPayloadAttributes,
    .nAttributes = numof(vPayloadAttributes),
    .pChildNodes = vPayloadChildNodes,
    .nChildNodes = numof(vPayloadChildNodes),
    .textHandler = NULL,
};

static int32_t
ContentRatingNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDNode *pSTSIDNode = (ENATSC3ROUTESLSSTSIDNode *)pUserData;
    ENATSC3ROUTESLSSTSIDContentRatingNode *pContentRatingNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_ContentRating);

    pContentRatingNode = (ENATSC3ROUTESLSSTSIDContentRatingNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ENATSC3ROUTESLSSTSIDContentRatingNodeAttribute_schemeIdUri:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSTSIDNode->dataBuffer, (char *)pValue, sValue,
                &pContentRatingNode->schemeIdUriDataOffset))
            {
                ERRORPRINT("Failed to store schemeIdUri\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sschemeIdUri: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDContentRatingNodeAttribute_value:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSTSIDNode->dataBuffer, (char *)pValue, sValue,
                &pContentRatingNode->valueDataOffset))
            {
                ERRORPRINT("Failed to store value\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*svalue: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", 
                xmlTextReaderDepth(reader) * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vContentRatingAttributes[] = 
{
    {
        .id = ENATSC3ROUTESLSSTSIDContentRatingNodeAttribute_schemeIdUri,
        .sName = sizeof("schedmeIdUri") - 1,
        .pName = "schemeIdUri",
        .handler = ContentRatingNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDContentRatingNodeAttribute_value,
        .sName = sizeof("value") - 1,
        .pName = "value",
        .handler = ContentRatingNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vContentRatingChildNodes[] =
{
};
static EnXMLParserNode contentRatingNodeData = 
{
    .pName = "ContentRating",
    .pAttributes = vContentRatingAttributes,
    .nAttributes = numof(vContentRatingAttributes),
    .pChildNodes = vContentRatingChildNodes,
    .nChildNodes = numof(vContentRatingChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessContentRatingNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDMediaInfoNode *pMediaInfoNode;
    ENATSC3ROUTESLSSTSIDContentRatingNode *pContentRatingNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_MediaInfo);

    pMediaInfoNode = (ENATSC3ROUTESLSSTSIDMediaInfoNode *)pParentNode;
    pContentRatingNode = &pMediaInfoNode->contentRating;
    pContentRatingNode->base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_ContentRating;

    VERBOSEPRINT("%*sContentRating\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&contentRatingNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)pContentRatingNode);

    return rval;
}

static int32_t
MediaInfoNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDNode *pSTSIDNode = (ENATSC3ROUTESLSSTSIDNode *)pUserData;
    ENATSC3ROUTESLSSTSIDMediaInfoNode *pMediaInfoNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_MediaInfo);

    pMediaInfoNode = (ENATSC3ROUTESLSSTSIDMediaInfoNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ENATSC3ROUTESLSSTSIDMediaInfoNodeAttribute_startup:
            if (EnXMLParserStringToBool((char *)pValue, sValue, 
                &pMediaInfoNode->startup))
            {
                ERRORPRINT("Failed to parse startup value %.*s\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sstartup: %.*s\n", xmlTextReaderDepth(reader) * 2, 
                "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDMediaInfoNodeAttribute_lang:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSTSIDNode->dataBuffer, (char *)pValue, sValue,
                &pMediaInfoNode->langDataOffset))
            {
                ERRORPRINT("Failed to store lang\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*slang: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDMediaInfoNodeAttribute_repId:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSTSIDNode->dataBuffer, (char *)pValue, sValue,
                &pMediaInfoNode->repIdDataOffset))
            {
                ERRORPRINT("Failed to store repId\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*srepId: %.*s\n", xmlTextReaderDepth(reader) * 2,
                "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDMediaInfoNodeAttribute_contentType:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSTSIDNode->dataBuffer, (char *)pValue, sValue,
                &pMediaInfoNode->contentTypeDataOffset))
            {
                ERRORPRINT("Failed to store contentType\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*scontentType: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", 
                xmlTextReaderDepth(reader) * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vMediaInfoAttributes[] = 
{
    {
        .id = ENATSC3ROUTESLSSTSIDMediaInfoNodeAttribute_startup,
        .sName = sizeof("startup") - 1,
        .pName = "startup",
        .handler = MediaInfoNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDMediaInfoNodeAttribute_lang,
        .sName = sizeof("lang") - 1,
        .pName = "lang",
        .handler = MediaInfoNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDMediaInfoNodeAttribute_repId,
        .sName = sizeof("repId") - 1,
        .pName = "repId",
        .handler = MediaInfoNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDMediaInfoNodeAttribute_contentType,
        .sName = sizeof("contentType") - 1,
        .pName = "contentType",
        .handler = MediaInfoNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vMediaInfoChildNodes[] =
{
    {
        .id = ENATSC3ROUTESLSSTSIDMediaInfoChildNode_ContentRating,
        .sName = sizeof("ContentRating") - 1,
        .pName = "ContentRating",
        .handler = ProcessContentRatingNode,
    },
};
static EnXMLParserNode mediaInfoNodeData = 
{
    .pName = "MediaInfo",
    .pAttributes = vMediaInfoAttributes,
    .nAttributes = numof(vMediaInfoAttributes),
    .pChildNodes = vMediaInfoChildNodes,
    .nChildNodes = numof(vMediaInfoChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessMediaInfoNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDContentInfoNode *pContentInfoNode;
    ENATSC3ROUTESLSSTSIDMediaInfoNode *pMediaInfoNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_ContentInfo);

    pContentInfoNode = (ENATSC3ROUTESLSSTSIDContentInfoNode *)pParentNode;
    pMediaInfoNode = &pContentInfoNode->mediaInfo;
    pMediaInfoNode->base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_MediaInfo;

    VERBOSEPRINT("%*sMediaInfo\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&mediaInfoNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)pMediaInfoNode);

    return rval;
}

static int32_t
ContentInfoNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue,
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData,
    EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDNode *pSTSIDNode = (ENATSC3ROUTESLSSTSIDNode *)pUserData;
    ENATSC3ROUTESLSSTSIDContentInfoNode *pContentInfoNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_ContentInfo);

    pContentInfoNode = (ENATSC3ROUTESLSSTSIDContentInfoNode *)pParentNode;

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(&pSTSIDNode->dataBuffer, 
        (char *)pValue, sValue, &pContentInfoNode->contentInfoDataOffset))
    {
        ERRORPRINT("Failed to store contentInfo\n");
        rval = -1;
        goto out;
    }

    VERBOSEPRINT("%*scontentInfo: %.*s\n", xmlTextReaderDepth(reader) * 2,
        "", sValue, pValue);

out:
    return rval;
}

static EnXMLParserAttribute vContentInfoAttributes[] = 
{
};
static EnXMLParserChildNode vContentInfoChildNodes[] =
{
    {
        .id = ENATSC3ROUTESLSSTSIDContentInfoChildNode_MediaInfo,
        .sName = sizeof("MediaInfo") - 1,
        .pName = "MediaInfo",
        .handler = ProcessMediaInfoNode,
    },
};
static EnXMLParserNode contentInfoNodeData = 
{
    .pName = "ContentInfo",
    .pAttributes = vContentInfoAttributes,
    .nAttributes = numof(vContentInfoAttributes),
    .pChildNodes = vContentInfoChildNodes,
    .nChildNodes = numof(vContentInfoChildNodes),
    .textHandler = ContentInfoNodeTextHandler,
};

static int32_t
FileNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDNode *pSTSIDNode = (ENATSC3ROUTESLSSTSIDNode *)pUserData;
    ENATSC3ROUTESLSSTSIDFileNode *pFileNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_File);

    pFileNode = (ENATSC3ROUTESLSSTSIDFileNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ENATSC3ROUTESLSSTSIDFileNodeAttribute_TOI:
            if (EnXMLParser_antoul((char *)pValue, sValue, &pFileNode->toi)) 
            {
                ERRORPRINT("Failed to parse toi %.*s\n", sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*stoi: %.*s\n", xmlTextReaderDepth(reader) * 2,
                "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDFileNodeAttribute_ContentLocation:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSTSIDNode->dataBuffer, (char *)pValue, sValue, 
                &pFileNode->contentLocationDataOffset))
            {
                ERRORPRINT("Failed to store contentLocation\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sContent-Location: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDFileNodeAttribute_ContentType:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSTSIDNode->dataBuffer, (char *)pValue, sValue, 
                &pFileNode->contentTypeDataOffset))
            {
                ERRORPRINT("Failed to store contentType\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sContent-Type: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDFileNodeAttribute_ContentEncoding:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSTSIDNode->dataBuffer, (char *)pValue, sValue, 
                &pFileNode->contentEncodingDataOffset))
            {
                ERRORPRINT("Failed to store contentEncoding\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sContent-Encoding: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDFileNodeAttribute_ContentLength:
            if (EnXMLParser_antoull((char *)pValue, sValue, 
                &pFileNode->contentLength)) 
            {
                ERRORPRINT("Failed to parse Content-Length %.*s\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sContent-Length: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDFileNodeAttribute_TransferLength:
            if (EnXMLParser_antoull((char *)pValue, sValue, 
                &pFileNode->transferLength)) 
            {
                ERRORPRINT("Failed to parse Transfer-Length %.*s\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sTransfer-Length: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", 
                xmlTextReaderDepth(reader) * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vFileAttributes[] = 
{
    {
        .id = ENATSC3ROUTESLSSTSIDFileNodeAttribute_TOI,
        .sName = sizeof("TOI") - 1,
        .pName = "TOI",
        .handler = FileNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDFileNodeAttribute_ContentLocation,
        .sName = sizeof("Content-Location") - 1,
        .pName = "Content-Location",
        .handler = FileNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDFileNodeAttribute_ContentType,
        .sName = sizeof("Content-Type") - 1,
        .pName = "Content-Type",
        .handler = FileNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDFileNodeAttribute_ContentEncoding,
        .sName = sizeof("Content-Encoding") - 1,
        .pName = "Content-Encoding",
        .handler = FileNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDFileNodeAttribute_ContentLength,
        .sName = sizeof("Content-Length") - 1,
        .pName = "Content-Length",
        .handler = FileNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDFileNodeAttribute_TransferLength,
        .sName = sizeof("Transfer-Length") - 1,
        .pName = "Transfer-Length",
        .handler = FileNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vFileChildNodes[] =
{
};
static EnXMLParserNode fileNodeData = 
{
    .pName = "File",
    .pAttributes = vFileAttributes,
    .nAttributes = numof(vFileAttributes),
    .pChildNodes = vFileChildNodes,
    .nChildNodes = numof(vFileChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessFDTFileNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDNode *pSTSIDNode = (ENATSC3ROUTESLSSTSIDNode *)pUserData;
    ENATSC3ROUTESLSSTSIDFDTParametersNode *pFDTParametersNode;
    ENATSC3ROUTESLSSTSIDFileNode fileNode;
    int32_t fileNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_FDTParameters);
    pFDTParametersNode = (ENATSC3ROUTESLSSTSIDFDTParametersNode *)pParentNode;

    memset(&fileNode, 0, sizeof(fileNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pSTSIDNode->dataBuffer, 
        sizeof(fileNode), &fileNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for file node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pFDTParametersNode->files,
        fileNodeOffset))
    {
        ERRORPRINT("Failed to store file node offset\n");
        rval = 1;
        goto out;
    }

    VERBOSEPRINT("%*sFile\n", xmlTextReaderDepth(reader) * 2, "");

    // Since the data buffer could be reallocated, we are using a temporary
    // ratingRegionTableNode that lives on the stack. Once we are done
    // processing the node it is committed to the data buffer.
    fileNode.base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_File;
    rval = EnXMLParserProcessNode(&fileNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)&fileNode);

    memcpy(pSTSIDNode->dataBuffer.pBuffer + fileNodeOffset, 
        &fileNode, sizeof(fileNode));

out:
    return rval;
}

static EnXMLParserAttribute vFDTParametersAttributes[] = 
{
};
static EnXMLParserChildNode vFDTParametersChildNodes[] =
{
    {
        .id = ENATSC3ROUTESLSSTSIDFDTParametersChildNode_FDTFile,
        .sName = sizeof("File") - 1,
        .pName = "File",
        .handler = ProcessFDTFileNode,
    },
};
static EnXMLParserNode fdtParametersNodeData = 
{
    .pName = "FDTParameters",
    .pAttributes = vFDTParametersAttributes,
    .nAttributes = numof(vFDTParametersAttributes),
    .pChildNodes = vFDTParametersChildNodes,
    .nChildNodes = numof(vFDTParametersChildNodes),
    .textHandler = NULL,
};

static int32_t
FileTemplateNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue,
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData,
    EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDNode *pSTSIDNode = (ENATSC3ROUTESLSSTSIDNode *)pUserData;
    ENATSC3ROUTESLSSTSIDFileTemplateNode *pFileTemplateNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_FileTemplate);

    pFileTemplateNode = (ENATSC3ROUTESLSSTSIDFileTemplateNode *)pParentNode;

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(&pSTSIDNode->dataBuffer, 
        (char *)pValue, sValue, &pFileTemplateNode->fileTemplateDataOffset))
    {
        ERRORPRINT("Failed to store fileTemplate\n");
        rval = -1;
        goto out;
    }

    VERBOSEPRINT("%*sfileTemplate: %.*s\n", 
        xmlTextReaderDepth(reader) * 2, "", sValue, pValue);

out:
    return rval;
}

static EnXMLParserAttribute vFileTemplateAttributes[] = 
{
};
static EnXMLParserChildNode vFileTemplateChildNodes[] =
{
};
static EnXMLParserNode fileTemplateNodeData = 
{
    .pName = "FileTemplate",
    .pAttributes = vFileTemplateAttributes,
    .nAttributes = numof(vFileTemplateAttributes),
    .pChildNodes = vFileTemplateChildNodes,
    .nChildNodes = numof(vFileTemplateChildNodes),
    .textHandler = FileTemplateNodeTextHandler,
};

static int32_t
FDTInstanceNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDNode *pSTSIDNode = (ENATSC3ROUTESLSSTSIDNode *)pUserData;
    ENATSC3ROUTESLSSTSIDFDTInstanceNode *pFDTInstanceNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_FDTInstance);

    pFDTInstanceNode = (ENATSC3ROUTESLSSTSIDFDTInstanceNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ENATSC3ROUTESLSSTSIDFDTInstanceNodeAttribute_efdtVersion:
            if (EnXMLParser_antol((char *)pValue, sValue, 
                &pFDTInstanceNode->efdtVersion)) 
            {
                ERRORPRINT("Failed to parse efdtVersion %.*s\n", 
                    sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sefdtVersion: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDFDTInstanceNodeAttribute_expires:
            if (EnXMLParser_antoll((char *)pValue, sValue, 
                &pFDTInstanceNode->expiresSecondsUTC)) 
            {
                ERRORPRINT("Failed to parse Expires %.*s\n", sValue, pValue);
                rval = -1;
                goto out;
            }
            if (EnXMLParser_ntpSecondsToSecondsUTC(
                pFDTInstanceNode->expiresSecondsUTC,
                &pFDTInstanceNode->expiresSecondsUTC))
            {
                ERRORPRINT("Failed to convert ntp seconds to utc\n");
                rval =-1;
                goto out;
            }
            VERBOSEPRINT("%*sExpires: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDFDTInstanceNodeAttribute_fileTemplate:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSTSIDNode->dataBuffer, (char *)pValue, sValue, 
                &pFDTInstanceNode->fileTemplateDataOffset))
            {
                ERRORPRINT("Failed to store fileTemplate\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sfileTemplate: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", 
                xmlTextReaderDepth(reader) * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static int32_t
ProcessFileNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDNode *pSTSIDNode = (ENATSC3ROUTESLSSTSIDNode *)pUserData;
    ENATSC3ROUTESLSSTSIDFDTInstanceNode *pFDTInstanceNode;
    ENATSC3ROUTESLSSTSIDFileNode fileNode;
    int32_t fileNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_FDTInstance);
    pFDTInstanceNode = (ENATSC3ROUTESLSSTSIDFDTInstanceNode *)pParentNode;

    memset(&fileNode, 0, sizeof(fileNodeOffset));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pSTSIDNode->dataBuffer, 
        sizeof(fileNode), &fileNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for file node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pFDTInstanceNode->files, 
        fileNodeOffset))
    {
        ERRORPRINT("Failed to store files offset\n");
        rval = 1;
        goto out;
    }

    VERBOSEPRINT("%*sFile\n", xmlTextReaderDepth(reader) * 2, "");

    // Since the data buffer could be reallocated, we are using a temporary
    // ratingValueAbbrevNode that lives on the stack. Once we are done
    // processing the node it is committed to the data buffer.
    fileNode.base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_File;
    rval = EnXMLParserProcessNode(&fileNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)&fileNode);

    memcpy(pSTSIDNode->dataBuffer.pBuffer + fileNodeOffset, &fileNode,
        sizeof(fileNode));

out:
    return rval;
}

static EnXMLParserAttribute vFDTInstanceAttributes[] = 
{
    {
        .id = ENATSC3ROUTESLSSTSIDFDTInstanceNodeAttribute_efdtVersion,
        .sName = sizeof("efdtVersion") - 1,
        .pName = "efdtVersion",
        .handler = FDTInstanceNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDFDTInstanceNodeAttribute_expires,
        .sName = sizeof("Expires") - 1,
        .pName = "Expires",
        .handler = FDTInstanceNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDFDTInstanceNodeAttribute_fileTemplate,
        .sName = sizeof("fileTemplate") - 1,
        .pName = "fileTemplate",
        .handler = FDTInstanceNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vFDTInstanceChildNodes[] =
{
    {
        .id = ENATSC3ROUTESLSSTSIDFDTInstanceChildNode_File,
        .sName = sizeof("File") - 1,
        .pName = "File",
        .handler = ProcessFileNode,
    },
};
static EnXMLParserNode fdtInstanceNodeData = 
{
    .pName = "FDTInstance",
    .pAttributes = vFDTInstanceAttributes,
    .nAttributes = numof(vFDTInstanceAttributes),
    .pChildNodes = vFDTInstanceChildNodes,
    .nChildNodes = numof(vFDTInstanceChildNodes),
    .textHandler = NULL,
};

static int32_t
EFDTNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDEFDTNode *pEFDTNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_EFDT);

    pEFDTNode = (ENATSC3ROUTESLSSTSIDEFDTNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ENATSC3ROUTESLSSTSIDEFDTNodeAttribute_version:
            if (EnXMLParser_antol((char *)pValue, sValue, &pEFDTNode->version)) 
            {
                ERRORPRINT("Failed to parse version %.*s\n", sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sversion: %.*s\n", xmlTextReaderDepth(reader) * 2, 
                "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", 
                xmlTextReaderDepth(reader) * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static int32_t
ProcessFileTemplateNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDEFDTNode *pEFDTNode;
    ENATSC3ROUTESLSSTSIDFileTemplateNode *pFileTemplateNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_EFDT);

    pEFDTNode = (ENATSC3ROUTESLSSTSIDEFDTNode *)pParentNode;
    pFileTemplateNode = &pEFDTNode->fileTemplate;
    pFileTemplateNode->base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_FileTemplate;

    VERBOSEPRINT("%*sFileTemplate\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&fileTemplateNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)pFileTemplateNode);

    return rval;
}

static int32_t
ProcessFDTParametersNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDNode *pSTSIDNode = (ENATSC3ROUTESLSSTSIDNode *)pUserData;
    ENATSC3ROUTESLSSTSIDEFDTNode *pEFDTNode;
    ENATSC3ROUTESLSSTSIDFDTParametersNode *pFDTParametersNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_EFDT);

    pEFDTNode = (ENATSC3ROUTESLSSTSIDEFDTNode *)pParentNode;
    pFDTParametersNode = &pEFDTNode->fdtParameters;
    pFDTParametersNode->base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_FDTParameters;

    if (EnXMLParserDataOffsetArrayInitialize(&pFDTParametersNode->files,
        &pSTSIDNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate titles array\n");
        rval = -1;
        goto out;
    }

    VERBOSEPRINT("%*sFDTParameters\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&fdtParametersNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)pFDTParametersNode);

out:
    return rval;
}

static int32_t
ProcessFDTInstanceNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDNode *pSTSIDNode = (ENATSC3ROUTESLSSTSIDNode *)pUserData;
    ENATSC3ROUTESLSSTSIDEFDTNode *pEFDTNode;
    ENATSC3ROUTESLSSTSIDFDTInstanceNode *pFDTInstanceNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_EFDT);

    pEFDTNode = (ENATSC3ROUTESLSSTSIDEFDTNode *)pParentNode;
    pFDTInstanceNode = &pEFDTNode->fdtInstance;

    pFDTInstanceNode->base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_FDTInstance;

    if (EnXMLParserDataOffsetArrayInitialize(&pFDTInstanceNode->files,
        &pSTSIDNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate files array\n");
        rval = -1;
        goto out;
    }

    VERBOSEPRINT("%*sFDT-Instance\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&fdtInstanceNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)pFDTInstanceNode);

out:
    return rval;
}

static EnXMLParserAttribute vEFDTAttributes[] = 
{
    {
        .id = ENATSC3ROUTESLSSTSIDEFDTNodeAttribute_version,
        .sName = sizeof("version") - 1,
        .pName = "version",
        .handler = EFDTNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vEFDTChildNodes[] =
{
    {
        .id = ENATSC3ROUTESLSSTSIDEFDTChildNode_FDTInstance,
        .sName = sizeof("FDT-Instance") - 1,
        .pName = "FDT-Instance",
        .handler = ProcessFDTInstanceNode,
    },
    // The following are in the Sinclair CES broadcast but are not to spec
    {
        .id = ENATSC3ROUTESLSSTSIDEFDTChildNode_FDTParameters,
        .sName = sizeof("FDTParameters") - 1,
        .pName = "FDTParameters",
        .handler = ProcessFDTParametersNode,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDEFDTChildNode_FileTemplate,
        .sName = sizeof("FileTemplate") - 1,
        .pName = "FileTemplate",
        .handler = ProcessFileTemplateNode,
    },
};
static EnXMLParserNode efdtNodeData = 
{
    .pName = "EFDT",
    .pAttributes = vEFDTAttributes,
    .nAttributes = numof(vEFDTAttributes),
    .pChildNodes = vEFDTChildNodes,
    .nChildNodes = numof(vEFDTChildNodes),
    .textHandler = NULL,
};

static int32_t
SrcFlowNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDSrcFlowNode *pSrcFlowNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_SrcFlow);

    pSrcFlowNode = (ENATSC3ROUTESLSSTSIDSrcFlowNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ENATSC3ROUTESLSSTSIDSrcFlowNodeAttribute_rt:
            if (EnXMLParserStringToBool((char *)pValue, sValue, 
                &pSrcFlowNode->rt))
            {
                ERRORPRINT("Failed to parse rt value %.*s\n", sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*srt: %.*s\n", xmlTextReaderDepth(reader) * 2, "", 
                sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDSrcFlowNodeAttribute_minBuffSize:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pSrcFlowNode->minBuffSize))
            {
                ERRORPRINT("Failed to parse minBuffSize %.*s\n", 
                    sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sminBuffSize: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", 
                xmlTextReaderDepth(reader) * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static int32_t
ProcessSrcFlowEFDTNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDSrcFlowNode *pSrcFlowNode;
    ENATSC3ROUTESLSSTSIDEFDTNode *pEFDTNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_SrcFlow);

    pSrcFlowNode = (ENATSC3ROUTESLSSTSIDSrcFlowNode *)pParentNode;
    pEFDTNode = &pSrcFlowNode->efdt;
    pEFDTNode->base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_EFDT;

    VERBOSEPRINT("%*sEFDT\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&efdtNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)pEFDTNode);

    return rval;
}

static int32_t
ProcessSrcFlowContentInfoNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDSrcFlowNode *pSrcFlowNode;
    ENATSC3ROUTESLSSTSIDContentInfoNode *pContentInfoNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_SrcFlow);

    pSrcFlowNode = (ENATSC3ROUTESLSSTSIDSrcFlowNode *)pParentNode;
    pContentInfoNode = &pSrcFlowNode->contentInfo;
    pContentInfoNode->base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_ContentInfo;

    VERBOSEPRINT("%*sContentInfo\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&contentInfoNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)pContentInfoNode);

    return rval;
}

static int32_t
ProcessSrcFlowPayloadNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDSrcFlowNode *pSrcFlowNode;
    ENATSC3ROUTESLSSTSIDPayloadNode *pPayloadNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_SrcFlow);

    pSrcFlowNode = (ENATSC3ROUTESLSSTSIDSrcFlowNode *)pParentNode;
    pPayloadNode = &pSrcFlowNode->payload;
    pPayloadNode->base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_Payload;

    VERBOSEPRINT("%*sPayload\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&payloadNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)pPayloadNode);

    return rval;
}

static EnXMLParserAttribute vSrcFlowAttributes[] = 
{
    {
        .id = ENATSC3ROUTESLSSTSIDSrcFlowNodeAttribute_rt,
        .sName = sizeof("rt") - 1,
        .pName = "rt",
        .handler = SrcFlowNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vSrcFlowChildNodes[] =
{
    {
        .id = ENATSC3ROUTESLSSTSIDSrcFlowChildNode_EFDT,
        .sName = sizeof("EFDT") - 1,
        .pName = "EFDT",
        .handler = ProcessSrcFlowEFDTNode,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDSrcFlowChildNode_ContentInfo,
        .sName = sizeof("ContentInfo") - 1,
        .pName = "ContentInfo",
        .handler = ProcessSrcFlowContentInfoNode,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDSrcFlowChildNode_Payload,
        .sName = sizeof("Payload") - 1,
        .pName = "Payload",
        .handler = ProcessSrcFlowPayloadNode,
    },
};
static EnXMLParserNode srcFlowNodeData = 
{
    .pName = "SrcFlow",
    .pAttributes = vSrcFlowAttributes,
    .nAttributes = numof(vSrcFlowAttributes),
    .pChildNodes = vSrcFlowChildNodes,
    .nChildNodes = numof(vSrcFlowChildNodes),
    .textHandler = NULL,
};

static int32_t
RepairFlowNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDRepairFlowNode *pRepairFlowNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_RepairFlow);

    pRepairFlowNode = (ENATSC3ROUTESLSSTSIDRepairFlowNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ENATSC3ROUTESLSSTSIDRepairFlowNodeAttribute_rt:
            if (EnXMLParserStringToBool((char *)pValue, sValue, 
                &pRepairFlowNode->rt))
            {
                ERRORPRINT("Failed to parse rt value %.*s\n", sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*srt: %.*s\n", xmlTextReaderDepth(reader) * 2, "",
                sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDRepairFlowNodeAttribute_minBuffSize:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pRepairFlowNode->minBuffSize))
            {
                ERRORPRINT("Failed to parse minBuffSize %.*s\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sminBuffSize: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", 
                xmlTextReaderDepth(reader) * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static int32_t
ProcessRepairFlowEFDTNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDRepairFlowNode *pRepairFlowNode;
    ENATSC3ROUTESLSSTSIDEFDTNode *pEFDTNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_RepairFlow);

    pRepairFlowNode = (ENATSC3ROUTESLSSTSIDRepairFlowNode *)pParentNode;
    pEFDTNode = &pRepairFlowNode->efdt;
    pEFDTNode->base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_EFDT;

    VERBOSEPRINT("%*sEFDT\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&efdtNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)pEFDTNode);

    return rval;
}

static int32_t
ProcessRepairFlowContentInfoNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDRepairFlowNode *pRepairFlowNode;
    ENATSC3ROUTESLSSTSIDContentInfoNode *pContentInfoNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_RepairFlow);

    pRepairFlowNode = (ENATSC3ROUTESLSSTSIDRepairFlowNode *)pParentNode;
    pContentInfoNode = &pRepairFlowNode->contentInfo;
    pContentInfoNode->base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_ContentInfo;

    VERBOSEPRINT("%*sContentInfo\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&contentInfoNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)pContentInfoNode);

    return rval;
}

static int32_t
ProcessRepairFlowPayloadNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDRepairFlowNode *pRepairFlowNode;
    ENATSC3ROUTESLSSTSIDPayloadNode *pPayloadNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_RepairFlow);

    pRepairFlowNode = (ENATSC3ROUTESLSSTSIDRepairFlowNode *)pParentNode;
    pPayloadNode = &pRepairFlowNode->payload;
    pPayloadNode->base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_Payload;

    VERBOSEPRINT("%*sPayload\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&payloadNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)pPayloadNode);

    return rval;
}

static EnXMLParserAttribute vRepairFlowAttributes[] = 
{
    {
        .id = ENATSC3ROUTESLSSTSIDRepairFlowNodeAttribute_rt,
        .sName = sizeof("rt") - 1,
        .pName = "rt",
        .handler = RepairFlowNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vRepairFlowChildNodes[] =
{
    {
        .id = ENATSC3ROUTESLSSTSIDRepairFlowChildNode_EFDT,
        .sName = sizeof("EFDT") - 1,
        .pName = "EFDT",
        .handler = ProcessRepairFlowEFDTNode,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDRepairFlowChildNode_ContentInfo,
        .sName = sizeof("ContentInfo") - 1,
        .pName = "ContentInfo",
        .handler = ProcessRepairFlowContentInfoNode,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDRepairFlowChildNode_Payload,
        .sName = sizeof("Payload") - 1,
        .pName = "Payload",
        .handler = ProcessRepairFlowPayloadNode,
    },
};
static EnXMLParserNode repairFlowNodeData = 
{
    .pName = "RepairFlow",
    .pAttributes = vRepairFlowAttributes,
    .nAttributes = numof(vRepairFlowAttributes),
    .pChildNodes = vRepairFlowChildNodes,
    .nChildNodes = numof(vRepairFlowChildNodes),
    .textHandler = NULL,
};

static int32_t
LSNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDLSNode *pLSNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_LS);

    pLSNode = (ENATSC3ROUTESLSSTSIDLSNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ENATSC3ROUTESLSSTSIDLSNodeAttribute_tsi:
            if (EnXMLParser_antoul((char *)pValue, sValue, &pLSNode->tsi))
            {
                ERRORPRINT("Failed to parse tsi %.*s\n", sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*stsi: %.*s\n", xmlTextReaderDepth(reader) * 2, "",
                sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDLSNodeAttribute_bw:
            if (EnXMLParser_antoul((char *)pValue, sValue, &pLSNode->bw))
            {
                ERRORPRINT("Failed to parse bw %.*s\n", sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sbw: %.*s\n", xmlTextReaderDepth(reader) * 2, "", 
                sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDLSNodeAttribute_startTime:
            if (EnXMLParserZTimeToMilliseconds((char *)pValue, sValue, 
                &pLSNode->startTimeSecondsUTC))
            {
                ERRORPRINT("Failed to parse dateTime \"%.*s\"\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sstartTime: %.*s\n", 
                xmlTextReaderDepth(reader) * 2, "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDLSNodeAttribute_endTime:
            if (EnXMLParserZTimeToMilliseconds((char *)pValue, sValue, 
                &pLSNode->endTimeSecondsUTC))
            {
                ERRORPRINT("Failed to parse dateTime \"%.*s\"\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sendTime: %.*s\n", xmlTextReaderDepth(reader) * 2,
                "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", 
                xmlTextReaderDepth(reader) * 2, "", 
                pAttribute->id, pAttribute->pName);
            goto out;
    }

out:
    return rval;
}

static int32_t
ProcessSrcFlowNode(EnXMLParserChildNode *pChildNode, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDLSNode *pLSNode;
    ENATSC3ROUTESLSSTSIDSrcFlowNode *pSrcFlowNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_LS);

    pLSNode = (ENATSC3ROUTESLSSTSIDLSNode *)pParentNode;
    pSrcFlowNode = &pLSNode->srcFlow;
    pSrcFlowNode->base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_SrcFlow;

    VERBOSEPRINT("%*sSrcFlow\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&srcFlowNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)pSrcFlowNode);

    return rval;
}

static int32_t
ProcessRepairFlowNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDLSNode *pLSNode;
    ENATSC3ROUTESLSSTSIDRepairFlowNode *pRepairFlowNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_LS);

    pLSNode = (ENATSC3ROUTESLSSTSIDLSNode *)pParentNode;
    pRepairFlowNode = &pLSNode->repairFlow;
    pRepairFlowNode->base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_RepairFlow;

    VERBOSEPRINT("%*sRepairFlow\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&repairFlowNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)pRepairFlowNode);

    return rval;
}

static EnXMLParserAttribute vLSAttributes[] = 
{
    {
        .id = ENATSC3ROUTESLSSTSIDLSNodeAttribute_tsi,
        .sName = sizeof("tsi") - 1,
        .pName = "tsi",
        .handler = LSNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDLSNodeAttribute_bw,
        .sName = sizeof("bw") - 1,
        .pName = "bw",
        .handler = LSNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDLSNodeAttribute_startTime,
        .sName = sizeof("startTime") - 1,
        .pName = "startTime",
        .handler = LSNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDLSNodeAttribute_endTime,
        .sName = sizeof("endTime") - 1,
        .pName = "endTime",
        .handler = LSNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vLSChildNodes[] =
{
    {
        .id = ENATSC3ROUTESLSSTSIDLSChildNode_SrcFlow,
        .sName = sizeof("SrcFlow") - 1,
        .pName = "SrcFlow",
        .handler = ProcessSrcFlowNode,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDLSChildNode_RepairFlow,
        .sName = sizeof("RepairFlow") - 1,
        .pName = "RepairFlow",
        .handler = ProcessRepairFlowNode,
    },
};
static EnXMLParserNode lsNodeData = 
{
    .pName = "LS",
    .pAttributes = vLSAttributes,
    .nAttributes = numof(vLSAttributes),
    .pChildNodes = vLSChildNodes,
    .nChildNodes = numof(vLSChildNodes),
    .textHandler = NULL,
};

static int32_t
RSNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDNode *pSTSIDNode = (ENATSC3ROUTESLSSTSIDNode *)pUserData;
    ENATSC3ROUTESLSSTSIDRSNode *pRSNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_RS);

    pRSNode = (ENATSC3ROUTESLSSTSIDRSNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ENATSC3ROUTESLSSTSIDRSNodeAttribute_sIpAddr:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSTSIDNode->dataBuffer, (char *)pValue, sValue, 
                &pRSNode->sIpAddrDataOffset))
            {
                ERRORPRINT("Failed to store sIpAddr\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*ssIpAddr: %.*s\n", xmlTextReaderDepth(reader) * 2,
                "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDRSNodeAttribute_dIpAddr:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSTSIDNode->dataBuffer, (char *)pValue, sValue, 
                &pRSNode->dIpAddrDataOffset))
            {
                ERRORPRINT("Failed to store dIpAddr\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sdIpAddr: %.*s\n", xmlTextReaderDepth(reader) * 2,
                "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSSTSIDRSNodeAttribute_dPort:
            if (EnXMLParser_antoul((char *)pValue, sValue, &pRSNode->dPort))
            {
                ERRORPRINT("Failed to parse dPort %.*s\n", sValue, pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sdPort: %.*s\n", xmlTextReaderDepth(reader) * 2,
                "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", 
                xmlTextReaderDepth(reader) * 2, "", 
                pAttribute->id, pAttribute->pName);
            goto out;
    }

out:
    return rval;
}

static int32_t
ProcessLSNode(EnXMLParserChildNode *pChildNode, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDNode *pSTSIDNode = (ENATSC3ROUTESLSSTSIDNode *)pUserData;
    ENATSC3ROUTESLSSTSIDRSNode *pRSNode;
    ENATSC3ROUTESLSSTSIDLSNode LSNode;
    int32_t LSNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_RS);
    pRSNode = (ENATSC3ROUTESLSSTSIDRSNode *)pParentNode;

    memset(&LSNode, 0, sizeof(LSNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pSTSIDNode->dataBuffer, 
        sizeof(LSNode), &LSNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for LS node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pRSNode->LSs, LSNodeOffset))
    {
        ERRORPRINT("Failed to store LS offset\n");
        rval = 1;
        goto out;
    }

    VERBOSEPRINT("%*sLS\n", xmlTextReaderDepth(reader) * 2, "");

    LSNode.base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_LS;
    rval = EnXMLParserProcessNode(&lsNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)&LSNode);

    memcpy(pSTSIDNode->dataBuffer.pBuffer + LSNodeOffset, 
        &LSNode, sizeof(LSNode));

out:
    return rval;
}

static EnXMLParserAttribute vRSAttributes[] = 
{
    {
        .id = ENATSC3ROUTESLSSTSIDRSNodeAttribute_sIpAddr,
        .sName = sizeof("sIpAddr") - 1,
        .pName = "sIpAddr",
        .handler = RSNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDRSNodeAttribute_dIpAddr,
        .sName = sizeof("dIpAddr") - 1,
        .pName = "dIpAddr",
        .handler = RSNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSSTSIDRSNodeAttribute_dPort,
        .sName = sizeof("dPort") - 1,
        .pName = "dPort",
        .handler = RSNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vRSChildNodes[] =
{
    {
        .id = ENATSC3ROUTESLSSTSIDRSChildNode_LS,
        .sName = sizeof("LS") - 1,
        .pName = "LS",
        .handler = ProcessLSNode,
    },
};
static EnXMLParserNode rsNodeData = 
{
    .pName = "RS",
    .pAttributes = vRSAttributes,
    .nAttributes = numof(vRSAttributes),
    .pChildNodes = vRSChildNodes,
    .nChildNodes = numof(vRSChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessRSNode(EnXMLParserChildNode *pChildNode, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSSTSIDNode *pSTSIDNode;
    ENATSC3ROUTESLSSTSIDRSNode RSNode;
    int32_t RSNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_STSID);
    pSTSIDNode = (ENATSC3ROUTESLSSTSIDNode *)pParentNode;

    memset(&RSNode, 0, sizeof(RSNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pSTSIDNode->dataBuffer,
        sizeof(RSNode), &RSNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for RS node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pSTSIDNode->RSs, RSNodeOffset))
    {
        ERRORPRINT("Failed to store RS node offset\n");
        rval = 1;
        goto out;
    }

    VERBOSEPRINT("%*sRS\n", xmlTextReaderDepth(reader) * 2, "");

    // Since the data buffer could be reallocated, we are using a temporary
    // ratingRegionTableNode that lives on the stack. Once we are done
    // processing the node it is committed to the data buffer.
    if (EnXMLParserDataOffsetArrayInitialize(&RSNode.LSs, 
        &pSTSIDNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate LSs array\n");
        rval = -1;
        goto out;
    }

    RSNode.base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_RS;
    rval = EnXMLParserProcessNode(&rsNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)&RSNode);

    memcpy(pSTSIDNode->dataBuffer.pBuffer + RSNodeOffset, &RSNode,
        sizeof(RSNode));

out:
    return rval;
}

static int32_t
STSIDNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSSTSIDNodeType_STSID);

    switch (pAttribute->id)
    {
        case ENATSC3ROUTESLSSTSIDNodeAttribute_XMLNS:
            VERBOSEPRINT("%*sxmlns: %.*s\n", xmlTextReaderDepth(reader) * 2,
                "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", 
                xmlTextReaderDepth(reader) * 2, "",
                pAttribute->id, pAttribute->pName);
            goto out;
    }

out:
    return rval;
}

static EnXMLParserAttribute vSTSIDAttributes[] = 
{
    {
        .id = ENATSC3ROUTESLSSTSIDNodeAttribute_XMLNS,
        .sName = sizeof("xmlns") - 1,
        .pName = "xmlns",
        .handler = STSIDNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vSTSIDChildNodes[] =
{
    {
        .id = ENATSC3ROUTESLSSTSIDChildNode_RS,
        .sName = sizeof("RS") - 1,
        .pName = "RS",
        .handler = ProcessRSNode,
    },
};
static EnXMLParserNode stsidNodeData = 
{
    .pName = "STSID",
    .pAttributes = vSTSIDAttributes,
    .nAttributes = numof(vSTSIDAttributes),
    .pChildNodes = vSTSIDChildNodes,
    .nChildNodes = numof(vSTSIDChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessSTSIDNode(xmlTextReaderPtr reader, ENATSC3ROUTESLSSTSIDNode *pSTSIDNode)
{
    int32_t rval = 0;

    VERBOSEPRINT("%*sSTSID\n", xmlTextReaderDepth(reader) * 2, "");

    if (EnXMLParserDataOffsetArrayInitialize(
        &pSTSIDNode->RSs, &pSTSIDNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate RS array\n");
        rval = -1;
        goto out;
    }

    rval = EnXMLParserProcessNode(&stsidNodeData, reader, pSTSIDNode, 
        (EnXMLParserNodeBase *)pSTSIDNode);

out:
    return rval;
}

ENATSC3Status
ENATSC3ROUTESLSSTSIDParse(const char *pXML, int32_t sXML, 
    ENATSC3ROUTESLSSTSIDNode *pParsedSTSID)
{
    xmlTextReaderPtr reader = NULL;
    int xmlret;
    int32_t fHaveDataBuffer = 0;
    ENATSC3Status rval = ENATSC3Status_Success;

    memset(pParsedSTSID, 0, sizeof(ENATSC3ROUTESLSSTSIDNode));
    reader = xmlReaderForMemory(pXML, sXML, NULL, NULL, 0);

    if (!reader)
    {
        ERRORPRINT("Failed to create xml reader\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    if (EnXMLParserDataBufferInitialize(&pParsedSTSID->dataBuffer, 4096, 4096))
    {
        ERRORPRINT("Failed to initialize MPD data buffer\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }
    fHaveDataBuffer = 1;

    xmlret = xmlTextReaderRead(reader);
    if (xmlret == 1)
    {
        int nodeType;
        const xmlChar *pLocalName;

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);

        if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "S-TSID"))
        {
            pParsedSTSID->base.nodeType = ENATSC3ROUTESLSSTSIDNodeType_STSID;
            if (ProcessSTSIDNode(reader, pParsedSTSID))
            {
                ERRORPRINT("Failed to process STSID\n");
                rval = ENATSC3Status_Failure;
                goto errout;
            }
        }
        else
        {
            ERRORPRINT("Top-level object isn't STSID (%d/%s)\n",
                nodeType, pLocalName);
            rval = ENATSC3Status_Failure;
            goto errout;
        }
    }
    else
    {
        ERRORPRINT("Couldn't find STSID\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }

out:
    if (reader)
    {
        xmlFreeTextReader(reader);
    }

    return rval;

errout:
    if (fHaveDataBuffer)
    {
        if (EnXMLParserDataBufferFree(&pParsedSTSID->dataBuffer))
        {
            ERRORPRINT("Failed to free STSID data buffer\n");
        }
    }

    goto out;
}

ENATSC3Status
ENATSC3ROUTESLSSTSIDFree(ENATSC3ROUTESLSSTSIDNode *pParsedSTSID)
{
    ENATSC3Status rval = ENATSC3Status_Success;

    if (EnXMLParserDataBufferFree(&pParsedSTSID->dataBuffer)) 
    { 
        ERRORPRINT("Failed to free STSID data buffer\n"); 
    }

    return rval;
}

