/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_defs.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3utils/enxml.h>

#include "atsc3sls_usbd.h"

static char *ensys = "sls";
static char *ensubsys = "usbd";

static int32_t fVerbose = 0;

#define VERBOSEPRINT(x, ...) \
    if (fVerbose) \
    { \
        LOGPRINT(x, ##__VA_ARGS__); \
    }

static int32_t
BasePatternNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue,
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData,
    EnXMLParserNodeBase *pParentNode)
{
    ENATSC3SLSUSBDBundleDescriptionNode *pUSBDNode = 
        (ENATSC3SLSUSBDBundleDescriptionNode *)pUserData;
    ENATSC3SLSUSBDBasePatternNode *pBasePatternNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3SLSUSBDNodeType_BasePattern);

    pBasePatternNode = (ENATSC3SLSUSBDBasePatternNode *)pParentNode;

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(&pUSBDNode->dataBuffer, 
        (char *)pValue, sValue, &pBasePatternNode->basePatternDataOffset))
    {
        ERRORPRINT("Failed to store basePattern\n");
        rval = -1;
        goto out;
    }
    VERBOSEPRINT("%*sbasePattern: %.*s\n", xmlTextReaderDepth(reader) * 2, "", 
        sValue, pValue);

out:
    return rval;
}

static EnXMLParserAttribute vBasePatternAttributes[] = 
{
};
static EnXMLParserChildNode vBasePatternChildNodes[] =
{
};
static EnXMLParserNode basePatternNodeData = 
{
    .pName = "basePattern",
    .pAttributes = vBasePatternAttributes,
    .nAttributes = numof(vBasePatternAttributes),
    .pChildNodes = vBasePatternChildNodes,
    .nChildNodes = numof(vBasePatternChildNodes),
    .textHandler = BasePatternNodeTextHandler,
};

static int32_t
ProcessBroadcastAppServiceBasePatternNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3SLSUSBDBundleDescriptionNode *pUSBDNode =
        (ENATSC3SLSUSBDBundleDescriptionNode *)pUserData;
    ENATSC3SLSUSBDBroadcastAppServiceNode *pBroadcastAppServiceNode;
    ENATSC3SLSUSBDBasePatternNode basePatternNode;
    int32_t basePatternNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3SLSUSBDNodeType_BroadcastAppService);
    pBroadcastAppServiceNode = 
        (ENATSC3SLSUSBDBroadcastAppServiceNode *)pParentNode;

    memset(&basePatternNode, 0, sizeof(basePatternNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pUSBDNode->dataBuffer,
        sizeof(basePatternNode), &basePatternNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for basePattern node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pBroadcastAppServiceNode->basePatterns, basePatternNodeOffset))
    {
        ERRORPRINT("Failed to store basePattern node offset\n");
        rval = 1;
        goto out;
    }

    basePatternNode.base.nodeType = ENATSC3SLSUSBDNodeType_BasePattern;

    VERBOSEPRINT("%*sbasePattern\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&basePatternNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)&basePatternNode);

    memcpy(pUSBDNode->dataBuffer.pBuffer + basePatternNodeOffset, 
        &basePatternNode, sizeof(basePatternNode));

out:
    return rval;
}

static EnXMLParserAttribute vBroadcastAppServiceAttributes[] = 
{
};
static EnXMLParserChildNode vBroadcastAppServiceChildNodes[] =
{
    {
        .id = ENATSC3SLSUSBDBroadcastAppServiceNodeAttribute_basePattern,
        .sName = sizeof("basePattern") - 1,
        .pName = "basePattern",
        .handler = ProcessBroadcastAppServiceBasePatternNode,
    },
};
static EnXMLParserNode broadcastAppServiceNodeData = 
{
    .pName = "basePattern",
    .pAttributes = vBroadcastAppServiceAttributes,
    .nAttributes = numof(vBroadcastAppServiceAttributes),
    .pChildNodes = vBroadcastAppServiceChildNodes,
    .nChildNodes = numof(vBroadcastAppServiceChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessUnicastAppServiceBasePatternNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3SLSUSBDBundleDescriptionNode *pUSBDNode =
        (ENATSC3SLSUSBDBundleDescriptionNode *)pUserData;
    ENATSC3SLSUSBDUnicastAppServiceNode *pUnicastAppServiceNode;
    ENATSC3SLSUSBDBasePatternNode basePatternNode;
    int32_t basePatternNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3SLSUSBDNodeType_UnicastAppService);
    pUnicastAppServiceNode = 
        (ENATSC3SLSUSBDUnicastAppServiceNode *)pParentNode;

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pUSBDNode->dataBuffer, 
        sizeof(basePatternNode), &basePatternNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for basePattern node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pUnicastAppServiceNode->basePatterns, basePatternNodeOffset))
    {
        ERRORPRINT("Failed to store basePattern offset\n");
        rval = 1;
        goto out;
    }

    basePatternNode.base.nodeType = ENATSC3SLSUSBDNodeType_BasePattern;

    VERBOSEPRINT("%*sbasePattern\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&basePatternNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)&basePatternNode);

    memcpy(pUSBDNode->dataBuffer.pBuffer + basePatternNodeOffset, 
        &basePatternNode, sizeof(basePatternNode));

out:
    return rval;
}

static EnXMLParserAttribute vUnicastAppServiceAttributes[] = 
{
};
static EnXMLParserChildNode vUnicastAppServiceChildNodes[] =
{
    {
        .id = ENATSC3SLSUSBDUnicastAppServiceNodeAttribute_basePattern,
        .sName = sizeof("basePattern") - 1,
        .pName = "basePattern",
        .handler = ProcessUnicastAppServiceBasePatternNode,
    },
};
static EnXMLParserNode unicastAppServiceNodeData = 
{
    .pName = "basePattern",
    .pAttributes = vUnicastAppServiceAttributes,
    .nAttributes = numof(vUnicastAppServiceAttributes),
    .pChildNodes = vUnicastAppServiceChildNodes,
    .nChildNodes = numof(vUnicastAppServiceChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessBroadcastAppServiceNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3SLSUSBDBundleDescriptionNode *pUSBDNode =
        (ENATSC3SLSUSBDBundleDescriptionNode *)pUserData;
    ENATSC3SLSUSBDDeliveryMethodNode *pDeliveryMethodNode;
    ENATSC3SLSUSBDBroadcastAppServiceNode broadcastAppServiceNode;
    int32_t broadcastAppServiceNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3SLSUSBDNodeType_DeliveryMethod);
    pDeliveryMethodNode = (ENATSC3SLSUSBDDeliveryMethodNode *)pParentNode;

    memset(&broadcastAppServiceNode, 0, sizeof(broadcastAppServiceNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace( &pUSBDNode->dataBuffer,
        sizeof(broadcastAppServiceNode), &broadcastAppServiceNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for broadcastAppService node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pDeliveryMethodNode->broadcastAppServices, 
        broadcastAppServiceNodeOffset))
    {
        ERRORPRINT("Failed to store broadcastAppService node offset\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayInitialize(
        &broadcastAppServiceNode.basePatterns, &pUSBDNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate basePatterns array\n");
        rval = -1;
        goto out;
    }

    broadcastAppServiceNode.base.nodeType = 
        ENATSC3SLSUSBDNodeType_BroadcastAppService;

    VERBOSEPRINT("%*sbroadcastAppService\n", xmlTextReaderDepth(reader) * 2, 
            "");

    rval = EnXMLParserProcessNode(&broadcastAppServiceNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)&broadcastAppServiceNode);

    memcpy(pUSBDNode->dataBuffer.pBuffer + broadcastAppServiceNodeOffset, 
        &broadcastAppServiceNode, sizeof(broadcastAppServiceNode));

out:
    return rval;
}

static int32_t
ProcessUnicastAppServiceNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3SLSUSBDBundleDescriptionNode *pUSBDNode =
        (ENATSC3SLSUSBDBundleDescriptionNode *)pUserData;
    ENATSC3SLSUSBDDeliveryMethodNode *pDeliveryMethodNode;
    ENATSC3SLSUSBDUnicastAppServiceNode unicastAppServiceNode;
    int32_t unicastAppServiceNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3SLSUSBDNodeType_DeliveryMethod);
    pDeliveryMethodNode = (ENATSC3SLSUSBDDeliveryMethodNode *)pParentNode;

    memset(&unicastAppServiceNode, 0, sizeof(unicastAppServiceNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pUSBDNode->dataBuffer,
        sizeof(unicastAppServiceNode), &unicastAppServiceNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for unicastAppService node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pDeliveryMethodNode->unicastAppServices, unicastAppServiceNodeOffset))
    {
        ERRORPRINT("Failed to store unicastAppServiceNodeOffsetnode offset\n");
        rval = 1;
        goto out;
    }

    // Since the data buffer could be reallocated, we are using a temporary
    // regionIdTextNode that lives on the stack. Once we are done
    // processing the node it is committed to the data buffer.
    unicastAppServiceNode.base.nodeType = 
        ENATSC3SLSUSBDNodeType_UnicastAppService;

    if (EnXMLParserDataOffsetArrayInitialize(
        &unicastAppServiceNode.basePatterns, &pUSBDNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate basePatterns array\n");
        rval = -1;
        goto out;
    }

    VERBOSEPRINT("%*sunicastAppService\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&unicastAppServiceNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)&unicastAppServiceNode);

    memcpy(pUSBDNode->dataBuffer.pBuffer + unicastAppServiceNodeOffset, 
        &unicastAppServiceNode, sizeof(unicastAppServiceNode));

out:
    return rval;
}

static EnXMLParserAttribute vDeliveryMethodAttributes[] = 
{
};
static EnXMLParserChildNode vDeliveryMethodChildNodes[] =
{
    {
        .id = ENATSC3SLSUSBDDeliveryMethodChildNode_broadcastAppService,
        .sName = sizeof("broadcastAppService") - 1,
        .pName = "broadcastAppService",
        .handler = ProcessBroadcastAppServiceNode,
    },
    {
        .id = ENATSC3SLSUSBDDeliveryMethodChildNode_unicastAppService,
        .sName = sizeof("unicastAppService") - 1,
        .pName = "unicastAppService",
        .handler = ProcessUnicastAppServiceNode,
    },
};
static EnXMLParserNode deliveryMethodNodeData = 
{
    .pName = "deliveryMethod",
    .pAttributes = vDeliveryMethodAttributes,
    .nAttributes = numof(vDeliveryMethodAttributes),
    .pChildNodes = vDeliveryMethodChildNodes,
    .nChildNodes = numof(vDeliveryMethodChildNodes),
    .textHandler = NULL,
};

static int32_t
NameNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3SLSUSBDBundleDescriptionNode *pUSBDNode = 
        (ENATSC3SLSUSBDBundleDescriptionNode *)pUserData;
    ENATSC3SLSUSBDNameNode *pNameNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3SLSUSBDNodeType_Name);

    pNameNode = (ENATSC3SLSUSBDNameNode *)pParentNode;

    switch (pAttribute->id)
    {
        case ENATSC3SLSUSBDNameNodeAttribute_lang:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pUSBDNode->dataBuffer, (char *)pValue, sValue, 
                &pNameNode->langDataOffset))
            {
                ERRORPRINT("Failed to store lang\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*slang: %.*s\n", xmlTextReaderDepth(reader) * 2, "", 
                sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", 
                xmlTextReaderDepth(reader) * 2, "", pAttribute->id, 
                pAttribute->pName);
            break;
    }

out:
    return rval;
}

static int32_t
NameNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue,
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData,
    EnXMLParserNodeBase *pParentNode)
{
    ENATSC3SLSUSBDBundleDescriptionNode *pUSBDNode = 
        (ENATSC3SLSUSBDBundleDescriptionNode *)pUserData;
    ENATSC3SLSUSBDNameNode *pNameNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == ENATSC3SLSUSBDNodeType_Name);

    pNameNode = (ENATSC3SLSUSBDNameNode *)pParentNode;

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(&pUSBDNode->dataBuffer, 
        (char *)pValue, sValue, &pNameNode->nameDataOffset))
    {
        ERRORPRINT("Failed to store basePattern\n");
        rval = -1;
        goto out;
    }
    VERBOSEPRINT("%*sname: %.*s\n", xmlTextReaderDepth(reader) * 2, "", 
        sValue, pValue);

out:
    return rval;
}

static EnXMLParserAttribute vNameAttributes[] = 
{
    {
        .id = ENATSC3SLSUSBDNameNodeAttribute_lang,
        .sName = sizeof("lang") - 1,
        .pName = "lang",
        .handler = NameNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vNameChildNodes[] =
{
};
static EnXMLParserNode nameNodeData = 
{
    .pName = "name",
    .pAttributes = vNameAttributes,
    .nAttributes = numof(vNameAttributes),
    .pChildNodes = vNameChildNodes,
    .nChildNodes = numof(vNameChildNodes),
    .textHandler = NameNodeTextHandler,
};

static int32_t
UserServiceDescriptionNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3SLSUSBDBundleDescriptionNode *pUSBDNode = 
        (ENATSC3SLSUSBDBundleDescriptionNode *)pUserData;
    ENATSC3SLSUSBDUserServiceDescriptionNode *pUserServiceDescriptionNode;
    int depth;
    int32_t rval = 0;

    assert(pParentNode->nodeType == 
        ENATSC3SLSUSBDNodeType_UserServiceDescription);

    pUserServiceDescriptionNode = 
        (ENATSC3SLSUSBDUserServiceDescriptionNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ENATSC3SLSUSBDUserServiceDescriptionNodeAttribute_globalServiceID:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pUSBDNode->dataBuffer, (char *)pValue, sValue, 
                &pUserServiceDescriptionNode->globalServiceIDDataOffset))
            {
                ERRORPRINT("Failed to store globalServiceID\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%sglobalServiceID: %.*s\n", depth * 2, "", 
                sValue, pValue);
            break;
        case ENATSC3SLSUSBDUserServiceDescriptionNodeAttribute_serviceId:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pUserServiceDescriptionNode->serviceId))
            {
                ERRORPRINT("Failed to parse serviceId %.*s\n", sValue, 
                    pValue);
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sserviceId: %.*s\n", depth * 2, "", sValue, pValue);
            break;
        case ENATSC3SLSUSBDUserServiceDescriptionNodeAttribute_sTSIDUri:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pUSBDNode->dataBuffer, (char *)pValue, sValue, 
                &pUserServiceDescriptionNode->STSIDUriDataOffset))
            {
                ERRORPRINT("Failed to store STSIDUri\n");
                rval = -1;
                goto out;
            }
            VERBOSEPRINT("%*sSTSIDUri: %.*s\n", depth * 2, "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", depth * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static int32_t
ProcessNameNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3SLSUSBDBundleDescriptionNode *pUSBDNode =
        (ENATSC3SLSUSBDBundleDescriptionNode *)pUserData;
    ENATSC3SLSUSBDUserServiceDescriptionNode *pUserServiceDescriptionNode;
    ENATSC3SLSUSBDNameNode nameNode;
    int32_t nameNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == 
        ENATSC3SLSUSBDNodeType_UserServiceDescription);
    pUserServiceDescriptionNode =
        (ENATSC3SLSUSBDUserServiceDescriptionNode *)pParentNode;

    memset(&nameNode, 0, sizeof(nameNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace( &pUSBDNode->dataBuffer, 
        sizeof(nameNode), &nameNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for name node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pUserServiceDescriptionNode->names, nameNodeOffset))
    {
        ERRORPRINT("Failed to store name node offset\n");
        rval = 1;
        goto out;
    }

    // Since the data buffer could be reallocated, we are using a temporary
    // nameNode that lives on the stack. Once we are done processing the node 
    // it is committed to the data buffer.
    nameNode.base.nodeType = ENATSC3SLSUSBDNodeType_Name;

    VERBOSEPRINT("%*sname\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&nameNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)&nameNode);

    memcpy(pUSBDNode->dataBuffer.pBuffer + nameNodeOffset, &nameNode,
        sizeof(nameNode));

out:
    return rval;
}

static int32_t
ProcessDeliveryMethodNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3SLSUSBDBundleDescriptionNode *pUSBDNode =
        (ENATSC3SLSUSBDBundleDescriptionNode *)pUserData;
    ENATSC3SLSUSBDUserServiceDescriptionNode *pUserServiceDescriptionNode;
    ENATSC3SLSUSBDDeliveryMethodNode deliveryMethodNode;
    int32_t deliveryMethodNodeOffset;
    int32_t rval = 0;

    assert(pParentNode->nodeType == 
        ENATSC3SLSUSBDNodeType_UserServiceDescription);
    pUserServiceDescriptionNode =
        (ENATSC3SLSUSBDUserServiceDescriptionNode *)pParentNode;

    memset(&deliveryMethodNode, 0, sizeof(deliveryMethodNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pUSBDNode->dataBuffer,
        sizeof(deliveryMethodNode), &deliveryMethodNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for deliveryMethod node\n");
        rval = 1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pUserServiceDescriptionNode->deliveryMethods, 
        deliveryMethodNodeOffset))
    {
        ERRORPRINT("Failed to store deliveryMethod node offset\n");
        rval = 1;
        goto out;
    }

    // Since the data buffer could be reallocated, we are using a temporary
    // ratingRegionTableNode that lives on the stack. Once we are done
    // processing the node it is committed to the data buffer.
    deliveryMethodNode.base.nodeType = ENATSC3SLSUSBDNodeType_DeliveryMethod;

    if (EnXMLParserDataOffsetArrayInitialize(
        &deliveryMethodNode.broadcastAppServices, &pUSBDNode->dataBuffer, 16, 
        16))
    {
        ERRORPRINT("Failed to allocate broacastAppServices array\n");
        rval = -1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayInitialize(
        &deliveryMethodNode.unicastAppServices, &pUSBDNode->dataBuffer, 16, 
        16))
    {
        ERRORPRINT("Failed to allocate uniscastAppServices array\n");
        rval = -1;
        goto out;
    }

    VERBOSEPRINT("%*sdeliveryMethod\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&deliveryMethodNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)&deliveryMethodNode);

    memcpy(pUSBDNode->dataBuffer.pBuffer + deliveryMethodNodeOffset, 
        &deliveryMethodNode, sizeof(deliveryMethodNode));

out:
    return rval;
}

static EnXMLParserAttribute vUserServiceDescriptionAttributes[] = 
{
    {
        .id = ENATSC3SLSUSBDUserServiceDescriptionNodeAttribute_globalServiceID,
        .sName = sizeof("globalServiceID") - 1,
        .pName = "globalServiceID",
        .handler = UserServiceDescriptionNodeAttributeHandler,
    },
    {
        .id = ENATSC3SLSUSBDUserServiceDescriptionNodeAttribute_serviceId,
        .sName = sizeof("serviceId") - 1,
        .pName = "serviceID",
        .handler = UserServiceDescriptionNodeAttributeHandler,
    },
    {
        .id = ENATSC3SLSUSBDUserServiceDescriptionNodeAttribute_sTSIDUri,
        .sName = sizeof("sTSIDUri") - 1,
        .pName = "sTSIDUri",
        .handler = UserServiceDescriptionNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vUserServiceDescriptionChildNodes[] =
{
    {
        .id = ENATSC3SLSUSBDUserServiceDescriptionChildNode_Name,
        .sName = sizeof("name") - 1,
        .pName = "name",
        .handler = ProcessNameNode,
    },
    {
        .id = ENATSC3SLSUSBDUserServiceDescriptionChildNode_DeliveryMethod,
        .sName = sizeof("deliveryMethod") - 1,
        .pName = "deliveryMethod",
        .handler = ProcessDeliveryMethodNode,
    },
};
static EnXMLParserNode userServiceDescriptionNodeData = 
{
    .pName = "UserServiceDescription",
    .pAttributes = vUserServiceDescriptionAttributes,
    .nAttributes = numof(vUserServiceDescriptionAttributes),
    .pChildNodes = vUserServiceDescriptionChildNodes,
    .nChildNodes = numof(vUserServiceDescriptionChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessUserServiceDescriptionNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3SLSUSBDBundleDescriptionNode *pUSBDNode =
        (ENATSC3SLSUSBDBundleDescriptionNode *)pUserData;
    ENATSC3SLSUSBDBundleDescriptionNode *pBundleDescriptionNode;
    ENATSC3SLSUSBDUserServiceDescriptionNode *pUserServiceDescriptionNode;
    int32_t rval = 0;

    assert(pParentNode->nodeType == 
        ENATSC3SLSUSBDNodeType_BundleDescriptionROUTE);

    pBundleDescriptionNode =
        (ENATSC3SLSUSBDBundleDescriptionNode *)pParentNode;
    pUserServiceDescriptionNode = 
        &pBundleDescriptionNode->userServiceDescription;
    pUserServiceDescriptionNode->base.nodeType = 
        ENATSC3SLSUSBDNodeType_UserServiceDescription;

    if (EnXMLParserDataOffsetArrayInitialize(
        &pUserServiceDescriptionNode->deliveryMethods, &pUSBDNode->dataBuffer, 
        16, 16))
    {
        ERRORPRINT("Failed to allocate deliveryMethods array\n");
        rval = -1;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayInitialize(
        &pUserServiceDescriptionNode->names, &pUSBDNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate names\n");
        rval = -1;
        goto out;
    }

    VERBOSEPRINT("%*suserServiceDescription\n", xmlTextReaderDepth(reader) * 2, 
        "");

    rval = EnXMLParserProcessNode(&userServiceDescriptionNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)pUserServiceDescriptionNode);

out:
    return rval;
}

static EnXMLParserAttribute vBundleDescriptionROUTEAttributes[] = 
{
};
static EnXMLParserChildNode vBundleDescriptionROUTEChildNodes[] =
{
    {
        .id = ENATSC3SLSUSBDNodeType_UserServiceDescription,
        .sName = sizeof("userServiceDescription") - 1,
        .pName = "userServiceDescription",
        .handler = ProcessUserServiceDescriptionNode,
    },
};
static EnXMLParserNode bundleDescriptionROUTENodeData = 
{
    .pName = "bundleDescriptionROUTE",
    .pAttributes = vBundleDescriptionROUTEAttributes,
    .nAttributes = numof(vBundleDescriptionROUTEAttributes),
    .pChildNodes = vBundleDescriptionROUTEChildNodes,
    .nChildNodes = numof(vBundleDescriptionROUTEChildNodes),
    .textHandler = NULL,
};

static int32_t
ProcessUSBDBundleDescriptionROUTENode(xmlTextReaderPtr reader, 
    ENATSC3SLSUSBDBundleDescriptionNode *pParsedUSBD)
{
    int32_t rval = 0;

    VERBOSEPRINT("%*sbundleDescriptionROUTE\n", xmlTextReaderDepth(reader) * 2, 
        "");

    rval = EnXMLParserProcessNode(&bundleDescriptionROUTENodeData, reader, 
        pParsedUSBD, (EnXMLParserNodeBase *)pParsedUSBD);

    return rval;
}

ENATSC3Status
ENATSC3SLSUSBDParse(const char *pXML, int32_t sXML, 
    ENATSC3SLSUSBDBundleDescriptionNode *pParsedUSBD)
{
    xmlTextReaderPtr reader = NULL;
    int32_t fHaveDataBuffer = 0;
    int xmlret;
    ENATSC3Status rval = ENATSC3Status_Success;

    memset(pParsedUSBD, 0, sizeof(ENATSC3SLSUSBDBundleDescriptionNode));
    reader = xmlReaderForMemory(pXML, sXML, NULL, NULL, 0);

    if (!reader)
    {
        ERRORPRINT("Failed to create xml reader\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }

    if (EnXMLParserDataBufferInitialize(&pParsedUSBD->dataBuffer, 4096, 4096))
    {
        ERRORPRINT("Failed to initialize USBD data buffer\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }
    fHaveDataBuffer = 1;

    xmlret = xmlTextReaderRead(reader);
    if (xmlret == 1)
    {
        int nodeType;
        const xmlChar *pLocalName;

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);

        if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "bundleDescriptionROUTE"))
        {
            pParsedUSBD->base.nodeType = 
                ENATSC3SLSUSBDNodeType_BundleDescriptionROUTE;
            ProcessUSBDBundleDescriptionROUTENode(reader, pParsedUSBD);
        }
        else
        {
            ERRORPRINT("Top-level object isn't USBD (%d/%s)\n",
                nodeType, pLocalName);
            rval = ENATSC3Status_Failure;
            goto errout;
        }
    }
    else
    {
        ERRORPRINT("Failed to parse USBD XML\n");
        rval = ENATSC3Status_Failure;
        goto errout;
    }

out:
    if (reader)
    {
        xmlFreeTextReader(reader);
    }

    return rval;

errout:
    if (fHaveDataBuffer)
    {
        if (EnXMLParserDataBufferFree(&pParsedUSBD->dataBuffer))
        {
            ERRORPRINT("Failed to free USBD data buffer\n");
        }
    }

    goto out;
}

ENATSC3Status
ENATSC3SLSUSBDFree(ENATSC3SLSUSBDBundleDescriptionNode *pParsedUSBD)
{
    ENATSC3Status rval = ENATSC3Status_Success;

    if (EnXMLParserDataBufferFree(&pParsedUSBD->dataBuffer))
    {
        ERRORPRINT("Failed to free USBD data buffer\n");
    }

    return rval;
}

