/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libendebug/libendebug.h>

#include "enxml.h"

static char *ensys = "enxml";
static char *ensubsys = "main";

int32_t fEnXMLVerboseG = 0;

#define MAX(x, y) ((x) > (y) ? (x) : (y))

#define STRING_LITERAL_COMPARE(pString, sString, literal) \
    (sString == (sizeof(literal) - 1) && \
    !strncmp((char *)pString, literal, sString))

int64_t
EnXMLParser_strntol(const char *pValue, int32_t sValue)
{
    int32_t rval = 0;

    while (sValue--)
    {
        rval *= 10;
        rval += *(pValue++) - '0';
    }

    return rval;
}

int32_t
EnXMLParser_antol(const char *num, int32_t n, int32_t *pValue)
{
    char text[16];
    int32_t rval = 0;

    strncpy(text, num, n);
    text[n] = 0;

    *pValue = strtol(text, NULL, 10);

    return rval;
}

int32_t
EnXMLParser_antoul(const char *num, int32_t n, uint32_t *pValue)
{
    char text[16];
    int32_t rval = 0;

    strncpy(text, num, n);
    text[n] = 0;

    *pValue = strtoul(text, NULL, 10);

    return rval;
}

int32_t
EnXMLParser_antoll(const char *num, int32_t n, int64_t *pValue)
{
    char text[16];
    int32_t rval = 0;

    strncpy(text, num, n);
    text[n] = 0;

    *pValue = strtoll(text, NULL, 10);

    return rval;
}

int32_t
EnXMLParser_antoull(const char *num, int32_t n, uint64_t *pValue)
{
    char text[16];
    int32_t rval = 0;

    strncpy(text, num, n);
    text[n] = 0;

    *pValue = strtoull(text, NULL, 10);

    return rval;
}

double
EnXMLParser_atofn(const char *num, int32_t n)
{
    char text[24];

    strncpy(text, num, sizeof(text));
    text[sizeof(text) - 1] = 0;

    return atof(text);
}

int32_t
EnXMLParserStringToBool(const char *pValue, int32_t sValue, int32_t *pBoolVal)
{
    int32_t rval = 0;

    if (STRING_LITERAL_COMPARE(pValue, sValue, "true"))
    {
        *pBoolVal = 1;
    }
    else if (STRING_LITERAL_COMPARE(pValue, sValue, "false"))
    {
        *pBoolVal = 0;
    }
    else
    {
        ERRORPRINT("Unhandled boolean value %.*s\n", sValue, pValue);
        rval = -1;
    }

    return rval;
}

int32_t
EnXMLParserStringToFraction(const char *pValue, int32_t sValue, 
    EnXMLFraction *pFraction)
{
    const char *pSlash;
    int32_t i;
    int32_t rval = 0;

    pSlash = pValue;
    for (i = 0; i < sValue; i++, pSlash++)
    {
        if (*pSlash == '/')
        {
            break;
        }
    }

    if (EnXMLParser_antol(pValue, i, &pFraction->num))
    {
        rval = -1;
        goto out;
    }

    if (i == sValue)
    {
        pFraction->den = 1;
    }
    else
    {
        if (EnXMLParser_antol(pValue + i + 1, sValue - i - 1, &pFraction->den))
        {
            rval = -1;
            goto out;
        }
    }

out:
    return rval;
}

int32_t
EnXMLParserStringToRatio(const char *pValue, int32_t sValue, 
    EnXMLRatio *pRatio)
{
    const char *pSlash;
    int32_t i;
    int32_t rval = 0;

    pSlash = pValue;
    for (i = 0; i < sValue; i++, pSlash++)
    {
        if (*pSlash == ':')
        {
            break;
        }
    }

    if (i == sValue)
    {
        rval = -1;
        goto out;
    }

    if (EnXMLParser_antol(pValue, i, &pRatio->w))
    {
        rval = -1;
        goto out;
    }

    if (EnXMLParser_antol(pValue + i + 1, sValue - i - 1, &pRatio->h))
    {
        rval = -1;
        goto out;
    }

out:
    return rval;
}

int32_t
EnXMLParserXMLDurationToMilliseconds(const char *pTime, int32_t sTime, 
    int64_t *pMilliseconds)
{
    char strTmp[24];
    int64_t tMilliseconds = 0;
    int32_t rval = 0;

    if (sTime < 2 || sTime > sizeof(strTmp) - 1 || 
        pTime[0] != 'P' || pTime[1] != 'T')
    {
        rval = -1;
        goto out;
    }

    pTime += 2;
    sTime -=2;

    // "PT" alone would just be zero
    if (sTime == 0)
    {
        goto out;
    }

    // Copy the time string to a temp buffer so we can zero-terminate it. This 
    // is required to prevent strtof from running beyond the string. Ideally 
    // we would use a version of strtof that takes a string length...
    strncpy(strTmp, pTime, sTime);
    strTmp[sTime] = 0;

    pTime = strTmp;

    while (sTime)
    {
        double val;
        char timeType;
        char *pEnd;

        val = strtod(pTime, &pEnd);
        if (pEnd == pTime)
        {
            ERRORPRINT("Failed to parse duration\n");
            rval = -1;
            goto out;
        }

        timeType = *pEnd;
        if (timeType == 'S')
        {
            tMilliseconds += val * 1000.0;
        }
        else if (timeType == 'M')
        {
            tMilliseconds += val * 60.0 * 1000.0;
        }
        else if (timeType == 'H')
        {
            tMilliseconds += val * 60.0 * 60.0 * 1000.0;
        }

        sTime -= pEnd - pTime + 1;
        pTime += pEnd - pTime + 1;
    }

    *pMilliseconds = tMilliseconds;

out:
    return rval;
}

int32_t
EnXMLParserDateToDaysSinceEpoch(int32_t year, int32_t month, int32_t day, 
    int64_t *pDays)
{
    struct tm tm_date;
    time_t tGMSecondsSinceEpoch;
    int32_t rval = 0;

    tm_date.tm_year = year - 1900;
    tm_date.tm_mon = month - 1;
    tm_date.tm_mday = day;
    tm_date.tm_sec = 0;
    tm_date.tm_min = 0;
    tm_date.tm_hour = 0;
    tm_date.tm_wday = 0;
    tm_date.tm_yday = 0;
    tm_date.tm_isdst = 0;

    tGMSecondsSinceEpoch = timegm(&tm_date);
    // Note that -1 is a valid return value for 23:59:59 12/31/1969.
    // However -1 is also the error return from mktime/timegm. Since we
    // always set tm_min and tm_sec to 0, -1 should always indicate an error.
    if (tGMSecondsSinceEpoch == -1)
    {
        ERRORPRINT("Failed to parse date: %d-%d-%d\n", year, month, day);
        rval = -1;
        goto out;
    }

    *pDays = tGMSecondsSinceEpoch / 24 / 60 / 60;

out:
    return rval;
}

int32_t
EnXMLParserZTimeToMilliseconds(const char *pTime, int32_t sTime, 
    int64_t *pMilliseconds)
{
    char strTmp[32];
    int64_t days;
    int64_t year;
    int64_t month;
    int64_t day;
    int64_t hour;
    int64_t minute;
    int64_t milliseconds;
    int32_t rval = 0;
    
    if (sTime < 19 || sTime > sizeof(strTmp) - 1 || pTime[4] != '-' ||
        pTime[7] != '-' || pTime[10] != 'T' || pTime[13] != ':' ||
        pTime[16] != ':')
    {
        ERRORPRINT("Invalid time string %.*s\n", sTime, pTime);
        rval = -1;
        goto out;
    }

    if (pTime[sTime - 1] != 'Z')
    {
        ERRORPRINT("WARN: Non-Z time string %.*s\n", sTime, pTime);
    }
    
    if (EnXMLParser_antoll(pTime + 0, 4, &year))
    {
        rval = -1;
        goto out;
    }
    if (EnXMLParser_antoll(pTime + 5, 2, &month))
    {
        rval = -1;
        goto out;
    }
    if (EnXMLParser_antoll(pTime + 8, 2, &day))
    {
        rval = -1;
        goto out;
    }
    if (EnXMLParser_antoll(pTime + 11, 2, &hour))
    {
        rval = -1;
        goto out;
    }
    if (EnXMLParser_antoll(pTime + 14, 2, &minute))
    {
        rval = -1;
        goto out;
    }
    if (EnXMLParser_antoll(pTime + 17, sTime - 18, &milliseconds))
    {
        rval = -1;
        goto out;
    }
    milliseconds *= 1000;

    if (EnXMLParserDateToDaysSinceEpoch(year, month, day, &days))
    {
        ERRORPRINT("Failed to get days since epoch\n");
        rval = -1;
        goto out;
    }

    *pMilliseconds = milliseconds + (minute * 60000) + (hour * 3600000) +
        days * 86400000;

out:
    return rval;
}

int32_t 
EnXMLParser_dateTimeToSecondsUTC(const char *pDateTime, 
    int32_t sDateTime, int64_t *pSecondsUTC)
{
    assert(0 && "Unimplemented - EnXMLParser_dateTimeToSecondsUTC");
}

int32_t 
EnXMLParser_ntpSecondsToSecondsUTC(int64_t ntpSeconds, int64_t *pSecondsUTC)
{
    int rval = 0;

    *pSecondsUTC = ntpSeconds - 2208988800;

    return rval;
}

int32_t
EnXMLParserProcessNode(EnXMLParserNode *pNodeData, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    int xmlret;
    int depth;
    int fEmpty;
    int32_t fDone = 0;
    int32_t rval = 0;

    depth = xmlTextReaderDepth(reader);
    fEmpty = xmlTextReaderIsEmptyElement(reader);

    pParentNode->presentFlags = 0;

    while (rval == 0)
    {
        const xmlChar *pLocalName;
        const xmlChar *pValue;
        int32_t sLocalName;
        int32_t i;

        xmlret = xmlTextReaderMoveToNextAttribute(reader);
        if (xmlret < 0)
        {
            ERRORPRINT("Failed to move to next attribute\n");
            rval = 1;
            break;
        }
        else if (xmlret == 0)
        {
            // no more attributes
            break;
        }

        if (xmlTextReaderNodeType(reader) != XML_READER_TYPE_ATTRIBUTE)
        {
            ERRORPRINT("Expected attribute - got %d\n", 
                xmlTextReaderNodeType(reader));
            rval = 1;
            break;
        }

        pLocalName = xmlTextReaderConstLocalName(reader);
        pValue = xmlTextReaderConstValue(reader);
        sLocalName = strlen((char *)pLocalName);

        for (i = 0; i < pNodeData->nAttributes; i++)
        {
            if (pNodeData->pAttributes[i].sName == sLocalName && 
                !strcmp(pNodeData->pAttributes[i].pName, (char *)pLocalName))
            {
                if (pNodeData->pAttributes[i].handler(
                    pNodeData->pAttributes + i, pValue, strlen((char *)pValue), 
                    reader, pUserData, pParentNode))
                 {
                     ERRORPRINT("Error handling attribute %s - ignoring\n", 
                         pLocalName);
                 }
                 else
                 {
                     pParentNode->presentFlags |= pNodeData->pAttributes[i].id;
                 }
                 break;
            }
        }

        if (i == pNodeData->nAttributes)
        {
            if (fEnXMLVerboseG)
            {
                ERRORPRINT("%*sUNHANDLED: {%s:%s}/{%s}\n", 
                    xmlTextReaderDepth(reader) * 2, "", 
                    xmlTextReaderConstPrefix(reader), pLocalName, pValue);
            }
        }
    }

    if (rval)
    {
        goto out;
    }

    if (fEmpty)
    {
        if (pNodeData->textHandler)
        {
            rval = pNodeData->textHandler(pNodeData, (xmlChar *)"", 0, reader, pUserData, 
                pParentNode);
        }
        if (fEnXMLVerboseG)
        {
            LOGPRINT("%*sEnd empty element (%"PRIx64")\n", depth * 2, "", 
                pParentNode->presentFlags);  
        }
        goto out;
    }

    while (rval == 0 && !fDone)
    {
        int nodeType;
        const xmlChar *pLocalName;
        const xmlChar *pValue;
        int32_t sLocalName;
        int32_t i;

        xmlret = xmlTextReaderRead(reader);
        if (xmlret  < 0)
        {
            ERRORPRINT("Error processing node\n");
            rval = 1;
            break;
        }
        else if (xmlret == 0)
        {
            // No more nodes
            break;
        }

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);
        sLocalName = strlen((char *)pLocalName);

        switch (nodeType)
        {
            case XML_READER_TYPE_ELEMENT: // Start element
                for (i = 0; i < pNodeData->nChildNodes; i++)
                {
                    if (pNodeData->pChildNodes[i].sName == sLocalName && 
                        !strcmp(pNodeData->pChildNodes[i].pName, 
                            (char *)pLocalName))
                    {
                        if (pNodeData->pChildNodes[i].handler(
                            pNodeData->pChildNodes + i, reader, pUserData, 
                            pParentNode))
                         {
                             ERRORPRINT(
                                 "Error handling child node %s\n",
                                 pLocalName);
                             rval = 1;
                         }
                         else
                         {
                             pParentNode->presentFlags |= 
                                  pNodeData->pChildNodes[i].id;
                         }
                         break;
                    }
                }
                
                if (i == pNodeData->nChildNodes)
                {
                    if (fEnXMLVerboseG)
                    {
                        LOGPRINT("UNHANDLED Start element %s attribute count %d on %s\n", 
                            xmlTextReaderConstLocalName(reader),
                            xmlTextReaderAttributeCount(reader),
                            pNodeData->pName); 
                        while (xmlTextReaderMoveToNextAttribute(reader))
                        {
                            LOGPRINT("  attribute type %d name %s value %s\n",
                                xmlTextReaderNodeType(reader),
                                xmlTextReaderConstLocalName(reader),
                                xmlTextReaderConstValue(reader));
                        }
                    }
                }
                break;
            case XML_READER_TYPE_TEXT: // text node
                if (pNodeData->textHandler)
                {
                    pValue = xmlTextReaderConstValue(reader);

                    rval = pNodeData->textHandler(pNodeData, pValue, 
                        strlen((char *)pValue), reader, pUserData, pParentNode);
                }
                else
                {
                    if (fEnXMLVerboseG)
                    {
                        if (fEnXMLVerboseG)
                        {
                            LOGPRINT("UNHANDLED text node %s {%s} on %s\n", 
                                xmlTextReaderConstLocalName(reader),
                                xmlTextReaderConstValue(reader), 
                                pNodeData->pName); 
                        }
                    }
                }
                break;
            case XML_READER_TYPE_SIGNIFICANT_WHITESPACE:
                break;
            case XML_READER_TYPE_END_ELEMENT: // End element
                if (fEnXMLVerboseG)
                {
                    if (fEnXMLVerboseG)
                    {
                        LOGPRINT("%*sEnd element %s (%"PRIx64")\n", depth * 2, 
                            "",  xmlTextReaderConstLocalName(reader), 
                            pParentNode->presentFlags); 
                    }
                }
                if (xmlTextReaderDepth(reader) == depth)
                {
                    fDone = 1;
                }
                break;
            default:
                if (fEnXMLVerboseG)
                {
                    LOGPRINT("Unhandled %d %s\n", nodeType, 
                        xmlTextReaderConstLocalName(reader)); 
                }
                break;
        }
    }

out:
    return rval;
}

int32_t 
EnXMLParserDataBufferInitialize(EnXMLParserDataBuffer *pDataBuffer, 
    int32_t sInitial, int32_t sIncrement)
{
    int32_t rval = 0;

    pDataBuffer->pBuffer = (uint8_t *)malloc(sInitial);
    if (!pDataBuffer->pBuffer)
    {
        rval = 1;
        goto out;
    }
    pDataBuffer->sBuffer = sInitial;
    pDataBuffer->nBufferBytesUsed = 0;
    pDataBuffer->increment = sIncrement;

out:
    return rval;
}

int32_t 
EnXMLParserDataBufferFree(EnXMLParserDataBuffer *pDataBuffer)
{
    int32_t rval = 0;

    free(pDataBuffer->pBuffer);
    pDataBuffer->pBuffer = NULL;
    
    return rval;
}

static int32_t 
EnXMLParserDataBufferGrow(EnXMLParserDataBuffer *pDataBuffer, int32_t nBytes)
{
    int32_t rval = 0;
    uint8_t *pNew;

    LOGPRINT("Growing buffer to %d\n", pDataBuffer->sBuffer + nBytes);
    pNew = (uint8_t *)realloc(pDataBuffer->pBuffer, pDataBuffer->sBuffer + 
        nBytes);
    if (!pNew)
    {
        ERRORPRINT("Failed to grow data buffer to %d\n", 
            pDataBuffer->sBuffer + nBytes);
        rval = 1;
        goto out;
    }
    pDataBuffer->pBuffer = pNew;
    pDataBuffer->sBuffer += nBytes;

out:
    return rval;
}

int32_t 
EnXMLParserDataBufferAddString(EnXMLParserDataBuffer *pDataBuffer, 
    const char *pString, int32_t sString, int32_t *pDataBufferOffset)
{
    int32_t rval = 0;

    if (pDataBuffer->sBuffer - pDataBuffer->nBufferBytesUsed < sString)
    {
        int32_t increment;
        
        increment = MAX(pDataBuffer->increment, 
            pDataBuffer->sBuffer - pDataBuffer->nBufferBytesUsed + sString);
        if (EnXMLParserDataBufferGrow(pDataBuffer, increment))
        {
            ERRORPRINT("Failed to grow data buffer\n");
            rval = 1;
            goto out;
        }
    }

    assert(pDataBuffer->sBuffer - pDataBuffer->nBufferBytesUsed >= sString);

    *pDataBufferOffset = pDataBuffer->nBufferBytesUsed;
    memcpy(pDataBuffer->pBuffer + pDataBuffer->nBufferBytesUsed, 
        pString, sString);
    pDataBuffer->nBufferBytesUsed += sString;

out:
    return rval;
}

int32_t 
EnXMLParserDataBufferAddStringAndZeroTerminate(
    EnXMLParserDataBuffer *pDataBuffer, const char *pString, 
    int32_t sString, int32_t *pDataBufferOffset)
{
    int32_t rval = 0;

    if (pDataBuffer->sBuffer - pDataBuffer->nBufferBytesUsed < sString + 1)
    {
        int32_t increment;
        
        increment = MAX(pDataBuffer->increment, 
            pDataBuffer->sBuffer - pDataBuffer->nBufferBytesUsed + 
            sString + 1);
        if (EnXMLParserDataBufferGrow(pDataBuffer, increment))
        {
            ERRORPRINT("Failed to grow data buffer\n");
            rval = 1;
            goto out;
        }
    }

    assert(pDataBuffer->sBuffer - pDataBuffer->nBufferBytesUsed >= 
        sString + 1);

    *pDataBufferOffset = pDataBuffer->nBufferBytesUsed;
    memcpy(pDataBuffer->pBuffer + pDataBuffer->nBufferBytesUsed, 
        pString, sString);
    pDataBuffer->nBufferBytesUsed += sString;
    pDataBuffer->pBuffer[pDataBuffer->nBufferBytesUsed++] = 0;

out:
    return rval;
}

int32_t 
EnXMLParserDataBufferAllocateSpace(EnXMLParserDataBuffer *pDataBuffer,
    int32_t nBytes, int32_t *pDataBufferOffset)
{
    int32_t rval = 0;

    if (pDataBuffer->sBuffer - pDataBuffer->nBufferBytesUsed < nBytes)
    {
        int32_t increment;
        
        increment = MAX(pDataBuffer->increment, 
            pDataBuffer->sBuffer - pDataBuffer->nBufferBytesUsed + nBytes);
        if (EnXMLParserDataBufferGrow(pDataBuffer, increment))
        {
            ERRORPRINT("Failed to grow data buffer\n");
            rval = 1;
            goto out;
        }
    }

    assert(pDataBuffer->sBuffer - pDataBuffer->nBufferBytesUsed >= nBytes);

    *pDataBufferOffset = pDataBuffer->nBufferBytesUsed;
    pDataBuffer->nBufferBytesUsed += nBytes;

out:
    return rval;
}

int32_t 
EnXMLParserDataOffsetArrayInitialize(
    EnXMLParserDataOffsetArray *pDataOffsetArray,
    EnXMLParserDataBuffer *pDataBuffer, int32_t sInitial, int32_t sIncrement)
{
    int32_t rval = 0;

    if (EnXMLParserDataBufferAllocateSpace(pDataBuffer, 
        sInitial * sizeof(int32_t), &pDataOffsetArray->dataBufferOffset))
    {
        ERRORPRINT("Failed to allocate DataOffsetArray space\n");
        rval = 1;
        goto out;
    }

    pDataOffsetArray->pDataBuffer = pDataBuffer;
    pDataOffsetArray->sInitialSize = sInitial;
    pDataOffsetArray->nOffsetsAllocated = sInitial;
    pDataOffsetArray->nOffsetsUsed = 0;
    pDataOffsetArray->increment = sIncrement;

out:
    return rval;
}

int32_t 
EnXMLParserDataOffsetArrayAppendOffset(
    EnXMLParserDataOffsetArray *pDataOffsetArray, int32_t offset)
{
    int32_t *pOffsetArray;
    int32_t rval = 0;

    if (pDataOffsetArray->nOffsetsAllocated == pDataOffsetArray->nOffsetsUsed)
    {
        int32_t newOffset;

        if (EnXMLParserDataBufferAllocateSpace(pDataOffsetArray->pDataBuffer,
            sizeof(int32_t) * (pDataOffsetArray->nOffsetsAllocated + 
            pDataOffsetArray->increment),
            &newOffset))
        {
            ERRORPRINT("Failed to grow data offset array\n");
            rval = 1;
            goto out;
        }

        pDataOffsetArray->nOffsetsAllocated += pDataOffsetArray->increment;
        pDataOffsetArray->dataBufferOffset = newOffset;
    }

    pOffsetArray = (int32_t *)(pDataOffsetArray->pDataBuffer->pBuffer + 
        pDataOffsetArray->dataBufferOffset);
    pOffsetArray[pDataOffsetArray->nOffsetsUsed++] = offset;

out:
    return rval;
}

