/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com).
 * All Rights Reserved.
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it 
 * and/or modify it under the terms of the GNU Affero General Public License 
 * as published by the Free Software Foundation, either version 3 of the 
 * License, or (at your option) any later version.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be 
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with the Entangle ATSC 3.0 Base Stack. If not, see 
 * <https://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <inttypes.h>
#include <string.h>
#include <pthread.h>
#include <errno.h>
#include <assert.h>
#include <sys/resource.h>
#include <sys/syscall.h>
#include <sys/types.h>

#include "libendebug.h"

static char *ensys ="enlog";
static char *ensubsys = "main";

#define STDOUT 1

#define NANOSECONDS_PER_SECOND 1000000000

#define ColorDefault    "\x1B[0m"
#define ColorRed        "\x1B[31m"
#define ColorGreen      "\x1B[32m"
#define ColorYellow     "\x1B[33m"
#define ColorBlue       "\x1B[34m"
#define ColorMagenta    "\x1B[35m"
#define ColorCyan       "\x1B[36m"
#define ColorWhite      "\x1B[37m"

static struct
{
    int32_t fInitialized;
    int32_t fNoConsole;
    int32_t fAllConsole;
} enlogGlobalsG = 
{
    .fInitialized = 0,
    .fNoConsole = 0,
    .fAllConsole = 0,
};

int32_t
EnLogInitialize(int32_t fLogFiles)
{
    int32_t rval = 0;

    if (enlogGlobalsG.fInitialized)
    {
        fprintf(stderr, ColorRed "%s:%s:%d Already initialized\n" ColorDefault,
            ensys, ensubsys, __LINE__);
        rval = 1;
        goto out;
    }

    if (getenv("ENLOG_ALLCONSOLE"))
    {
        enlogGlobalsG.fAllConsole = atol(getenv("ENLOG_ALLCONSOLE"));
    }

    enlogGlobalsG.fInitialized = 1;

out:
    return rval;
}

int32_t
EnLogShutdown(void)
{
    int32_t rval = 0;

    if (!enlogGlobalsG.fInitialized)
    {
        fprintf(stderr, ColorRed "%s:%s:%d Not initialized\n" ColorDefault,
            ensys, ensubsys, __LINE__);
        rval = 1;
        goto out;
    }

out:
    return rval;
}

int32_t
EnLogFlush(void)
{
    return 0;
}

int32_t
EnLog_Internal(int32_t fSendToConsole, int32_t fLogToFile, EnLogFile log,
    const char *pMessage, int32_t sMessage)
{
    char *pStartColor = NULL;
    int32_t sStartColor;
    char *pEndColor = NULL;
    int32_t sEndColor;
    int32_t rval = 0;

    switch (log)
    {
        case EnLogFile_Debug:
            break;
        case EnLogFile_Info:
            break;
        case EnLogFile_Error:
            pStartColor = ColorRed;
            sStartColor = sizeof(ColorRed);
            pEndColor = ColorDefault;
            sEndColor = sizeof(ColorDefault);
            break;
        default:
            fprintf(stderr, ColorRed "%s:%s:%d UNHANDLED LOG LEVEL %d\n"
                ColorDefault, ensys, ensubsys, __LINE__, log);
            break;
    }

    if ((fSendToConsole && !enlogGlobalsG.fNoConsole) || 
         enlogGlobalsG.fAllConsole)
    {
        if (pStartColor)
        {
            write(STDOUT, pStartColor, sStartColor);
        }

        write(STDOUT, pMessage, sMessage);

        if (pEndColor)
        {
            write(STDOUT, pEndColor, sEndColor);
        }
    }

    return rval;
}

int32_t
EnLog(int32_t fSendToConsole, int32_t fLogToFile, EnLogFile log,
    const char *sys, const char *subsys, int32_t line,
    const char *format, ...)
{
    struct timespec utcnow;
    char strMessage[8192];
    int32_t nb = 0;
    va_list ap;
    int32_t rval = 0;

    clock_gettime(CLOCK_REALTIME, &utcnow);

    nb = snprintf(strMessage, sizeof(strMessage), "%ld.%06ld %s:%s:%d ",
        utcnow.tv_sec, utcnow.tv_nsec / 1000, sys, subsys, line);
    assert(nb >= 0);
    if (nb > sizeof(strMessage))
    {
        nb = sizeof(strMessage);
    }

    va_start(ap, format);

    nb += vsnprintf(strMessage + nb, sizeof(strMessage) - nb, format, ap);
    assert(nb >= 0);
    if (nb > sizeof(strMessage))
    {
        nb = sizeof(strMessage);
    }

    EnLog_Internal(fSendToConsole, fLogToFile, log, strMessage, nb);

    va_end(ap);

    return rval;
}

int32_t
EnLogNoPrefix(int32_t fSendToConsole, int32_t fLogToFile, EnLogFile log,
    const char *format, ...)
{
    char strMessage[8192];
    int32_t nb = 0;
    va_list ap;
    int32_t rval = 0;

    va_start(ap, format);

    nb = vsnprintf(strMessage, sizeof(strMessage), format, ap);
    if (nb >= sizeof(strMessage))
    {
        nb = sizeof(strMessage);
    }

    EnLog_Internal(fSendToConsole, fLogToFile, log, strMessage, nb);

    va_end(ap);

    return rval;
}

int32_t
EnLogThrottle(struct timespec *pTNextMessage, struct timespec throttle,
    int32_t fSendToConsole, int32_t fLogToFile, EnLogFile log,
    const char *sys, const char *subsys, int32_t line,
    const char *format, ...)
{
    struct timespec now;

    clock_gettime(CLOCK_MONOTONIC, &now);

    if (now.tv_sec > pTNextMessage->tv_sec ||
        ((now.tv_sec == pTNextMessage->tv_sec) &&
         (now.tv_nsec > pTNextMessage->tv_nsec)))
    {
        struct timespec utcnow;
        char strMessage[8192];
        va_list ap;
        int32_t nb = 0;

        clock_gettime(CLOCK_REALTIME, &utcnow);

        nb = snprintf(strMessage, sizeof(strMessage), "%ld.%06ld %s:%s:%d ",
            utcnow.tv_sec, utcnow.tv_nsec / 1000, sys, subsys, line);
        assert(nb >= 0);
        if (nb > sizeof(strMessage))
        {
            nb = sizeof(strMessage);
        }

        va_start(ap, format);

        nb += vsnprintf(strMessage + nb, sizeof(strMessage) - nb, format, ap);
        assert(nb >= 0);
        if (nb > sizeof(strMessage))
        {
            nb = sizeof(strMessage);
        }

        if (!pTNextMessage->tv_sec)
        {
            *pTNextMessage = now;
        }

        EnLog_Internal(fSendToConsole, fLogToFile, log, strMessage, nb);

        va_end(ap);

        pTNextMessage->tv_sec += throttle.tv_sec;
        pTNextMessage->tv_nsec += throttle.tv_nsec;
        if (pTNextMessage->tv_nsec > NANOSECONDS_PER_SECOND)
        {
            pTNextMessage->tv_sec++;
            pTNextMessage->tv_nsec -= NANOSECONDS_PER_SECOND;
        }

        if (pTNextMessage->tv_sec < now.tv_sec ||
            (pTNextMessage->tv_sec == now.tv_sec &&
             pTNextMessage->tv_nsec < now.tv_nsec))
        {
            *pTNextMessage = now;
        }
    }

    return 0;
}

int32_t
EnLogNoPrefixThrottle(struct timespec *pTNextMessage,
    struct timespec throttle, int32_t fSendToConsole, int32_t fLogToFile,
    EnLogFile log, const char *format, ...)
{
    struct timespec now;

    clock_gettime(CLOCK_MONOTONIC, &now);

    if (now.tv_sec > pTNextMessage->tv_sec ||
        ((now.tv_sec == pTNextMessage->tv_sec) &&
         (now.tv_nsec > pTNextMessage->tv_nsec)))
    {
        int32_t nb = 0;
        char strMessage[8192];
        va_list ap;

        va_start(ap, format);

        nb += vsnprintf(strMessage + nb, sizeof(strMessage) - nb, format, ap);
        assert(nb >= 0);
        if (nb > sizeof(strMessage))
        {
            nb = sizeof(strMessage);
        }

        EnLog_Internal(fSendToConsole, fLogToFile, log, strMessage, nb);

        va_end(ap);

        if (!pTNextMessage->tv_sec)
        {
            *pTNextMessage = now;
        }

        pTNextMessage->tv_sec += throttle.tv_sec;
        pTNextMessage->tv_nsec += throttle.tv_nsec;
        if (pTNextMessage->tv_nsec > NANOSECONDS_PER_SECOND)
        {
            pTNextMessage->tv_sec++;
            pTNextMessage->tv_nsec -= NANOSECONDS_PER_SECOND;
        }

        if (pTNextMessage->tv_sec < now.tv_sec ||
            (pTNextMessage->tv_sec == now.tv_sec &&
             pTNextMessage->tv_nsec < now.tv_nsec))
        {
            *pTNextMessage = now;
        }
    }

    return 0;
}
